#!/opt/cloudlinux/venv/bin/python3 -bb
# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import os
import uuid
from urllib.parse import urlparse
from datetime import time
from unittest import TestCase
import mock

import sqlalchemy.sql.schema
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.pool import StaticPool

from lvestats.lib.dbengine import clear_rows, create_db, make_db_engine, cfg_get_uri
from lvestats.lib.commons.func import init_database_rnd, drop_database
from lvestats.orm.history_gov import history_gov
from lvestats.orm.history import history, history_x60
from lvestats.orm.servers import servers


class BaseDb(object):
    # default database config; sqlite in memory
    _cfg_sqlite = {
        'db_type': 'sqlite',
        'sqlite_db_path': '/var/lve/test_lvestats2.sqlite',
        'connect_string': 'root:root@localhost',
        'server_id': 'localhost',
        'debug': False
    }
    _cfg_mysql = {
        'db_type': 'mysql',
        'connect_string': 'lvestats2:lvestats2_passwd@localhost:5432/db_lvestats2',
        'server_id': 'localhost'
    }
    _cfg_pgsql = {
        'db_type': 'postgresql',
        'connect_string': 'lvestat:passw@localhost/dblvestat',
        'server_id': 'localhost'
    }
    cfg = _cfg_sqlite

    engine = None  # link to sqlalchemy database engine
    _drop_database = False  # need delete temporary database (using in tearDownClass)
    _temporary_db = False  # for controlling present temporary database
    need_clear_rows = True  # need delete rows in tables; checked in setUp before every unit test
    _database_prefix = 'tmp_test_'
    _base_db_test_engine = None

    @classmethod
    def drop_database(cls):
        cls.engine.dispose()
        if cls.engine.url.drivername == 'sqlite':
            try:
                if cls.engine.url.database:
                    os.unlink(cls.engine.url.database)
            except OSError as e:
                print('WARNING: {}'.format(e))
        else:
            drop_database(cls.cfg_get_uri())

    @classmethod
    def drop_all_tables(cls):
        meta = sqlalchemy.sql.schema.MetaData(cls.engine)
        meta.reflect()
        meta.drop_all()

    @classmethod
    def create_tables(cls):
        create_db(cls.engine)

    @classmethod
    def recreate_tables(cls):
        cls.drop_all_tables()
        cls.create_tables()

    @classmethod
    def clear_rows(cls, server_id=None):
        try:
            clear_rows(cls.engine, server_id=server_id)
        except (OperationalError, ProgrammingError):
            cls.recreate_tables()

    @classmethod
    def set_superglobal_engine(cls, engine):
        BaseDb._base_db_test_engine = engine

    @classmethod
    def get_superglobal_engine(cls):
        if hasattr(BaseDb, '_base_db_test_engine'):
            return getattr(BaseDb, '_base_db_test_engine')

    @classmethod
    def cfg_update_uri(cls, uri):
        uri_ = urlparse(uri)
        if uri_.scheme:
            # uri_.scheme can be either in 'mysql' or 'mysql+pymysql' format
            db_type = uri_.scheme.split('+')[0]
            cls.cfg['db_type'] = db_type
        if uri_.scheme == 'sqlite':
            cls.cfg['sqlite_db_path'] = uri_.path[1:]
        else:
            cls.cfg['connect_string'] = uri_.netloc + uri_.path

    @classmethod
    def cfg_update_db_name(cls, db_name):
        if cls.cfg['db_type'] == 'sqlite':
            cls.cfg['sqlite_db_path'] = '/' + db_name
        else:
            uri_cfg_ = urlparse(cls.cfg['db_type'] + '://' + cls.cfg['connect_string'])
            cls.cfg['connect_string'] = uri_cfg_.netloc + '/' + db_name

    @classmethod
    def cfg_get_db_name(cls):
        if cls.cfg['db_type'] == 'sqlite':
            return cls.cfg['sqlite_db_path']
        else:
            return urlparse(cls.cfg['db_type'] + '://' + cls.cfg['connect_string']).path

    @classmethod
    def cfg_get_uri(cls):
        return cfg_get_uri(cls.cfg)

    @classmethod
    def init_database(cls, uri=None, drop_database_=True):
        """
        Create database and configure sqlalchemy engine
        If uri=None its nitialise sqlite database in memory
        :param uri: uri to database
        :param drop_database_: delete temporary database after all tests
        :return:
        """
        cls._drop_database = drop_database_
        if uri is not None:
            cls.cfg_update_uri(uri)
        db_name = cls.cfg_get_db_name()
        pool_class = StaticPool if cls.cfg['db_type'] != 'sqlite' else None
        if cls.cfg['db_type'] != 'sqlite' and not db_name:  # create database if need
            # create engine without database
            cls.engine = make_db_engine(cfg=cls.cfg, pool_class=pool_class, check_support=False)
            db_name = init_database_rnd(cls.engine, cls._database_prefix)  # create temporary database
            cls._temporary_db = True
            cls.engine.dispose()  # disconnect
            cls.cfg_update_db_name(db_name)
        else:
            if os.path.isdir(db_name):
                cls.cfg_update_db_name(os.path.join(db_name, cls._database_prefix + str(uuid.uuid4())[0:8] + '.sqlite'))
                cls._temporary_db = True

        if cls.cfg['db_type'] == 'sqlite' and not os.path.isfile(cls.cfg['sqlite_db_path']):
            cls._temporary_db = True  # drop sqlite database if need
        # create engine without database
        cls.engine = make_db_engine(cfg=cls.cfg, pool_class=pool_class, check_support=False)

    @classmethod
    def drop_trash(cls):
        """
        Force clear rows and delete temporary database
        :return:
        """
        try:
            cls.clear_rows(server_id=cls.cfg['server_id'])  # clear tables (delete rows only)
        except (OperationalError, ProgrammingError):
            pass
        if cls._temporary_db:
            cls.drop_database()
            cls._temporary_db = False  # set that temporary database not present

    def mk_db_session(self):
        return sessionmaker(bind=self.engine)()

    def save_dbgov_line(self, server_id, uid, _time, **kwargs):
        insert_dict = {'server_id': server_id, 'uid': uid, 'ts': _time}
        insert_dict.update(kwargs)
        self.engine.execute(history_gov.__table__.insert(insert_dict))

    def save_servers_line(self, server_id, version='1'):
        self.engine.execute(servers.__table__.insert({'server_id': server_id, 'lve_version': version}))

    def bulk_history_start(self):
        connection = self.engine.connect()
        return connection, connection.begin()

    def bulk_history_add(self, connection, server_id, uid, _time, **kwargs):
        connection.execute(history.__table__.insert(self.prep_history_line(server_id, uid, _time, **kwargs)))

    @staticmethod
    def prep_history_line(server_id, uid, _time, **kwargs):
        insert_dict = {val[0]: 0 for val in list(history.__table__.columns.items())}  # setup 0 for every column
        insert_dict.update(kwargs)
        insert_dict.update({'server_id': server_id, 'id': uid, 'created': _time})
        return insert_dict

    def save_history_x60_line(self, server_id, uid, _time, **kwargs):
        insert_dict = {val[0]: 0 for val in list(history_x60.__table__.columns.items())}  # setup 0 for every column
        insert_dict.update(kwargs)
        insert_dict.update({'server_id': server_id, 'id': uid, 'created': _time})
        self.engine.execute(history_x60.__table__.insert(insert_dict))

    def save_history_line(self, server_id, uid, _time, keepalive=False, **kwargs):
        self.engine.execute(history.__table__.insert(self.prep_history_line(server_id, uid, _time, **kwargs)))
        if keepalive:
            # save keep alive record
            self.engine.execute(history.__table__.insert(self.prep_history_line(server_id, 0, _time)))

    def idle_history_data(self):
        create_db(self.engine)
        start = int(time.time())
        for t in range(start - 3600, start, 60):
            self.save_history_line('localhost', uid=0, _time=t)
            self.save_history_line('localhost', uid=506, _time=t + 1)
            self.save_history_line('localhost', uid=507, _time=t + 1)


class BaseDbTest(TestCase, BaseDb):
    HIDE_MAX_UIDS = 'false'

    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        import logging
        from lvestats.lib.commons.logsetup import init_console_log
        logging.root.setLevel(logging.DEBUG)
        init_console_log(logging.WARNING)

    @classmethod
    def setUpClass(cls):
        TestCase.setUpClass()
        if os.environ.get('CLTEST_DB') == 'mysql':
            cls.cfg = cls._cfg_mysql.copy()
        elif os.environ.get('CLTEST_DB') == 'pgsql':
            cls.cfg = cls._cfg_pgsql.copy()
        else:
            cls.cfg = cls._cfg_sqlite.copy()
        cls.engine = cls.get_superglobal_engine()  # link super global database engine to class attribute engine
        if cls.engine is None:  # initial database engine if need
            uri = os.environ.get('CLTEST_DBURI')
            cls.init_database(uri)  # create temporary database and create config mock
        else:
            cls.clear_rows()

        cls.read_config_proceed_patcher = mock.patch('lvestats.lib.lveinfolib.read_config', return_value={
            'hide_lve_more_than_maxuid': cls.HIDE_MAX_UIDS
        })
        cls.read_config_proceed_patcher.start()

    @classmethod
    def tearDownClass(cls):
        TestCase.tearDownClass()
        if cls._temporary_db and cls._drop_database:  # drop temporary database if need
            cls.drop_database()
            cls._temporary_db = False  # set that temporary database not present

        if hasattr(cls, 'read_config_proceed_patcher'):
            cls.read_config_proceed_patcher.stop()

    def setUp(self):
        TestCase.setUp(self)
        if self.need_clear_rows:
            self.clear_rows()  # clear all rows if need

    def tearDown(self):
        TestCase.tearDown(self)
