#!/opt/cloudlinux/venv/bin/python3 -bb
# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
"""
Generating valid test lve-stats >= 2.1-15 database
"""
import sys
import os
import csv
import random
import json
from datetime import datetime
from urllib.parse import urlparse

from sqlalchemy.exc import OperationalError, IntegrityError
from typing import Dict, Iterable, Tuple, Optional, Generator  # pylint: disable=unused-import

from lvestats.lib import dbengine, config, ustate
from lvestats.lib.lveinfolib import FIELD_TO_TABLE_COLUMN, FIELD_AVERAGE, FIELD_LIMIT, FIELD_FAULT, FIELD_MAX
from lvestats.orm import history, history_gov, servers, incident
from lvestats.lib.commons.dateutil import gm_datetime_to_unixtimestamp, local_to_gm
from lvestats.lib.commons.progress import ProgressCallback

# this need only for command line
if __name__ == '__main__':
    import argparse

MEMORY_PAGE = 1024 * 4
INCIDENT_LENGTH_BOUNDARIES = (50, 270)
YEAR = datetime.now().year


class BaseDataCase(object):
    AGGREGATION_PERIOD = 60
    table = None

    def __init__(self, csv_dict):
        self.csv_dict = csv_dict  # original input data
        self._csv_dict = self._normalize_and_validate_data()  # corrected and validated data
        self.ts_from = self._csv_dict['From']
        self.ts_to = self._csv_dict['To']
        self._table_rows = max(1, (self.ts_to - self.ts_from) // self.AGGREGATION_PERIOD)
        self._unprocessed_rows = self._table_rows

    def __getitem__(self, k):
        return self._csv_dict[k]

    def get(self, k, d=None):
        return self._csv_dict.get(k, d)

    @staticmethod
    def _str_to_timestamp(time_str):
        # type: (str) -> int
        try:
            dt = datetime.strptime(time_str, '%m-%d %H:%M').replace(year=YEAR)
        except ValueError:
            dt = datetime.strptime(time_str, '%Y-%m-%d %H:%M')
        gm_datetime = local_to_gm(dt)
        return gm_datetime_to_unixtimestamp(gm_datetime)

    def _normalize_and_validate_data(self):
        raise NotImplementedError()

    def get_insert_dict_by_timestamp(self, ts, server_id):
        raise NotImplementedError()

    @classmethod
    def get_service_alive_db_record(cls, ts, server_id):
        # type: () -> Optional[Dict]
        return None

    def _get_table_params_dict(self):
        insert_dict = {column: 0 for column in list(self.table.__table__.c.keys())}
        return insert_dict

    def is_empty(self):
        return self._unprocessed_rows == 0


class DataCase(BaseDataCase):
    """
    Class to storing one test case data
    """
    table = history

    def __init__(self, csv_dict):
        super(DataCase, self).__init__(csv_dict)

        self._control_avg_sum = {item: int(self._csv_dict[item]) * self._table_rows for item in FIELD_AVERAGE}
        self._control_fault_sum = {item: int(self._csv_dict[item]) for item in FIELD_FAULT}

    def _normalize_and_validate_data(self):
        """
        Convert and validate input raw data
        """
        _csv_dict = dict()

        for item, val in list(self.csv_dict.items()):

            # convert string to timestamp
            if item in ('From', 'To'):
                _csv_dict[item] = self._str_to_timestamp(val)

            # correct cpu
            elif item in ('aCPU', 'mCPU', 'lCPU'):
                _csv_dict[item] = int(float(val) * 100)

            # correct and validate memory values
            elif item in ('aVMem', 'mVMem', 'lVMem', 'aPMem', 'mPMem', 'lPMem'):
                val_corrected = float(val) / MEMORY_PAGE
                if val_corrected != int(val_corrected):  # check if value is multiple of MEMORY_PAGE
                    raise ValidateError('Value "{}" must be a multiple of {}'.format(val, MEMORY_PAGE))
                _csv_dict[item] = int(val_corrected)

            # correct IO
            elif item in ('aIO', 'mIO', 'lIO'):
                _csv_dict[item] = int(float(val) * 1024)

            # convert all variables to integer (all columns in table history is integer)
            else:
                _csv_dict[item] = int(val)

        # set zero for non present items (exclude max items)
        _csv_dict.update({item: 0 for item in (FIELD_AVERAGE + FIELD_LIMIT + FIELD_FAULT) if item not in _csv_dict})

        # validate max; if fault present max >= limit
        for limit_item, fault_item, max_item in zip(FIELD_LIMIT, FIELD_FAULT, FIELD_MAX):
            if max_item in _csv_dict and _csv_dict[fault_item] > 0 and _csv_dict[max_item] < _csv_dict[limit_item]:
                raise ValidateError(
                    'fault > 0 if it should be satisfied max >= limit; '
                    '{fault_item}={fault_val} but {max_item}={max_value} '
                    'and {limit_item}={limit_value}'.format(
                        fault_item=fault_item, fault_val=self.csv_dict[fault_item],
                        max_item=max_item, max_value=self.csv_dict[max_item],
                        limit_item=limit_item, limit_value=self.csv_dict[limit_item]
                    ))

        return _csv_dict

    def get_insert_dict_by_timestamp(self, ts, server_id):
        insert_dict = self._get_table_params_dict()
        items_info = list(zip(FIELD_AVERAGE, FIELD_LIMIT, FIELD_FAULT, FIELD_MAX))

        for avg_item, limit_item, fault_item, max_item in items_info:
            avg_col = FIELD_TO_TABLE_COLUMN[avg_item]
            limit_col = FIELD_TO_TABLE_COLUMN[limit_item]
            fault_col = FIELD_TO_TABLE_COLUMN[fault_item]
            if self._control_avg_sum[avg_item] <= 0 and self._control_fault_sum[fault_item] <= 0:
                insert_dict[limit_col] = self[limit_item]
                continue
            # algorithm generate average columns
            avg_random_max = min(self.get(max_item) or self[limit_item],
                                 self._control_avg_sum[avg_item])
            if self[limit_item]:
                avg_random = self._control_avg_sum[avg_item] - \
                             self.get(max_item, self[limit_item]) * (self._unprocessed_rows - 1)
                avg_random_min = max(0, avg_random)
            else:
                avg_random_min = 0
            avg_data = random.randrange(avg_random_min, avg_random_max + 1)

            # max must be present
            if self.get(max_item) is not None and not self[fault_item] and (
                    self._unprocessed_rows == self._table_rows):
                avg_data = self[max_item]

            # algorithm generate fault columns
            fault_random_max = max(0, int(self._control_fault_sum[fault_item]))
            if self._unprocessed_rows == 1 and int(self._control_fault_sum[fault_item]) > 0:
                fault_random_min = int(self._control_fault_sum[fault_item])
            else:
                fault_random_min = 0
            fault_data = random.randrange(fault_random_min, fault_random_max // self._unprocessed_rows + 1)

            # for controlling columns sum
            self._control_avg_sum[avg_item] -= avg_data
            self._control_fault_sum[fault_item] -= fault_data
            insert_dict[avg_col] = avg_data
            insert_dict[limit_col] = self[limit_item]
            insert_dict[fault_col] = fault_data

        self._unprocessed_rows -= 1  # controlling tables rows for specific data case
        if any(insert_dict.values()):  # don't write data if all is zero
            insert_dict['created'] = ts
            insert_dict['id'] = self['ID']
            insert_dict['server_id'] = server_id
            return insert_dict

    @classmethod
    def get_service_alive_db_record(cls, ts, server_id):
        # add row with id=0 for write keep alive data
        insert_dict = {col: 0 for col in list(cls.table.__table__.c.keys())}
        insert_dict['created'] = ts
        insert_dict['server_id'] = server_id
        return insert_dict


class DbGovDataCase(BaseDataCase):
    table = history_gov

    FIELD_TO_TABLE_COLUMN = {
        'CPU': 'sum_cpu',
        'WRITE': 'sum_write',
        'READ': 'sum_read',
        'lCPU': 'limit_cpu_on_period_end',
        'lREAD': 'limit_read_on_period_end',
        'lWRITE': 'limit_write_on_period_end',
        'RESTRICT': 'cause_of_restrict'
    }

    AVG_FIELDS = ['CPU', 'READ', 'WRITE']

    def __init__(self, csv_dict):
        super(DbGovDataCase, self).__init__(csv_dict)

        self.control_avg_sum = {item: int(self._csv_dict[item]) * self._table_rows for item in self.AVG_FIELDS}

    def _normalize_and_validate_data(self):
        _csv_dict = dict()

        for item, val in list(self.csv_dict.items()):

            # convert string to timestamp
            if item in ('From', 'To'):
                _csv_dict[item] = self._str_to_timestamp(val)
            # convert all variables to integer (all columns in table history is integer)
            else:
                _csv_dict[item] = int(val)

        return _csv_dict

    def get_insert_dict_by_timestamp(self, ts, server_id):
        insert_dict = self._get_table_params_dict()
        items_info = list(zip(['CPU', 'READ', 'WRITE'], ['lCPU', 'lREAD', 'lWRITE']))

        for avg_item, limit_item in items_info:
            avg_col = self.FIELD_TO_TABLE_COLUMN[avg_item]
            limit_col = self.FIELD_TO_TABLE_COLUMN[limit_item]

            # algorithm generate average columns
            avg_random_max = min(self[limit_item],
                                 self.control_avg_sum[avg_item])
            if self[limit_item]:
                avg_random = self.control_avg_sum[avg_item] - \
                             self[limit_item] * (self._unprocessed_rows - 1)
                avg_random_min = max(0, avg_random)
            else:
                avg_random_min = 0
            avg_data = random.randrange(avg_random_min, avg_random_max + 1)

            # for controlling columns sum
            self.control_avg_sum[avg_item] -= avg_data
            insert_dict[avg_col] = avg_data
            insert_dict[limit_col] = self[limit_item]

        self._unprocessed_rows -= 1  # controlling tables rows for specific data case
        if any(insert_dict.values()):  # don't write data if all is zero
            insert_dict['ts'] = ts
            insert_dict['uid'] = self['ID']
            insert_dict['server_id'] = server_id
            insert_dict[self.FIELD_TO_TABLE_COLUMN['RESTRICT']] = self['RESTRICT']
            return insert_dict


class ValidateError(ValueError):
    pass


class DatabaseGeneratorError(Exception):
    pass


class DatabaseGenerator(object):

    def __init__(self, db_uri, server_id='localhost', lve_version=8, progress=None,
                 lveps_data_path=None, dbgov_data_path=None, lve_data_path=None):
        self.db_uri = db_uri
        self.server_id = server_id
        self.lve_version = lve_version
        self.dbengine = None
        self.progress = progress or self.progress_callback
        self.lve_data_path = lve_data_path
        self.lveps_data_path = lveps_data_path
        self.dbgov_data_path = dbgov_data_path

    def configure_database(self):
        """
        Create and/or validate database
        """
        db_uri_parsed = urlparse(self.db_uri)
        conf = {'server_id': self.server_id, 'db_type': db_uri_parsed.scheme}
        if db_uri_parsed.scheme == 'sqlite':
            conf.update({'sqlite_db_path': db_uri_parsed.path + db_uri_parsed.netloc})
        else:
            conf.update({'connect_string': db_uri_parsed.netloc + db_uri_parsed.path})

        self.dbengine = dbengine.make_db_engine(conf, check_support=False)
        dbengine.create_db(self.dbengine)

    def _load_csv(self, csv_path):
        # type: (str) -> Generator[Dict]
        """Load test data from csv file"""
        with open(csv_path) as csv_stream:
            for dict_ in csv.DictReader(csv_stream):
                yield dict_

    def progress_callback(self, *args, **kwargs):
        """
        Callback function for output progress
        """
        pass

    def generate_lve_version(self, conn_):
        # add lve version to database
        try:
            conn_.execute(servers.__table__.insert(), {'server_id': self.server_id, "lve_version": self.lve_version})
        except IntegrityError:
            pass

    def generate_snapshots(self, conn_):
        if self.lveps_data_path is None:
            return
        # read lve data from given lve_data_path
        data_list = self._load_csv(self.lve_data_path)

        # reading snapshots plugin configuration
        cfg = config.read_config()
        snapshots_configuration = config.read_plugin_config(cfg, plugin='SnapshotSaver')
        period_between_incidents = int(snapshots_configuration['period_between_incidents'])
        snapshots_per_minute = int(snapshots_configuration['snapshots_per_minute'])
        max_snapshots_per_incident = int(snapshots_configuration['max_snapshots_per_incident'])

        # parsing data for snapshots
        with open(self.lveps_data_path) as lveps_out_file:
            snapshots_lveps_output = ''.join(lveps_out_file.readlines())
        lveps_data = ustate._pars_lveps(snapshots_lveps_output)

        # generating incidents
        incidents = []
        incidents_to_data_cases = dict()
        for data_case in data_list:
            faults_count = sum([int(data_case.csv_dict[key]) for key in FIELD_FAULT])
            if faults_count == 0:
                continue

            common_info = {'uid': data_case['ID'], 'server_id': self.server_id}
            i = data_case.ts_from
            while i < data_case.ts_to:
                incident_length = random.randrange(*INCIDENT_LENGTH_BOUNDARIES)
                incident_timeframe = (i, i + incident_length)
                i += incident_length + period_between_incidents
                incident_info = dict(
                    incident_start_time=incident_timeframe[0],
                    incident_end_time=incident_timeframe[1]
                )
                incident_info.update(common_info)
                incidents.append(incident_info)
                incidents_to_data_cases[frozenset(list(incident_info.items()))] = data_case

        # generating snapshots
        snapshots = []
        for incident_info in incidents:
            snapshots_per_incident = []
            for snapshot_time in range(incident_info['incident_start_time'],
                                       incident_info['incident_end_time'],
                                       60 // snapshots_per_minute):
                data_case = incidents_to_data_cases[frozenset(list(incident_info.items()))]
                snapshot_ = {
                    'incident_start_time': incident_info['incident_start_time'],
                    'dump_time': snapshot_time,
                    'uid': incident_info['uid'],
                    'server_id': incident_info['server_id']
                }
                processes_snapshot_for_user = lveps_data[int(incident_info['uid'])]

                processes_snapshot = {}
                for k, v in list(processes_snapshot_for_user['TID'].items()):
                    snap_proc = dict(CPU=v["CPU"], MEM=v["MEM"], CMD=v['CMD'], PID=k)
                    processes_snapshot[k] = snap_proc

                snapshot_['snap_proc'] = json.dumps(processes_snapshot)
                snapshot_['snap_sql'] = '""'
                snapshot_['snap_http'] = '[]'
                untranslated_faults = dict([(key, int(data_case.csv_dict[key])) for key in FIELD_FAULT])

                # translating aCPU, CPUf... to cpu_usage, cpu_faults
                faults = dict([(FIELD_TO_TABLE_COLUMN[k], v)
                               for k, v in list(untranslated_faults.items())
                               if not (v == 0 or str(v).startswith('0'))])
                snapshot_['snap_faults'] = json.dumps(faults)
                snapshots_per_incident.append(snapshot_)

                if len(snapshots_per_incident) > max_snapshots_per_incident:
                    break
            snapshots.extend(snapshots_per_incident)

        for incident_info in incidents:
            conn_.execute(incident.__table__.insert(), incident_info)

        # TODO: Lets not actually save snapshots, as they would go somewhere else on disk, not in a test db file

    def _generate_history(self, conn_, data_list, data_class):
        items = [data_class(item) for item in data_list]
        ts_from, ts_to = self._get_from_to_and_table_dict(items)

        if self.progress:
            self.progress(0)  # start progress 0%

        insert_list = []
        for ts_ in range(ts_from, ts_to, BaseDataCase.AGGREGATION_PERIOD):
            ts = random.randrange(ts_, ts_ + BaseDataCase.AGGREGATION_PERIOD - 1)  # randomise timestamp

            # insert keep alive row if data_class needs it
            keep_alive = data_class.get_service_alive_db_record(ts, self.server_id)
            if keep_alive:
                insert_list.append(keep_alive)

            data_cases = self._iter_datacase_by_timestamp(items, ts)
            for data_case in data_cases:
                item = data_case.get_insert_dict_by_timestamp(ts, self.server_id)
                if item:
                    insert_list.append(item)
            if self.progress:
                percentage = (1 - float(ts - ts_from) / (ts_to - ts_from)) * 100
                self.progress(percentage)

            # saving ALL dict's to insert_list causes a huge memory usage
            if len(insert_list) > 250:
                conn_.execute(data_class.table.__table__.insert(), insert_list)
                insert_list = []
        if insert_list:
            conn_.execute(data_class.table.__table__.insert(), insert_list)

        if self.progress and __name__ == '__main__':
            self.progress(100, force=True)
            if hasattr(self.progress, 'stop'):
                self.progress.stop()

    def generate_history(self, conn_):
        data_list = self._load_csv(self.lve_data_path)
        self._generate_history(conn_, data_list, DataCase)

    def generate_dbgov_history(self, conn_):
        if self.dbgov_data_path is None:
            return
        data_list = self._load_csv(self.dbgov_data_path)
        self._generate_history(conn_, data_list, DbGovDataCase)

    @staticmethod
    def _get_from_to_and_table_dict(data_items):
        # type: (Iterable[BaseDataCase]) -> Tuple[int, int]
        ts_from = min(data_.ts_from for data_ in data_items)
        ts_to = max(data_.ts_to for data_ in data_items)
        return ts_from, ts_to

    @staticmethod
    def _iter_datacase_by_timestamp(data_list, ts):
        data_case_list = [data_case for data_case in data_list if
                          data_case.ts_from <= ts < data_case.ts_to and
                          not data_case.is_empty()]
        return data_case_list

    def generate_database(self):
        if self.dbengine is None:  # check if database initialised
            raise DatabaseGeneratorError('You must first call configure_database method to (re)initial database')
        conn_ = self.dbengine.connect()
        trans = conn_.begin()  # generate database in one transaction
        try:
            self.generate_lve_version(conn_)
            self.generate_history(conn_)
            self.generate_dbgov_history(conn_)
            self.generate_snapshots(conn_)
        except OperationalError as e:
            trans.rollback()
            sys.exit(' Error!, Data was rollback: {}...'.format(str(e)[:255]))

        trans.commit()
        conn_.close()


# functions for running DatabaseGenerator from console
def setup_parser():
    parser = argparse.ArgumentParser(
        description="%(prog)s - Utility to generate valid test database for lve-stats >= 2.1-15",
        epilog="Run %(prog)s",
        prog=os.path.basename(__file__).replace('.pyc', '.py'))
    parser.add_argument(
        'csv', type=str,
        help='Input csv file, for generating lve history database')
    parser.add_argument(
        'dburi', type=str,
        help='uri to database; for example sqlite:/var/lve/test_lvestats2.db')
    parser.add_argument(
        '--lve-version', type=int, default=8, choices=[8, 6],
        help='LVE version for server_id; default "%(default)s"')
    parser.add_argument(
        '--server-id', '--server_id', type=str, default='localhost',
        help='Server id which data generated; default "%(default)s"')
    parser.add_argument(
        '--progress', action='store_true', default=False,
        help='Show database generation progress interactively')
    parser.add_argument(
        '--lveps-data-path', type=str, default=None,
        help='Path to file with data for snapshots (lveps output)')
    parser.add_argument(
        '--dbgov-data-path', type=str, default=None,
        help='Input csv file, for generating governor history database')
    return parser


def main(cnf, argv_=None):
    if argv_ is None:
        argv_ = sys.argv[1:]
    parser = setup_parser()
    namespace = parser.parse_args(argv_)

    if namespace.progress:
        progress = ProgressCallback(time_change_step=0.1)
    else:
        progress = None

    # check if input file present
    if not os.path.exists(namespace.csv):
        sys.exit('Can\'t find csv file "{}"'.format(namespace.csv))

    BaseDataCase.AGGREGATION_PERIOD = cnf.get('aggregation_period', 60)

    database_generator = DatabaseGenerator(
        db_uri=namespace.dburi,
        server_id=namespace.server_id,
        lve_version=namespace.lve_version,
        progress=progress,
        lveps_data_path=namespace.lveps_data_path,
        dbgov_data_path=namespace.dbgov_data_path,
        lve_data_path=namespace.csv)

    try:
        database_generator.configure_database()
        database_generator.generate_database()
    except KeyboardInterrupt:
        sys.exit('\nWarning: data generation is stopped by the user; all the generated data written to database')
    except dbengine.MakeDbException as e:
        sys.exit('\n{}'.format(str(e)))


if __name__ == '__main__':
    try:
        cnf = config.read_config()
        main(cnf)
    except config.ConfigError as ce:
        ce.log_and_exit()
