# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import os
import sys
import time
import io
import logging
from contextlib import contextmanager
from subprocess import check_output, STDOUT, CalledProcessError
from pwd import getpwnam


@contextmanager
def capture_all(command, *args, **kwargs):
    log_disable_backup = logging.root.manager.disable
    logging.disable(logging.WARNING)

    old_stdout, old_stderr = sys.stdout, sys.stderr
    sys.stdout = mock_stdout = io.StringIO()
    sys.stderr = mock_stderr = io.StringIO()

    root_logger = logging.getLogger()
    log_handler = logging.StreamHandler(mock_stdout)
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    log_handler.setFormatter(formatter)
    root_logger.addHandler(log_handler)

    result = command(*args, **kwargs)

    sys.stdout, sys.stderr = old_stdout, old_stderr
    logging.disable(log_disable_backup)
    root_logger.removeHandler(log_handler)
    mock_stdout.flush()
    yield {"result": result, "stdout": mock_stdout.getvalue(), "stderr": mock_stderr.getvalue()}


class DummyFile(object):
    def write(self, x):
        pass


@contextmanager
def nostdout():
    save_stdout = sys.stdout
    sys.stdout = DummyFile()
    yield
    sys.stdout = save_stdout


class temporary_user(object):
    def __init__(self, username):
        self.username = username

    def __enter__(self):
        try:
            check_output(['/usr/sbin/useradd', self.username], stderr=STDOUT)
        except CalledProcessError as e:
            if e.returncode == 9:
                # TODO what if user exists?
                pass
            else:
                raise e
        os.seteuid(getpwnam(self.username).pw_uid)

    def __exit__(self, type, value, traceback):
        os.seteuid(os.getuid())
        check_output(['/usr/sbin/userdel', self.username], stderr=STDOUT)


class change_timezone(object):
    """Temporary change time zone for testing"""
    def __init__(self, timezone='Europe/London'):
        self.timezone = timezone
        self._original_tz = None

    def __enter__(self):
        self._original_tz = os.environ.get('TZ')
        os.environ['TZ'] = self.timezone
        time.tzset()  # change timezone for 'time' module

    def __exit__(self, type, value, traceback):
        if self._original_tz is None:
            del os.environ['TZ']
        else:
            os.environ['TZ'] = self._original_tz
        time.tzset()


class import_fail(object):
    """Mock failed import"""
    def __init__(self, package_name=None):
        self.package = package_name
        self.realimport = globals()['__builtins__']['__import__']

    def __enter__(self):
        # noinspection PyDefaultArgument
        # pylint:disable=dangerous-default-value
        def mock_import(name, globals={}, locals={}, fromlist=[], level=0):
            if self.package in name:
                raise ImportError
            return self.realimport(name, globals, locals, fromlist, level)
        globals()['__builtins__']['__import__'] = mock_import

    def __exit__(self, exc_type, exc_val, exc_tb):
        globals()['__builtins__']['__import__'] = self.realimport
