# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import ddt
import mock
import unittest
from sqlalchemy import Column, String, Integer, Index, Float, Table, MetaData

from lvestats.lib import uidconverter
from lvestats.lib.commons.func import merge_dicts
from lvestats.utils.dbmigrator import alembic_migrate
from test.base.base_db_test import BaseDbTest

HISTORY_GOV_TABLE_NAME = 'lve_stats2_history_gov'


old_table = Table(
    HISTORY_GOV_TABLE_NAME, MetaData(),
    Column('ts', Integer, index=True, primary_key=True, autoincrement=False),
    Column('username', String(64), primary_key=True, autoincrement=False),
    Column('sum_cpu', Float),
    Column('sum_write', Float),
    Column('sum_read', Float),
    Column('limit_cpu_on_period_end', Integer),
    Column('limit_read_on_period_end', Integer),
    Column('limit_write_on_period_end', Integer),
    Column('cause_of_restrict', Integer),
    Column('server_id', String(255), primary_key=True, autoincrement=False)
)


new_table = Table(
    HISTORY_GOV_TABLE_NAME, MetaData(),
    Column('ts', Integer, primary_key=True, autoincrement=False),
    Index('ix_{}_ts_2'.format(HISTORY_GOV_TABLE_NAME), 'ts'),
    Column('uid', Integer, primary_key=True, autoincrement=False),
    Column('sum_cpu', Float),
    Column('sum_write', Float),
    Column('sum_read', Float),
    Column('limit_cpu_on_period_end', Integer),
    Column('limit_read_on_period_end', Integer),
    Column('limit_write_on_period_end', Integer),
    Column('cause_of_restrict', Integer),
    Column('server_id', String(255), primary_key=True, autoincrement=False)
)

test_case_default_data = {
    u'sum_cpu': 7.77, u'sum_write': 6.66, u'sum_read': 1.11,
    u'limit_cpu_on_period_end': 3, u'limit_read_on_period_end': 4,
    u'limit_write_on_period_end': 5, u'cause_of_restrict': 1}

test_case_upgrade = (
    ([{u'server_id': u'localhost', u'ts': 11111, u'username': u'wordpress'}],
     [{u'server_id': u'localhost', u'ts': 11111, u'uid':        777}]),

    ([{u'server_id': u'localhost', u'ts': 11111, u'username': u'wordpress'},
      {u'server_id': u'localhost', u'ts': 22222, u'username': u'wordpress'}],
     [{u'server_id': u'localhost', u'ts': 11111, u'uid':        777},
      {u'server_id': u'localhost', u'ts': 22222, u'uid':        777}]),

    ([{u'server_id': u'localhost', u'ts': 11111, u'username': u'wordpress'},
      {u'server_id': u'localhost', u'ts': 22222, u'username': u'wordpress2'}],
     [{u'server_id': u'localhost', u'ts': 11111, u'uid':        777},
      {u'server_id': u'localhost', u'ts': 22222, u'uid':        666}]),

    ([{u'server_id': u'localhost', u'ts': 11111, u'username': u'wordpress'},
      {u'server_id': u'localhost', u'ts': 11111, u'username': u'root'}],
     [{u'server_id': u'localhost', u'ts': 11111, u'uid':        777}]),

    ([{u'server_id': u'localhost', u'ts': 11111, u'username': u'no_such_user'},
      {u'server_id': u'localhost', u'ts': 11111, u'username': u'root'}],
     [])
)

test_case_downgrade = (
    ([{u'server_id': u'localhost', u'ts': 11111, u'uid':       777}],
     [{u'server_id': u'localhost', u'ts': 11111, u'username': u'wordpress'}]),

    ([{u'server_id': u'localhost', u'ts': 11111, u'uid':        777},
      {u'server_id': u'localhost', u'ts': 22222, u'uid':        777}],
     [{u'server_id': u'localhost', u'ts': 11111, u'username': u'wordpress'},
      {u'server_id': u'localhost', u'ts': 22222, u'username': u'wordpress'}]),

    ([{u'server_id': u'localhost', u'ts': 11111, u'uid':        777},
      {u'server_id': u'localhost', u'ts': 22222, u'uid':        666}],
     [{u'server_id': u'localhost', u'ts': 11111, u'username': u'wordpress'},
      {u'server_id': u'localhost', u'ts': 22222, u'username': u'wordpress2'}]),

    ([{u'server_id': u'localhost', u'ts': 11111, u'uid':        777},
      {u'server_id': u'localhost', u'ts': 11111, u'uid':        0}],
     [{u'server_id': u'localhost', u'ts': 11111, u'username': u'wordpress'}]),

    ([{u'server_id': u'localhost', u'ts': 11111, u'uid':        888},
      {u'server_id': u'localhost', u'ts': 11111, u'uid':        0}],
     [])
)


def mocked_username_to_uid(username, local_server_id, server_id, db_engine):
    username_to_uid_dict = {'wordpress': 777, 'wordpress2': 666}
    return username_to_uid_dict.get(username)


def mocked_uid_to_username(uid, local_server_id, server_id, db_engine):
    uid_to_username_dict = {777: 'wordpress', 666: 'wordpress2'}
    return uid_to_username_dict.get(uid)


@ddt.ddt
class TestDBMigrator(BaseDbTest):
    @classmethod
    def setUpClass(cls):
        BaseDbTest.setUpClass()

    @classmethod
    def tearDownClass(cls):
        cls.recreate_tables()
        BaseDbTest.tearDownClass()

    @ddt.data(*test_case_upgrade)
    @ddt.unpack
    @mock.patch.object(uidconverter, 'username_to_uid', mocked_username_to_uid)
    def test_upgrade_15fed017ca0b_add_uid_column(self, insert_data, output):
        # extend test case data
        insert_data = [merge_dicts(d, test_case_default_data) for d in insert_data]
        output = [merge_dicts(d, test_case_default_data) for d in output]

        alembic_migrate(self.engine, alembic_version='48efb1674139',
                        lve_stats_cfg=self.cfg)

        self.engine.execute(old_table.insert(insert_data))
        alembic_migrate(self.engine, alembic_version='15fed017ca0b',
                        lve_stats_cfg=self.cfg)
        output_dict = list(map(dict, self.engine.execute(new_table.select())))
        self.assertListEqual(output, output_dict)

    @ddt.data(*test_case_downgrade)
    @ddt.unpack
    @mock.patch.object(uidconverter, 'uid_to_username', mocked_uid_to_username)
    def test_downgrade_15fed017ca0b_add_uid_column(self, insert_data, output):
        # extend test case data
        insert_data = [merge_dicts(d, test_case_default_data) for d in insert_data]
        output = [merge_dicts(d, test_case_default_data) for d in output]

        alembic_migrate(self.engine, alembic_version='15fed017ca0b',
                        lve_stats_cfg=self.cfg)
        self.engine.execute(new_table.insert(insert_data))
        alembic_migrate(self.engine, alembic_version='48efb1674139',
                        lve_stats_cfg=self.cfg)
        output_dict = list(map(dict, self.engine.execute(old_table.select())))
        self.assertListEqual(output, output_dict)
