# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import os
import shutil
import unittest

from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.pool import NullPool

from test.base.base_db_test import BaseDbTest
from lvestats.lib.dbengine import make_db_engine
from lvestats.utils.dbmigrator import alembic_migrate

OLD_ALEMBIC_VERSION = '15fed017ca0b'
NEW_ALEMBIC_VERSION = '54d3ae3043dd'
DATA_DIR = os.path.join(os.path.dirname(__file__), 'data/alembic')
ORIG_DB = os.path.join(DATA_DIR, OLD_ALEMBIC_VERSION + '.db')
NEW_DB = os.path.join(DATA_DIR, NEW_ALEMBIC_VERSION + '.db')
TEST_DB = os.path.join(DATA_DIR, 'test.db')

cfg_sqlite = {
    'db_type': 'sqlite',
    'sqlite_db_path': TEST_DB,
    'connect_string': 'root:root@localhost',
    'server_id': 'localhost',
    'debug': False
}

g_engine = BaseDbTest.get_superglobal_engine()
sqlite_on_disk = g_engine and str(g_engine.url) != 'sqlite:///' and str(g_engine.url).startswith('sqlite:///')


@unittest.skipUnless(sqlite_on_disk, 'Run test for sqlite on disk only')
class TestDBMigrator(unittest.TestCase):

    def setUp(self):
        shutil.copy(ORIG_DB, TEST_DB)
        self.dbengine = make_db_engine(cfg=cfg_sqlite, pool_class=NullPool)

    def tearDown(self):
        os.remove(TEST_DB)

    def test_upgrade(self):
        alembic_migrate(self.dbengine, alembic_version=NEW_ALEMBIC_VERSION, lve_stats_cfg=cfg_sqlite)

        inspector = Inspector.from_engine(self.dbengine)
        self.assertNotIn('lve_stats2_snapshot', inspector.get_table_names())
        incident_columns = [x['name'] for x in inspector.get_columns('lve_stats2_incident')]
        self.assertIn('dump_time', incident_columns)
        self.assertIn('snapshot_count', incident_columns)

    def test_downgrade(self):
        alembic_migrate(self.dbengine, alembic_version=OLD_ALEMBIC_VERSION, lve_stats_cfg=cfg_sqlite)
        inspector = Inspector.from_engine(self.dbengine)
        self.assertIn('lve_stats2_snapshot', inspector.get_table_names())
        incident_columns = [x['name'] for x in inspector.get_columns('lve_stats2_incident')]
        self.assertNotIn('dump_time', incident_columns)
        self.assertNotIn('snapshot_count', incident_columns)

    def test_upgrade_downgrade_upgrade(self):
        alembic_migrate(self.dbengine, alembic_version=NEW_ALEMBIC_VERSION, lve_stats_cfg=cfg_sqlite)
        alembic_migrate(self.dbengine, alembic_version=OLD_ALEMBIC_VERSION, lve_stats_cfg=cfg_sqlite)
        alembic_migrate(self.dbengine, alembic_version=NEW_ALEMBIC_VERSION, lve_stats_cfg=cfg_sqlite)
        inspector = Inspector.from_engine(self.dbengine)
        self.assertNotIn('lve_stats2_snapshot', inspector.get_table_names())
        incident_columns = [x['name'] for x in inspector.get_columns('lve_stats2_incident')]
        self.assertIn('dump_time', incident_columns)
        self.assertIn('snapshot_count', incident_columns)
