# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
from time import tzset

import datetime
import json
import csv
import os
import unittest
from collections import deque, defaultdict
from sqlalchemy import insert
from lvestats.lib.commons.func import get_chunks
import ddt
import mock
from clcommon.clproc import LIMIT_LVP_ID
from sys import getsizeof, stderr
from itertools import chain
from clcommon.lib import MySQLGovException

from test.base.base_db_test import BaseDbTest
from test.helpers import change_timezone
from lvestats.lib.commons import dateutil
from lvestats.lib.commons.func import serialize_lve_id
from lvestats.orm import history, history_gov
from lvestats.lib.parsers.lveinfoargparse import time_unit_type
from lvestats.lib.parsers.cloudlinux_statistics_parser import setup_statistics_parser
from lvestats.lib import cloudlinux_statistics


_USERMAP = {503: 'user503', 501: 'user501', 1000000503: 'reseller503'}
_INV_USERMAP = {v: k for k, v in _USERMAP.items()}
RESELLER_ID = 1000000503


# noinspection PyUnusedParameter
def mock_uid_to_username(uid, local_server_id, server_id, db_engine):
    if uid in _USERMAP:
        return _USERMAP[uid]
    return None


# noinspection PyUnusedParameter
def mock_username_to_uid_local(username):
    if username in list(_USERMAP.values()):
        return _INV_USERMAP[username]
    return None


# noinspection PyDefaultArgument
# pylint:disable=dangerous-default-value
def total_size(o, handlers={}, verbose=False):
    """ Returns the approximate memory footprint an object and all of its contents.

    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, deque, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:

        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}

    """
    def dict_handler(d):
        return chain.from_iterable(list(d.items()))
    all_handlers = {
        cloudlinux_statistics.LVEList: lambda l: (l.stat._lve_results, l.stat._gov_results, l.stat._resellers_results),
        tuple: iter,
        list: iter,
        deque: iter,
        dict: dict_handler,
        set: iter,
        frozenset: iter}
    all_handlers.update(handlers)     # user handlers take precedence
    seen = set()                      # track which object id's have already been seen
    default_size = getsizeof(0)       # estimate sizeof object without __sizeof__

    def sizeof(o):
        if id(o) in seen:       # do not double count the same object
            return 0
        seen.add(id(o))
        s = getsizeof(o, default_size)

        if verbose:
            print(s, type(o), repr(o), file=stderr)

        for typ, handler in list(all_handlers.items()):
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)


def check_timestamps_equal_or_none(lve_list, gov_list):
    """
    elements of lve_results and gov_results are either None or lists with timestamps and stats
    in lve_results elements from and to have indexes 1 and 2
    in gov_results elements from and to have indexes 7 and 8
    """
    result = True
    for i, pair in enumerate(zip(lve_list, gov_list)):
        if not(pair[0] is None or pair[1] is None
                or (pair[0][1] == pair[1][7] and pair[0][2] == pair[1][8])):
            print('equal timestamps or none check failed at index {} with values {}'.format(i, pair))
            result = False
    return result


@ddt.ddt
class TestCloudLinuxStatistics(BaseDbTest):
    HIDE_MAX_UIDS = 'true'

    def setupGov(self):
        for time in range(0, 10 * 60, 60):
            self.save_dbgov_line(self.server_id, uid=501, _time=time, sum_cpu=1,
                                 limit_cpu_on_period_end=5, limit_read_on_period_end=20,
                                 limit_write_on_period_end=30,
                                 sum_read=300, sum_write=200)
            self.save_dbgov_line(self.server_id, uid=503, _time=time, sum_cpu=2, sum_read=30, sum_write=40,
                                 limit_write_on_period_end=100, limit_read_on_period_end=100)

    def setupGov_from_csv(self, filename):
        with open(filename) as f:
            # csv should contain result of 'select * from lve_stats2_history_gov' in csv format
            dataset = list(csv.reader(f))
        for row in dataset:
            # row[8] is cause_of_restrict
            self.save_dbgov_line(_time=row[0], uid=row[1], sum_cpu=row[2], sum_write=row[3], sum_read=row[4],
                                 limit_cpu_on_period_end=row[5], limit_read_on_period_end=row[6],
                                 limit_write_on_period_end=row[7], server_id=row[9])

    def setupLVE(self, with_reseller=False):
        for time in range(0, 10 * 60, 60):
            self.save_history_line(self.server_id, uid=0, _time=time, cpu=100)
            self.save_history_line(
                self.server_id, uid=501, _time=time,
                nproc=8, lnproc=9, nproc_fault=2,
                io=10, io_limit=19, io_fault=25,
                memphy=5, lmemphy=21, memphy_fault=8,
                mep=4, mep_limit=18, mep_fault=12,
                mem=23, mem_limit=20, mem_fault=10,
                cpu=2, cpu_limit=17, cpu_fault=1000,
                iops=8, liops=22, iops_fault=16)
            self.save_history_line(
                self.server_id, uid=503, _time=time,
                iops=8, liops=22, iops_fault=16,
                io=6, io_limit=19, io_fault=15,
                memphy=5, lmemphy=21, memphy_fault=11,
                mep=4, mep_limit=18, mep_fault=12,
                cpu=160000, cpu_limit=200000, cpu_fault=14,
                lnproc=9, nproc=7, nproc_fault=13,
                mem=3, mem_limit=20, mem_fault=10,)
            self.save_history_line(
                self.server_id, uid=self.max_uid_lve_id, _time=time,
                iops=8, liops=22, iops_fault=16,
                io=6, io_limit=19, io_fault=15,
                memphy=5, lmemphy=21, memphy_fault=11,
                mep=4, mep_limit=18, mep_fault=12,
                cpu=100, cpu_limit=200000, cpu_fault=14,
                lnproc=9, nproc=7, nproc_fault=13,
                mem=3, mem_limit=20, mem_fault=10, )
            if with_reseller:
                self.save_history_line(
                    self.server_id, uid=serialize_lve_id(LIMIT_LVP_ID, 503), _time=time,
                    iops=8, liops=22, iops_fault=16,
                    cpu=160000, cpu_limit=200000, cpu_fault=14,
                    io=6, io_limit=19, io_fault=15,
                    memphy=5, lmemphy=21, memphy_fault=11,
                    mep=4, mep_limit=18, mep_fault=12,
                    nproc=7, lnproc=9, nproc_fault=13,
                    mem=3, mem_limit=20, mem_fault=10)

    def setupLVE_from_csv(self, filename):
        with open(filename) as f:
            # csv should contain result of 'select * from lve_stats2_history' in csv format
            dataset = list(csv.reader(f))
        for row in dataset:
            self.save_history_line(keepalive=True,
                                    uid=row[0], cpu=row[1], cpu_limit=row[2], cpu_fault=row[3],
                                    mep=row[4], mep_limit=row[5], io=row[6], io_limit=row[7], io_fault=row[8],
                                    mem=row[9], mem_limit=row[10], mem_fault=row[11], mep_fault=row[12],
                                    _time=row[13], server_id=row[14], lmemphy=row[15], memphy=row[16], memphy_fault=row[17],
                                    lnproc=row[18], nproc=row[19], nproc_fault=row[20],
                                    liops=row[21], iops=row[22], iops_fault=row[23])

    def setUp(self):
        self.server_id = self.cfg['server_id']
        self.max_uid_lve_id = 60001
        BaseDbTest.setUp(self)
        os.environ['TZ'] = 'UTC'
        tzset()
        self.mockSetup()

    def tearDown(self):
        BaseDbTest.tearDown(self)

        del os.environ['TZ']
        tzset()

    def mockSetup(self):
        patcher1 = mock.patch('lvestats.lib.cloudlinux_statistics.uid_to_username', side_effect=mock_uid_to_username)
        patcher1.start()
        self.addCleanup(patcher1.stop)
        patcher3 = mock.patch('lvestats.lib.commons.func.get_governor_status', return_value=('success', None))
        patcher3.start()
        self.addCleanup(patcher3.stop)
        patcher4 = mock.patch('lvestats.lib.cloudlinux_statistics.getuser', return_value='root')  # for build-system mock environment
        patcher4.start()
        self.addCleanup(patcher4.stop)
        patcher5 = mock.patch('clcommon.lib.MySQLGovernor.get_governor_status_by_username', return_value='watched')
        patcher5.start()
        self.addCleanup(patcher5.stop)
        patcher7 = mock.patch('lvestats.lib.cloudlinux_statistics.username_to_uid_local', side_effect=mock_username_to_uid_local)
        patcher7.start()
        self.addCleanup(patcher7.stop)
        patcher8 = mock.patch('lvestats.lib.config.read_config', return_value={})
        patcher8.start()
        self.addCleanup(patcher8.stop)

    def test_generate_statistics_no_lve(self):
        self.setupGov()
        options = setup_statistics_parser(["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json"], self.server_id)
        with mock.patch.object(cloudlinux_statistics.CloudLinuxStatistics, "_show_mysql", return_value=True):
            result = cloudlinux_statistics.get_statistics(
                engine=self.engine,
                id_=None,
                options=options,
                timestamp=15,
                log=None)
        self.maxDiff = None
        self.assertEqual({
            'timestamp': 15, 'mySqlGov': 'success', 'resellers': [], 'result': 'success', 'users': [
                {'reseller': 'N/A', 'username': 'user501', 'domain': 'N/A', 'id': 501,
                 'faults': {'ep': {'lve': 0},  'vmem': {'lve': 0},    'iops': {'lve': 0},   'io': {'lve': 0},
                            'nproc': {'lve': 0}, 'cpu': {'lve': 0},               'pmem': {'lve': 0}},
                 'limits': {'ep': {'lve': 0.0}, 'vmem': {'lve': 0.0}, 'iops': {'lve': 0.0},
                            'io': {'lve': 0.0, 'mysql': 50.0 * 1024 * 1024}, 'nproc': {'lve': 0.0}, 'cpu': {'lve': 0.0, 'mysql': 5.0},
                            'pmem': {'lve': 0.0}},
                 'usage':  {'ep': {'lve': 0.0}, 'vmem': {'lve': 0.0}, 'iops': {'lve': 0.0},
                            'io': {'lve': 0.0, 'mysql': 500 * 1024 * 1024}, 'nproc': {'lve': 0.0}, 'cpu': {'lve': 0.0, 'mysql': 1},
                            'pmem': {'lve': 0.0}}},
                {'reseller': 'N/A', 'username': 'user503', 'domain': 'N/A', 'id': 503,
                 'faults': {'ep': {'lve': 0},   'vmem': {'lve': 0},   'iops': {'lve': 0},   'io': {'lve': 0},
                            'nproc': {'lve': 0}, 'cpu': {'lve': 0},               'pmem': {'lve': 0}},
                 'limits': {'ep': {'lve': 0.0}, 'vmem': {'lve': 0.0}, 'iops': {'lve': 0.0},
                            'io': {'lve': 0.0, 'mysql': 200.0 * 1024 * 1024}, 'nproc': {'lve': 0.0}, 'cpu': {'lve': 0.0, 'mysql': 0.0},
                            'pmem': {'lve': 0.0}},
                 'usage':  {'ep': {'lve': 0.0}, 'vmem': {'lve': 0.0}, 'iops': {'lve': 0.0},
                            'io': {'lve': 0.0, 'mysql': 70 * 1024 * 1024}, 'nproc': {'lve': 0.0}, 'cpu': {'lve': 0.0, 'mysql': 2},
                            'pmem': {'lve': 0.0}}}]},
                         result)

    @mock.patch('clcommon.lib.MySQLGovernor.is_governor_present', return_value=True)
    @mock.patch('clcommon.lib.MySQLGovernor.get_limits_by_user', return_value=(1234, 1234))
    def test_generate_statistics_no_gov(self, *args):
        self.setupLVE()
        with change_timezone('UTC'):
            options = setup_statistics_parser(["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json"], self.server_id)
            with mock.patch.object(cloudlinux_statistics.CloudLinuxStatistics, "_show_mysql", return_value=True):
                result = cloudlinux_statistics.get_statistics(
                    engine=self.engine,
                    id_=None,
                    options=options,
                    timestamp=15,
                    log=None)
        self.maxDiff = None
        self.assertEqual({'timestamp': 15, 'resellers': [], 'mySqlGov': 'success', 'result': 'success', 'users': [
            {'reseller': 'N/A', 'username': 'user501',
             'faults': {'vmem': {'lve': 100}, 'iops': {'lve': 160}, 'io': {'lve': 250}, 'nproc': {'lve': 20},
                        'cpu': {'lve': 10000}, 'ep': {'lve': 120}, 'pmem': {'lve': 80}}, 'domain': 'N/A',
             'limits': {'vmem': {'lve': 81920}, 'iops': {'lve': 22}, 'io': {'lve': 19.0, 'mysql': 1234 * 1024}, 'nproc': {'lve': 9},
                        'cpu': {'lve': 0.17, 'mysql': 1234}, 'ep': {'lve': 18}, 'pmem': {'lve': 86016}},
             'usage': {'vmem': {'lve': 94208}, 'iops': {'lve': 8}, 'io': {'lve': 10, 'mysql': 0}, 'nproc': {'lve': 8},
                       'cpu': {'lve': 0.02, 'mysql': 0.0}, 'ep': {'lve': 4}, 'pmem': {'lve': 20480}}, 'id': 501},
            {'reseller': 'N/A', 'username': 'user503',
             'faults': {'vmem': {'lve': 100}, 'iops': {'lve': 160}, 'io': {'lve': 150}, 'nproc': {'lve': 130},
                        'cpu': {'lve': 140}, 'ep': {'lve': 120}, 'pmem': {'lve': 110}}, 'domain': 'N/A',
             'limits': {'vmem': {'lve': 81920}, 'iops': {'lve': 22}, 'io': {'lve': 19.0, 'mysql': 1234 * 1024}, 'nproc': {'lve': 9},
                        'cpu': {'lve': 2000.0, 'mysql': 1234}, 'ep': {'lve': 18}, 'pmem': {'lve': 86016}},
             'usage': {'vmem': {'lve': 12288}, 'iops': {'lve': 8}, 'io': {'lve': 6.0, 'mysql': 0}, 'nproc': {'lve': 7},
                       'cpu': {'lve': 1600.0, 'mysql': 0}, 'ep': {'lve': 4}, 'pmem': {'lve': 20480}}, 'id': 503}]},
                         result)

    @ddt.data(
        ({'user503', 'unknown2', 'user501'}, 0),
        ({'user503'}, 1),
        ({'user501'}, 1),
        ({'unknown2'}, 2),
        (set(), 2))
    @ddt.unpack
    @mock.patch('clcommon.lib.MySQLGovernor.is_governor_present', return_value=True)
    @mock.patch('clcommon.lib.MySQLGovernor.get_limits_by_user', return_value=(1234, 1234))
    def test_generate_statistics_no_gov_without_admin(self, admins, test_result, *args):
        self.setupLVE()
        with change_timezone('UTC'), \
                mock.patch('lvestats.lib.cloudlinux_statistics.cpapi.admins', return_value=admins):
            options = setup_statistics_parser(["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json"], self.server_id)
            with mock.patch.object(cloudlinux_statistics.CloudLinuxStatistics, "_show_mysql", return_value=True):
                result = cloudlinux_statistics.get_statistics(
                    engine=self.engine,
                    id_=None,
                    options=options,
                    timestamp=15,
                    log=None)
        self.assertEqual(len(result['users']), test_result)

    @mock.patch('clcommon.lib.MySQLGovernor.is_governor_present', return_value=True)
    @mock.patch('clcommon.lib.MySQLGovernor.get_limits_by_user', return_value=(1234, 1234))
    def test_generate_statistics_with_reseller(self, *args):
        self.setupLVE(with_reseller=True)
        with change_timezone('UTC'):
            args = ['--json', '--id=501', '--from=1970-01-01 00:00', '--to=1970-01-01 00:10']
            options = setup_statistics_parser(args, self.server_id)
            result = cloudlinux_statistics.get_statistics(
                engine=self.engine,
                id_=None,
                options=options,
                timestamp=15,
                log=None)
        self.maxDiff = None
        check_list = self.convert_defaultdicts_to_dicts(result['resellers'])

        self.assertListEqual([
            {'name': 'reseller503',
             'faults': {'vmem': {'lve': 100}, 'iops': {'lve': 160}, 'io': {'lve': 150}, 'nproc': {'lve': 130},
                        'cpu': {'lve': 140}, 'ep': {'lve': 120}, 'pmem': {'lve': 110}},
             'limits': {'vmem': {'lve': 81920}, 'iops': {'lve': 22}, 'io': {'lve': 19.0, 'mysql': 1234 * 1024},
                        'nproc': {'lve': 9}, 'cpu': {'lve': 2000.0, 'mysql': 1234}, 'ep': {'lve': 18}, 'pmem': {'lve': 86016}},
             'usage': {'vmem': {'lve': 12288}, 'iops': {'lve': 8}, 'io': {'lve': 6.0, 'mysql': 0},
                       'nproc': {'lve': 7}, 'cpu': {'lve': 1600.0, 'mysql': 0}, 'ep': {'lve': 4}, 'pmem': {'lve': 20480}},
             'id': 1000000503}], check_list)

    @staticmethod
    def convert_defaultdicts_to_dicts(data):
        def process_field(field):
            """
            Convert default dict to dict
            """
            return dict(field) if isinstance(field, defaultdict) else field

        def process_item(item):
            """
            Apply process_field to each key in dict
            """
            return {field: process_field(item[field]) for field in item}

        return list(map(process_item, data))

    @ddt.data((15, datetime.datetime(1970, 1, 1, 0, 0), datetime.datetime(1970, 1, 1, 0, 10, 0),
               {
                   'timestamp': 15,
                   'mySqlGov': 'success',
                   'result': 'success',
                   'resellers': [],
                   'users': [
                       {
                           'reseller': 'N/A', 'username': 'user501',
                           'faults': {
                               'vmem': {'lve': 100},
                               'iops': {'lve': 160},
                               'io': {'lve': 250},
                               'nproc': {'lve': 20},
                               'cpu': {'lve': 10000},
                               'ep': {'lve': 120},
                               'pmem': {'lve': 80}},
                           'domain': 'N/A',
                           'limits': {
                               'vmem': {'lve': 81920},
                               'iops': {'lve': 22},
                               'io': {'lve': 19.0, 'mysql': 50 * 1024 * 1024},
                               'nproc': {'lve': 9},
                               'cpu': {'lve': 0.17, 'mysql': 5.0},
                               'ep': {'lve': 18},
                               'pmem': {'lve': 86016}},
                           'usage': {
                               'vmem': {'lve': 94208},
                               'iops': {'lve': 8},
                               'io': {'lve': 10.0, 'mysql': 500 * 1024 * 1024},
                               'nproc': {'lve': 8},
                               'cpu': {'lve': 0.02, 'mysql': 1.0},
                               'ep': {'lve': 4},
                               'pmem': {'lve': 20480}},
                           'id': 501},
                       {
                           'reseller': 'N/A', 'username': 'user503',
                           'faults': {
                               'vmem': {'lve': 100},
                               'iops': {'lve': 160},
                               'io': {'lve': 150},
                               'nproc': {'lve': 130},
                               'cpu': {'lve': 140},
                               'ep': {'lve': 120},
                               'pmem': {'lve': 110}},
                           'domain': 'N/A',
                           'limits': {
                               'vmem': {'lve': 81920},
                               'iops': {'lve': 22},
                               'io': {'lve': 19.0, 'mysql': 200 * 1024 * 1024},
                               'nproc': {'lve': 9},
                               'cpu': {'lve': 2000.0, 'mysql': 0.0},
                               'ep': {'lve': 18},
                               'pmem': {'lve': 86016}},
                           'usage': {
                               'vmem': {'lve': 12288},
                               'iops': {'lve': 8},
                               'io': {'lve': 6.0, 'mysql': 70 * 1024 * 1024},
                               'nproc': {'lve': 7},
                               'cpu': {'lve': 1600.0, 'mysql': 2.0},
                               'ep': {'lve': 4},
                               'pmem': {'lve': 20480}},
                           'id': 503}]}))
    @ddt.unpack
    @mock.patch('clcommon.lib.MySQLGovernor.is_governor_present', return_value=False)
    def test_generate_statistics_both(self, timestamp, period_from, period_to, valid_result, mock_governor):
        self.setupLVE()
        self.setupGov()
        options = setup_statistics_parser(["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json"], self.server_id)
        with mock.patch.object(cloudlinux_statistics.CloudLinuxStatistics, "_show_mysql", return_value=True):
            result = cloudlinux_statistics.get_statistics(
                engine=self.engine,
                id_=None,
                options=options,
                timestamp=15,
                log=None)
        self.maxDiff = None
        self.assertEqual(valid_result, result)

    @ddt.data('gov48.csv', 'gov28.csv', 'gov3.csv')
    def test_generate_statistics_both_data_amount_differ(self, gov_csv_filename):
        """
        This test checks that data entries from gov and lve are aligned correctly after process_data()
        lve_results always has 19 records here, so
        gov48 checks case when len(gov_results) > len(lve_results) * 2,
        gov28 checks case when len(gov_results) ~ len(lve_results),
        gov3 checks case when len(lve_results) > len(gov_results) * 2
        """
        data_dir = os.path.join(os.path.dirname(__file__), 'data')
        self.setupLVE_from_csv(os.path.join(data_dir, 'lve29.csv'))
        self.setupGov_from_csv(os.path.join(data_dir, gov_csv_filename))
        # server_id and uid from csv loaded above
        test_server_id = 'da67e379-6'
        test_uid = 1065
        options = setup_statistics_parser(["--from=2022-07-26 00:00", "--to=2022-07-28 00:00", "--time-unit=1h", "--json"], test_server_id)
        with mock.patch.object(cloudlinux_statistics.CloudLinuxStatistics, "_show_mysql", return_value=True):
            stat = cloudlinux_statistics.CloudLinuxStatistics(
                engine=self.engine,
                show_columns=options.show,
                order_by=options.order_by,
                server_id=options.server_id,
                by_usage=options.by_usage,
                by_usage_percentage=options.percentage / 100.,
                by_fault=options.by_fault,
                threshold=options.threshold,
                limit=options.limit,
                uid=test_uid,
                time_unit=options.time_unit)
            stat.generate_statistics(getattr(options, 'from'), options.to)
            self.assertEqual(len(stat._lve_results), len(stat._gov_results))
            self.assertTrue(check_timestamps_equal_or_none(stat._lve_results, stat._gov_results))

    @ddt.data(True, False)
    def test_generate_statistics_skip_maxuid_lve(self, skip_maxuid_lve):
        self.setupLVE()
        options = setup_statistics_parser(["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json"],
                                          self.server_id)
        with mock.patch.object(cloudlinux_statistics.CloudLinuxStatistics, "_show_mysql", return_value=True), \
                mock.patch('lvestats.lib.lveinfolib.str_to_bool', return_value=skip_maxuid_lve):
            result = cloudlinux_statistics.get_statistics(
                engine=self.engine,
                id_=None,
                options=options,
                timestamp=15,
                log=None)
        ids = [user.get('id') for user in result['users']]
        self.assertEqual(self.max_uid_lve_id not in ids, skip_maxuid_lve)

    def test_generate_statistics_both_wrong_server_id(self):
        """LVES-909"""
        self.setupLVE()
        self.setupGov()
        options = setup_statistics_parser(["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json"], "5555555555555555555555555")
        with mock.patch.object(cloudlinux_statistics.CloudLinuxStatistics, "_show_mysql", return_value=True):
            result = cloudlinux_statistics.get_statistics(
                engine=self.engine,
                id_=None,
                options=options,
                timestamp=15,
                log=None)
        self.maxDiff = None
        self.assertEqual(
            {
                'timestamp': 15,
                'mySqlGov': 'success',
                'result': 'success',
                'resellers': [],
                'users': []},
            result)

    @ddt.data(('cpu',         '90',  '10', '',              None,   '1',    []),
              ('cpu',         '80',  '10', '',              None,   '1',    ['user503']),
              ('cpu',         '10',  '10', 'cpu',           None,   '1',    ['user503', 'user501']),
              ('cpu',         '10',  '1',  'cpu',           None,   '1',    ['user503']),
              ('cpu',         '10',  '10', 'any_faults',    None,   '1',    ['user501', 'user503']),
              (None,          '90',  '10', '',              'cpu',  '5000', ['user501']),
              (None,          '90',  '10', 'cpu',           'cpu',  '100',  ['user503', 'user501']),
              (None,          '90',  '10', 'cpu',           'cpu',  '100',  ['user503', 'user501']),
              (None,          '90',  '10', 'nproc',         None,   '1',    ['user501', 'user503']),
              (None,          '90',  '10', 'mysql_cpu',     None,   '1',    ['user503', 'user501']),
              (None,          '90',  '10', 'mysql_cpu',     None,   '1',    ['user503', 'user501']),
              (None,          '90',  '10', 'mysql_io',      None,   '1',    ['user501', 'user503']),
              ('mysql_io',    '90',  '10', 'any_faults',    None,   '1',    ['user501']),
              ('mysql_io',    '30',  '10', 'any_faults',    None,   '1',    ['user501', 'user503']),
              ('mysql_cpu',   '10',  '10', 'any_faults',    None,   '1',    ['user501']),
              ('mysql_cpu',   '90',  '10', 'any_faults',    None,   '1',    []),
              (None,          '90',  '10', 'pmem_faults',   None,   '1',    ['user503', 'user501']),
              (None,          '90',  '10', 'io_faults',     None,   '1',    ['user501', 'user503']),
              (None,          '90',  '10', 'nproc_faults',  None,   '1',    ['user503', 'user501']),
              (None,          '90',  '10', 'io',            None,   '1',    ['user501', 'user503']),
              (None,          '90',  '10', 'vmem',          None,   '1',    ['user501', 'user503']),
              (None,          '90',  '10', '',              'any', '20000', []),
              (None,          '90',  '10', '',              'any', '10000', ['user501']),
              (None,          '90',  '10', '',              'any', '1000',  ['user501']),
              (None,          '90',  '10', '',              'any', '900',   ['user501', 'user503']),
              (None,          '90',  '10', '',              'nproc', '100', ['user503']),
              (None,          '90',  '10', '',              'nproc', '10',  ['user503', 'user501']))
    @ddt.unpack
    def test_order(self, by_usage, by_usage_percentage, limit, order_by, by_fault, threshold, users_in_order):
        self.setupGov()
        self.setupLVE()
        args = ["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json", "--limit=" + limit]
        if by_usage is not None:
            args.append("--by-usage=" + by_usage)
            args.append("--percentage=" + by_usage_percentage)
        if order_by:
            args.append("--order-by=" + order_by)
        if by_fault is not None:
            args.append("--by-fault=" + by_fault)
            args.append("--threshold=" + threshold)
        options = setup_statistics_parser(args, self.server_id)
        with change_timezone('UTC'):
            with mock.patch.object(cloudlinux_statistics.CloudLinuxStatistics, "_show_mysql", return_value=True):
                result = cloudlinux_statistics.get_statistics(
                    engine=self.engine,
                    id_=None,
                    options=options,
                    timestamp=15,
                    log=None)
        users = [entry['username'] for entry in result['users']]
        self.assertEqual(users_in_order, users)

    @ddt.data(
        ['cpu',        '30', [503]],
        ['cpu',        '10', [501, 503]],
        ['vmem',       '10', [501, 503]],
        ['vmem',       '50', [501]],
        ['io',         '30', [501, 503]],
        ['io',         '40', [501]],
        ['mysql_cpu',  '10', [501]],
        ['mysql_cpu',  '30', []],
        ['mysql_io',   '10', [501, 503]],
        ['mysql_io',   '30', [501, 503]],
        ['mysql_io',   '40', [501]])
    @ddt.unpack
    def test_by_usage(self, by_usage, by_usage_percentage, users):
        self.setupLVE()
        self.setupGov()
        args = ["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json",
                "--by-usage=" + by_usage, "--percentage=" + by_usage_percentage]
        options = setup_statistics_parser(args, self.server_id)
        with change_timezone('UTC'):
            with mock.patch.object(cloudlinux_statistics.CloudLinuxStatistics, "_show_mysql", return_value=True):
                result = cloudlinux_statistics.get_statistics(
                    engine=self.engine,
                    id_=None,
                    options=options,
                    timestamp=15,
                    log=None)
        result_users = [user['id'] for user in result['users']]
        self.assertEqual(result_users, users)

        # check if we got real mysql data
        for user_data in result['users']:
            data = {
                'usage': {
                    'cpu': {
                        501: 1,
                        503: 2},
                    'io': {
                        501: 500 * 1024 * 1024,
                        503: 70 * 1024 * 1024}},
                'limits': {
                    'cpu': {
                        501: 5,
                        503: 0},
                    'io': {
                        501: 50 * 1024 * 1024,
                        503: 200 * 1024 * 1024}}}
            self.assertEqual(user_data['usage']['cpu']['mysql'], data['usage']['cpu'][user_data['id']])
            self.assertEqual(user_data['usage']['io']['mysql'], data['usage']['io'][user_data['id']])
            self.assertEqual(user_data['limits']['cpu']['mysql'], data['limits']['cpu'][user_data['id']])
            self.assertEqual(user_data['limits']['io']['mysql'], data['limits']['io'][user_data['id']])

    def fill_singleuser_data(self):
        d_from = datetime.datetime(2014, 12, 12, 10, 0, 0)
        t_from = dateutil.gm_datetime_to_unixtimestamp(dateutil.local_to_gm(d_from))

        # fill data for 30m
        insert_list = []
        db_insert_list = []
        for time in range(0, 1800, 60):
            db_insert_list.append(dict(
                server_id=self.server_id, uid=501, ts=t_from + time, sum_cpu=1,
                limit_cpu_on_period_end=5, limit_read_on_period_end=20,
                limit_write_on_period_end=30,
                sum_read=300, sum_write=200))
            insert_list.extend([
                dict(
                    server_id=self.server_id, id=0, created=t_from + time, cpu=0, lnproc=0, iops=0, io=0,
                    cpu_limit=0, io_limit=0, io_fault=0, memphy_fault=0, memphy=0,
                    mep_limit=0, mep_fault=0, lmemphy=0, nproc_fault=0, mem=0,
                    cpu_fault=0, mem_fault=0, mep=0, nproc=0, liops=0,
                    mem_limit=0, iops_fault=0),
                dict(
                    server_id=self.server_id, id=501, created=t_from + time, lnproc=9, iops=8, io=6,
                    cpu_limit=17, io_limit=19, io_fault=15, memphy_fault=11, memphy=5,
                    mep_limit=18, mep_fault=12, lmemphy=21, nproc_fault=13, mem=3,
                    cpu_fault=1000, mem_fault=10, mep=4, nproc=7, liops=22,
                    mem_limit=20, iops_fault=16, cpu=20)])
        for chunk in get_chunks(insert_list):
            self.engine.execute(insert(history), chunk)
        for chunk in get_chunks(db_insert_list):
            self.engine.execute(insert(history_gov), chunk)

    @mock.patch('clcommon.lib.MySQLGovernor.is_governor_present', return_value=True)
    def test_cloudlinux_statistics_singleuser(self, _):
        d_from = datetime.datetime(2014, 12, 12, 10, 0, 0)
        t_from = dateutil.gm_datetime_to_unixtimestamp(dateutil.local_to_gm(d_from))

        self.save_dbgov_line(self.server_id, uid=501, _time=t_from, sum_cpu=1,
                             limit_cpu_on_period_end=5, limit_read_on_period_end=20,
                             limit_write_on_period_end=30,
                             sum_read=300, sum_write=200)
        self.save_history_line(self.server_id, uid=0, _time=t_from, cpu=100)
        self.save_history_line(
            self.server_id, uid=501, _time=t_from, lnproc=9, iops=8, io=6,
            cpu_limit=17, io_limit=19, io_fault=15, memphy_fault=11, memphy=5,
            mep_limit=18, mep_fault=12, lmemphy=21, nproc_fault=13, mem=3,
            cpu_fault=1000, mem_fault=10, mep=4, nproc=7, created=0, liops=22,
            mem_limit=20, iops_fault=16, cpu=20)
        with mock.patch.object(cloudlinux_statistics, '_get_uid_for_select', return_value=501):
            options = setup_statistics_parser(
                ['--json', '--id=501', '--from=2014-12-12 10:00', '--to=2014-12-12 10:10'], self.server_id)
            json_data, _ = cloudlinux_statistics.execute(
                self.engine, options)

        response = json.loads(json_data)
        self.assertEqual(len(response['user']), 1)

        user_info = response['user'][0]  # get first element
        self.check_contain_dbgov_lve_data(user_info)

    def check_contain_dbgov_lve_data(self, user_info):
        """
        Check if all data (lve and dbgov) exists.
        :type user_info: dict
        """
        self.assertIn('mysql', user_info['usage']['io'])
        self.assertIn('mysql', user_info['usage']['cpu'])
        self.assertIn('mysql', user_info['limits']['io'])
        self.assertIn('mysql', user_info['limits']['cpu'])
        self.assertIn('ep', user_info['usage'])
        self.assertIn('pmem', user_info['usage'])
        self.assertIn('ep', user_info['faults'])
        self.assertIn('pmem', user_info['faults'])
        self.assertIn('ep', user_info['faults'])
        self.assertIn('io', user_info['faults'])
        self.assertIsInstance(user_info['from'], int)
        self.assertIsInstance(user_info['to'], int)

    @ddt.data(['1m', 30],
              ['2m', 15])
    @ddt.unpack
    @mock.patch('clcommon.lib.MySQLGovernor.is_governor_present', return_value=True)
    def test_cloudlinux_statistics_singleuser_order_time_unit(self, time_unit, reference_count, _):
        self.fill_singleuser_data()
        # from and to are set in fill_singleuser_data
        with mock.patch.object(cloudlinux_statistics, '_get_uid_for_select', return_value=501):
            options = setup_statistics_parser(['--json', '--id=501',
                                               '--from=2014-12-12 10:00', '--to=2014-12-12 10:30',
                                               '--time-unit=%s' % time_unit], self.server_id)
            json_data, _ = cloudlinux_statistics.execute(self.engine, options)
        response = json.loads(json_data)
        self.assertEqual(len(response['user']), reference_count)
        # check whether time-unit works
        for i in range(0, len(response['user']) - 1):
            user_info = response['user'][i]
            self.check_contain_dbgov_lve_data(user_info)
            self.assertEqual(response['user'][i + 1]['from'] - response['user'][i]['from'], time_unit_type(time_unit))

    def test_cloudlinux_statistics_singleuser_limit(self):
        self.fill_singleuser_data()
        # from and to are set in fill_singleuser_data
        with mock.patch.object(cloudlinux_statistics, '_get_uid_for_select', return_value=501):
            options = setup_statistics_parser(['--json', '--id=501',
                                               '--from=2014-12-12 10:00', '--to=2014-12-12 10:30',
                                               '--limit=10'], self.server_id)
            json_data, _ = cloudlinux_statistics.execute(self.engine, options)
        response = json.loads(json_data)
        self.assertEqual(len(response['user']), 10)

        # check whether time-unit works
        for i in range(0, len(response['user']) - 1, 1):
            self.assertEqual(response['user'][i + 1]['from'] - response['user'][i]['from'], 60)

    def test_json_output(self):
        options = setup_statistics_parser(['--json'], self.server_id)
        json_data, exitcode = cloudlinux_statistics.execute(self.engine, options)
        unpack = json.loads(json_data)

        self.assertIsInstance(unpack, dict)
        self.assertEqual(exitcode, 0)

    def test_json_output_with_error(self):
        # don't move this line to the top, cause mock won't work
        from lvestats.lib.lveinfolib import HistoryShow

        with mock.patch.object(HistoryShow, "proceed", side_effect=ValueError("Some error occurred")):
            options = setup_statistics_parser(['--json'], self.server_id)
            json_data, exitcode = cloudlinux_statistics.execute(self.engine, options)
        unpack = json.loads(json_data)

        self.assertIsInstance(unpack, dict)
        self.assertIn("timestamp", unpack)
        self.assertIn("mySqlGov", unpack)
        self.assertEqual(unpack["result"], "Some error occurred")
        self.assertEqual(exitcode, 0)

    @mock.patch('clcommon.lib.MySQLGovernor.is_governor_present', return_value=True)
    @mock.patch('clcommon.lib.MySQLGovernor.get_limits_by_user', return_value=(1234, 5678))
    def test_realtime_dbgov_limits(self, *args):
        """Test if we return realtime data for all users if governor present"""
        self.setupLVE()
        with change_timezone('UTC'):
            options = setup_statistics_parser(["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json"], self.server_id)
            result = cloudlinux_statistics.get_statistics(
                engine=self.engine,
                id_=None,
                options=options,
                timestamp=15,
                log=None)
        self.maxDiff = None
        for user in result['users']:
            self.assertEqual(user['limits']['cpu']['mysql'], 1234)
            self.assertEqual(user['limits']['io']['mysql'], 5678 * 1024)

    @mock.patch('clcommon.lib.MySQLGovernor.is_governor_present', return_value=True)
    @mock.patch('clcommon.lib.MySQLGovernor.get_limits_by_user', side_effect=MySQLGovException('some exception'))
    def test_realtime_dbgov_limits_exception(self, *args):
        """Check case when dbgov raises an exception"""
        self.setupLVE()
        with change_timezone('UTC'):
            options = setup_statistics_parser(["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json"], self.server_id)
            result = cloudlinux_statistics.get_statistics(
                engine=self.engine,
                id_=None,
                options=options,
                timestamp=15,
                log=None)
        self.maxDiff = None
        for user in result['users']:
            self.assertEqual(user['limits']['cpu']['mysql'], 0)
            self.assertEqual(user['limits']['io']['mysql'], 0)

    @ddt.unpack
    @ddt.data(
        ([501, 503],                  [501, 503]),
        ([501],                       [501]),
        ([502],                       []),
        ([501, 502],                  [501]),
        (list(range(500, 1000)),      [501, 503]),
        ([],                          []))
    def test_multi_uids(self, uids, uids_output):
        self.setupGov()
        self.setupLVE()
        with change_timezone('UTC'):
            options = setup_statistics_parser(["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json"], self.server_id)
            valid_results = cloudlinux_statistics.get_statistics(
                engine=self.engine,
                id_=None,  # get all results (tested in other tests)
                options=options,
                timestamp=15,
                log=None)
        interesting_results = [_ for _ in valid_results['users'] if _['id'] in uids_output]

        with change_timezone('UTC'):
            options = setup_statistics_parser(["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json"], self.server_id)
            result = cloudlinux_statistics.get_statistics(
                engine=self.engine,
                id_=uids,
                options=options,
                timestamp=15,
                log=None)
        result_uids = [_['id'] for _ in result.get('users', {'id': list()})]
        self.assertListEqual(uids_output, result_uids)
        self.assertListEqual(result['users'], interesting_results)

    # LVESTATS-97
    @ddt.unpack
    @ddt.data(
        (list(_USERMAP.values())[0], list(_USERMAP.keys())[1], True),
        (list(_USERMAP.values())[0], 502, False))
    def test_get_stat_for_resellers_user(self, reseller_name, user_id, owner):
        self.setupLVE(with_reseller=True)
        self.setupGov()
        reseller_id = _INV_USERMAP[reseller_name]
        if owner:
            uid_for_select = [reseller_id, user_id, serialize_lve_id(LIMIT_LVP_ID, reseller_id)]
            resellers_users = [reseller_name, list(_USERMAP.values())[1]]
            result = u'success'
        else:
            uid_for_select = []
            resellers_users = [reseller_name]
            result = u'Permission denied. User with id {} ' \
                     u'does not belong reseller `{}`'.format(user_id, reseller_name)
        with mock.patch.object(cloudlinux_statistics, 'get_users_for_reseller', return_value=resellers_users):
            with mock.patch.object(cloudlinux_statistics, '_get_uid_for_select', return_value=uid_for_select):
                options = setup_statistics_parser(
                    ['--json', '--id={}'.format(user_id), '--for-reseller={}'.format(reseller_name)], self.server_id)
                json_data, _ = cloudlinux_statistics.execute(
                    self.engine, options)
                response = json.loads(json_data)
                self.assertEqual(response['result'], result)

    @ddt.data((501, True), (502, False), (503, True))
    @ddt.unpack
    def test_get_stat_for_resellers_user2(self, user_id, user_in_response):
        self.setupLVE(with_reseller=True)
        self.setupGov()
        reseller_name = 'reseller503'
        resellers_users = list(_USERMAP.values())
        with mock.patch.object(cloudlinux_statistics, 'get_users_for_reseller', return_value=resellers_users):
            with mock.patch('lvestats.lib.cloudlinux_statistics.reseller_to_uid', return_value=1000001000):
                options = setup_statistics_parser(
                    ['--json', '--id={}'.format(user_id), '--for-reseller={}'.format(reseller_name)], self.server_id)
                json_data, _ = cloudlinux_statistics.execute(self.engine, options)
        response = json.loads(json_data)
        if user_in_response:
            self.assertIn(u'user', response)
        else:
            self.assertNotEqual(response[u'result'], u'success')

    # LVESTATS-97 fixes
    @ddt.unpack
    @ddt.data(
        ('root', True, list(_USERMAP.values())[0], list(_USERMAP.keys())[0], list(_USERMAP.keys())[0]),
        ('root', True, list(_USERMAP.values())[0], None, list(_USERMAP.keys()) + [1000001000]),
        ('root', True, None, list(_USERMAP.keys())[0], list(_USERMAP.keys())[0]),
        ('user', False, list(_USERMAP.values())[0], 1000, 1001),
        ('user', False, list(_USERMAP.values())[0], None, 1001),
        ('user', False, None, list(_USERMAP.keys())[0], 1001)
    )
    def test_get_uid_for_select(self, user, is_admin, for_reseller, uid, result):
        self.setupLVE(with_reseller=True)
        self.setupGov()
        resellers_users = list(_USERMAP.values())

        def mocked_username_to_uid(name):
            for k, v in _USERMAP.items():
                if v == name:
                    return k
            return 0

        with \
                mock.patch('lvestats.lib.cloudlinux_statistics.getuser',
                           return_value=user), \
                mock.patch('lvestats.lib.cloudlinux_statistics.os.getuid', return_value=result), \
                mock.patch('lvestats.lib.cloudlinux_statistics.cpapi.is_admin', return_value=is_admin), \
                mock.patch.object(cloudlinux_statistics, 'get_users_for_reseller', return_value=resellers_users), \
                mock.patch('lvestats.lib.cloudlinux_statistics.username_to_uid_local',
                           side_effect=mocked_username_to_uid), \
                mock.patch('lvestats.lib.cloudlinux_statistics.reseller_to_uid', return_value=1000001000):
            res = cloudlinux_statistics._get_uid_for_select(
                for_reseller=for_reseller, uid=uid)
            if isinstance(result, int):
                self.assertEqual(res, result, (res, result))
            else:
                self.assertCountEqual(res, result, (res, result))

    def test_size(self):
        self.fill_singleuser_data()
        args = ['--json', '--id=501', '--from=2014-12-12 10:00', '--to=2014-12-12 10:30']
        options = setup_statistics_parser(args, self.server_id)
        statistics = cloudlinux_statistics.get_statistics(self.engine, 501, None, options)
        tsize = total_size(statistics)
        self.assertLess(tsize, 80000)

    @mock.patch('clcommon.lib.MySQLGovernor.is_governor_present', return_value=True)
    def test_default_mysql_limits(self, *args):
        """Test if we return correct mysql limit when there is no mysql data for this point"""
        self.setupLVE()

        def mock_get_limits_by_user(user):
            if user == 'user501':
                return 1234, 1234
            elif user == 'user503':
                return 4321, 4321
            else:
                return 0, 0

        with mock.patch('clcommon.lib.MySQLGovernor.get_limits_by_user', side_effect=mock_get_limits_by_user):
            options = setup_statistics_parser(['--json', '--from=1970-01-01 00:00', '--to=1970-01-01 00:10'], self.server_id)
            json_data, _ = cloudlinux_statistics.execute(
                self.engine, options)
            response = json.loads(json_data)
            self.assertEqual(
                sorted((
                    (response['users'][0]['username'],
                     response['users'][0]['limits']['cpu']['mysql']),
                    (response['users'][1]['username'],
                     response['users'][1]['limits']['cpu']['mysql']))),
                sorted((('user501', 1234), ('user503', 4321))))

            options = setup_statistics_parser(
                ['--json', '--id', '501', '--from=1970-01-01 00:00', '--to=1970-01-01 00:10'], self.server_id)
            json_data, _ = cloudlinux_statistics.execute(
                self.engine, options)
            response = json.loads(json_data)
            self.assertEqual(response['user'][0]['limits']['cpu']['mysql'], 1234)

            options = setup_statistics_parser(
                ['--json', '--id', '503', '--from=1970-01-01 00:00', '--to=1970-01-01 00:10'], self.server_id)
            json_data, _ = cloudlinux_statistics.execute(
                self.engine, options)
            response = json.loads(json_data)
            self.assertEqual(response['user'][0]['limits']['cpu']['mysql'], 4321)

    def test_server_id(self):
        server_id_1 = 'asdfgfh'
        server_id_2 = 'qwerty'

        self.save_history_line(server_id_1, uid=0, _time=0, cpu=100)
        self.save_history_line(server_id_2, uid=0, _time=0, cpu=100)
        self.save_history_line(
            server_id_1, uid=501, _time=0,
            nproc=8, lnproc=9, nproc_fault=2,
            io=10, io_limit=19, io_fault=25,
            memphy=5, lmemphy=21, memphy_fault=8,
            mep=4, mep_limit=18, mep_fault=12,
            mem=23, mem_limit=20, mem_fault=10,
            cpu=2, cpu_limit=17, cpu_fault=1000,
            iops=8, liops=22, iops_fault=16)
        self.save_history_line(
            server_id_2, uid=503, _time=0,
            iops=8, liops=22, iops_fault=16,
            io=6, io_limit=19, io_fault=15,
            memphy=5, lmemphy=21, memphy_fault=11,
            mep=4, mep_limit=18, mep_fault=12,
            cpu=160000, cpu_limit=200000, cpu_fault=14,
            lnproc=9, nproc=7, nproc_fault=13,
            mem=3, mem_limit=20, mem_fault=10, )

        options = setup_statistics_parser(['--json', '--from=1970-01-01 00:00', '--to=1970-01-01 00:10'], server_id_1)
        json_data, _ = cloudlinux_statistics.execute(
            self.engine, options)
        response = json.loads(json_data)
        self.assertEqual(len(response['users']), 1)
        self.assertEqual(response['users'][0]['id'], 501)

        options = setup_statistics_parser(['--json', '--from=1970-01-01 00:00', '--to=1970-01-01 00:10'], server_id_2)
        json_data, _ = cloudlinux_statistics.execute(
            self.engine, options)
        response = json.loads(json_data)
        self.assertEqual(len(response['users']), 1)
        self.assertEqual(response['users'][0]['id'], 503)

    @ddt.data(
        (set(), set()),
        ({'unknown'}, set()),
        ({'user503', 'unknown2', 'user501'}, {501, 503}))
    @ddt.unpack
    def test_get_admin_uids(self, admins, result):
        with mock.patch('lvestats.lib.cloudlinux_statistics.cpapi.admins', return_value=admins):
            self.assertEqual(
                cloudlinux_statistics.CloudLinuxStatistics._get_admin_uids(),
                result)

    @ddt.data(
        (AttributeError(), set()))
    @ddt.unpack
    def test_get_admin_uids_errors(self, admins, result):
        with mock.patch('lvestats.lib.cloudlinux_statistics.cpapi.admins', side_effect=admins):
            self.assertEqual(
                cloudlinux_statistics.CloudLinuxStatistics._get_admin_uids(),
                result)

    @mock.patch('lvestats.lib.config.read_config', return_value={})
    @mock.patch('lvestats.lib.cloudlinux_statistics.make_db_engine')
    def test_get_resellers_with_faults(self, mdbe, rc):
        mdbe.return_value = self.engine
        self.assertEqual(cloudlinux_statistics.get_users_and_resellers_with_faults(), (0, 0))


if __name__ == '__main__':
    unittest.main()
