# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#

from time import tzset

import os
import unittest

import ddt
import jsonschema
import mock
from clcommon.clproc import LIMIT_LVP_ID
from clcommon.lib import MySQLGovernor

from test.base.base_db_test import BaseDbTest
from test.helpers import change_timezone
from lvestats.lib.cloudlinux_statistics import get_statistics
from lvestats.lib.commons.func import serialize_lve_id
from lvestats.lib.parsers.cloudlinux_statistics_parser import setup_statistics_parser

_USERMAP = {503: 'user503', 501: 'user501', 1000000504: 'reseller504'}


# noinspection PyUnusedParameter
def mock_uid_to_username(uid, local_server_id, server_id, db_engine):
    if uid in _USERMAP:
        return _USERMAP[uid]
    return None


# json schema collection (http://spacetelescope.github.io/understanding-json-schema/) for validate json output

schema_case_users = dict()
schema_case_users[('cpu',)] = {
    "type": "object", "required": ["users", "resellers"], "properties": {"users": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["username", "domain", "id", "faults", "limits", "usage"],
             "properties": {
                 "username": {"type": "string"},
                 "reseller": {"type": "string"},
                 "domain":   {"type": "string"},
                 "id":       {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["cpu"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["cpu"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["cpu"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}},
        "resellers": {"type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["name", "id", "faults", "limits", "usage"],
             "properties": {
                 "name": {"type": "string"},
                 "id":   {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["cpu"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["cpu"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["cpu"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}}}}

schema_case_users[('io',)] = {
    "type": "object", "required": ["users", "resellers"], "properties": {"users": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["username", "domain", "id", "faults", "limits", "usage"],
             "properties": {
                 "username": {"type": "string"},
                 "reseller": {"type": "string"},
                 "domain":   {"type": "string"},
                 "id":       {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["io"], "properties": {
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["io"], "properties": {
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["io"], "properties": {
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}},
        "resellers": {"type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["name", "id", "faults", "limits", "usage"],
             "properties": {
                 "name": {"type": "string"},
                 "id": {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["io"], "properties": {
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["io"], "properties": {
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["io"], "properties": {
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}}
    }}

schema_case_users[('iops',)] = {
    "type": "object", "required": ["users", "resellers"], "properties": {"users": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["username", "domain", "id", "faults", "limits", "usage"],
             "properties": {
                 "username": {"type": "string"},
                 "reseller": {"type": "string"},
                 "domain":   {"type": "string"},
                 "id":       {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["iops"], "properties": {
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["iops"], "properties": {
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["iops"], "properties": {
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}},
        "resellers": {"type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["name", "id", "faults", "limits", "usage"],
             "properties": {
                 "name": {"type": "string"},
                 "id": {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["iops"], "properties": {
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["iops"], "properties": {
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["iops"], "properties": {
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}}
    }}

schema_case_users[('ep',)] = {
    "type": "object", "required": ["users", "resellers"], "properties": {"users": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["username", "domain", "id", "faults", "limits", "usage"],
             "properties": {
                 "username": {"type": "string"},
                 "reseller": {"type": "string"},
                 "domain":   {"type": "string"},
                 "id":       {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["ep"], "properties": {
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["ep"], "properties": {
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["ep"], "properties": {
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}},
        "resellers": {"type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["name", "id", "faults", "limits", "usage"],
             "properties": {
                 "name": {"type": "string"},
                 "id": {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["ep"], "properties": {
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["ep"], "properties": {
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["ep"], "properties": {
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}}
    }}

schema_case_users[('nproc',)] = {
    "type": "object", "required": ["users", "resellers"], "properties": {"users": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["username", "domain", "id", "faults", "limits", "usage"],
             "properties": {
                 "username": {"type": "string"},
                 "reseller": {"type": "string"},
                 "domain":   {"type": "string"},
                 "id":       {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["nproc"], "properties": {
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["nproc"], "properties": {
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["nproc"], "properties": {
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}},
        "resellers": {"type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["name", "id", "faults", "limits", "usage"],
             "properties": {
                 "name": {"type": "string"},
                 "id": {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["nproc"], "properties": {
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["nproc"], "properties": {
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["nproc"], "properties": {
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}}
    }}

schema_case_users[('pmem',)] = {
    "type": "object", "required": ["users", "resellers"], "properties": {"users": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["username", "domain", "id", "faults", "limits", "usage"],
             "properties": {
                 "username": {"type": "string"},
                 "reseller": {"type": "string"},
                 "domain":   {"type": "string"},
                 "id":       {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["pmem"], "properties": {
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["pmem"], "properties": {
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["pmem"], "properties": {
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}},
        "resellers": {"type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["name", "id", "faults", "limits", "usage"],
             "properties": {
                 "name": {"type": "string"},
                 "id": {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["pmem"], "properties": {
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["pmem"], "properties": {
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["pmem"], "properties": {
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}}
    }}

schema_case_users[('vmem',)] = {
    "type": "object", "required": ["users", "resellers"], "properties": {"users": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["username", "domain", "id", "faults", "limits", "usage"],
             "properties": {
                 "username": {"type": "string"},
                 "reseller": {"type": "string"},
                 "domain":   {"type": "string"},
                 "id":       {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["vmem"], "properties": {
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["vmem"], "properties": {
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["vmem"], "properties": {
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}},
        "resellers": {"type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["name", "id", "faults", "limits", "usage"],
             "properties": {
                 "name": {"type": "string"},
                 "id": {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["vmem"], "properties": {
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["vmem"], "properties": {
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["vmem"], "properties": {
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}}
    }}

schema_case_users[('mysql',)] = {
    "type": "object", "required": ["users", "resellers"], "properties": {"users": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["username", "domain", "id", "limits", "usage"],
             "properties": {
                 "username": {"type": "string"},
                 "reseller": {"type": "string"},
                 "domain":   {"type": "string"},
                 "id":       {"type": "integer"},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                         "mysql": {"type": "number"}}},
                      "io": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                          "mysql": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                         "mysql": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                         "mysql": {"type": "number"}}}}}}
             }},
        "resellers": {"type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["name", "id", "faults", "limits", "usage"],
             "properties": {
                 "name": {"type": "string"},
                 "id": {"type": "integer"},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                         "mysql": {"type": "number"}}},
                      "io": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                          "mysql": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                         "mysql": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                         "mysql": {"type": "number"}}}}}}}}}}


schema_case_users[('mysql', 'cpu')] = {
    "type": "object", "required": ["users", "resellers"], "properties": {"users": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["username", "domain", "id", "limits", "usage"],
             "properties": {
                 "username": {"type": "string"},
                 "reseller": {"type": "string"},
                 "domain":   {"type": "string"},
                 "id":       {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["cpu"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                         "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                      "io": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                          "mysql": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                         "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                         "mysql": {"type": "number"}}}}}}
             }},
        "resellers": {"type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["name", "id", "faults", "limits", "usage"],
             "properties": {
                 "name": {"type": "string"},
                 "id": {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["cpu"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                         "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                      "io": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                          "mysql": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                         "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                         "mysql": {"type": "number"}}}}}}}}}}

schema_case_users[('cpu', 'io')] = {
    "type": "object", "required": ["users", "resellers"], "properties": {"users": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["username", "domain", "id", "faults", "limits", "usage"],
             "properties": {
                 "username": {"type": "string"},
                 "reseller": {"type": "string"},
                 "domain":   {"type": "string"},
                 "id":       {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                 }},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                 }},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                 }}}}}},
    "resellers": {"type": "array", "items":
        {"type": "object", "additionalProperties": False,
         "required": ["name", "id", "faults", "limits", "usage"],
         "properties": {
             "name": {"type": "string"},
             "id": {"type": "integer"},
             "faults": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
             }},
             "limits": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                 "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                     "lve": {"type": "number"}}},
                 "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                     "lve": {"type": "number"}}},
             }},
             "usage": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                 "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                     "lve": {"type": "number"}}},
                 "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                     "lve": {"type": "number"}}},
             }}}}}}

schema_case_users[('all',)] = {
    "type": "object", "required": ["users", "resellers"], "properties": {"users": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["username", "domain", "id", "limits", "usage"],
             "properties": {
                 "username": {"type": "string"},
                 "reseller": {"type": "string"},
                 "domain":   {"type": "string"},
                 "id":       {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False,
                            "required": ["cpu", "io", "iops", "ep", "nproc", "pmem", "vmem"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                 }},
                 "limits": {"type": "object", "additionalProperties": False,
                            "required": ["cpu", "io", "iops", "ep", "nproc", "pmem", "vmem"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                         "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                          "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                          "lve": {"type": "number"}}},
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}
                 }},
                 "usage": {"type": "object", "additionalProperties": False,
                           "required": ["cpu", "io", "iops", "ep", "nproc", "pmem", "vmem"], "properties": {
                    "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                         "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                          "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                          "lve": {"type": "number"}}},
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}
             }}}}},
        "resellers": {"type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["name", "id", "faults", "limits", "usage"],
             "properties": {
                 "name": {"type": "string"},
                 "id": {"type": "integer"},
                 "faults": {"type": "object", "additionalProperties": False,
                            "required": ["cpu", "io", "iops", "ep", "nproc", "pmem", "vmem"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                 }},
                 "limits": {"type": "object", "additionalProperties": False,
                            "required": ["cpu", "io", "iops", "ep", "nproc", "pmem", "vmem"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                         "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                          "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                          "lve": {"type": "number"}}},
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}
                 }},
                 "usage": {"type": "object", "additionalProperties": False,
                           "required": ["cpu", "io", "iops", "ep", "nproc", "pmem", "vmem"], "properties": {
                    "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                         "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                          "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                          "lve": {"type": "number"}}},
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}}}}


schema_case_one_user = dict()
schema_case_one_user[('cpu',)] = {
    "type": "object", "required": ["user"], "properties": {"user": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["from", "to", "faults", "limits", "usage"],
             "properties": {
                 "from": {"type": "number"},
                 "to":   {"type": "number"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["cpu"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["cpu"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["cpu"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}}}}

schema_case_one_user[('io',)] = {
    "type": "object", "required": ["user"], "properties": {"user": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["from", "to", "faults", "limits", "usage"],
             "properties": {
                 "from": {"type": "number"},
                 "to":   {"type": "number"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["io"], "properties": {
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["io"], "properties": {
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["io"], "properties": {
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}}}}

schema_case_one_user[('iops',)] = {
    "type": "object", "required": ["user"], "properties": {"user": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["from", "to", "faults", "limits", "usage"],
             "properties": {
                 "from": {"type": "number"},
                 "to":   {"type": "number"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["iops"], "properties": {
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["iops"], "properties": {
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["iops"], "properties": {
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}}}}

schema_case_one_user[('ep',)] = {
    "type": "object", "required": ["user"], "properties": {"user": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["from", "to", "faults", "limits", "usage"],
             "properties": {
                 "from": {"type": "number"},
                 "to":   {"type": "number"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["ep"], "properties": {
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["ep"], "properties": {
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["ep"], "properties": {
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}}}}

schema_case_one_user[('nproc',)] = {
    "type": "object", "required": ["user"], "properties": {"user": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["from", "to", "faults", "limits", "usage"],
             "properties": {
                 "from": {"type": "number"},
                 "to":   {"type": "number"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["nproc"], "properties": {
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["nproc"], "properties": {
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["nproc"], "properties": {
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}}}}


schema_case_one_user[('pmem',)] = {
    "type": "object", "required": ["user"], "properties": {"user": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["from", "to", "faults", "limits", "usage"],
             "properties": {
                 "from": {"type": "number"},
                 "to":   {"type": "number"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["pmem"], "properties": {
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["pmem"], "properties": {
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["pmem"], "properties": {
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}}}}

schema_case_one_user[('vmem',)] = {
    "type": "object", "required": ["user"], "properties": {"user": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["from", "to", "faults", "limits", "usage"],
             "properties": {
                 "from": {"type": "number"},
                 "to":   {"type": "number"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["vmem"], "properties": {
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["vmem"], "properties": {
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["vmem"], "properties": {
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}}}}}}}


schema_case_one_user[('mysql',)] = {
    "type": "object", "required": ["user"], "properties": {"user": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["from", "to", "limits", "usage"],
             "properties": {
                "from": {"type": "number"},
                 "to":   {"type": "number"},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                         "mysql": {"type": "number"}}},
                      "io": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                          "mysql": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                         "mysql": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                         "mysql": {"type": "number"}}}}}}
             }}}}


schema_case_one_user[('mysql', 'cpu')] = {
    "type": "object", "required": ["user"], "properties": {"user": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["from", "to", "limits", "usage"],
             "properties": {
                 "from": {"type": "number"},
                 "to":   {"type": "number"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["cpu"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}}},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                         "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                      "io": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                          "mysql": {"type": "number"}}}}},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                         "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["mysql"], "properties": {
                         "mysql": {"type": "number"}}}}}}
             }}}}

schema_case_one_user[('cpu', 'io')] = {
    "type": "object", "required": ["user"], "properties": {"user": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["from", "to", "faults", "limits", "usage"],
             "properties": {
                 "from": {"type": "number"},
                 "to":   {"type": "number"},
                 "faults": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                 }},
                 "limits": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                 }},
                 "usage": {"type": "object", "additionalProperties": False, "required": ["cpu", "io"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                 }}}}}}}

schema_case_one_user[('all',)] = {
    "type": "object", "required": ["user"], "properties": {"user": {
        "type": "array", "items":
            {"type": "object", "additionalProperties": False,
             "required": ["from", "to", "faults", "usage", "limits"],
             "properties": {
                 "from": {"type": "number"},
                 "to":   {"type": "number"},
                 "faults": {"type": "object", "additionalProperties": False,
                            "required": ["cpu", "io", "iops", "ep", "nproc", "pmem", "vmem"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                 }},
                 "limits": {"type": "object", "additionalProperties": False,
                            "required": ["cpu", "io", "iops", "ep", "nproc", "pmem", "vmem"], "properties": {
                     "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                         "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                          "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                          "lve": {"type": "number"}}},
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}
                 }},
                 "usage": {"type": "object", "additionalProperties": False,
                           "required": ["cpu", "io", "iops", "ep", "nproc", "pmem", "vmem"], "properties": {
                    "cpu": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                         "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                     "io": {"type": "object", "additionalProperties": False, "required": ["mysql", "lve"], "properties": {
                          "mysql": {"type": "number"}, "lve": {"type": "number"}}},
                     "iops": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                          "lve": {"type": "number"}}},
                     "ep": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "nproc": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "pmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}},
                     "vmem": {"type": "object", "additionalProperties": False, "required": ["lve"], "properties": {
                         "lve": {"type": "number"}}}
             }}}}}}}


@ddt.ddt
class TestCloudLinuxStatisticsJsonValidation(BaseDbTest):
    patcher1 = None
    patcher2 = None
    patcher3 = None
    patcher4 = None
    patcher5 = None

    def setupGov(self):
        for time in range(3*60, 13 * 60, 60):
            self.save_dbgov_line(self.server_id, uid=501, _time=time, sum_cpu=1,
                                 limit_cpu_on_period_end=5, limit_read_on_period_end=20,
                                 limit_write_on_period_end=30,
                                 sum_read=300, sum_write=200)
            self.save_dbgov_line(self.server_id, uid=502, _time=time, sum_cpu=2, sum_read=30, sum_write=40,
                                 limit_write_on_period_end=100, limit_read_on_period_end=100)
            self.save_dbgov_line(self.server_id, uid=503, _time=time, sum_cpu=2, sum_read=30, sum_write=40,
                                 limit_write_on_period_end=100, limit_read_on_period_end=100)

    def setupLVE(self):
        for time in range(0, 10 * 60, 60):
            self.save_history_line(self.server_id, uid=0, _time=time, cpu=100)
            self.save_history_line(
                self.server_id, uid=502, _time=time, lnproc=9, iops=8, io=10,
                cpu_limit=17, io_limit=19, io_fault=25, memphy_fault=8, memphy=5,
                mep_limit=18, mep_fault=12, lmemphy=21, nproc_fault=2, mem=23,
                cpu_fault=1000, mem_fault=10, mep=4, nproc=8, created=0, liops=22,
                mem_limit=20, iops_fault=16, cpu=2)
            self.save_history_line(
                self.server_id, uid=503, _time=time, lnproc=9, iops=8, io=6,
                cpu_limit=200000, io_limit=19, io_fault=15, memphy_fault=11, memphy=5,
                mep_limit=18, mep_fault=12, lmemphy=21, nproc_fault=13, mem=3,
                cpu_fault=14, mem_fault=10, mep=4, nproc=7, created=0, liops=22,
                mem_limit=20, iops_fault=16, cpu=160000)
            self.save_history_line(
                self.server_id, uid=504, _time=time, lnproc=9, iops=8, io=6,
                cpu_limit=0, io_limit=19, io_fault=15, memphy_fault=11, memphy=5,
                mep_limit=18, mep_fault=12, lmemphy=21, nproc_fault=13, mem=3,
                cpu_fault=14, mem_fault=10, mep=4, nproc=7, created=0, liops=22,
                mem_limit=20, iops_fault=16, cpu=160000)

            # save data about reseller
            self.save_history_line(
                self.server_id, uid=serialize_lve_id(LIMIT_LVP_ID, 504), _time=time, lnproc=9, iops=8, io=6,
                cpu_limit=0, io_limit=19, io_fault=15, memphy_fault=11, memphy=5,
                mep_limit=18, mep_fault=12, lmemphy=21, nproc_fault=13, mem=3,
                cpu_fault=14, mem_fault=10, mep=4, nproc=7, created=0, liops=22,
                mem_limit=20, iops_fault=16, cpu=160000)

    def setUp(self):
        self.server_id = self.cfg['server_id']
        BaseDbTest.setUp(self)
        os.environ['TZ'] = 'UTC'
        tzset()
        self.mockSetup()
        self.setupLVE()
        self.setupGov()

    def tearDown(self):
        # if exception will be thrown, patcher will not unpatch, nor BaseDbTest.tearDown will not run
        self.patcher1.stop()
        self.patcher3.stop()
        self.patcher4.stop()
        self.patcher5.stop()
        BaseDbTest.tearDown(self)

        del os.environ['TZ']
        tzset()

    def mockSetup(self):
        self.patcher1 = mock.patch('lvestats.lib.cloudlinux_statistics.uid_to_username', side_effect=mock_uid_to_username)
        self.patcher3 = mock.patch("lvestats.lib.commons.func.get_governor_status", return_value=('success', None))
        self.patcher4 = mock.patch("getpass.getuser", return_value='root')  # for build-system mock environment
        self.patcher5 = mock.patch('clcommon.lib.MySQLGovernor.get_governor_status_by_username', return_value="watched")
        self.patcher1.start()
        self.patcher3.start()
        self.patcher4.start()
        self.patcher5.start()
        patcher6 = mock.patch('lvestats.lib.config.read_config', return_value={})
        patcher6.start()
        self.addCleanup(patcher6.stop)

    @ddt.data(*list(schema_case_users.items()))
    @ddt.unpack
    def test_validate_json_multiple_users(self, show_columns, schema):
        options = setup_statistics_parser(
            ["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json", "--show=" + ",".join(show_columns)], self.server_id)
        with change_timezone('UTC'):
            with mock.patch.object(MySQLGovernor, "is_governor_present", return_value=True):
                result = get_statistics(engine=self.engine, id_=None, options=options, timestamp=15, log=None)
        jsonschema.validate(result, schema)

    @ddt.data(*list(schema_case_one_user.items()))
    @ddt.unpack
    def test_validate_json_one_user_501(self, show_columns, schema):
        options = setup_statistics_parser(
            ["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json", "--show=" + ",".join(show_columns)], self.server_id)
        with change_timezone('UTC'):
            with mock.patch.object(MySQLGovernor, "is_governor_present", return_value=True):
                result = get_statistics(engine=self.engine, id_=501, options=options, timestamp=15, log=None)
        jsonschema.validate(result, schema)

    @ddt.data(*list(schema_case_one_user.items()))
    @ddt.unpack
    def test_validate_json_one_user_503(self, show_columns, schema):
        options = setup_statistics_parser(
            ["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json", "--show=" + ",".join(show_columns)], self.server_id)
        with change_timezone('UTC'):
            with mock.patch.object(MySQLGovernor, "is_governor_present", return_value=True):
                result = get_statistics(engine=self.engine, id_=503, options=options, timestamp=15, log=None)
        jsonschema.validate(result, schema)

    @ddt.data(*list(schema_case_one_user.items()))
    @ddt.unpack
    def test_validate_json_one_user_504(self, show_columns, schema):
        options = setup_statistics_parser(
            ["--from=1970-01-01 00:00", "--to=1970-01-01 00:10", "--json", "--show=" + ",".join(show_columns)], self.server_id)
        with change_timezone('UTC'):
            with mock.patch.object(MySQLGovernor, "is_governor_present", return_value=True):
                result = get_statistics(engine=self.engine, id_=504, options=options, timestamp=15, log=None)
        jsonschema.validate(result, schema)


if __name__ == '__main__':
    unittest.main()
