# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import unittest
import ddt
import sys

from datetime import datetime

from lvestats.lib.parsers.cloudlinux_statistics_parser import setup_statistics_parser


@ddt.ddt
class CloudLinuxStatisticsParserTestCase(unittest.TestCase):
    @ddt.data((['--json'], True),
              ([], False))
    @ddt.unpack
    def test_parser_json(self, argl, value):
        parsed = setup_statistics_parser(argl)
        self.assertEqual(parsed.json, value)

    @ddt.data((['--order-by', 'pmEm'], (None, 1, 'pmem', None, 90)),
              (['--by-usage', 'iops'], (None, 1, None, 'iops', 90)),
              (['--by-fault=io', '--threshold=10'], ('io', 10, None, None, 90)),
              (['--by-fault', 'cpu'], ('cpu', 1, None, None, 90)),
              (['--by-fault', 'ep', '--order-by=cpu'], ('ep', 1, 'cpu', None, 90)))
    @ddt.unpack
    def test_parser_by_fault(self, argl, value):
        parsed = setup_statistics_parser(argl)
        self.assertEqual((parsed.by_fault, parsed.threshold, parsed.order_by,
                          parsed.by_usage, parsed.percentage), value)

    @ddt.data(['--order-by', 'pmEmpmEm'],
              ['--by-usage', 'iops1'],
              ['--by-fault=acpu', '--threshold=10'],
              ['--by-usage=xx', '--percentage=10'],
              ['--by-fault', 'mysql_cpu'],
              ['--by-fault', 'epep', '--order-by=cpu'])
    def test_parser_incorrect_argument(self, test_value):
        err = sys.stderr
        sys.stderr = open("/dev/null", "w")
        with self.assertRaises(SystemExit) as cm:
            setup_statistics_parser(test_value, {})
        sys.stderr = err
        self.assertEqual(cm.exception.code, 2)

    @ddt.data((['--period=10m'], datetime(2016, 8, 13, 10, 20)),
              (['--period=4h'], datetime(2016, 8, 13, 6, 30)),
              (['--period=4d'], datetime(2016, 8, 9, 10, 30)),
              (['--period=100m'], datetime(2016, 8, 13, 8, 50)),
              (['--from=2016-08-10'], datetime(2016, 8, 10, 0, 0)),
              (['--from=2016-08-10 15:15'], datetime(2016, 8, 10, 15, 15)),
              (['--from=2016-08-20'], datetime(2016, 8, 20, 0, 0)),
              (['--from=2015-08-20'], datetime(2015, 8, 20, 0, 0)),
              (['--from=15-08-20'], datetime(2015, 8, 20, 0, 0)),
              )
    @ddt.unpack
    def test_parser_period_or_from(self, argl, start_date):
        datetime_now = datetime(2016, 8, 13, 10, 30)
        parsed = setup_statistics_parser(argl, _datetime_now=datetime_now)
        self.assertEqual(getattr(parsed, 'from'), start_date)
        self.assertEqual(parsed.to, datetime_now)

    @ddt.data((['--from=2016-08-10', '--to=2016-08-15'], datetime(2016, 8, 10), datetime(2016, 8, 15)),
              (['--from=2016-08-10 22:15', '--to=2016-08-15 21:13'],
               datetime(2016, 8, 10, 22, 15), datetime(2016, 8, 15, 21, 13)),
              (['--to=2016-08-15 21:13'], datetime(2016, 8, 13, 10, 20), datetime(2016, 8, 15, 21, 13)),
              )
    @ddt.unpack
    def test_parser_from_and_to(self, argl, start_date, to_date):
        datetime_now = datetime(2016, 8, 13, 10, 30)
        parsed = setup_statistics_parser(argl, _datetime_now=datetime_now)
        self.assertEqual(getattr(parsed, 'from'), start_date)
        self.assertEqual(parsed.to, to_date)

    def test_server_id(self):
        parsed = setup_statistics_parser([], 'server_id')
        self.assertEqual(parsed.server_id, 'server_id')

        parsed = setup_statistics_parser(['--server_id', '123'], 'server_id')
        self.assertEqual(parsed.server_id, '123')


if __name__ == '__main__':
    unittest.main(verbosity=2)
