# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import argparse
import unittest
import mock
import ddt

from typing import NamedTuple
from lvestats.lib.cloudlinux_statsnotifier import execute
from lvestats.plugins.generic.statsnotifier import get_stats_notifier_parameters


def mock_jsonify(d):
    return d


@ddt.ddt
class TestClStatsNotifier(unittest.TestCase):

    @ddt.data((True, True, 0, 1005),
              (True, True, 1005, 1005),
              (False, True, 1005, 1005),
              (True, False, 0, 1005))
    @ddt.unpack
    def test_notify_user_get(self, should_be_notified, is_allowed_notifications, current_user_uid, target_user):
        """
        Checks whether 'notifyUser' parameter is present in json
        """
        options = argparse.Namespace(json=True, id=target_user, notifyUser=None)
        with mock.patch('lvestats.lib.cloudlinux_statsnotifier.os.getuid', return_value=current_user_uid), \
                mock.patch('lvestats.lib.cloudlinux_statsnotifier.json.dumps', side_effect=mock_jsonify), \
                mock.patch('lvestats.lib.cloudlinux_statsnotifier.user_should_be_notified',
                           return_value=should_be_notified), \
                mock.patch('lvestats.lib.cloudlinux_statsnotifier.is_notifications_allowed',
                           return_value=is_allowed_notifications), \
                mock.patch('lvestats.lib.cloudlinux_statsnotifier.pwd.getpwuid',
                           return_value=mock.Mock(pw_name="user")):
            result, _ = execute(options, mock.Mock())
            self.assertEqual(result.get('result'), 'success')
            self.assertEqual(result.get('notifyUser'), should_be_notified)
            self.assertEqual(result.get('notificationsAllowed'), is_allowed_notifications)

    @ddt.data((argparse.Namespace(json=None, id=None, notifyUser=None), 0,
               'Only JSON mode supported for now'),
              (argparse.Namespace(json=True, id=None, notifyUser=None), 0,
               'Please, specify user id via --id option'),
              (argparse.Namespace(json=True, id=1005, notifyUser=None), 0,
               dict(result='success', timestamp=1, notifyUser=True)),
              (argparse.Namespace(json=True, id=1005, notifyUser=None), 1005,
               dict(result='success', timestamp=1, notifyUser=True)),
              (argparse.Namespace(json=True, id=1005, notifyUser=True), 0,
               dict(result='success', timestamp=1)),
              (argparse.Namespace(json=True, id=1005, notifyUser=False), 1005,
               dict(result='success', timestamp=1))
              )
    @ddt.unpack
    def test_check_opts(self, opts, current_user, result):
        with mock.patch('lvestats.lib.cloudlinux_statsnotifier.os.getuid', return_value=current_user), \
                mock.patch('lvestats.lib.cloudlinux_statsnotifier.json.dumps', side_effect=mock_jsonify), \
                mock.patch('lvestats.lib.cloudlinux_statsnotifier.get_notifications_settings',
                           return_value=result) as get_notifications_settings_mock, \
                mock.patch('lvestats.lib.cloudlinux_statsnotifier.manage_user_notifications',
                           return_value=result) as manage_notifications_mock:
            res, _ = execute(opts, mock.Mock())
            self.assertEqual(manage_notifications_mock.called, opts.notifyUser is not None and isinstance(res, dict))
            self.assertEqual(get_notifications_settings_mock.called, opts.notifyUser is None and isinstance(res, dict))
            self.assertEqual(res, result)

    @ddt.data((True, False),
              (False, False),
              (False, True))
    @ddt.unpack
    def test_stats_notification_parameter(self, is_root, is_default_reseller_settings):
        settings = NamedTuple('Settings', [('NOTIFY_CUSTOMERS_ON_FAULTS', bool)])
        reseller = [('root',)] if is_root else [('anyreseller',)]
        reseller_settings = settings(NOTIFY_CUSTOMERS_ON_FAULTS=False) if not is_default_reseller_settings else None
        with mock.patch('lvestats.plugins.generic.statsnotifier.cpinfo', return_value=reseller), \
            mock.patch('lvestats.plugins.generic.statsnotifier.StatsNotifier') as stats_notifier_mock, \
            mock.patch('lvestats.plugins.generic.statsnotifier.admins', return_value=['root']):

            reseller_set_mock = stats_notifier_mock.return_value.get_reseller_settings
            reseller_set_mock.return_value = reseller_settings
            default_reseller_set_mock = stats_notifier_mock.return_value.get_default_reseller_settings
            default_reseller_set_mock.return_value = settings(NOTIFY_CUSTOMERS_ON_FAULTS=False)
            admin_set_mock = stats_notifier_mock.return_value.get_admin_settings
            admin_set_mock.return_value = settings(NOTIFY_CUSTOMERS_ON_FAULTS=True)
            setting = get_stats_notifier_parameters('any')
            self.assertEqual(setting, is_root)
            self.assertEqual(default_reseller_set_mock.called, is_default_reseller_settings)
            self.assertEqual(reseller_set_mock.called, not is_root)
            self.assertEqual(admin_set_mock.called, is_root)


if __name__ == '__main__':
    unittest.main()
