# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import json
import os
import threading

import clcommon.cpapi
import clcommon.lib
import ddt
import mock

import lvestats.lib.commons.server_status
import lvestats.lib.info.cloudlinux_top
from test.base.base_db_test import BaseDbTest
from lvestats.lib.commons import func
from lvestats.lib.commons.users_manager import g_users_manager, UserNotFoundError
from lvestats.plugins.generic.persistors import CloudLinuxTopFileSaver

DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')


@ddt.ddt
class TestCloudLinuxTop(BaseDbTest):
    uid_to_username_backup = None
    get_domain_backup = None

    MAIN_FAIL_ARGS = [
        ["--max", "-4", "--json"],
        ["--max", "sdsds", "--json"],
        ["--max", "--json"],
        ["--rrrr", "--json"],
        ["-l", "--json"],
        ["qwert", "--json"],
        ["--json", "-o", "epq"],
        ["-l", "--json", "-o", "mysql_io"]]

    MAIN_CORRECT_ARGS = [
        ["--max", "4", "--json"],
        ["--hide-mysql", "--json"],
        ["--hide-mysql", "--max", "1", "--json"],
        ["--username", "qwert", "--domain", ".com", "--json"],
        ["--username", "484", "--domain", "example", "--hide-mysql", "--max", "0", "--json"],
        ["--max", "4", "--json", "-o", "cpu"],
        ["--hide-mysql", "--json", "-o", "io"],
        ["--hide-mysql", "--max", "1", "--json", "-o", "mysql_cpu"],
        ["--max", "4", "--json", "-o", "mysql_cpu"],
        ["--username", "qwert", "--domain", ".com", "--json", "-o", "mysql_io"],
        ["--username", "484", "--domain", "example", "--hide-mysql", "--max", "0", "--json", "-o", "iops"],
        ["--max", "4", "--json", "-o", "ep"],
        ["--max", "4", "--json", "-o", "nproc"],
        ["--max", "4", "--json", "-o", "pMem"],
        ["--max", "4", "--json", "-o", "MySQL_io"]]

    ORDER_BY_ARGS = [
        ("cpu",       [500, 505, 503, 502, 504, 501, 506], [1000000500, 1000000502, 1000000501]),
        ("CPU",       [500, 505, 503, 502, 504, 501, 506], [1000000500, 1000000502, 1000000501]),
        ("mySQL_cpu", [501, 503, 500, 502, 504, 505, 506], [1000000500, 1000000501, 1000000502]),
        ("io",        [500, 501, 502, 503, 504, 505, 506], [1000000500, 1000000501, 1000000502]),
        ("mysql_io",  [503, 501, 500, 502, 504, 505, 506], [1000000500, 1000000501, 1000000502]),
        ("iops",      [506, 500, 501, 502, 503, 504, 505], [1000000500, 1000000501, 1000000502]),
        ("ep",        [506, 505, 504, 503, 502, 501, 500], [1000000502, 1000000501, 1000000500]),
        ("nproc",     [502, 505, 501, 500, 504, 503, 506], [1000000502, 1000000501, 1000000500]),
        ("pmem",      [500, 505, 501, 506, 502, 503, 504], [1000000500, 1000000501, 1000000502])]

    USER_DOMAINS_RESELLERS_LIST = {
        ("mock", 'mock', "test.com"),
        ("first", 'qwert', "example.net"),
        ("second", 'mock', "example.com"),
        ("test_user", 'mock', "example.com"),
        ("user7", 'qwert', "example.net"),
        ("qwert", 'qwert', None)}

    @staticmethod
    def mock_get_username_by_id(uid):
        user_dict = {
            500: "mock",
            501: "first",
            502: "second",
            503: "test_user",
            504: "user7",
            505: "qwert",
            506: None,
            60001: None,
            # resellers
            1000000500: "reseller0",
            1000000501: "reseller1",
            1000000502: "reseller2"
        }
        return user_dict.get(uid, None)

    @staticmethod
    def mock_get_limits_by_user(username):
        user_dict = {
            "test_user": (200, 25),
            "first": (1000, 2500)}
        return user_dict.get(username, (100, 1000))

    # pylint: disable=arguments-differ
    # noinspection PyMethodOverriding
    @mock.patch("clcommon.lib.MySQLGovernor")
    def setUp(self, mock_governor):
        BaseDbTest.setUp(self)
        instance = mock_governor.return_value
        instance.get_limits_by_user.return_value = (100, 1000)
        self.clear_rows = True
        self.cl_top = lvestats.lib.info.cloudlinux_top.CloudLinuxTop(cnf=self.cfg, engine=self.engine)
        self.cl_top.fname = os.path.join(DATA_DIR, 'cloudlinux_top.json')
        self.cl_top.fname_db = os.path.join(DATA_DIR, 'cloudlinux_dbtop.json')
        self.cl_top.governor_status = "enabled"
        self.cl_top.governor_mode = "on"
        self.cl_top._get_username_by_id = self.mock_get_username_by_id
        self.cl_top.mysql_governor.get_limits_by_user = self.mock_get_limits_by_user

        self.mock_getuser = mock.patch('lvestats.lib.info.cloudlinux_top.getuser', return_value='root')
        self.mock_getuser.start()
        self.mock_get_domains = mock.patch.object(g_users_manager, '_iter_panel_users_tuples',
                                                  return_value=self.USER_DOMAINS_RESELLERS_LIST)
        self.mock_get_domains.start()
        self.mock_skip_maxuids_lve = mock.patch('clcommon.clfunc.uid_max',
                                                return_value=60000)
        self.mock_skip_maxuids_lve.start()

    def tearDown(self):
        self.mock_skip_maxuids_lve.stop()
        self.mock_getuser.stop()
        self.mock_get_domains.stop()
        g_users_manager.users_cache = dict()  # clear cache

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    def test_get_data_limit(self, mock_server_status):
        data, _ = self.cl_top._get_data()
        self.assertEqual(len(data), 7)
        data, _ = self.cl_top._get_data(limit=2)
        self.assertEqual(len(data), 2)
        data, _ = self.cl_top._get_data(limit=200)
        self.assertEqual(len(data), 7)

    @ddt.data((True, []), (False, [60001]))
    @ddt.unpack
    def test_get_data_filter_uids(self, to_filter_by_maxuid, expected_list):
        self.cl_top.hide_maxuid_lve = to_filter_by_maxuid
        with mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0):
            data, _ = self.cl_top._get_data()
        filtered = [user for user in data if user['id'] > 60000]
        filtered_ids = [usr.get('id') for usr in filtered]
        self.assertEqual(filtered_ids, expected_list)

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    def test_get_data_without_admins(self, mock_server_status):
        with mock.patch('lvestats.lib.info.cloudlinux_top.cpapi.admins', return_value={'first', 'second'}):
            self.cl_top._admins = self.cl_top._get_admins()
            data, _ = self.cl_top._get_data()
            self.assertEqual(len(data), 5)
            data, _ = self.cl_top._get_data(limit=5)
            self.assertEqual(len(data), 5)
            data, _ = self.cl_top._get_data(limit=4)
            self.assertEqual(len(data), 4)

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    def test_get_data_limit_resellers(self, mock_server_status):
        """Resellers should not be limited"""
        _, data = self.cl_top._get_data()
        self.assertEqual(len(data), 3)
        _, data = self.cl_top._get_data(limit=2)
        self.assertEqual(len(data), 3)
        _, data = self.cl_top._get_data(limit=200)
        self.assertEqual(len(data), 3)

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    def test_get_data_limit_governor_mode(self, mock_server_status):
        self.cl_top.governor_status = "enabled"
        data, _ = self.cl_top._get_data(limit=2, show_mysql=True)
        self.assertEqual(len(data), 2)

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    def test_get_data_username(self, mock_server_status):
        data, _ = self.cl_top._get_data(username_filter="user")
        self.assertEqual(len(data), 2)
        self.assertTrue(data[0]["id"] in (503, 504))
        self.assertTrue(data[1]["id"] in (503, 504))
        data, _ = self.cl_top._get_data(limit=3, username_filter="s")
        self.assertEqual(len(data), 3)

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    def test_get_data_domain(self, mock_server_status):
        data, _ = self.cl_top._get_data(domain_filter=".com")
        self.assertEqual(len(data), 3)
        data, _ = self.cl_top._get_data(domain_filter="test")
        self.assertEqual(len(data), 1)
        data, _ = self.cl_top._get_data(domain_filter="example.com", username_filter="user")
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]["id"], 503)
        data, _ = self.cl_top._get_data(
            domain_filter="example",
            username_filter="user",
            limit=1)
        self.assertEqual(len(data), 1)

    @ddt.data(
        ("all", 'test_user', (10, 20, "-", "-")),
        ("all", 'first', (268.5008795263993, 19, "-", "-")),
        ("abusers", 'test_user', (10, 21, 200, 25600)),
        ("abusers", 'first', (1000, 19, 1000, 2560000)),
        ("on", 'test_user', (10, 21, 200, 25600)),
        ("on", 'first', (1000, 19, 1000, 2560000)),
        ("all", 'unknown', (0, 0, "-", "-")),
        ("abusers", 'unknown', (0, 0, 100, 1024000)))
    @ddt.unpack
    def test_add_mysql_data(self, governor_mode, username, result):
        cl_data = {
            u'dbgov_data': {
                u'test_user': {u'cause_of_restrict': u'-', u'time_of_restrict': 0, u'cpu': 10, u'io': 21},
                u'first': {u'cause_of_restrict': u'write', u'time_of_restrict': 60, u'cpu': 1000, u'io': 19}}}
        user = {
            u'usage': {u'io': {u'all': 20.0}, u'mem': 43.0, u'ep': 1.0, u'iops': 0.0, u'pno': 3.0, u'cpu': {u'all': 268.5008795263993}},
            'username': 'mock', 'domain': 'test.com',
            u'limit': {u'io': {u'all': 1048576}, u'mem': 0, u'ep': 20, u'iops': 1024, u'pno': 100, u'cpu': {u'all': 2700}}, u'id': 500}
        self.cl_top.governor_mode = governor_mode
        self.cl_top._add_mysql_data(user, cl_data, username)
        self.assertEqual(
            (
                user["usage"]["cpu"]["mysql"],
                user["usage"]["io"]["mysql"],
                user["limit"]["cpu"]["mysql"],
                user["limit"]["io"]["mysql"]),
            result)

    def assertHaveMysqlData(self, user):
        error_msg = "user {0} (id={1}) have no mysql data".format(user["username"], user["id"])
        self.assertTrue("restricted" in user, error_msg)
        self.assertTrue("mysql" in user["usage"]["cpu"], error_msg)
        self.assertTrue("mysql" in user["usage"]["io"], error_msg)
        self.assertTrue("mysql" in user["limit"]["cpu"], error_msg)
        self.assertTrue("mysql" in user["limit"]["io"], error_msg)
        if user["username"] in ("first", "test_user"):
            self.assertNotEqual(user["limit"]["cpu"]["mysql"], 100)
            self.assertNotEqual(user["limit"]["io"]["mysql"], 1000 * 1024)
        else:
            self.assertEqual(user["limit"]["cpu"]["mysql"], 100)
            self.assertEqual(user["limit"]["io"]["mysql"], 1000 * 1024)

    def assertDontHaveMysqlData(self, user):
        error_msg = "user {0} (id={1}) have mysql data".format(user["username"], user["id"])
        self.assertFalse("restricted" in user, error_msg)
        self.assertFalse("mysql" in user["usage"]["cpu"], error_msg)
        self.assertFalse("mysql" in user["usage"]["io"], error_msg)
        self.assertFalse("mysql" in user["limit"]["cpu"], error_msg)
        self.assertFalse("mysql" in user["limit"]["io"], error_msg)

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    def test_get_data_mysql(self, mock_server_status):
        self.cl_top.governor_status = "enabled"
        data, _ = self.cl_top._get_data(show_mysql=True)
        for user in data:
            self.assertHaveMysqlData(user)
            if user["id"] in (501, 503):
                self.assertGreater(user["usage"]["cpu"]["mysql"], 0)
                self.assertGreater(user["usage"]["io"]["mysql"], 0)
                if user["id"] == 501:
                    self.assertTrue(user["restricted"])
                else:
                    self.assertFalse(user["restricted"])
            else:
                self.assertEqual(user["usage"]["cpu"]["mysql"], 0)
                self.assertEqual(user["usage"]["io"]["mysql"], 0)
                self.assertFalse(user["restricted"])

                data, _ = self.cl_top._get_data(show_mysql=False)
        for user in data:
            self.assertDontHaveMysqlData(user)

        self.cl_top.governor_status = "disabled"
        data, _ = self.cl_top._get_data(show_mysql=True)
        for user in data:
            self.assertDontHaveMysqlData(user)

        data, _ = self.cl_top._get_data(show_mysql=False)
        for user in data:
            self.assertDontHaveMysqlData(user)

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    @ddt.data(*MAIN_FAIL_ARGS)
    def test_main_fail(self, value, mock_server_status):
        result, _ = self.cl_top.main(*value)
        result = json.loads(result)
        self.assertTrue("timestamp" in result)
        self.assertNotEqual(result["result"], "success")
        self.assertNotIn("mySqlGovMode", result)

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    @ddt.data(*MAIN_CORRECT_ARGS)
    def test_main_success(self, value, mock_server_status):
        self.cl_top.governor_mode = "abusers"
        result, _ = self.cl_top.main(*value)
        result = json.loads(result)
        self.assertTrue("timestamp" in result)
        self.assertEqual(result["result"], "success")
        self.assertEqual(result["mySqlGovMode"], "abusers")

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    @ddt.data(*ORDER_BY_ARGS)
    @ddt.unpack
    def test_main_order_by(self, value, order, resellers_order, mock_server_status):
        result, _ = self.cl_top.main("--json", "-o", value)
        result = json.loads(result)
        self.assertTrue("timestamp" in result)
        self.assertEqual(result["result"], "success")
        self.assertEqual(result["mySqlGovMode"], "on")
        self.assertEqual([item["id"] for item in result["users"]], order)
        self.assertEqual([item["id"] for item in result["resellers"]], resellers_order)

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=1)
    @ddt.data(*MAIN_CORRECT_ARGS)
    def test_main_stopped_server(self, value, mock_server_status):
        result, _ = self.cl_top.main(*value)
        result = json.loads(result)
        self.assertTrue("timestamp" in result)
        self.assertNotEqual(result["result"], "success")
        self.assertNotIn("mySqlGovMode", result)

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=2)
    @ddt.data(*MAIN_CORRECT_ARGS)
    def test_main_dead_server(self, value, mock_server_status):
        result, _ = self.cl_top.main(*value)
        result = json.loads(result)
        self.assertTrue("timestamp" in result)
        self.assertNotEqual(result["result"], "success")
        self.assertNotIn("mySqlGovMode", result)

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    @ddt.data(*MAIN_CORRECT_ARGS)
    def test_main_domain_exeption(self, value, *args):
        def mock_get_domain(username, raise_exc=True):
            if raise_exc:
                raise UserNotFoundError({
                           'message': "An error occured while getting domain for user %(username)s: %(error)s",
                           'context': {'username': 'user001', 'error': 'very bad error'}})

        with mock.patch.object(g_users_manager, 'get_domain', side_effect=mock_get_domain):
            result, _ = self.cl_top.main(*value)
        result = json.loads(result)
        self.assertTrue("timestamp" in result)
        if "--domain" in value:
            self.assertNotEqual(result["result"], "success")
        else:
            self.assertEqual(result["result"], "success")
            for user in result["users"]:
                self.assertEqual(user["domain"], "N/A")

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    def test_wrong_db_type(self, mock_server_status):
        cfg = {
            'db_type': 'sqlite1111',
            'sqlite_db_path': '',
            'connect_string': 'root:root@localhost',
            'server_id': 'localhost'
        }
        cl_top = lvestats.lib.info.cloudlinux_top.CloudLinuxTop(cnf=cfg)
        result, _ = cl_top.main("--json")
        result = json.loads(result)
        self.assertIn("sqlite1111", result["result"] % result["context"])

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    @ddt.data(*MAIN_CORRECT_ARGS)
    def test_without_json(self, value, mock_server_status):
        """LVES-490"""
        value.remove("--json")
        result, exitcode = self.cl_top.main(*value)
        self.assertIn("Use --json option, other modes currently unsupported", result)
        self.assertEqual(1, exitcode)

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    def test_get_statistics_about_other_reseller_under_reseller(self, mock_server_status):
        with mock.patch('lvestats.lib.info.cloudlinux_top.getuser', return_value='reseller1'):
            result, exitcode = self.cl_top.main('--json', '--for-reseller', 'reseller2')
        self.assertIn("Permission denied. Reseller can view information only about himself. "
                      "Option --for-reseller=reseller2 is forbidden.", result)
        self.assertEqual(1, exitcode)

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    @mock.patch('lvestats.lib.info.cloudlinux_top.get_users_for_reseller', return_value=['test_user', 'second', 'mock'])
    def test_get_statistics_about_current_reseller_under_reseller(self, mock_server_status, mock_get_users_for_reseller):
        with mock.patch('lvestats.lib.info.cloudlinux_top.getuser', return_value='reseller0'):
            result, exitcode = self.cl_top.main('--json', '--for-reseller', 'reseller0')
        self.assertNotIn("Permission denied.", result)
        self.assertEqual(0, exitcode)

        result = json.loads(result)
        self.assertEqual(len(result['resellers']), 0)
        self.assertEqual(len(result['users']), 3)

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    def test_get_statistics_about_other_reseller_under_root(self, mock_server_status):
        with mock.patch('lvestats.lib.info.cloudlinux_top.getuser', return_value='root'):
            with mock.patch('lvestats.lib.info.cloudlinux_top.CloudLinuxTop._get_data', return_value=([], [])) as mock_get_data:
                result, exitcode = self.cl_top.main('--json', '--for-reseller', 'reseller2')
                self.assertEqual(mock_get_data.call_args[1]['reseller'], 'reseller2')
        self.assertNotIn("Permission denied.", result)
        self.assertEqual(0, exitcode)

    @mock.patch.object(clcommon.lib.MySQLGovernor, "is_governor_present", return_value=True)
    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    @ddt.data(
        (
            clcommon.lib.MySQLGovException({
                'message': 'Message. %(error)s',
                'context': {'error': 'very bad error'}}),
            ('Message. %(error)s', {'error': 'very bad error'})))
    @ddt.unpack
    def test_mysql_gov_exception(self, dbgov_error, expected_result, *args):
        """LVES-665"""
        with mock.patch.object(clcommon.lib.MySQLGovernor, "_load_info", side_effect=dbgov_error):
            cl_top = lvestats.lib.info.cloudlinux_top.CloudLinuxTop(cnf=self.cfg, engine=self.engine)
            cl_top.mysql_governor._is_governor_present = True
            cl_top._get_data = mock.MagicMock(return_value=([], []))
            result, exitcode = cl_top.main("--json")
            result = json.loads(result)
            self.assertEqual(0, exitcode)
            self.assertEqual((result["warning"], result["context"]), expected_result)

    @ddt.data(('root', None, 7),  # root gets all users
              ('root', 'mock', 3),  # root gets all users for reseller
              ('mock', 'mock', 3),  # only own users
              ('qwert', 'qwert', 2),  # only own users
              ('user7', 'user7', 1),  # current user sees only himself
              ('root', 'empty_reseller', 0),  # root gets all users for reseller with no users
              ('empty_reseller', 'empty_reseller', 0))  # resellers with no users does not get any
    @ddt.unpack
    def test_get_resellers_users_data(self, current_user, reseller, expected_count):
        def mocked_get_users_for_reseller(reseller_name):
            if reseller_name == 'mock':
                return ['test_user', 'second', 'mock']
            elif reseller_name == 'qwert':
                return ['user7', 'qwert']
            elif reseller_name == 'empty_reseller':
                return []
            else:
                return [reseller_name]

        with mock.patch('lvestats.lib.info.cloudlinux_top.getuser', return_value=current_user), \
                mock.patch('lvestats.lib.info.cloudlinux_top.get_users_for_reseller',
                           side_effect=mocked_get_users_for_reseller):
            data, _ = self.cl_top._get_data(reseller=reseller)
            self.assertEqual(len(data), expected_count)

    @mock.patch('lvestats.lib.info.cloudlinux_top.get_users_for_reseller')
    def test_get_resellers_containers_data(self, mock_get_users_for_reseller):
        mock_get_users_for_reseller.return_value = []

        # get value under root, get all resellers
        _, data = self.cl_top._get_data(reseller=None)
        self.assertEqual(len(data), 3)

        # get value under first reseller
        _, data = self.cl_top._get_data(reseller='reseller0')
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['name'], 'reseller0')

        # get value under other reseller
        _, data = self.cl_top._get_data(reseller='reseller1')
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['name'], 'reseller1')

        # get value under non-reseller
        _, data = self.cl_top._get_data(reseller='nonreseller')
        self.assertEqual(len(data), 0)

    @ddt.data(
        (
            [OSError(), OSError()],
            {"users": [], "dbgov_data": {}, 'resellers': []}),
        (
            [IOError(), IOError()],
            {"users": [], "dbgov_data": {}, 'resellers': []}),
        (
            [IOError(), {"dbgov_data": {"result1": "result2"}}],
            {"users": [], 'resellers': [], "dbgov_data": {"result1": "result2"}}),
        (
            [{"users": ["result1", "result2"], 'resellers': []}, OSError()],
            {"users": ["result1", "result2"], 'resellers': [], "dbgov_data": {}}),
        (
            [{"users": ["result1", "result2"], 'resellers': []}, {"dbgov_data": {"result1": "result2"}}],
            {"users": ["result1", "result2"], 'resellers': [], "dbgov_data": {"result1": "result2"}}))
    @ddt.unpack
    def test_load_from_json_file(self, side_effects, expected_result):
        with mock.patch("json.load", side_effect=side_effects):
            result = self.cl_top._load_data()
        self.assertEqual(result, expected_result)

    def test_atomic_write(self):
        atomic_test_filename = os.path.join(DATA_DIR, 'cloudlinux_top_atomic_test.json')
        cl_top_filesaver = CloudLinuxTopFileSaver(fname=atomic_test_filename)
        cl_top = lvestats.lib.info.cloudlinux_top.CloudLinuxTop(cnf=self.cfg, engine=self.engine)
        cl_top.fname = atomic_test_filename

        def writer(result):
            data_for_saving = [{1: 6, "qwerty": "asdfhgjhgjhg", 864: 928921, 123456: 987654321}] * 50000
            try:
                for _ in range(0, 25):
                    cl_top_filesaver._output({"users": data_for_saving})
            except Exception as e:
                result["writer_exception"] = e

        def reader(result):
            try:
                for _ in range(0, 25):
                    cl_top._load_from_json_file(fname=atomic_test_filename)
            except Exception as e:
                result["reader_exception"] = e

        treads_results = {}
        writer_tread = threading.Thread(target=writer, args=(treads_results,))
        reader_tread = threading.Thread(target=reader, args=(treads_results,))
        writer_tread.start()
        reader_tread.start()
        writer_tread.join()
        reader_tread.join()
        if "writer_exception" in treads_results:
            raise treads_results["writer_exception"]
        if "reader_exception" in treads_results:
            raise treads_results["reader_exception"]

    def test_get_data_reseller_field_exists(self):
        users_data, _ = self.cl_top._get_data()
        user_reseller_map = {user: reseller for user, reseller, _ in self.USER_DOMAINS_RESELLERS_LIST}
        for user in users_data:
            self.assertEqual(user['reseller'],
                             user_reseller_map.get(user['username'], 'N/A'))

    @mock.patch('lvestats.lib.commons.server_status.check_server_status', return_value=0)
    def test_get_statistics_about_current_user_under_user(self, mock_server_status):
        with mock.patch('lvestats.lib.info.cloudlinux_top.getuser', return_value='test_user'):
            result, exitcode = self.cl_top.main('--json')
        self.assertNotIn("Permission denied.", result)
        self.assertEqual(0, exitcode)

        result = json.loads(result)
        self.assertEqual(len(result['resellers']), 0)
        self.assertEqual(len(result['users']), 1)


class TestFunc(BaseDbTest):
    @mock.patch('clcommon.cpapi.userdomains', side_effect=TypeError())
    def test_get_domain(self, _):
        self.assertRaises(func.DomainException, func.get_domain, "mock")
