# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#

# pylint: disable=no-absolute-import
# pylint: disable=old-division

import json
import unittest
import mock
import ddt

from lvestats.lib.commons.sizeutil import mempages_to_bytes
from lvestats.plugins.generic.cm_collector import CMCollector, CM_DISABLED_MSG
from lvestats.plugins.generic.analyzers import LVEUsage
from lvestats.lib.commons.func import DomainException
from clcommon.lib import MySQLGovException, MySQLGovernorAbsent

USER_NAME_UID_MAP = (
    ('user1', 1001),
    ('user2', 1002),
    ('user3', 1003),
    ('user4', 1004),
)

USER_DOMAINS = {
    'user1': 'user1.com',
    'user2': 'user2.com',
    'user3': 'user3.com',
    'user4': 'user4.com',
}


def pwd_getpwuid_mock(uid):
    for name_, id_ in USER_NAME_UID_MAP:
        if uid == id_:
            pwd_mock = mock.MagicMock()
            pwd_mock.pw_name = name_
            return pwd_mock
    raise KeyError()


def get_domains_mock(username, raise_exception=True):
    try:
        domain = USER_DOMAINS[username]
    except KeyError:
        raise DomainException("No domains for user {}".format(username))
    return [domain]


_warn_msg = ''


def _add_to_warn_log(msg, *args):
    global _warn_msg
    _warn_msg += msg


@ddt.ddt
class TestCMCollector(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None
        self.plugin = CMCollector()
        self.plugin.domains_cache.cache = {}
        lve_usage_1001 = LVEUsage(8)
        lve_usage_1001.cpu_usage = 3984             # CPU average
        lve_usage_1001.lcpu = 4000                  # CPU limit
        lve_usage_1001.cpu_fault = 4                # CPU faults
        lve_usage_1001.mep = 0                      # EP average
        lve_usage_1001.lep = 20                     # EP limit
        lve_usage_1001.mep_fault = 0                # EP faults
        lve_usage_1001.memphy = 110522              # PMEM average in 4K pages
        lve_usage_1001.lmemphy = 262144             # PMEM limit in 4K pages
        lve_usage_1001.memphy_fault = 0             # PMEM faults
        lve_usage_1001.nproc = 25                   # NPROC average
        lve_usage_1001.lnproc = 100                 # NPROC limit
        lve_usage_1001.nproc_fault = 0              # NPROC faults
        lve_usage_1001.io_usage = 0                 # IO average
        lve_usage_1001.io = 1048576                 # IO limit
        lve_usage_1001.io_fault = 0                 # IO faults
        lve_usage_1001.iops = 3                     # IOPS average
        lve_usage_1001.liops = 1024                 # IOPS limit
        lve_usage_1001.iops_fault = 0               # IOPS faults

        lve_usage_1002 = LVEUsage(8)
        lve_usage_1002.cpu_usage = 4500             # CPU average
        lve_usage_1002.lcpu = 5000                  # CPU limit
        lve_usage_1002.cpu_fault = 56                # CPU faults
        lve_usage_1002.mep = 6                      # EP average
        lve_usage_1002.lep = 80                     # EP limit
        lve_usage_1002.mep_fault = 1               # EP faults
        lve_usage_1002.memphy = 110512              # PMEM average in 4K pages
        lve_usage_1002.lmemphy = 262200             # PMEM limit in 4K pages
        lve_usage_1002.memphy_fault = 4             # PMEM faults
        lve_usage_1002.nproc = 30                   # NPROC average
        lve_usage_1002.lnproc = 50                 # NPROC limit
        lve_usage_1002.nproc_fault = 4              # NPROC faults
        lve_usage_1002.io_usage = 100                 # IO average
        lve_usage_1002.io = 1108576                 # IO limit
        lve_usage_1002.io_fault = 6                 # IO faults
        lve_usage_1002.iops = 789                     # IOPS average
        lve_usage_1002.liops = 2048                 # IOPS limit
        lve_usage_1002.iops_fault = 20               # IOPS faults

        self.const_cpu_l, self.const_io_l, self.const_pmem_l = 10000, 1024, 262144
        self.const_ep_l, self.const_nproc_l, self.const_iops_l = 20, 100, 1024
        self.const_mysql_cpu_l, self.const_mysql_io_l = 100, 1024

        # content of /proc/lve/list to be read by lvestats
        self.proclve_lines = [
            f'0\t0\t{self.const_cpu_l}\t{self.const_io_l}\t0\t0\t0\t0\t{self.const_ep_l}\t1'
            f'\t0\t0\t{self.const_pmem_l}\t100\t{self.const_nproc_l}\t0'
            f'\t0\t0\t0\t{self.const_iops_l}\t0\t0\t0\t0\t0',

            f'1001\t0\t{lve_usage_1001.lcpu}\t{lve_usage_1001.io}\t0\t0\t0\t0\t{lve_usage_1001.lep}\t1'
            f'\t0\t0\t{lve_usage_1001.lmemphy}\t100\t{lve_usage_1001.lnproc}\t0'
            f'\t0\t0\t0\t{lve_usage_1001.liops}\t0\t0\t0\t0\t0',

            f'1002\t0\t{lve_usage_1002.lcpu}\t{lve_usage_1002.io}\t0\t0\t0\t0\t{lve_usage_1002.lep}\t1'
            f'\t0\t0\t{lve_usage_1002.lmemphy}\t100\t{lve_usage_1002.lnproc}\t0'
            f'\t0\t0\t0\t{lve_usage_1002.liops}\t0\t0\t0\t0\t0',

            f'1003\t0\t{self.const_cpu_l}\t{self.const_io_l}\t0\t0\t0\t0\t{self.const_ep_l}\t1'
            f'\t0\t0\t{self.const_pmem_l}\t100\t{self.const_nproc_l}\t0'
            f'\t0\t0\t0\t{self.const_iops_l}\t0\t0\t0\t0\t0',

            f'1004\t0\t{self.const_cpu_l}\t{self.const_io_l}\t0\t0\t0\t0\t{self.const_ep_l}\t1'
            f'\t0\t0\t{self.const_pmem_l}\t100\t{self.const_nproc_l}\t0'
            f'\t0\t0\t0\t{self.const_iops_l}\t0\t0\t0\t0\t0',
        ]

        self.dbgov_data = {
            USER_NAME_UID_MAP[1][1]: {
                "cpu_usage": 7.8, "cpu_limit": 100,
                "io_usage": 10151, "io_limit": 562036736
            },
        }
        self.lve_data = {
            "lve_usage": {
                USER_NAME_UID_MAP[0][1]: lve_usage_1001,
                USER_NAME_UID_MAP[1][1]: lve_usage_1002,
                10000: {}
            },
            "dbgov_data_for_cm": self.dbgov_data
        }

        self.expected_dict_file = {
            "lve_stats": [
                {"lve_id": USER_NAME_UID_MAP[0][1], "username": USER_NAME_UID_MAP[0][0],
                 "domain": get_domains_mock(USER_NAME_UID_MAP[0][0])[0],
                 "cpu": {"a": int(round(lve_usage_1001.cpu_usage))/100,                 # pylint: disable=round-builtin
                         "l": lve_usage_1001.lcpu/100,
                         "f": lve_usage_1001.cpu_fault},
                 "ep": {"a": lve_usage_1001.mep, "l": lve_usage_1001.lep,
                        "f": lve_usage_1001.mep_fault},
                 "pmem": {"a": mempages_to_bytes(int(round(lve_usage_1001.memphy))),    # pylint: disable=round-builtin
                          "l": mempages_to_bytes(lve_usage_1001.lmemphy),
                          "f": lve_usage_1001.memphy_fault},
                 "nproc": {"a": int(round(lve_usage_1001.nproc)),                       # pylint: disable=round-builtin
                           "l": lve_usage_1001.lnproc,
                           "f": lve_usage_1001.nproc_fault},
                 "io": {"a": int(round(lve_usage_1001.io_usage)),                       # pylint: disable=round-builtin
                        "l": lve_usage_1001.io,
                        "f": lve_usage_1001.io_fault},
                 "iops": {"a": int(round(lve_usage_1001.iops)),                         # pylint: disable=round-builtin
                          "l": lve_usage_1001.liops,
                          "f": lve_usage_1001.iops_fault},
                 "mysql_cpu": {"l": self.const_mysql_cpu_l},
                 "mysql_io": {"l": self.const_mysql_io_l * 1024}
                 },
                {"lve_id": USER_NAME_UID_MAP[1][1], "username": USER_NAME_UID_MAP[1][0],
                 "domain": get_domains_mock(USER_NAME_UID_MAP[1][0])[0],
                 "cpu": {"a": int(round(lve_usage_1002.cpu_usage)) / 100,               # pylint: disable=round-builtin
                         "l": lve_usage_1002.lcpu / 100,
                         "f": lve_usage_1002.cpu_fault},
                 "ep": {"a": lve_usage_1002.mep, "l": lve_usage_1002.lep,
                        "f": lve_usage_1002.mep_fault},
                 "pmem": {"a": mempages_to_bytes(int(round(lve_usage_1002.memphy))),    # pylint: disable=round-builtin
                          "l": mempages_to_bytes(lve_usage_1002.lmemphy),
                          "f": lve_usage_1002.memphy_fault},
                 "nproc": {"a": int(round(lve_usage_1002.nproc)),                       # pylint: disable=round-builtin
                           "l": lve_usage_1002.lnproc,
                           "f": lve_usage_1002.nproc_fault},
                 "io": {"a": int(round(lve_usage_1002.io_usage)),                       # pylint: disable=round-builtin
                        "l": lve_usage_1002.io,
                        "f": lve_usage_1002.io_fault},
                 "iops": {"a": int(round(lve_usage_1002.iops)),                         # pylint: disable=round-builtin
                          "l": lve_usage_1002.liops,
                          "f": lve_usage_1002.iops_fault},
                 "mysql_cpu": {"a": self.dbgov_data[USER_NAME_UID_MAP[1][1]]["cpu_usage"],
                               "l": self.dbgov_data[USER_NAME_UID_MAP[1][1]]["cpu_limit"]},
                 "mysql_io": {"a": self.dbgov_data[USER_NAME_UID_MAP[1][1]]["io_usage"],
                              "l": self.dbgov_data[USER_NAME_UID_MAP[1][1]]["io_limit"]}
                 }

            ]
        }

        self.lve_data_without_usage = {
            'cpu': {
                'l': self.const_cpu_l / 100
            },
            'ep': {
                'l': self.const_ep_l
            },
            'pmem': {
                'l': mempages_to_bytes(self.const_pmem_l)
            },
            'nproc': {
                'l': self.const_nproc_l
            },
            'io': {
                'l': self.const_io_l * 1024
            },
            'iops': {
                'l': self.const_iops_l
            },
            'mysql_cpu': {
                'l': self.const_mysql_cpu_l
            },
            'mysql_io': {
                'l': self.const_mysql_io_l * 1024
            }
        }

        for user in [USER_NAME_UID_MAP[2], USER_NAME_UID_MAP[3]]:
            user_data = {
                'lve_id': user[1],
                'username': user[0],
                'domain': get_domains_mock(user[0])[0]
            }
            user_data.update(self.lve_data_without_usage)
            self.expected_dict_file['lve_stats'].append(user_data)

    def test_01_execute(self):
        """
        Test for execute function
        """
        get_domains_mock_calls = [mock.call(user[0], raise_exception=False)for user in USER_NAME_UID_MAP]
        _write_file_mock_calls = [mock.call(json.dumps(self.expected_dict_file), '/var/lve/cm_lve.json', 0o600)]
        with mock.patch("lvestats.plugins.generic.cm_collector.pwd.getpwuid",
                        side_effect=pwd_getpwuid_mock) as _get_pwuid_mock, \
                mock.patch("lvestats.plugins.generic.cm_collector.jwt_token_check",
                           return_value=(True, "OK", None)) as _jwt_token_mock, \
                mock.patch("lvestats.plugins.generic.cm_collector.get_domains",
                           side_effect=get_domains_mock) as _get_domains_mock, \
                mock.patch("lvestats.plugins.generic.cm_collector.ProcLve.lines",
                           return_value=self.proclve_lines), \
                mock.patch("lvestats.plugins.generic.cm_collector.ProcLve.version",
                           return_value=9), \
                mock.patch("lvestats.plugins.generic.cm_collector.is_cmt_disabled",
                           return_value=False), \
                mock.patch("lvestats.plugins.generic.cm_collector.is_client_enabled",
                           return_value=True), \
                mock.patch("lvestats.plugins.generic.cm_collector.MySQLGovernor.get_limits_by_user",
                           return_value=(self.const_mysql_cpu_l, self.const_mysql_io_l)), \
                mock.patch("lvestats.plugins.generic.cm_collector.write_file_via_tempfile") as _write_file_mock:
            self.plugin.execute(self.lve_data)
            self.assertListEqual(_get_domains_mock.mock_calls, get_domains_mock_calls)
            self.assertListEqual(_write_file_mock.mock_calls, _write_file_mock_calls)
            self.assertFalse('dbgov_data_for_cm' in self.lve_data)
            _jwt_token_mock.assert_called_once_with()

    def test_02_execute_bad_jwt_token(self):
        """
        Test for execute function. Error case. Invalid jwt token
        """
        global _warn_msg
        _warn_msg = ''
        token_msg = "Bad token"
        self.plugin.log = mock.Mock()
        self.plugin.log.info = _add_to_warn_log
        with mock.patch("lvestats.plugins.generic.cm_collector.jwt_token_check",
                        return_value=(False, token_msg, None)) as _jwt_token_mock, \
                mock.patch("lvestats.plugins.generic.cm_collector.get_domains") as _get_domains_mock, \
                mock.patch("lvestats.plugins.generic.cm_collector.write_file_via_tempfile") as _write_file_mock:
            self.plugin.execute(self.lve_data)
            _jwt_token_mock.assert_called_once_with()
            _get_domains_mock.assert_not_called()
            _write_file_mock.assert_not_called()
            self.assertTrue(self.plugin._is_jwt_token_error)
            self.assertEqual(_warn_msg, "JWT token error: '%s'. CM Collector plugin will not work, "
                                        "it is expected behaviour for non CL+ clients")

    @ddt.data(
        (True, True),
        (False, False),
    )
    @ddt.unpack
    def test_03_execute_cm_is_not_enabled(self, _is_cmt_disabled,
                                          _is_client_enabled):
        """
        Test for execute function. Error case. CM isn't enabled
        """
        mock_log_info = self.plugin.log.info = mock.Mock()
        with mock.patch("lvestats.plugins.generic.cm_collector.jwt_token_check",
                        return_value=(True, "OK", None)), \
                mock.patch("lvestats.plugins.generic.cm_collector.is_cmt_disabled",
                           return_value=_is_cmt_disabled) as _cmt_dis_mock, \
                mock.patch("lvestats.plugins.generic.cm_collector.is_client_enabled",
                           return_value=_is_client_enabled) as _client_mock, \
                mock.patch("lvestats.plugins.generic.cm_collector.get_domains") as _get_domains_mock, \
                mock.patch("lvestats.plugins.generic.cm_collector.write_file_via_tempfile") as _write_file_mock:
            self.plugin.execute(self.lve_data)
            _cmt_dis_mock.assert_called_once()
            # is_client_enabled will be called only if cmt isn't disabled
            if not _is_cmt_disabled:
                _client_mock.assert_called_once()
            else:
                _client_mock.assert_not_called()
            _get_domains_mock.assert_not_called()
            _write_file_mock.assert_not_called()
            self.assertTrue(self.plugin._is_cm_enabled_error)
            mock_log_info.assert_called_once_with(CM_DISABLED_MSG)

    @ddt.data(
        (MySQLGovernorAbsent(), True),
        (MySQLGovException("some exception"), False),
    )
    @ddt.unpack
    def test_04_prepare_dbgov_limits_exception(self, _exc, _is_governor_absent):
        """
        Checks that _prepare_dbgov_limits processes exceptions correctly

        :param _exception: exception instance.
            - If there will be multiple MySQLGovernorAbsent errors: log
             should be called only once.
            - If there will be multiple basic MySQLGovException errors:
             log should work every.
        :param _is_governor_absent: expected result of _is_governor_absent_error
        for a plugin
        """
        # Count of method calls
        attempts = 2
        mock_log_warn = self.plugin.log.warning = mock.Mock()
        with mock.patch("lvestats.plugins.generic.cm_collector.MySQLGovernor.get_limits_by_user",
                        side_effect=_exc):
            # Make multiple calls to _prepare_dbgov_limits
            for _ in range(attempts):
                self.assertIsNone(self.plugin._prepare_dbgov_limits("user"))
            self.assertEqual(_is_governor_absent,
                             self.plugin._is_governor_absent_error)
            if _is_governor_absent:
                # Ensure that warning log was called only once
                mock_log_warn.assert_called_once()
            else:
                self.assertEqual(attempts, mock_log_warn.call_count)

    def test_user_domains_cached(self):
        with mock.patch("lvestats.plugins.generic.cm_collector.get_domains", side_effect=get_domains_mock) \
                as _get_domains_mock:
            self.plugin.domains_cache.get('user1')
            self.plugin.domains_cache.get('user1')

            self.plugin.domains_cache.get('unknown')

        # only 2 calls
        self.assertTrue(_get_domains_mock.call_count, 2)
        self.assertTrue(_get_domains_mock.calls, [mock.call('user1', raise_exception=False),
                                                  mock.call('unknown', raise_exception=False)])
        print(_get_domains_mock.mock_calls)


if __name__ == '__main__':
    unittest.main()
