# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import time
import unittest
from io import StringIO

import mock
from lvestat import LVEStat
from sqlalchemy.pool import AssertionPool

from lvestats.lib.dbengine import make_db_engine, create_db
from lvestats.plugins import default_plugins
from lvestats.plugins.generic.persistors import FileSaver, CloudLinuxTopFileSaver
from lvestats.plugins.generic.statsnotifier import StatsNotifier
from lvestats.plugins.generic.collectors import LVECollector, CPUInfoCollector
from lvestats.plugins.generic.lvestats_memory_usage import LVEStatsMemoryUsage
from lvestats.plugins.generic.burster import LveLimitsBurster


LVE_VERSION = 8


class MockLVECollector(LVECollector):
    now = 0

    def execute(self, lve_data):
        stats = {}
        lve_data['LVE_VERSION'] = LVE_VERSION
        for lve_id in range(1, 5):
            stat = LVEStat(None, LVE_VERSION)

            stat.id = lve_id
            stat.mep = 1
            stat.cpu = 10  # cpu limit in [1...10000*ncpu]
            stat.ncpu = 1
            stat.cpu_usage = self.now
            stat.nproc_fault = self.now

            stats[stat.id] = stat
        lve_data['old_stats'] = lve_data.get('stats', {})
        lve_data['stats'] = stats


class MockCpuInfoCollector(CPUInfoCollector):
    def get_cpuinfo_file(self):
        cpunfo = ('processor	: 0\n'
                  'cpu MHz		: 100.000\n'  # doesn't really matter for LVE_VERSION > 4
                  'processor	: 1\n'
                  'cpu MHz		: 200.000\n'  # doesn't really matter for LVE_VERSION > 4
                  )

        return StringIO(cpunfo)


class MockFileSaver(FileSaver):
    def __init__(self, fname='/tmp/info'):
        super(MockFileSaver, self).__init__(fname)


class MockStatsNotifier(StatsNotifier):
    def __init__(self):
        super(MockStatsNotifier, self).__init__()

    def execute(self, lve_data):
        pass

    def _load_plugin_settings(self):
        pass


class MockCLTopFileSaver(CloudLinuxTopFileSaver):
    def _output(self, result):
        pass


class MockLVEStatsMemoryUsage(object):
    def execute(self, lve_data):
        pass


class MockLveLimitsBurster:
    def execute(self, _lve_data):
        pass


# List of mocks and stubs to replace real classes in order to not touch external world
replacements = {
    LVECollector: MockLVECollector,
    CPUInfoCollector: MockCpuInfoCollector,
    FileSaver: MockFileSaver,
    CloudLinuxTopFileSaver: MockCLTopFileSaver,
    StatsNotifier: MockStatsNotifier,
    LVEStatsMemoryUsage: MockLVEStatsMemoryUsage,
    LveLimitsBurster: MockLveLimitsBurster,
}


class TestConnectionPoolOverflow(unittest.TestCase):
    cfg = {
        'timeout': '5',  # 5 seconds is one iteration timeout
        'db_timeout': '60',  # writes to db should be done each time
        'server_id': 'localhost',
        'db_type': 'sqlite',
        'sqlite_db_path': ''}

    def mockSetup(self):
        patcher1 = mock.patch('lvestats.lib.config.read_config', return_value={})
        patcher1.start()
        self.addCleanup(patcher1.stop)

    def setUp(self):
        self.plugins = [replacements.get(p, p) for p, _ in default_plugins()]
        #  AssertionPool will fail the test if there is a checked out connection that is not returned to pool.
        #  Besides that it will point to the place where it was checked out in a stack trace.

        # super(TestConnectionPoolOverflow, self).setUp(self)
        self.engine = make_db_engine(self.cfg, pool_class=AssertionPool)
        create_db(self.engine)
        self.mockSetup()

    @mock.patch("lvestats.plugins.generic.snapshot_saver.get_lveps", return_value={})
    def testConnectionPoolOverflow(self, get_lveps):
        lve_data = {'stats': {}}

        db_timeout = int(self.cfg['db_timeout'])
        timeout = int(self.cfg['timeout'])
        next_invocation = {p: 0 for p in self.plugins}
        plugin_instances = {p: p() for p in self.plugins}

        start = int(time.time())
        for now in range(start, start + db_timeout * 5, timeout):
            for plugin_cls in self.plugins:
                plugin_instance = plugin_instances[plugin_cls]
                if hasattr(plugin_instance, 'set_config'):
                    plugin_instance.set_config(self.cfg)

                if hasattr(plugin_instance, 'set_db_engine'):
                    plugin_instance.set_db_engine(self.engine)

                plugin_instance.now = now

                if next_invocation[plugin_cls] <= now:
                    with mock.patch('lvestats.plugins.generic.cleaners.reboot_lock'), \
                            mock.patch('lvestats.plugins.generic.dbgov_saver.reboot_lock'), \
                            mock.patch('lvestats.plugins.generic.dbsaver.reboot_lock'), \
                            mock.patch('lvestats.plugins.generic.dbsaver_x60.reboot_lock'), \
                            mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
                        plugin_instance.execute(lve_data)
                    next_invocation[plugin_cls] = now + getattr(plugin_instance, 'period',
                                                                getattr(plugin_cls, 'period', timeout))


if __name__ == '__main__':
    unittest.main()
