# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import csv
import os
import time
import unittest
from io import StringIO

import mock
from clcommon.clproc import ProcLve
from lvestat import LVEStat

from lvestats.orm import history
from lvestats.plugins.generic.aggregators import LveUsageAggregator, LveUsageAggregator5S
from lvestats.plugins.generic.dbsaver import DBSaver
from lvestats.plugins.generic.analyzers import LVEUsageAnalyzer
from lvestats.plugins.generic.collectors import LVECollector, CPUInfoCollector
from lvestats.plugins.generic.persistors import FileSaver
from test.base.base_db_test import BaseDbTest

MEGAHERZ = 1000000
GIGAHERZ = 1000000000

BASE_PATH = os.path.dirname(os.path.abspath(__file__))


class MockProcLve(ProcLve):
    def __init__(self):
        super(MockProcLve, self).__init__()
        self.mock_resellers_lines = {
            # 1000000000 is cpu usage in ticks. For lve_version > 4 it is 1Gtick per second per core if usage was 100%
            0: ['8:LVE	EP	lCPU	lIO	CPU	MEM	IO	lMEM	lEP	nCPU	fMEM	fEP	lMEMPHY	lCPUW	lNPROC	MEMPHY	fMEMPHY	NPROC	fNPROC	lIOPS	iOPS',
                '501	0	2500	1024	1000000000	0	0	262144	20	1	0	0	262144	100	0	0	0	0	0	10	0']
        }  # type: dict[int, list[str]]

    def lines(self, lvp_id=0, without_limits=True):
        with mock.patch('clcommon.clproc.open', return_value=mock.MagicMock(), create=True) as mock_en:
            io = StringIO(os.linesep.join(self.mock_resellers_lines[lvp_id]))
            io.seek(0)
            mock_en.return_value.__enter__.return_value = io
            lines = list(super(MockProcLve, self).lines(lvp_id, without_limits))
        return lines

    def version(self):
        """
        Return /proc/lve filesystem version
        """
        line = self.mock_resellers_lines[0][0][:3]  # we need only first three symbols for parse version
        return int(line.rsplit(':', 1)[0])


class MockLVECollector(LVECollector):
    def __init__(self):
        super(MockLVECollector, self).__init__()
        self.min_uid = 500  # value for tests
        self._proc_lve = MockProcLve()

    def set_entry(self, values_dict):
        """
        For backward-compatibility with old tests;
        :param dict values_dict:
        """
        lve_stat = LVEStat(self._proc_lve.mock_resellers_lines[0][1], self._proc_lve.version())
        for attr, value in values_dict.items():
            setattr(lve_stat, attr, value)
        line = '\t'.join((str(attr) for attr in lve_stat._get_attributes()))
        self._proc_lve.mock_resellers_lines[0][1] = line


class MockCpuInfoCollector(CPUInfoCollector):
    def get_cpuinfo_file(self):
        cpunfo = ('processor	: 0\n'
                  'cpu MHz		: 100.000\n'  # doesn't really matter for LVE_VERSION > 4
                  'processor	: 1\n'
                  'cpu MHz		: 200.000\n'  # doesn't really matter for LVE_VERSION > 4
                  )
        return StringIO(cpunfo)


class MockFileSaver(FileSaver):
    def __init__(self, fname='/var/lve/info'):
        super(MockFileSaver, self).__init__(fname)
        self.file = StringIO('')

    def write(self, output):
        self.file.truncate(0)
        writer = csv.writer(self.file)
        writer.writerows(output)


class DataValidityTest(BaseDbTest):
    def setUp(self):
        BaseDbTest.setUp(self)
        self.cfg.update({
            'timeout': '5',  # 5 seconds is one iteration timeout
            'db_timeout': '0',  # writes to db should be done each time
            }
        )
        self.data = {}

        mock_proclve = mock.patch('clcommon.clproc.ProcLve.proc_lve_list', return_value=os.path.join(BASE_PATH, "data/proc/lve"))
        self.addCleanup(mock_proclve.stop)
        mock_proclve.start()

        mock_proclve_resellers = mock.patch('clcommon.clproc.ProcLve.resellers_supported', return_value=False)
        self.addCleanup(mock_proclve_resellers.stop)
        mock_proclve_resellers.start()

        mock_proclve_version = mock.patch('clcommon.clproc.ProcLve.version', return_value=8)
        self.addCleanup(mock_proclve_version.stop)
        mock_proclve_version.start()

        patcher1 = mock.patch('lvestats.lib.config.read_config', return_value={})
        patcher1.start()
        self.addCleanup(patcher1.stop)

        self.lve_collector = MockLVECollector()

        self.cpu_collector = MockCpuInfoCollector()

        self.lve_usage_analyzer = LVEUsageAnalyzer()
        self.lve_usage_analyzer.set_config(self.cfg)

        self.lve_usage_aggregator = LveUsageAggregator()
        self.lve_usage_aggregator_5s = LveUsageAggregator5S()

        # in_memory_db
        self.dbsaver = DBSaver()
        self.dbsaver.set_config(self.cfg)
        self.dbsaver.set_db_engine(self.engine)
        self.fsaver = MockFileSaver()

    def run_cycle(self, now=None):
        if now is None:
            now = time.time()

        plugins = [
            self.lve_collector,
            self.cpu_collector,
            self.lve_usage_analyzer,
            self.lve_usage_aggregator_5s,
            self.lve_usage_aggregator,
            self.dbsaver,
            self.fsaver, ]

        for plugin in plugins:
            if not hasattr(plugin, '__lastrun__'):
                plugin.__lastrun__ = 0

            if not hasattr(plugin, 'period') or now - plugin.__lastrun__ > plugin.period:
                plugin.now = now
                with mock.patch('lvestats.plugins.generic.dbsaver.reboot_lock'):
                    plugin.execute(self.data)
                plugin.__lastrun__ = now

    def test_aggregator(self):
        self.lve_collector.set_entry({'cpu_usage': 0})
        self.run_cycle(now=1)
        self.lve_collector.set_entry({'cpu_usage': 1 * GIGAHERZ})
        self.run_cycle(now=2)
        self.lve_collector.set_entry({'cpu_usage': 2 * GIGAHERZ})
        self.run_cycle(now=3)
        self.lve_collector.set_entry({'cpu_usage': 5 * GIGAHERZ})
        self.run_cycle(now=4)
        self.run_cycle(now=65)
        self.assertTrue('lve_usage' in self.data)
        self.assertListEqual(self.data['lve_usages'], [])
        self.assertEqual(3, self.data['lve_usage'][501].cpu_fault)

    def test_cpu_usage_calculated_and_stored(self):
        self.lve_collector.set_entry({'cpu_usage': 0})
        self.run_cycle(now=1)

        # limit_cpu = 1000
        # self.lve_collector.set_entry('cpu', limit_cpu)
        self.lve_collector.set_entry({'cpu_usage': int(2.0 * GIGAHERZ)})

        self.run_cycle(now=61)

        # CPU speed should be 2 GHz (2 cores)
        self.assertEqual(self.data['totalHz'], 2.0 * GIGAHERZ)

        usage_expected = int((2.0 * GIGAHERZ) / 60 / 100000)
        usage_actual = self.data['lve_usage'][501].cpu_usage

        self.assertEqual(usage_expected, usage_actual)

        # There should be 2 writes in db -- first time and second time when fault appears
        result = list(self.engine.execute(history.__table__.select()))
        self.assertEqual(1, len(result))
        self.assertEqual(usage_expected, result[0].cpu)

        self.assertEqual(1, self.fsaver.file.getvalue().count('\n'))

    def test_aggregators(self):
        self.lve_collector.set_entry({'cpu_usage': 0, 'cpu': 5000})
        self.run_cycle(1)

        self.assertEqual(len(self.data['lve_usage_5s']), 0)
        # FIXME(vlebedev): Dirty hack to get raw usage measurement in plugins after this one.
        # self.assertEqual(len(self.data['lve_usages_5s']), 1)
        self.assertEqual(len(self.data['lve_usages']), 1)
        self.assertTrue('lve_usage' not in self.data)
        fsaver_value = self.fsaver.file.getvalue()
        self.assertEqual(0, fsaver_value.count('\n'))

        for i in range(1, 12):
            self.lve_collector.set_entry({'cpu_usage': int(i * GIGAHERZ), 'mep': i})
            self.run_cycle(now=1 + i * 5)
            self.assertEqual(self.data['lve_usage_5s'][501].cpu_usage, 2000)
            self.assertEqual(self.data['lve_usage_5s'][501].cpu_fault, 0)
            self.assertEqual(self.data['lve_usage_5s'][501].mep, i)
            self.assertEqual(len(self.data['lve_usage_5s']), 1)
            # FIXME(vlebedev): Dirty hack to get raw usage measurement in plugins after this one.
            # self.assertEqual(len(self.data['lve_usages_5s']), 0)
            self.assertEqual(len(self.data['lve_usages']), i + 1)
            self.assertTrue('lve_usage' not in self.data)
            old_fsaver_value, fsaver_value = fsaver_value, self.fsaver.file.getvalue()
            self.assertEqual(1, fsaver_value.count('\n'))
            self.assertNotEqual(old_fsaver_value, fsaver_value)

        self.lve_collector.set_entry({'cpu_usage': int(24 * GIGAHERZ), 'nproc': 12, 'mep': 12})
        self.run_cycle(now=61)

        self.assertEqual(self.data['lve_usage_5s'][501].cpu_usage, 5000)
        self.assertEqual(self.data['lve_usage_5s'][501].cpu_fault, 1)
        self.assertEqual(self.data['lve_usage_5s'][501].mep, 12)
        self.assertEqual(len(self.data['lve_usage_5s']), 1)
        # FIXME(vlebedev): Dirty hack to get raw usage measurement in plugins after this one.
        # self.assertEqual(len(self.data['lve_usages_5s']), 0)
        self.assertEqual(len(self.data['lve_usages']), 0)
        self.assertTrue('lve_usage' in self.data)
        self.assertEqual(self.data['lve_usage'][501].cpu_usage, 2250)  # (11 * 2000 * 5 + 5000 * 5) / 60
        self.assertEqual(self.data['lve_usage'][501].cpu_fault, 1)
        self.assertEqual(self.data['lve_usage'][501].mep, 6.5)
        self.assertEqual(self.data['lve_usage'][501].nproc, 1)
        self.assertEqual(1, self.fsaver.file.getvalue().count('\n'))

        result = list(self.engine.execute(history.__table__.select()))
        self.assertEqual(1, len(result))
        self.assertEqual(2250, result[0].cpu)

    def test_increasing_cpu_use(self):
        self.lve_collector.set_entry({'cpu_usage': 0})
        self.run_cycle(now=1)
        limit_cpu = 25
        self.lve_collector.set_entry({'cpu': limit_cpu})
        self.lve_collector.set_entry({'cpu_usage': 2 * GIGAHERZ * 5})  # 5 seconds of 2Ghz ticks means 100.0%
        self.run_cycle(now=61)
        self.assertEqual(limit_cpu, self.data['lve_usage'][501].cpu_usage)  # 25 is lcpu
        self.assertEqual(1, self.data['lve_usage'][501].cpu_fault)

    def test_increasing_io_use(self):
        self.lve_collector.set_entry({'io_usage': 0})
        self.run_cycle(now=1)
        io_limit = 42  # KB
        self.lve_collector.set_entry({'io': io_limit})
        self.lve_collector.set_entry({'io_usage': io_limit * 2 * 60})  # 60 seconds of double io usage
        self.run_cycle(now=61)
        self.assertEqual(1024 * io_limit, self.data['lve_usage'][501].io_usage)
        self.assertEqual(1, self.data['lve_usage'][501].io_fault)

    def test_increasing_iops_use(self):
        self.lve_collector.set_entry({'iops': 0})
        self.run_cycle(now=1)
        liops = 142
        self.lve_collector.set_entry({'liops': liops})
        self.lve_collector.set_entry({'iops': liops * 2 * 60})  # 60 seconds of double iops usage
        self.run_cycle(now=61)
        self.assertEqual(liops, self.data['lve_usage'][501].iops)
        self.assertEqual(1, self.data['lve_usage'][501].iops_fault)

    def test_store_if_nothing_changes_but_faults_are_present(self):
        self.run_cycle(now=1)

        self.lve_collector.set_entry({'liops': 1})
        self.lve_collector.set_entry({'iops': 2})
        self.run_cycle(now=61)
        self.assertEqual(1, len(self.data['lve_usage']))

        self.lve_collector.set_entry({'liops': 1})
        self.lve_collector.set_entry({'iops': 2 + 2 * 5})  # 2 per 5 seconds
        self.run_cycle(now=121)
        self.assertEqual(1, len(self.data['lve_usage']))

    def test_lvectl_destroy_should_not_produce_negative_values(self):
        self.lve_collector.set_entry({'iops': 200000})
        self.run_cycle(now=1)
        self.lve_collector.set_entry({'liops': 1})  # to make it storeable
        self.lve_collector.set_entry({'iops': 0})
        self.run_cycle(now=61)
        self.assertEqual(0, self.data['lve_usage'][501].iops)

    def test_uid_filtration(self):
        self.lve_collector.min_uid = 500

        # pylint: disable=line-too-long
        self.lve_collector._proc_lve.mock_resellers_lines = {
            0: ['8:LVE	EP	lCPU	lIO	CPU	MEM	IO	lMEM	lEP	nCPU	fMEM	fEP	lMEMPHY	lCPUW	lNPROC	MEMPHY	fMEMPHY	NPROC	fNPROC  lIOPS   iOPS',
                '0	0	2500	1024	1000000000	0	0	262144	20	1	0	0	262144	100	0	0	0	0	0   10   0',
                '100	0	2500	1024	1000000000	0	0	262144	20	1	0	0	262144	100	0	0	0	0	0   10   0',
                '500	0	2500	1024	1000000000	0	0	262144	20	1	0	0	262144	100	0	0	0	0	0   10   0',
                '501	0	2500	1024	1000000000	0	0	262144	20	1	0	0	262144	100	0	0	0	0	0   10   0',
                '111501	0	2500	1024	1000000000	0	0	262144	20	1	0	0	262144	100	0	0	0	0	0   10   0',
                '1	0	2500	1024	1000000000	0	0	262144	20	1	0	0	262144	100	0	0	0	0	0   10   0',
                '499	0	2500	1024	1000000000	0	0	262144	20	1	0	0	262144	100	0	0	0	0	0   10   0',
                '50	0	2500	1024	1000000000	0	0	262144	20	1	0	0	262144	100	0	0	0	0	0   10   0',
        ]}

        self.lve_collector.execute(self.data)
        self.assertEqual(len(self.data['stats']), 4)
        self.assertNotIn(100, self.data['stats'])
        self.assertNotIn(1, self.data['stats'])
        self.assertNotIn(499, self.data['stats'])
        self.assertNotIn(50, self.data['stats'])
        self.assertIn(0, self.data['stats'])
        self.assertIn(500, self.data['stats'])
        self.assertIn(501, self.data['stats'])
        self.assertIn(111501, self.data['stats'])


if __name__ == '__main__':
    unittest.main()
