# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import os
import unittest
import ddt
import mock
from time import time

from test.base.base_db_test import BaseDbTest
from lvestats.lib.commons import dateutil
from lvestats.lib.lveinfolib import OutputFormatter
from lvestats.lib.lveinfolib_gov import HistoryShowDBGov

DBGOVERNOR_TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), 'var_lve_dbgovernor')


def add_table_prefix():
    pass


def mocked_uid_to_username(uid, local_server_id, server_id, db_engine):
    uid_to_username_dict = {500: 'wordpress'}
    return uid_to_username_dict.get(uid, uid)


@ddt.ddt
class TestDBGov(BaseDbTest):
    fields = ['TS', 'USER', 'CPU', 'READ', 'WRITE', 'CON', 'lCPU', 'lREAD',
              'lWRITE', 'RESTRICT']
    timeunit_fields = ['FROM', 'TO'] + fields[1:]

    def setUp(self):
        BaseDbTest.setUp(self)
        self.save_dbgov_line('localhost', uid=500, _time=0, sum_cpu=0.0009)
        self.save_dbgov_line('localhost', uid=500, _time=60, sum_cpu=0.001)
        self.save_dbgov_line('localhost', uid=500, _time=120, sum_cpu=3.0009)

        patcher1 = mock.patch('lvestats.lib.config.read_config', return_value={})
        patcher1.start()
        self.addCleanup(patcher1.stop)

    @mock.patch('lvestats.lib.lveinfolib_gov.uidconverter.uid_to_username', side_effect=mocked_uid_to_username)
    def test_dbgoverner_list(self, mocked_uid_to_username):
        self.save_dbgov_line('localhost', uid=1, _time=0, sum_cpu=0, sum_write=1, sum_read=1,
                             cause_of_restrict=7)
        self.save_dbgov_line('localhost', uid=1, _time=60, sum_cpu=1, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=1, _time=120, sum_cpu=2, sum_write=1, sum_read=1,
                             cause_of_restrict=2, limit_cpu_on_period_end=100,
                             limit_read_on_period_end=200, limit_write_on_period_end=300)
        self.save_dbgov_line('localhost', uid=1, _time=180, sum_cpu=3, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=1, _time=240, sum_cpu=4, sum_write=1, sum_read=1)

        self.save_dbgov_line('localhost', uid=4,  _time=0, sum_cpu=1, sum_write=1, sum_read=1)

        self.save_dbgov_line('localhost', uid=2, _time=0,  sum_cpu=3, sum_write=1)
        self.save_dbgov_line('localhost', uid=2, _time=120,  sum_cpu=4, sum_write=1, sum_read=1)

        rows = HistoryShowDBGov(
            self.engine,
            dateutil.unixtimestamp_to_gm_datetime(0),
            dateutil.unixtimestamp_to_gm_datetime(241),
            show_columns=self.fields,
            server_id='localhost',
            cfg={}
        ).history_dbgov_show()
        self.assertEqual(
            rows,
            [
                [240, 1,           2.0,     1.0,  1.0,  0, 100,  200,  300,  u'CRW'],
                [120, 2,           1.4,     0.2,  0.4,  0, None, None, None, u''],
                [0,   4,           0.2,     0.2,  0.2,  0, None, None, None, u''],
                [120, 'wordpress', 0.60056, None, None, 0, None, None, None, u'']
            ])

    dbgov_restrict_data = (
        ((1, 2, 3), "CR"),
        ((1, 2, 2), "CR"),
        ((1, 2, 0, 0, 1), "CR"),
        ((1, 1, 0, 0, 1), "C"),
        ((2, 2, 0, 0, 5), "CRW"),
        ((1, 1, 0, 1, 5), "CW"),
        ((5, 6), "CRW"),
        ((0, 3, 0, 0, 5), "CRW"),
        ((0, 2, 0, 0), "R"),
        ((0, 0, 0, 0, 0), ""),
    )

    @ddt.data(*dbgov_restrict_data)
    @ddt.unpack
    def test_dbgoverner_restrict(self, test_data, result):
        for t, cause in enumerate(test_data):
            self.save_dbgov_line('localhost', uid=1, _time=t, sum_cpu=1, cause_of_restrict=cause)
        rows = HistoryShowDBGov(
            self.engine,
            dateutil.unixtimestamp_to_gm_datetime(0),
            dateutil.unixtimestamp_to_gm_datetime(10),
            show_columns=self.fields,
            time_unit=1
        ).history_dbgov_show()
        self.assertEqual(rows[0][9], result)

    dbgov_restrict_data_with_uid = (
        ((1, 2, 3), ("C", "R", "CR")),
        ((1, 2, 2), ("C", "R", "R")),
        ((1, 2, 0, 0, 1), ("C", "R", "", "", "C")),
        ((1, 1, 0, 0, 1), ("C", "C", "", "", "C")),
        ((2, 2, 0, 0, 5), ("R", "R", "", "", "CW")),
        ((1, 1, 0, 1, 5), ("C", "C", "", "C", "CW")),
        ((5, 6), ("CW", "RW")),
        ((0, 3, 0, 0, 5), ("", "CR", "", "", "CW")),
        ((0, 2, 0, 0), ("", "R", "", "")),
        ((0, 0, 0, 0, 0), ("", "", "", "", "")),
    )

    @ddt.data(*dbgov_restrict_data_with_uid)
    @ddt.unpack
    def test_dbgoverner_restrict_with_uid(self, test_data, result):
        for t, cause in enumerate(test_data):
            self.save_dbgov_line('localhost', uid=1, _time=t, sum_cpu=1, cause_of_restrict=cause)
        rows = HistoryShowDBGov(
            self.engine,
            dateutil.unixtimestamp_to_gm_datetime(0),
            dateutil.unixtimestamp_to_gm_datetime(10),
            uid=1,
            time_unit=1,
            show_columns=self.fields).history_dbgov_show()
        restricts = list(zip(*rows))[9]
        self.assertEqual(restricts, result)

    @ddt.data((None, [0.7507]), (500, [0.0, 0.001, 3.0009]))
    @ddt.unpack
    def test_dbgoverner_formated_CPU(self, uid, valid_result):
        rows = HistoryShowDBGov(
            self.engine,
            dateutil.unixtimestamp_to_gm_datetime(0),
            dateutil.unixtimestamp_to_gm_datetime(181),
            uid=uid,
            show_columns=self.fields,
            cfg=self.cfg,
            time_unit=1
        ).history_dbgov_show()
        output_formatted = OutputFormatter(
            fields=self.fields,
            orders=[
                (['TS'], 'datetime'),
                (['TS'], lambda dt: OutputFormatter.strftime(
                    dt, '%m-%d %H:%M:%S')),
                (['CPU'], lambda x: 0 if x < 0.001 else x),
                (['READ', 'WRITE'], lambda x: 0 if x < 0.001 else x),
            ]
        )
        output_formatted.set_rows(rows)
        results = list(output_formatted)
        self.assertListEqual([r[2] for r in results], valid_result)

    @unittest.skip('In development mode')
    def test_dbgoverner_order_by(self):
        self.save_dbgov_line('localhost', uid=501, _time=4, sum_cpu=0.0009, sum_read=100, sum_write=100)
        self.save_dbgov_line('localhost', uid=501, _time=5, sum_cpu=0.001, sum_read=50, sum_write=180)
        self.save_dbgov_line('localhost', uid=501, _time=6, sum_cpu=3.0009, sum_read=30, sum_write=150)
        rows = HistoryShowDBGov(
            self.engine,
            dateutil.unixtimestamp_to_gm_datetime(4),
            dateutil.unixtimestamp_to_gm_datetime(8),
            uid=501,
            time_unit=1,
            show_columns=['READ', 'WRITE'], order_by='io').history_dbgov_show()
        self.assertEqual(rows[0][0], 50.0)
        self.assertEqual(rows[1][0], 100.0)
        self.assertEqual(rows[2][0], 30.0)

        rows = HistoryShowDBGov(
            self.engine,
            dateutil.unixtimestamp_to_gm_datetime(4),
            dateutil.unixtimestamp_to_gm_datetime(6),
            uid=501,
            show_columns=['READ', 'WRITE'], order_by='read').history_dbgov_show()
        self.assertEqual(rows[0][0], 100.0)
        self.assertEqual(rows[1][0], 50.0)
        self.assertEqual(rows[2][0], 30.0)

        rows = HistoryShowDBGov(
            self.engine,
            dateutil.unixtimestamp_to_gm_datetime(4),
            dateutil.unixtimestamp_to_gm_datetime(6),
            uid=501,
            show_columns=['READ', 'WRITE']).history_dbgov_show()
        self.assertEqual(len(rows), 3)

    @ddt.data(-500, -1, 0, 501)
    def test_no_such_id(self, uid):
        rows = HistoryShowDBGov(
            self.engine,
            dateutil.unixtimestamp_to_gm_datetime(0),
            dateutil.unixtimestamp_to_gm_datetime(3),
            uid=uid, show_columns=self.fields, cfg=self.cfg,
            time_unit=1
        ).history_dbgov_show()
        self.assertListEqual(list(), list(rows))

    @ddt.data([60 * 10, 60, 10],
              [60 * 10, 600, 2],
              [60 * 60, 60, 60],
              [60 * 60 * 60, 120, 1801],
              [60 * 10, 120, 6])
    @ddt.unpack
    def test_time_unit(self, time_interval, time_unit, rows_len):
        # get current timestamp (do not remove int()!)
        now = int(time())

        # insert values with one-minute interval
        for i in range(0, time_interval, 60):
            self.save_dbgov_line('localhost', uid=500, _time=now - i, sum_cpu=60)

        rows = HistoryShowDBGov(
            self.engine,
            dateutil.unixtimestamp_to_gm_datetime(now - time_interval),
            dateutil.unixtimestamp_to_gm_datetime(now),
            uid=500,
            show_columns=self.timeunit_fields,
            cfg=self.cfg,
            time_unit=time_unit
        ).history_dbgov_show()

        output_formatted = OutputFormatter(
            fields=self.timeunit_fields,
            orders=[
                (['FROM', 'TO'], 'datetime'),
                (['FROM', 'TO'], lambda dt: OutputFormatter.strftime(
                    dt, '%m-%d %H:%M:%S')),
                (['CPU'], lambda x: 0 if x < 0.001 else x),
                (['READ', 'WRITE'], lambda x: 0 if x < 0.001 else x),
            ]
        )
        output_formatted.set_rows(rows)
        results = list(output_formatted)

        self.assertEqual(len(results), rows_len)

    dbgov_restrict_data_with_uid = (
        ((1, 2, 3), (("C", "R", "CR"), ("CR", "CR"), ("CR",))),
        ((1, 2, 2), (("C", "R", "R"), ("CR", "R"), ("CR",))),
        ((1, 2, 0, 0, 1), (("C", "R", "", "", "C"), ("CR", "", "C"), ("CR", "C"), ("CR", "C"), ("CR",))),
        ((1, 1, 0, 0, 1), (("C", "C", "", "", "C"), ("C", "", "C"), ("C", "C"), ("C", "C"), ("C",))),
        ((2, 2, 0, 0, 5), (("R", "R", "", "", "CW"), ("R", "", "CW"), ("R", "CW"), ("R", "CW"), ("CRW",))),
        ((1, 1, 0, 2, 5), (("C", "C", "", "R", "CW"), ("C", "R", "CW"), ("C", "CRW"), ("CR", "CW"), ("CRW",))),
        ((5, 6), (("CW", "RW"), ("CRW",))),
        ((0, 3, 0, 0, 5), (("", "CR", "", "", "CW"), ("CR", "", "CW"), ("CR", "CW"), ("CR", "CW"), ("CRW",))),
        ((0, 2, 0, 0), (("", "R", "", ""), ("R", ""), ("R", ""), ("R",))),
        ((0, 0, 0, 0, 0), (("", "", "", "", ""), ("", "", ""), ("", ""), ("", ""), ("",))),
        ((0, 1, 2, 3, 4, 5, 6, 7), (
            ("", "C", "R", "CR", "W", "CW", "RW", "CRW"), ("C", "CR", "CW", "CRW"), ("CR", "CRW", "CRW"),
            ("CR", "CRW"), ("CRW", "CRW"), ("CRW", "CRW"), ("CRW", "CRW"), ("CRW",))),
    )

    @ddt.data(*dbgov_restrict_data_with_uid)
    @ddt.unpack
    def test_time_unit_with_resrtict(self, test_data, result_list):
        for t, cause in enumerate(test_data):
            self.save_dbgov_line('localhost', uid=1, _time=t, sum_cpu=1, cause_of_restrict=cause)
        for time_unit, result in enumerate(result_list):
            rows = HistoryShowDBGov(
                self.engine,
                dateutil.unixtimestamp_to_gm_datetime(0),
                dateutil.unixtimestamp_to_gm_datetime(10),
                uid=1,
                show_columns=self.timeunit_fields,
                cfg=self.cfg,
                time_unit=time_unit + 1
            ).history_dbgov_show()
            restricts = list(zip(*rows))[10]
            from_to_list = []
            for row in rows:
                from_to_list.append(row[:2])
            self.assertEqual(
                restricts, result,
                "test_data: {0} \ntest result: {1} \nactual result: {2} \ntime unit: {3} \nfrom-to: {4}".format(
                    str(test_data), str(result), str(restricts), str(time_unit + 1), str(from_to_list)))

    @ddt.unpack
    @ddt.data(([1001, 1003],                   [1001, 1003]),
              ((1001, 1002),                   [1001]),
              ([1003],                         [1003]),
              ([1002],                         []),
              ([1000, 1002],                   []),
              ([],                             []),
              (tuple(),                        []),
              (list(range(600, 1500)),         [1001, 1003, 1004]))
    def test_multiple_uids(self, uids, uids_output):
        self.save_dbgov_line('hostname', uid=1001, _time=4, sum_cpu=0.0009, sum_read=100, sum_write=100)
        self.save_dbgov_line('hostname', uid=1003, _time=5, sum_cpu=0.001,  sum_read=50,  sum_write=180)
        self.save_dbgov_line('hostname', uid=1004, _time=6, sum_cpu=3.0009, sum_read=30,  sum_write=150)

        valid_result = HistoryShowDBGov(
            self.engine,
            dateutil.unixtimestamp_to_gm_datetime(0),
            dateutil.unixtimestamp_to_gm_datetime(10),
            uid=None,
            show_columns=['TS', 'ID', 'CPU', 'READ', 'WRITE', 'CON', 'lCPU', 'lREAD', 'lWRITE', 'RESTRICT'],
            cfg=self.cfg,
            time_unit=1,
            server_id='hostname'
        ).history_dbgov_show()
        interested_values = [item for item in valid_result if item[1] in uids]

        result = HistoryShowDBGov(
            self.engine,
            dateutil.unixtimestamp_to_gm_datetime(0),
            dateutil.unixtimestamp_to_gm_datetime(10),
            uid=uids, show_columns=['TS', 'ID', 'CPU', 'READ', 'WRITE', 'CON', 'lCPU', 'lREAD', 'lWRITE', 'RESTRICT'],
            cfg=self.cfg,
            time_unit=1,
            server_id='hostname'
        ).history_dbgov_show()
        result_uids = [line[1] for line in result]
        self.assertEqual(result, interested_values)
        self.assertListEqual(uids_output, result_uids)

    def test_multiple_uids_no_such_serverid(self):
        self.save_dbgov_line('hostname', uid=1001, _time=4, sum_cpu=0.0009, sum_read=100, sum_write=100)
        self.save_dbgov_line('hostname', uid=1003, _time=5, sum_cpu=0.001,  sum_read=50,  sum_write=180)
        self.save_dbgov_line('hostname', uid=1004, _time=6, sum_cpu=3.0009, sum_read=30,  sum_write=150)
        result = HistoryShowDBGov(
            self.engine,
            dateutil.unixtimestamp_to_gm_datetime(0),
            dateutil.unixtimestamp_to_gm_datetime(10),
            uid=None,
            show_columns=['TS', 'ID', 'CPU', 'READ', 'WRITE', 'CON', 'lCPU', 'lREAD', 'lWRITE', 'RESTRICT'],
            cfg=self.cfg,
            time_unit=1,
            server_id='hostname2'
        ).history_dbgov_show()
        self.assertEqual([], result)

    @ddt.data(
        [['cpu'], 0.5, [1]],
        [['cpu'], 0.1, [1, 2]],
        [['io'], 0.6, []],
        [['io'], 0.3, [2]],
        [['io'], 0.2, [2]],
        [['io'], 0.1, [1, 2]],
        [['io', 'cpu'], 0.5, [1, 2]],  # check 'or' statement
        [['io', 'cpu'], 0.6, [1]],
        [['io', 'cpu'], 0.7, []])
    @ddt.unpack
    @mock.patch('lvestats.lib.lveinfolib_gov.uidconverter.uid_to_username', side_effect=mocked_uid_to_username)
    def test_by_usage(self, by_usage, percentage, result, *args):
        self.save_dbgov_line('localhost', uid=1, _time=300,
                             sum_cpu=65, limit_cpu_on_period_end=100,
                             sum_write=10, sum_read=10, limit_read_on_period_end=50, limit_write_on_period_end=50)
        self.save_dbgov_line('localhost', uid=2, _time=300,
                             sum_cpu=11, limit_cpu_on_period_end=100,
                             sum_write=1, sum_read=5, limit_read_on_period_end=5, limit_write_on_period_end=5)
        rows = HistoryShowDBGov(
            self.engine,
            dateutil.unixtimestamp_to_gm_datetime(300),
            dateutil.unixtimestamp_to_gm_datetime(350),
            show_columns=self.fields,
            cfg=self.cfg,
            by_usage=by_usage,
            by_usage_percentage=percentage
        ).history_dbgov_show()

        users = [user_info[1] for user_info in rows]
        self.assertEqual(users, result)
if __name__ == '__main__':
    unittest.main()
