# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import glob
import os
import shutil
import tempfile
import unittest
import mock
import ddt

from sqlalchemy import select

from test.base.base_db_test import BaseDbTest
from lvestats.orm.history_gov import history_gov
from lvestats.plugins.generic.dbgov_saver import DBGovSaver

DBGOVERNOR_TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), 'var_lve_dbgovernor')


@ddt.ddt
class TestDBGov(BaseDbTest):
    def setUp(self):
        self.dbgovernor_dir = tempfile.mkdtemp(prefix='tmp_unittest_')

        # copy all files from DBGOVERNOR_TEST_DATA_DIR to self.dbgovernor_dir
        list(map(lambda f_: shutil.copy(f_, self.dbgovernor_dir),
                 glob.iglob(os.path.join(DBGOVERNOR_TEST_DATA_DIR, '*'))))
        BaseDbTest.setUp(self)

        self.plugin = DBGovSaver()
        self.plugin.min_uid = 500
        # mocking for prevent system api call
        self.plugin.set_config(self.cfg)
        self.plugin.DBSTAT_DIR = self.dbgovernor_dir
        self.plugin.set_db_engine(self.engine)

    @mock.patch.object(DBGovSaver, 'get_user_id',
                       lambda _, usr: {'root': 0, 'second_user': 500, 'cltest1': 501}.get(usr, -1))
    def test_dbgoverner_should_save_all_entries(self):
        self.plugin.min_uid = 500
        with mock.patch('lvestats.plugins.generic.dbgov_saver.reboot_lock'):
            self.plugin.execute({})
        entries = list(self.engine.execute(select([history_gov.ts, history_gov.server_id, history_gov.uid])))
        self.assertEqual(4, len(entries))

    @mock.patch.object(DBGovSaver, 'get_user_id',
                       lambda _, usr: {'root': 0, 'second_user': 500, 'cltest1': 501}.get(usr, -1))
    def test_dbgoverner_should_save_all_entries2(self):
        with mock.patch.object(os, "listdir", lambda _: ["governor.1411996306", "governor.1412001047"]), \
                mock.patch('lvestats.plugins.generic.dbgov_saver.reboot_lock'):
            self.plugin.execute({})
        entries = list(self.engine.execute(select([history_gov.ts, history_gov.server_id, history_gov.uid])))
        self.assertEqual(len(entries), 2)
        flist = list(filter(self.plugin.FILE_PATTERN.search, os.listdir(self.plugin.DBSTAT_DIR)))
        self.assertEqual(len(flist), 4)

    @mock.patch.object(DBGovSaver, 'get_user_id', lambda _, usr: {'root': 0}.get(usr, -1))
    def test_dbgovernor_should_delete_all_files_when_nothing_saved(self):
        os.unlink(os.path.join(self.plugin.DBSTAT_DIR, "governor.1412002047"))  # in this file we have uid
        with mock.patch('lvestats.plugins.generic.dbgov_saver.reboot_lock'):
            self.plugin.execute({})
        entries = list(self.engine.execute(select([history_gov.ts, history_gov.server_id, history_gov.uid])))
        self.assertEqual(len(entries), 0)
        flist = list(filter(self.plugin.FILE_PATTERN.search, os.listdir(self.plugin.DBSTAT_DIR)))
        self.assertEqual(len(flist), 0)

    @mock.patch.object(DBGovSaver, 'get_user_id', lambda _, usr: {'root': 0}.get(usr, -1))
    def test_dbgovernor_should_delete_all_files_when_nothing_saved2(self):
        with mock.patch.object(os, "listdir", lambda _: ["governor.1411996306", "governor.1412001047"]), \
                mock.patch('lvestats.plugins.generic.dbgov_saver.reboot_lock'):
            self.plugin.execute({})
        entries = list(self.engine.execute(select([history_gov.ts, history_gov.server_id, history_gov.uid])))
        self.assertEqual(len(entries), 0)
        flist = list(filter(self.plugin.FILE_PATTERN.search, os.listdir(self.plugin.DBSTAT_DIR)))
        self.assertEqual(len(flist), 4)

    def test_dbgovernor_should_process_nondecodable_files_correctly(self):
        with mock.patch.object(os, 'listdir', lambda _: ['governor.1412004047']):
            result = self.plugin.scan_dir()
            self.assertCountEqual(result, [(os.path.join(self.plugin.DBSTAT_DIR, 'governor.1412004047'), [])])

    @mock.patch.object(DBGovSaver, 'get_user_id',
                       lambda _, usr: {'root': 0, 'second_user': 5, 'cltest1': 1}.get(usr, -1))
    def test_dbgoverner_filter_min_uid1(self):
        self.plugin.min_uid = 501
        with mock.patch('lvestats.plugins.generic.dbgov_saver.reboot_lock'):
            self.plugin.execute({})
        entries = list(self.engine.execute(select([history_gov.ts, history_gov.server_id, history_gov.uid])))
        self.assertEqual(0, len(entries))

    @mock.patch.object(DBGovSaver, 'get_user_id',
                       lambda _, usr: {'root': 0, 'second_user': 599, 'cltest1': 1000}.get(usr, -1))
    def test_dbgoverner_filter_min_uid2(self):
        self.plugin.min_uid = 600
        with mock.patch('lvestats.plugins.generic.dbgov_saver.reboot_lock'):
            self.plugin.execute({})
        entries = list(self.engine.execute(select([history_gov.ts, history_gov.server_id, history_gov.uid])))
        self.assertEqual(1, len(entries))

    @mock.patch.object(DBGovSaver, 'get_user_id',
                       lambda _, usr: {'root': 0, 'second_user': 1499, 'cltest1': 100}.get(usr, -1))
    def test_dbgoverner_filter_min_uid3(self):
        self.plugin.min_uid = 500
        with mock.patch('lvestats.plugins.generic.dbgov_saver.reboot_lock'):
            self.plugin.execute({})
        entries = list(self.engine.execute(select([history_gov.ts, history_gov.server_id, history_gov.uid])))
        self.assertEqual(3, len(entries))

    @ddt.data(
        {1001: {'cpu_limit': 400, 'io_limit': 2097152000, 'cpu_usage': 226.4, 'io_usage': 125631}},
        {}
    )
    def test_execute_governer_data_absent(self, _dbgov_data_for_cm):
        """
        Test execute method. Case governor data absent
        """
        connect_obj_mock = mock.MagicMock()
        old_governor_data = {'a': 1}
        lve_data = {"dbgov_data_for_cm": old_governor_data}
        scan_dir_result = [('file1', ['line1', 'line2'])]
        with mock.patch.object(self.plugin, "scan_dir", return_value=scan_dir_result) as _scan_dir_mock, \
                mock.patch.object(self.engine, "connect", return_value=connect_obj_mock) as _connect_mock, \
                mock.patch.object(self.plugin, "write_to_db", return_value=_dbgov_data_for_cm) as _write_mock:
            self.plugin.execute(lve_data)
            _connect_mock.assert_called_once()
            _scan_dir_mock.assert_called_once()
            _write_mock.assert_called_once_with(connect_obj_mock, scan_dir_result)
            if _dbgov_data_for_cm:
                # new data from governor absent
                self.assertEqual(lve_data["dbgov_data_for_cm"], _dbgov_data_for_cm)
            else:
                # new data from governor absent
                self.assertEqual(lve_data["dbgov_data_for_cm"], old_governor_data)

    def tearDown(self):
        shutil.rmtree(self.dbgovernor_dir)
        BaseDbTest.tearDown(self)


if __name__ == '__main__':
    unittest.main()
