# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import os
import unittest
import shutil
from unittest import mock

import ddt
from xml.dom import minidom

from clcommon import cpapi
from clcommon.cpapi import NotSupported
from clcommon.cpapi.cpapiexceptions import CPAPIExternalProgramFailed

from lvestats.lib.commons.func import get_lve_version, LVEVersionError, \
    get_governor_ignore_for_user, get_chunks, \
    get_reseller_domains, reboot_lock, get_domains, DomainException, touch, \
    serialize_lve_id, deserialize_lve_id
from lvestats.utils.dbmigrator.dbgovernor_config import set_governor_save_uid


DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')


class GetLVEVersionTestCase(unittest.TestCase):

    def test_01(self):
        # print DATA_DIR
        self.assertEqual(8, get_lve_version(_lve_file=os.path.join(DATA_DIR, 'lve_file.txt')))

    def test_02(self):
        self.assertEqual(4, get_lve_version(_lve_file=os.path.join(DATA_DIR, 'lve_file2.txt')))

    def test_03(self):
        self.assertRaises(LVEVersionError, get_lve_version, _lve_file=os.path.join(DATA_DIR, 'no_such_file'))

    def test_04(self):
        self.assertRaises(LVEVersionError, get_lve_version, _lve_file=os.path.join(DATA_DIR, 'lve_file_bad.txt'))


class TestDBGovernorConfig(unittest.TestCase):
    MYSQLCONFIGORIGINAL = os.path.join(DATA_DIR, 'mysql-governor-config.xml')
    MYSQLCONFIGTEST = os.path.join(DATA_DIR, 'test.xml')

    def setUp(self):
        shutil.copy(self.MYSQLCONFIGORIGINAL, self.MYSQLCONFIGTEST)

    def tearDown(self):
        os.remove(self.MYSQLCONFIGTEST)

    def test_set_uid_on_off(self):
        governor_config = minidom.parse(self.MYSQLCONFIGTEST)
        self.assertFalse(governor_config.getElementsByTagName("statistic")[0].hasAttribute("save_uid"))

        set_governor_save_uid(self.MYSQLCONFIGTEST, True)
        governor_config = minidom.parse(self.MYSQLCONFIGTEST)
        self.assertTrue(governor_config.getElementsByTagName("statistic")[0].hasAttribute("save_uid"))
        self.assertEqual(governor_config.getElementsByTagName("statistic")[0].getAttribute("save_uid"), "On")

        set_governor_save_uid(self.MYSQLCONFIGTEST, False)
        governor_config = minidom.parse(self.MYSQLCONFIGTEST)
        self.assertFalse(governor_config.getElementsByTagName("statistic")[0].hasAttribute("save_uid"))

    def test_set_uid_off(self):
        set_governor_save_uid(self.MYSQLCONFIGTEST, False)
        governor_config = minidom.parse(self.MYSQLCONFIGTEST)
        self.assertFalse(governor_config.getElementsByTagName("statistic")[0].hasAttribute("save_uid"))


def mock_dbgov_config(f):
    return minidom.parseString("""
        <governor>
            <lve use="all"></lve>
            <intervals short="15" mid="60" long="300"></intervals>
            <log file="/var/log/dbgovernor-error.log" mode="ERROR"></log>
            <restrict level1="60s" level2="15m" level3="1h" level4="1d" timeout="1h" log="/var/log/dbgovernor-restrict.log" user_max_connections="30"></restrict>
            <statistic mode="on"></statistic>
            <default>
              <limit name="cpu" current="400" short="380" mid="350" long="300"></limit>
              <limit name="read" current="1000000000" short="830000000" mid="760000000" long="590000000"></limit>
              <limit name="write" current="1000000000" short="830000000" mid="760000000" long="590000000"></limit>
            </default>
            <user mysql_name="root" mode="ignore"></user>
            <user mysql_name="mysql" mode="ignore"></user>
            <user mysql_name="#mysql" mode="ignore"></user>
            <user mysql_name="#mysql_system#" mode="ignore"></user>
            <user mysql_name="eximstats" mode="ignore"></user>
            <user mysql_name="dbgovernor" mode="ignore"></user>
            <user mysql_name="leechprotect" mode="ignore"></user>
            <user mysql_name="unauthenticated user" mode="ignore"></user>
            <user mysql_name="logaholic" mode="ignore"></user>
            <user mysql_name="cphulkd" mode="ignore"></user>
            <user mysql_name="pp" mode="ignore"></user>
            <user mysql_name="DELAYED" mode="ignore"></user>
            <user mysql_name="admin" mode="ignore"></user>
            <user name="user1" mode="ignore"></user>
            <user name="user2" mode="restrict"></user>
            <user name="user3" mode="norestrict"></user>
        </governor>
    """)


class TestDBGOVIgnoreConfig(unittest.TestCase):
    @mock.patch("lvestats.lib.commons.func.get_governor_config_path", return_value="/var/run/mysql-governor-config.xml")
    @mock.patch('xml.dom.minidom.parse', side_effect=mock_dbgov_config)
    def test_get_ignore_value(self, mock_parse_config_file, get_governor_config_path):
        result = get_governor_ignore_for_user("user1")
        self.assertEqual(result, True)
        result = get_governor_ignore_for_user("user2")
        self.assertEqual(result, False)
        result = get_governor_ignore_for_user("user3")
        self.assertEqual(result, False)
        result = get_governor_ignore_for_user("user4")
        self.assertEqual(result, False)

    @mock.patch("lvestats.lib.commons.func.get_governor_config_path", return_value=None)
    @mock.patch('xml.dom.minidom.parse', side_effect=mock_dbgov_config)
    def test_get_ignore_value_without_dbgov(self, mock_parse_config_file, get_governor_config_path):
        result = get_governor_ignore_for_user("user1")
        self.assertEqual(result, None)


@ddt.ddt
class TestFunctions(unittest.TestCase):
    @ddt.data(
        [[], 1000, 0],
        [[1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 10, 1],
        [[1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 1, 10],
        [[1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 10, 1],
        [[1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 3, 4],
        [[1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 2, 5])
    @ddt.unpack
    def test_get_chunks(self, test_data, chunk_length, expected_number_of_chunks):
        number_of_chunks = 0
        for chunk in get_chunks(test_data, chunk_length):
            number_of_chunks += 1
            self.assertLessEqual(len(chunk), chunk_length)
        self.assertEqual(number_of_chunks, expected_number_of_chunks)

    @mock.patch('clcommon.cpapi.reseller_domains', return_value=dict(x='y'))
    def test_get_reseller_domains(self, mock_reseller_domains):
        self.assertEqual(get_reseller_domains('reseller'), dict(x='y'))

    @mock.patch('clcommon.cpapi.reseller_domains')
    @ddt.data((NotSupported("fail"), dict()),
              (CPAPIExternalProgramFailed("fail"), dict()))
    @ddt.unpack
    def test_error_get_reseller_domains(self, side_effect, result, mock_reseller_domains):
        mock_reseller_domains.side_effect = side_effect
        self.assertEqual(get_reseller_domains('reseller'), result)

    @ddt.data(
        KeyError,
        IOError,
        TypeError,
        cpapi.cpapiexceptions.NotSupported,
        cpapi.cpapiexceptions.NoPanelUser,
        cpapi.cpapiexceptions.ParsingError,
    )
    def test_get_domains(self, exc):
        with mock.patch('clcommon.cpapi.userdomains', side_effect=exc):
            with self.assertRaises(DomainException):
                get_domains('someuser')


@ddt.ddt
class TestRebootMock(unittest.TestCase):
    def setUp(self):
        self.temp_fname = '/var/lve/lvestats.lock'

    @ddt.data(
        '/temp/test.txt',
        None)
    def test_constructor_with_arg(self, fname):
        with mock.patch('lvestats.lib.commons.func.reboot_lock._build_lock_filename',
                        return_value=self.temp_fname) as _build_fname, \
                mock.patch('lvestats.lib.commons.func.os.open',
                           return_value=self.temp_fname) as _open:
            reboot_lock(fname)
            if not fname:
                _build_fname.assert_called_once()
                _open.assert_called_with(self.temp_fname, os.O_WRONLY | os.O_CREAT, 0o600)
            else:
                _build_fname.assert_not_called()
                _open.assert_called_with(fname, os.O_WRONLY | os.O_CREAT, 0o600)

    def test_constructor_wo_arg(self):
        with mock.patch('lvestats.lib.commons.func.reboot_lock._build_lock_filename',
                        return_value=self.temp_fname) as _build_fname, \
                mock.patch('lvestats.lib.commons.func.os.open',
                           return_value=self.temp_fname) as _open:
            reboot_lock()
            _build_fname.assert_called_once()
            _open.assert_called_with(self.temp_fname, os.O_WRONLY | os.O_CREAT, 0o600)

    @ddt.data(
        False,
        True)
    def test__create_lock(self, _is_path_exists):
        _dir = '/var/lve'      # This is hardcoded in reboot_lock._create_lock function
        with mock.patch('lvestats.lib.commons.func.mkdir_p') as _mkdir_p, \
                mock.patch('lvestats.lib.commons.func.os.path.exists', return_value=_is_path_exists):
            self.assertEqual(reboot_lock._build_lock_filename(), self.temp_fname)
            if _is_path_exists:
                _mkdir_p.assert_not_called()
            else:
                _mkdir_p.assert_called_with(_dir)


@ddt.ddt
class TestMiscelaneous(unittest.TestCase):
    @ddt.data(
        None,
        123456)
    def test_01_touch(self, _times):
        """
        Test for touch function
        """
        touch_fname = 'touch.txt'
        with mock.patch('lvestats.lib.commons.func.open',
                        mock.mock_open(read_data=None)) as _mock_open, \
             mock.patch('lvestats.lib.commons.func.os.utime') as _mock_utime, \
             mock.patch('lvestats.lib.commons.func.os.chmod') as _mock_chmod:
            touch(touch_fname, _times)
            _mock_open.assert_called_once_with(touch_fname, 'a', encoding='utf-8')
            _mock_chmod.assert_called_once_with(touch_fname, 0o644)
            _mock_utime.assert_called_once_with(touch_fname, _times)

    @ddt.data(
        (1001, 0, 10**9, 1001),
        (1001, 0, 2**31-1, 1001),
        (1002, 1003, 10**9, 1002),
        (1002, 1003, 2**31-1, 1002),
        (2**32 - 1, 1004, 10**9, 1004 + 10**9),
        (2**32 - 1, 1004, 2**31-1, 1004 + 2**31 - 1),
    )
    @ddt.unpack
    def test_02_serialize_lve_id(
            self,
            lve_id: int,
            lvp_id: int,
            max_lve_id: int,
            result: int,
    ):
        with mock.patch(
                'lvestats.lib.commons.func.MAX_LVE_ID',
                max_lve_id,
        ):
            real_result = serialize_lve_id(
                lve_id=lve_id,
                lvp_id=lvp_id,
            )
            self.assertEqual(
                real_result,
                result,
            )

    @ddt.data(
        ((1001, False), 10**9, 1001),
        ((1001, False), 2**31-1, 1001),
        ((1002, False), 10**9, 1002),
        ((1002, False), 2**31-1, 1002),
        ((1004, True), 10**9, 1004 + 10**9),
        ((1004, True), 2**31-1, 1004 + 2**31 - 1),
    )
    @ddt.unpack
    def test_03_deserialize_lve_id(
            self,
            result: int,
            max_lve_id: int,
            serialized_value: int,
    ):
        with mock.patch(
                'lvestats.lib.commons.func.MAX_LVE_ID',
                max_lve_id,
        ):
            real_result = deserialize_lve_id(
                serialized_value=serialized_value,
            )
            self.assertEqual(
                real_result,
                result,
            )


if __name__ == '__main__':
    unittest.main()
