# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import os
import random
import time
import unittest

import ddt
import numpy as np

from lvestats.lib.commons.argparse_utils import period_type
from lvestats.lib.chart import Renderer

GENERATE_IMAGES = __name__ == '__main__'  # if run as script
IMAGES_OUTPUT = '/tmp'
IMAGE_BASE_NAME = os.path.splitext(os.path.basename(__file__))[0]


@ddt.ddt
class TestOptimise(unittest.TestCase):

    def generate_points(self, period='30d'):
        period_seconds = period_type(period)
        points_len = period_seconds // 60

        cycles_numb = 30  # one per twenty-four hours
        x = np.linspace(0, np.pi * 2 * cycles_numb, points_len)
        y_ = np.abs(np.sin(x) - (np.sin(np.divide(x, 10.)) + 1) / 4.)

        # randomise
        y = y_ + np.random.rand(y_.size) * (1 - np.sin(np.divide(x, 45)))  # randomise y
        return np.column_stack((x, y)).tolist()

    @ddt.data(('1h', 0.2), ('4h', 0.5), ('1d', 3), ('7d', 20), ('14d', 40), ('30d', 90))
    @ddt.unpack
    def test_optimise_many_points(self, period, timeout):
        points = self.generate_points(period)
        render = Renderer()
        time_start = time.time()
        points_optimised = render.optimise_points(points)
        optimise_time = time.time() - time_start
        # generate images for a visual comparison if run test as script
        if __name__ == '__main__':
            self.generate_svg(points, '{}_{}_raw'.format(IMAGE_BASE_NAME, period))
            self.generate_svg(points_optimised, '{}_{}'.format(IMAGE_BASE_NAME, period))

        # check needed points
        if len(points) >= 700:
            self.assertTrue(700 <= len(points_optimised) <= 800)
        else:
            self.assertEqual(len(points), len(points_optimised))
        # optimise not more than 1 minute
        self.assertGreater(timeout, optimise_time)
        # NOTE(vlebedev): Ensure that native types are returned instead of those of numpy.
        #                 See CLOS-1947 for details.
        self.assertIsInstance(points_optimised, list)
        some_point = random.choice(points_optimised)
        self.assertIsInstance(some_point, list)
        self.assertEqual([type(c) for c in some_point], [float, float])

if __name__ == '__main__':
    unittest.main()
