# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import os
import unittest
from unittest.mock import patch

import ddt
from sqlalchemy import Column, String, Integer
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import delete, and_, between, select

import lvestats.lib.dbengine
from test.generate_lvestats2db import DatabaseGenerator
from lvestats.lib.dbengine import cfg_get_uri, DatabaseConfigError
from lvestats.orm import history, history_x60, history_gov, incident, servers, user
from lvestats.orm.base import JsonSerializableBase
from lvestats.orm.const import LVE_STATS_2_TABLENAME_PREFIX, SERVER_ID_LENGTH
from lvestats.utils.dbmigrator import alembic_migrate
from test.base.base_db_test import BaseDbTest

MYSQL_SOCKET_PATH = '/var/lib/mysql/mysql.sock'
MYSQL_SOCKET_PATH_ENCODED = MYSQL_SOCKET_PATH.replace('/', '%2F')

MockBase = declarative_base(cls=(JsonSerializableBase,))


class MockServers(MockBase):
    __tablename__ = LVE_STATS_2_TABLENAME_PREFIX + 'servers'

    server_id = Column('server_id', String(SERVER_ID_LENGTH), primary_key=True, autoincrement=False)
    lve_version = Column('lve_version', Integer)
    test_column = Column('test_column', String)


@ddt.ddt
class TestDBEngine(BaseDbTest):
    TABLES = [history, history_x60, history_gov, incident, servers, user]

    def tearDown(self):
        self.recreate_tables()

    def test_validate_database_correct(self):
        result = lvestats.lib.dbengine.validate_database(self.engine)
        self.assertFalse(result["table_error"])
        self.assertFalse(result["column_error"])

    @ddt.data(*TABLES)
    def test_validate_database_no_table(self, table):
        table.__table__.drop(self.engine)
        result = lvestats.lib.dbengine.validate_database(self.engine)
        self.assertTrue(result["table_error"])
        self.assertFalse(result["column_error"])

    def test_validate_database_no_column(self):
        result = lvestats.lib.dbengine.validate_database(self.engine, base=MockBase)
        self.assertFalse(result["table_error"])
        self.assertTrue(result["column_error"])


class TestKeepAliveRecovery(BaseDbTest):
    def setUp(self):
        BaseDbTest.setUp(self)
        self.session = sessionmaker(bind=self.engine)()

    def tearDown(self):
        BaseDbTest.tearDown(self)
        self.session.close()

    def test_validate_and_fix_lost_keep_alive(self):
        # generate database
        db_generator = DatabaseGenerator(
            db_uri=self.engine.url, server_id=self.cfg['server_id'],
            lve_data_path=os.path.join(os.path.dirname(__file__), 'data', 'lveinfo_time_unit_dyn.csv'))
        db_generator.dbengine = self.engine
        db_generator.generate_database()
        self.session.commit()  # using for clear cache in mariadb between queries
        # check that all keep alive present
        self.assertListEqual(
            lvestats.lib.dbengine.find_lost_keep_alive(
                self.session, server_id=self.cfg['server_id']), list())

        # delete some keep alive records
        ts_list = list(self.engine.execute(select([history.created]).where(and_(
            history.server_id == self.cfg['server_id'],
            history.id == 0)).order_by(history.created)))
        ts_from, ts_to = ts_list[len(ts_list)//2][0], ts_list[len(ts_list)//2 + 100][0]
        self.engine.execute(delete(history, and_(
            history.server_id == self.cfg['server_id'],
            history.id == 0,
            between(history.created, ts_from, ts_to)
        )))

        # test fix incorrect database
        self.session.commit()
        self.assertTrue(
            lvestats.lib.dbengine.find_lost_keep_alive(
                self.session, server_id=self.cfg['server_id']))
        lvestats.lib.dbengine.fix_db(self.engine, cfg=self.cfg)
        self.session.commit()
        self.assertListEqual(
            lvestats.lib.dbengine.find_lost_keep_alive(
                self.session, server_id=self.cfg['server_id']), list())


class TestDBMigrator(BaseDbTest):

    def test_downgrade_upgrade(self):
        alembic_migrate(self.engine, alembic_version='base',
                        lve_stats_cfg=self.cfg)
        alembic_migrate(self.engine, lve_stats_cfg=self.cfg)
        validate_result = lvestats.lib.dbengine.validate_database(
            self.engine, hide_logging=True)
        self.assertDictContainsSubset(
            {"table_error": False, "column_error": False}, validate_result)


correct_config_test_case = (
    ({'db_type': 'sqlite', 'connect_string': 'localhost:root:root'},
     'sqlite:////var/lve/lvestats2.db'),

    ({'db_type': 'sqlite', 'sqlite_db_path': '/var/lve/lvestats2.sqlite'},
     'sqlite:////var/lve/lvestats2.sqlite'),

    ({'db_type': 'mysql', 'connect_string': 'dbuser:dbpass@localhost:12345/lvestats_db'},
     f'mysql+pymysql://dbuser:dbpass@localhost:12345/lvestats_db?unix_socket={MYSQL_SOCKET_PATH_ENCODED}'),

    ({'db_type': 'mysql', 'connect_string': 'dbuser:dbpass@127.0.0.1:12345/lvestats_db'},
     f'mysql+pymysql://dbuser:dbpass@127.0.0.1:12345/lvestats_db'),

    ({'db_type': 'postgresql',
      'connect_string': 'dbuser:dbpass@localhost:12345/lvestats_db'},
     'postgresql+psycopg2://dbuser:dbpass@localhost:12345/lvestats_db'),
)

incorrect_config_test_case = (
    ({'db_type': 'mariadb', 'connect_string': 'localhost:root:root'},
     DatabaseConfigError),  # incorrect database type
    ({'connect_string': 'localhost:root:root'},
     DatabaseConfigError),  # no 'db_type' present
    ({'db_type': 'postgresql'},
     DatabaseConfigError),  # no 'connect_string' present
)


@ddt.ddt
class TestConfig(unittest.TestCase):

    @ddt.data(*correct_config_test_case)
    @ddt.unpack
    def test_config(self, cfg, dburi):
        with patch('clcommon.mysql_lib.get_unix_socket_path', return_value=MYSQL_SOCKET_PATH):
            self.assertEqual(dburi, cfg_get_uri(cfg))

    @ddt.data(*incorrect_config_test_case)
    @ddt.unpack
    def test_incorrect_config(self, cfg, exception):
        with patch('clcommon.mysql_lib.get_unix_socket_path', return_value=MYSQL_SOCKET_PATH):
            self.failureException(exception, cfg_get_uri, cfg)


if __name__ == '__main__':
    unittest.main()
