# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import os
import sys
import shutil
import unittest
import json

import ddt
import mock
from contextlib import contextmanager

from lvestats.lib.snapshot import Snapshot

TEST_FILES = ['a.snapshot', '45.snapshot', '22.snapshot', '3213.snapshot', '2a2.snapshot']

TEST_PATH = '/tmp/lve_stats_test/snapshots'


@contextmanager
def mock_drop_privileges(self):
    yield


@ddt.ddt
class TestLibSnapshot(unittest.TestCase):
    def setUp(self):
        shutil.rmtree(TEST_PATH, ignore_errors=True)
        self.patcher1 = mock.patch('lvestats.lib.snapshot.SNAPSHOT_PATH', TEST_PATH)
        self.patcher1.start()
        if os.geteuid() != 0:
            self.patcher2 = mock.patch('lvestats.lib.snapshot.drop_privileges', mock_drop_privileges)
            self.patcher2.start()
        self._test_snapshot_string = '{"snap_sql": "", "server_id": "localhost", "uid": 500, "incident_start_time": ' \
            '1467197361, "snap_faults": {"cpu_fault": 3}, "snap_http": [], "snap_proc": {"24645": ' \
            '{"MEM": "1", "CMD": "cat /dev/urandom", "PID": "24645", "CPU": "27%"}, "24644": {"MEM": ' \
            '"1", "CMD": "lve_suwrapper 600 cat /dev/urandom", "PID": "24644", "CPU": "0%"}}, ' \
            '"dump_time": 1467197361}'
        self.snapshot = Snapshot({
            "uid": 500,
            "incident_start_time": 1467197361,
            "incident_end_time": None})

    def tearDown(self):
        self.patcher1.stop()
        if os.geteuid() != 0:
            self.patcher2.stop()

    @ddt.data(['22.snapshot', 22],
              ['31321.snapshot', 31321],
              ['400.snapshot', 400],
              ['2.snapshot', 2],
              ['440.', None],
              ['440', None],
              ['440.snapsho', None],
              ['.440.snapshot', None],
              ['440..snapshot', None],
              ['440.SNAPSHOT', None],
              ['', None],
              ['.', None],
              ['abs.snapshot', None],
              ['.snapshot', None])
    @ddt.unpack
    def test_get_ts(self, file_, result):
        self.assertEqual(Snapshot.get_ts(file_), result)

    @ddt.data([[], None, None, []],
              [TEST_FILES, None, None, [22, 45, 3213]],
              [TEST_FILES, 33, None, [45, 3213]],
              [TEST_FILES, 30, 50, [45]],
              [TEST_FILES, None, 7000, [22, 45, 3213]])
    @ddt.unpack
    def test_snapshot_filter(self, files, from_ts, to_ts, result):
        self.assertEqual(Snapshot.snapshot_filter(files, from_ts, to_ts), result)

    @ddt.data((10, 100, [22, 45]), (10, None, [22, 45, 3213]), (None, None, [22, 45, 3213]))
    @ddt.unpack
    @mock.patch('lvestats.lib.snapshot.Snapshot.get_file_list')
    def test_get_ts_list(self, from_ts, to_ts, result, mock_get_file_list):
        mock_get_file_list.return_value = TEST_FILES
        sn = Snapshot({"uid": 100})
        self.assertEqual(sn.get_ts_list(from_ts, to_ts), result)

    def save_snapshot(self, time):
        snapshot_dump_time = time
        proc_json = json.loads('{"24644": {"CPU": "0%", "MEM": "1", "CMD": "lve_suwrapper 600 cat /dev/urandom", '
                               '"PID": "24644"}, "24645": {"CPU": "27%", "MEM": "1", "CMD": "cat /dev/urandom", '
                               '"PID": "24645"}}')
        fault_json = json.loads('{"cpu_fault": 3}')
        data = {
            'uid': 500,
            'server_id': 'localhost',
            'dump_time': snapshot_dump_time,
            'incident_start_time': self.snapshot.incident["incident_start_time"],
            'snap_proc': proc_json,
            'snap_sql': json.loads('""'),
            'snap_http': json.loads('[]'),
            'snap_faults': fault_json}
        return self.snapshot.save(data)

    @unittest.skipUnless(os.getegid() == 0, 'Can run only on live test system')
    def test_save_snapshot_race_condition(self):
        snapshot_timestamp = 1467197361
        root_file = os.path.join(TEST_PATH, 'important_root_file')
        snapshot_dir = os.path.join(TEST_PATH, '500')
        snapshot_file = os.path.join(snapshot_dir, str(snapshot_timestamp) + '.snapshot')

        os.makedirs(snapshot_dir)
        open(root_file, 'a').close()
        file_stat = os.stat(root_file)
        os.symlink(root_file, snapshot_file)

        with self.assertRaises(IOError):
            self.save_snapshot(snapshot_timestamp)
        new_file_stat = os.stat(root_file)
        self.assertEqual(file_stat.st_uid, new_file_stat.st_uid)
        self.assertEqual(file_stat.st_gid, new_file_stat.st_gid)
        self.assertEqual(file_stat.st_mode, new_file_stat.st_mode)
        self.assertEqual(file_stat.st_size, new_file_stat.st_size)

    def test_save_snapshot(self):
        dump_time = 1467197361
        with mock.patch('lvestats.lib.snapshot.Snapshot._replace_unicode_data') as mock_replace_unicode_data:
            filename = self.save_snapshot(dump_time)

            # check if unicode was removed from data
            self.assertTrue(mock_replace_unicode_data.called)

        # lets check the name
        self.assertEqual(filename, os.path.join(TEST_PATH, '500', str(dump_time) + '.snapshot'))

        with open(filename, 'rb') as _file:
            self.assertNotEqual(
                _file.read(),
                b'{"snap_sql": "", "server_id": "localhost", "uid": 500, "incident_start_time": '
                b'1467197361, "snap_faults": {"cpu_fault": 3}, "snap_http": [], "snap_proc": {"24645": '
                b'{"MEM": "1", "CMD": "cat /dev/urandom", "PID": "24645", "CPU": "27%"}, "24644": {"MEM": '
                b'"1", "CMD": "lve_suwrapper 600 cat /dev/urandom", "PID": "24644", "CPU": "0%"}}, '
                b'"dump_time": 1467197361}')

        self.assertEqual(
            json.loads(self.snapshot.read_file_content(filename)),
            json.loads(self._test_snapshot_string))

        # ownership and permissions on the file (user owned, user & group readable)
        st = os.stat(filename)
        if os.geteuid() == 0:  # only if we are running as root, like server does.
            self.assertEqual(st.st_uid, 500)
            self.assertEqual(st.st_mode & 0o777, 0o400)

            # and parent dir should also be user owned, user / group readable
            st = os.stat(os.path.dirname(filename))
            self.assertEqual(st.st_mode & 0o777, 0o751)

    def test_read_file_content(self):
        os.makedirs(TEST_PATH)
        filename = os.path.join(TEST_PATH, 'test_file.snapshot')
        with open(filename, 'w') as f:
            f.write(self._test_snapshot_string)
        self.assertEqual(
            self.snapshot.read_file_content(filename),
            self._test_snapshot_string)
        with open(filename, 'rb') as f:
            file_content = f.read()
        self.assertNotEqual(
            file_content,
            self._test_snapshot_string.encode())
        self.assertLess(
            sys.getsizeof(file_content),
            sys.getsizeof(self._test_snapshot_string))
        self.assertEqual(
            self.snapshot.read_file_content(filename),
            self._test_snapshot_string)

    def test_get_incident_snapshots(self):
        dump_time = 1467197361
        self.save_snapshot(dump_time)
        self.save_snapshot(dump_time + 60)
        self.save_snapshot(dump_time + 120)
        self.save_snapshot(dump_time + 180)

        self.snapshot.incident["incident_end_time"] = None
        snapshots = self.snapshot.get_incident_snapshots()
        self.assertEqual(len(snapshots), 4)

        self.snapshot.incident["incident_end_time"] = dump_time + 120
        snapshots = self.snapshot.get_incident_snapshots()
        self.assertEqual(len(snapshots), 3)

        self.snapshot.incident["incident_start_time"] = dump_time + 60
        snapshots = self.snapshot.get_incident_snapshots()
        self.assertEqual(len(snapshots), 2)

    def test_get_snapshots(self):
        dump_time = self.snapshot.incident["incident_start_time"]
        self.save_snapshot(dump_time)
        self.save_snapshot(dump_time + 60)
        self.save_snapshot(dump_time + 120)
        self.save_snapshot(dump_time + 180)

        snapshots = self.snapshot.get_snapshots(None, None)
        self.assertEqual(len(snapshots), 4)
        snapshots = self.snapshot.get_snapshots(None, dump_time+200)
        self.assertEqual(len(snapshots), 4)
        snapshots = self.snapshot.get_snapshots(dump_time+1, None)
        self.assertEqual(len(snapshots), 3)
        snapshots = self.snapshot.get_snapshots(dump_time+1, dump_time+130)
        self.assertEqual(len(snapshots), 2)

    def test_load_bad_snapshot(self):
        dump_time = 1467197361
        filename = self.snapshot.get_file_name(self.snapshot.ts_to_name(dump_time))

        # lets try empty file
        self.snapshot.create_file(dump_time).close()
        self.snapshot.get_incident_snapshots()  # if we don't throw exception -- we are good

        # lets try no permission
        os.chmod(filename, 000)
        self.snapshot.get_incident_snapshots()

    def test_delete_old(self):
        dump_time = 1467197361
        self.save_snapshot(dump_time)
        self.save_snapshot(dump_time + 60)
        self.save_snapshot(dump_time + 120)
        self.save_snapshot(dump_time + 180)

        self.snapshot.delete_old(dump_time + 60)
        self.assertEqual(len(os.listdir(os.path.join(TEST_PATH, '500'))), 2)

        self.snapshot.delete_old(dump_time + 180)
        self.assertFalse(os.path.exists(os.path.join(TEST_PATH, '500')))
