# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
# pylint: disable-all

import unittest
import urllib.request
import urllib.error
import urllib.parse
from io import BytesIO

import ddt
import mock

from lvestats.lib.commons.litespeed import LiteSpeed, LiteSpeedException, LiteSpeedInvalidCredentials, \
    LiteSpeedDisabledException

ADMIN_REQUEST = [b'127.0.0.1	0	PR	0.0	235	235	0	-2	_AdminVHost	Status					"GET /status\0HTTP/1.1']
FIRST_DOMAIN = [b'46.98.54.77	0	PR	26.9	158	158	0	-2	example.com	APVH_coolwind_Suphp56:	26	/tmp/lshttpd/APVH_coolwind_Suphp56.sock	85970	0	"GET / HTTP/1.1"']
FIRST_DOMAIN_BAD_TIME = [b'46.98.54.77	0	PR	0.-9	158	158	0	-2	example.com	APVH_coolwind_Suphp56:	26	/tmp/lshttpd/APVH_coolwind_Suphp56.sock	85970	0	"GET / HTTP/1.1"']
FIRST_DOMAIN_WITH_LONG_URL = [b'46.98.54.77	0	PR	26.9	158	158	0	-2	example.com	APVH_coolwind_Suphp56:	26	/tmp/lshttpd/APVH_coolwind_Suphp56.sock	85970	0	"GET /some_long/url/that/will/be/truncated..."']
OTHER_DOMAIN = [b'46.98.54.77	0	PR	24.9	158	158	0	-2	notexample.com	APVH_coolwind_Suphp56:	26	/tmp/lshttpd/APVH_coolwind_Suphp56.sock	85970	0	"POST /test.php HTTP/2.0"']
NOT_VALID_INFO = [b'127.0.0.1	0	PR	0.0	235	235	0	-2		Status		"GET /status\0HTTP/1.1']
WWW_DOMAIN = [b'46.98.54.77	0	PR	0.-9	158	158	0	-2	www.example1.com	APVH_coolwind_Suphp56:	26	/tmp/lshttpd/APVH_coolwind_Suphp56.sock	85970	0	"GET / HTTP/1.1"']
SOME_BAD_STRING = [b'something that cannot_be_parsed11"']
FIRST_DOMAIN_WITH_ADMIN = ADMIN_REQUEST + FIRST_DOMAIN
FIRST_DOMAIN_OTHER_DOMAIN_WITH_ADMIN = ADMIN_REQUEST + OTHER_DOMAIN
ALL_REQUESTS_TOGETHER = ADMIN_REQUEST + FIRST_DOMAIN + OTHER_DOMAIN + NOT_VALID_INFO + WWW_DOMAIN


@ddt.ddt
class TestLiteSpeed(unittest.TestCase):
    @ddt.data([ADMIN_REQUEST, [], 0],
              [FIRST_DOMAIN, ['example.com'], 1],
              [FIRST_DOMAIN_BAD_TIME, ['example.com'], 1],
              [OTHER_DOMAIN, ['example.com'], 0],
              [FIRST_DOMAIN_OTHER_DOMAIN_WITH_ADMIN, ['example.com'], 0],
              [FIRST_DOMAIN_WITH_LONG_URL, ['example.com'], 1],
              [ALL_REQUESTS_TOGETHER, [], 0],
              [SOME_BAD_STRING, [], 0],
              [WWW_DOMAIN, ['example1.com'], 1],
              [ALL_REQUESTS_TOGETHER, ['example.com', 'example1.com', 'notexample.com'], 3])
    @ddt.unpack
    def test_domains(self, requests, hosts, result_count):
        with mock.patch.multiple('lvestats.lib.commons.litespeed.LiteSpeed',
                                 _get_requests=mock.DEFAULT,
                                 _get_litespeed_pid=mock.DEFAULT) as mocks, \
                mock.patch('lvestats.lib.commons.litespeed.get_all_user_domains',
                           return_value=hosts):
            mocks['_get_requests'].return_value = requests
            mocks['_get_litespeed_pid'].return_value = 100

            # login-password is unused
            litespeed = LiteSpeed('admin', 'password')
            self.assertEqual(len(litespeed.get_user_data(b'')), result_count)

    @ddt.data([FIRST_DOMAIN,          ['example.com'],    26.9, b'GET', b'/',          b'HTTP/1.1'],
              [FIRST_DOMAIN_BAD_TIME, ['example.com'],    -1,   b'GET', b'/',          b'HTTP/1.1'],
              [OTHER_DOMAIN,          ['notexample.com'], 24.9, b'POST', b'/test.php', b'HTTP/2.0'])
    @ddt.unpack
    def test_domain_time(self, requests, hosts, time, request_type, request_url, request_version):
        with mock.patch.multiple('lvestats.lib.commons.litespeed.LiteSpeed',
                                 _get_requests=mock.DEFAULT,
                                 _get_litespeed_pid=mock.DEFAULT) as mocks, \
             mock.patch('lvestats.lib.commons.litespeed.get_all_user_domains',
                        return_value=hosts):
            mocks['_get_requests'].return_value = requests
            mocks['_get_litespeed_pid'].return_value = 100

            # login-password is unused
            litespeed = LiteSpeed('admin', 'password')
            user_data = litespeed.get_user_data('')
            self.assertEqual(len(user_data), 1)

            user_data_row = user_data[0]

            http_type = user_data_row[2]
            self.assertEqual(http_type, request_type)

            domain_time = user_data_row[5]
            self.assertEqual(domain_time, time)

            domain_url = user_data_row[3]
            self.assertEqual(domain_url, request_url)

            domain_version = user_data_row[4]
            self.assertEqual(domain_version, request_version)

    def test_litespeed_running(self):
        with mock.patch.multiple('lvestats.lib.commons.litespeed.LiteSpeed',
                                 _get_litespeed_pid=mock.DEFAULT) as mocks, \
                mock.patch('lvestats.lib.commons.litespeed.get_all_user_domains'):

            # litespeed connection is alive
            mocks['_get_litespeed_pid'].return_value = None
            self.assertEqual(LiteSpeed.is_litespeed_running(), False)

            # litespeed connection is dead
            mocks['_get_litespeed_pid'].return_value = 12345
            self.assertEqual(LiteSpeed.is_litespeed_running(), True)

    @mock.patch('lvestats.lib.commons.litespeed.urllib.request.urlopen')
    def test_litespeed_connection(self, mock_urlopen):
        litespeed = LiteSpeed('admin', 'password')

        with mock.patch.object(litespeed, '_get_litespeed_webadmin_port', return_value='7080'):
            # connection is ok
            mock_urlopen.return_value = BytesIO()
            self.assertEqual(litespeed._get_requests(), [])

            # bad request (e.g. license is dead)
            mock_urlopen.side_effect = IOError()
            self.assertRaises(LiteSpeedException, litespeed._get_requests)

            mock_urlopen.side_effect = urllib.error.HTTPError("http://somesite.com", 401, "Ooops", None, None)
            self.assertRaises(LiteSpeedInvalidCredentials, litespeed._get_requests)

            mock_urlopen.side_effect = urllib.error.HTTPError("http://somesite.com", 400, "Ooops", None, None)
            self.assertRaises(LiteSpeedDisabledException, litespeed._get_requests)

    @mock.patch('lvestats.lib.commons.litespeed.urllib.request.urlopen')
    def test_litespeed_https_connection(self, mock_urlopen):
        litespeed = LiteSpeed('admin', 'password')
        mock_urlopen.return_value = BytesIO()
        with mock.patch.object(litespeed, '_get_litespeed_webadmin_port', return_value='7080'):
            litespeed._get_requests()
            self.assertIsNotNone(mock_urlopen.call_args[1]['context'])

    @ddt.data([b'"GET /someurl/ HTTP1.1"', (b'GET', b'/someurl/', b'HTTP1.1')],
              [b'"GET /sometruncatedurl..."', (b'GET', b'/sometruncatedurl...', b'')],
              [b'someunparsableinfo', None])
    @ddt.unpack
    def test_parse_request_info(self, row, valid_info):
        litespeed = LiteSpeed('login', 'password')
        self.assertEqual(litespeed._parse_request_info(row), valid_info)

    @ddt.data(
        [b'0.1',  0.1],
        [b'0.-1',  -1],
        [b'0.-5',  -1],
        [b'abc',   -1])
    @ddt.unpack
    def test_get_request_time(self, time, result):
        self.assertEqual(LiteSpeed.to_float(time), result)

    @ddt.data(
        '7000',
        '7080',
        '7100'
    )
    def test_get_ls_webadmin_port(self, _port):
        ls_admin_content_content_part = """<?xml version="1.0" encoding="UTF-8"?>
<adminConfig>
    <listenerList>
        <listener>
            <name>adminListener</name>
            <address>*:{}</address>
            <secure>0</secure>
        </listener>
    </listenerList>
</adminConfig>
        """.format(_port)
        with mock.patch('lvestats.lib.commons.litespeed.open', mock.mock_open(read_data=ls_admin_content_content_part)) as _mock_open:
            litespeed = LiteSpeed('login', 'password')
            ls_web_admin_port = litespeed._get_litespeed_webadmin_port()
            self.assertEqual(ls_web_admin_port, _port)
            _mock_open.assert_called_once_with(litespeed.LS_ADMIN_CONFIG, 'r', encoding='utf-8')

    @ddt.data(
        [AttributeError],
        [IndexError],
        [ValueError],
        [OSError],
        [IOError]
    )
    @ddt.unpack
    def test_get_ls_webadmin_port_exceptions(self, _exception_class):
        error_message = "Test"
        exc = _exception_class(error_message)
        with mock.patch('builtins.open', side_effect=exc):
            litespeed = LiteSpeed('login', 'password')
            try:
                litespeed._get_litespeed_webadmin_port()
                print("Exception should be thrown")
                self.assertFalse(True)
            except LiteSpeedException as e:
                self.assertEqual(str(e), "Can't determine current LiteSpeed webadmin console port from config {}: {}".
                                 format(litespeed.LS_ADMIN_CONFIG, error_message))


if __name__ == "__main__":
    unittest.main()
