# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import logging
import os
import time
import unittest
import warnings

import ddt
import mock
import uuid
from raven.handlers.logging import SentryHandler

import lvestats.lib.commons.logsetup as logsetup
from lvestats.eventloop.main_loop import MainLoop
from lvestats.lib.commons.decorators import no_sigpipe
from lvestats.lib.commons.sentry import LveStatsSentryFilter
from test.helpers import capture_all

RECORD_LIST_FILE = os.path.join(os.path.dirname(__file__), 'data', 'errors_record_list')


def _emit_mock(self, record, **kwargs):
    """
    :type self: SentryHandler
    """
    self._emits_count += 1


def get_sentry_handler_remove_others(log):
    """
    :param logging.Logger log:
    :return lvestats.lib.commons.sentry.LveStatsSentryHandler:
    """
    result = None
    for handler in log.root.handlers[:]:
        if isinstance(handler, SentryHandler):
            result = handler
        else:
            log.root.removeHandler(handler)
    return result


def patch_sentry_handler(log, no_emit=True):
    """
    :param logging.Logger log:
    :return lvestats.lib.commons.sentry.LveStatsSentryHandler:
    """
    handler = get_sentry_handler_remove_others(log)
    handler.client.sample_rate = 1.0  # capture all messages
    if no_emit:
        SentryHandler._emit = _emit_mock
        handler._emits_count = 0
    LveStatsSentryFilter._record_list_file = RECORD_LIST_FILE
    return handler


class ExceptionPlugin(object):
    def execute(self, lve_data):
        raise Exception('I am broken, sorry')


@ddt.ddt
class TestLogsetup(unittest.TestCase):
    @staticmethod
    def reload_logsetup():
        logging.shutdown()
        # TODO: Find where disabled changed to True and fix
        log = logging.getLogger("lvestats")
        log.disabled = 0
        log.root.disabled = 0
        for handler in log.root.handlers[:]:
            log.root.removeHandler(handler)
        logsetup.CONSOLE_LOG = None
        logsetup.SYS_LOG = None
        logsetup.FILE_LOG = None
        logsetup.SENTRY_LOG = None

    def setUp(self):
        self.reload_logsetup()
        self.remove_records()

    def tearDown(self):
        self.remove_records()

    def remove_records(self):
        LveStatsSentryFilter.clear_record_list()
        try:
            os.remove("/tmp/sent")
        except OSError:
            pass

    @classmethod
    def setUpClass(cls):
        cls.log_disable = logging.root.manager.disable
        logging.disable(0)

    @classmethod
    def tearDownClass(cls):
        logging.disable(cls.log_disable)
        cls.reload_logsetup()

    def assert_all_handlers_is_none(self):
        self.assertEqual(logsetup.CONSOLE_LOG, None)
        self.assertEqual(logsetup.SYS_LOG, None)
        self.assertEqual(logsetup.FILE_LOG, None)
        self.assertEqual(logsetup.SENTRY_LOG, None)

    @ddt.data(
        (dict(config={}, console_level=logging.ERROR, file_level=None), logging.ERROR, logging.CRITICAL),
        (dict(config={}), logging.ERROR, logging.CRITICAL),
        (dict(config={"logging_level": "error"}, console_level=logging.INFO, file_level=None), logging.INFO, logging.ERROR),
        (dict(config={"logging_level": "error"}, console_level=logging.WARN, file_level=logging.WARN), logging.WARN, logging.WARN))
    @ddt.unpack
    def test_setup(self, test_data, console_level, file_level):
        self.assert_all_handlers_is_none()
        log = logsetup.setup_logging(**test_data)
        self.assertNotEqual(logsetup.CONSOLE_LOG, None)
        self.assertNotEqual(logsetup.SENTRY_LOG, None)

        if os.getegid() == 0:
            # We can't create this handlers if user is not root
            self.assertNotEqual(logsetup.SYS_LOG, None)
            self.assertNotEqual(logsetup.FILE_LOG, None)
            self.assertEqual(len(log.root.handlers), 4, log.root.handlers)
            self.assertEqual(logsetup.CONSOLE_LOG.level, console_level)
            self.assertEqual(logsetup.FILE_LOG.level, file_level)
        else:
            self.assertEqual(len(log.root.handlers), 3, log.root.handlers)
            self.assertEqual(logsetup.CONSOLE_LOG.level, console_level)

    def test_sentry_handler_log_in_cycle(self):
        self.assert_all_handlers_is_none()
        log = logsetup.setup_logging({})
        handler = patch_sentry_handler(log)
        for _ in range(10):
            log.critical("THIS IS EXTREMELY DANGEROUS ERROR")
            log.error("And may attack at any time")
            log.warning("So we must deal with it")
        self.assertEqual(handler._emits_count, 2)
        self.assertEqual(len(handler.filters[0]._record_list_cache), 2)

    def test_sentry_handler_log_in_continuous_cycle(self):
        self.assert_all_handlers_is_none()
        log = logsetup.setup_logging({})
        handler = patch_sentry_handler(log)
        log.critical("Same text")
        log.critical("Same text")
        self.assertEqual(len(handler.filters[0]._record_list_cache), 2)
        for _ in range(10):
            LveStatsSentryFilter._last_clean = time.time() - 43201
            log.critical("Another bad bug")
        self.assertEqual(handler._emits_count, 12)
        self.assertEqual(len(handler.filters[0]._record_list_cache), 1)

    def test_sentry_handler_log_in_continuous_cycle2(self):
        self.assert_all_handlers_is_none()
        log = logsetup.setup_logging({})
        handler = patch_sentry_handler(log)
        log.critical("Same text")
        log.critical("Same text")
        self.assertEqual(len(handler.filters[0]._record_list_cache), 2)
        for _ in range(10):
            LveStatsSentryFilter._last_clean = time.time() - 43100
            log.critical("Another bad bug")
        self.assertEqual(handler._emits_count, 3)
        self.assertEqual(len(handler.filters[0]._record_list_cache), 3)

    def test_sentry_handler_log_without_cache(self):
        self.assert_all_handlers_is_none()
        log = logsetup.setup_logging({})
        handler = patch_sentry_handler(log)
        for _ in range(10):
            handler.filters[0]._record_list_cache.clear()
            log.critical("Bug #1")
            log.critical("Bug #2")
            log.critical("Bug\n #3 with \n and \t and \r")
        self.assertEqual(handler._emits_count, 3)
        self.assertEqual(len(handler.filters[0]._record_list_cache), 3)

    def test_sentry_handler_same_line_different_text(self):
        self.assert_all_handlers_is_none()
        log = logsetup.setup_logging({})
        handler = patch_sentry_handler(log)
        for i in range(10):
            log.critical("Bug #" + str(i))
        self.assertEqual(handler._emits_count, 10)

    def test_sentry_clear_record_list(self):
        sentry_filter = LveStatsSentryFilter()
        sentry_filter._record_list_file = "/tmp/" + str(uuid.uuid4())
        sentry_filter.clear_record_list()

    def test_main_loop(self):
        """
        1. Checks that events are grouped correctly by Exception class name
        2. Checks that PluginExecutionException considered as Exception and is not sent
        """
        log = logsetup.setup_logging({})
        handler = patch_sentry_handler(log, False)
        with mock.patch("lvestats.lib.commons.func.get_lve_version", return_value=8):
            main_loop = MainLoop(
                {
                    'db_type': 'sqlite',
                    'sqlite_db_path': '/var/lve/test_lvestats2.sqlite',
                    'connect_string': 'root:root@localhost',
                    'server_id': 'localhost',
                    'debug': False},
                interval=1,
                plugins=[ExceptionPlugin],
                plugin_timeout=1)

        def send_mock(*args, **kwargs):
            exception_values = kwargs['exception']['values'][0]
            error = exception_values["type"] + "\n"
            # some debug info to figure out other possible exceptions
            print('Exception class:', exception_values['type'])
            print('Exception value:', exception_values['value'])
            print('Stack trace:')
            for row in exception_values['stacktrace']['frames']:
                print(row['function'], ' at line', row['lineno'], 'in', row['abs_path'])
            with open("/tmp/sent", "a") as f:
                f.write(str(error))

        handler.client.send = send_mock
        main_loop.run(times=1)
        with open("/tmp/sent") as f:
            result = f.readlines()
        # check that grouping works correctly and there is only one record of class Exception
        exception_records = len([row for row in result if row.strip() == 'Exception'])
        self.assertEqual(exception_records, 1)
        self.assertNotIn('PluginExecutionException\n', result)

    @mock.patch.object(logging.Logger, "warning")
    def test_utility_decorator(self, logger_warning):
        @no_sigpipe
        def test_function():
            warnings.warn("Some warning")

        logsetup.setup_logging({})
        if not (os.getegid() == 0 or os.path.exists('/dev/log')):
            # We can't create this handler if user is not root
            logsetup.FILE_LOG = True
        with capture_all(test_function) as result:
            self.assertEqual("", result["stdout"] + result["stderr"])
        self.assertEqual(logger_warning.call_count, 1)
        self.assertIn("Some warning", logger_warning.call_args[0][0])


if __name__ == '__main__':
    unittest.main()
