# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2023 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT

import unittest

from ddt import ddt, data, unpack

from lvestats.lib.bursting import info
from lvestats.orm import BurstingEventType, bursting_events_table


class RowProxy:
    """
    Helper class that is used instead of sqlalchemy.engine.RowProxy object,
    which doesn't provide a public API.
    """

    def __init__(self, lve_id, timestamp, event_type):
        self.data = {
            bursting_events_table.c.lve_id: lve_id,
            bursting_events_table.c.timestamp: timestamp,
            bursting_events_table.c.event_type: event_type,
        }

    def __getitem__(self, key):
        return self.data[key]


@ddt
class TestLveBurstingInfo(unittest.TestCase):
    LVE_ID1 = 1001
    LVE_ID2 = 1002
    TS_FROM = 1698951664.0
    TS_DELTA = 10  # seconds
    TS_TO = TS_FROM + TS_DELTA

    @data(
        # The input data represents rows retrieved from DB
        # and the expected number of table rows to be constructed from them
        ([RowProxy(LVE_ID1, TS_FROM + 1, BurstingEventType.STOPPED.value),
          RowProxy(LVE_ID1, TS_FROM + 3, BurstingEventType.STARTED.value),
          RowProxy(LVE_ID1, TS_FROM + 5, BurstingEventType.STOPPED.value)],
         1),
        ([RowProxy(LVE_ID1, TS_FROM - 5, BurstingEventType.STARTED.value),
          RowProxy(LVE_ID1, TS_FROM + 1, BurstingEventType.STOPPED.value),
          RowProxy(LVE_ID1, TS_FROM + 3, BurstingEventType.STARTED.value),
          RowProxy(LVE_ID1, TS_FROM + 5, BurstingEventType.STOPPED.value)],
         2),
        ([RowProxy(LVE_ID1, TS_FROM - 5, BurstingEventType.STOPPED.value),
          RowProxy(LVE_ID1, TS_FROM + 1, BurstingEventType.STOPPED.value),
          RowProxy(LVE_ID1, TS_FROM + 3, BurstingEventType.STARTED.value),
          RowProxy(LVE_ID1, TS_FROM + 5, BurstingEventType.STOPPED.value)],
         1),
        ([RowProxy(LVE_ID1, TS_FROM - 5, BurstingEventType.STOPPED.value),
          RowProxy(LVE_ID1, TS_FROM + 3, BurstingEventType.STARTED.value)],
         1),
        ([RowProxy(LVE_ID1, TS_FROM - 5, BurstingEventType.STOPPED.value)],
         0),
        ([RowProxy(LVE_ID1, TS_FROM - 5, BurstingEventType.STARTED.value)],
         1),
        ([RowProxy(LVE_ID1, TS_FROM - 5, BurstingEventType.STARTED.value),
          RowProxy(LVE_ID1, TS_FROM + 3, BurstingEventType.STARTED.value),
          RowProxy(LVE_ID1, TS_FROM + 4, BurstingEventType.STARTED.value),
          RowProxy(LVE_ID1, TS_FROM + 5, BurstingEventType.STOPPED.value)],
         1),
        ([RowProxy(LVE_ID1, TS_FROM - 5, BurstingEventType.STOPPED.value),
          RowProxy(LVE_ID1, TS_FROM + 3, BurstingEventType.STARTED.value),
          RowProxy(LVE_ID1, TS_FROM + 4, BurstingEventType.STARTED.value),
          RowProxy(LVE_ID1, TS_FROM + 5, BurstingEventType.STOPPED.value),
          RowProxy(LVE_ID1, TS_FROM + 6, BurstingEventType.STARTED.value)],
         2),
        ([RowProxy(LVE_ID1, TS_FROM - 5, BurstingEventType.STOPPED.value),
          RowProxy(LVE_ID1, TS_FROM + 3, BurstingEventType.STARTED.value),
          RowProxy(LVE_ID1, TS_FROM + 5, BurstingEventType.STOPPED.value),
          RowProxy(LVE_ID1, TS_FROM + 6, BurstingEventType.STARTED.value),
          RowProxy(LVE_ID2, TS_FROM - 4, BurstingEventType.STOPPED.value),
          RowProxy(LVE_ID2, TS_FROM + 3, BurstingEventType.STARTED.value),
          RowProxy(LVE_ID2, TS_FROM + 5, BurstingEventType.STOPPED.value)],
         3),
    )
    @unpack
    def test_build_table_rows(self, rows, expected_table_length):
        table_rows = info._build_table_rows(rows, self.TS_FROM, self.TS_TO)
        self.assertEqual(len(table_rows), expected_table_length)


if __name__ == '__main__':
    unittest.main(verbosity=2)
