# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import time
import os
import unittest

import mock
import ddt

from test.base.base_db_test import BaseDbTest
from lvestats.lib.commons import dateutil
from lvestats.lib.lveinfolib import HistoryShow


@ddt.ddt
class TestHistoryShow(BaseDbTest):
    def setUp(self):
        BaseDbTest.setUp(self)
        os.environ['TZ'] = 'Europe/London'  # set the zero time zone
        time.tzset()
        self.clear_rows = True

        patcher1 = mock.patch('lvestats.lib.config.read_config', return_value={})
        patcher1.start()
        self.addCleanup(patcher1.stop)

    @ddt.data(-500, -1, 0, 5001)
    def test_no_such_id(self, uid):
        self.save_history_line('localhost', uid=0,   _time=3)
        self.save_history_line('localhost', uid=500, _time=3, cpu=100)

        history_show = HistoryShow(
            dbengine=self.engine,
            period_from=dateutil.unixtimestamp_to_gm_datetime(0),
            period_to=dateutil.unixtimestamp_to_gm_datetime(6),
            uid=uid,
            time_unit=1,
            server_id='localhost',
            show_columns=('From', 'To', 'ID', 'aCPU'))

        self.assertListEqual(list(), list(history_show.proceed()))

    def test_use_heartbeat(self):
        # The idea is following: we have 6 records for id=0,
        # therefore avg value for usage should be (100 + 100 + 100) / 6
        self.save_history_line('localhost', uid=0, _time=0)
        self.save_history_line('localhost', uid=0, _time=1)
        self.save_history_line('localhost', uid=0, _time=2)
        self.save_history_line('localhost', uid=0, _time=3)
        self.save_history_line('localhost', uid=0, _time=4)
        self.save_history_line('localhost', uid=0, _time=5)

        # For id=1, there are only 3 records, all have 100% cpu,
        # but avg should be 100 + 100 + 100 / 6
        self.save_history_line('localhost', uid=1, _time=3, cpu=100)
        self.save_history_line('localhost', uid=1, _time=4, cpu=100)
        self.save_history_line('localhost', uid=1, _time=5, cpu=100)

        result = HistoryShow(
            self.engine,
            period_from=dateutil.unixtimestamp_to_gm_datetime(0),
            period_to=dateutil.unixtimestamp_to_gm_datetime(6),
            server_id='localhost',
            show_columns=('From', 'To', 'ID', 'aCPU'),
            time_unit=6).proceed()
        # we divide by 100 here because show_history has 0..1 range for aCPU
        self.assertListEqual([[0, 6, 1, 300 / 100.0 / 6.0]], list(result))

    def test_aggregate_works_when_t_from_is_not_queried(self):
        self.save_history_line('localhost', uid=0, _time=1, cpu=0)
        self.save_history_line('localhost', uid=1, _time=1, cpu=100)
        result = HistoryShow(
            self.engine,
            period_from=dateutil.unixtimestamp_to_gm_datetime(0),
            period_to=dateutil.unixtimestamp_to_gm_datetime(6),
            show_columns=('ID', 'aCPU'),
            time_unit=3).proceed()

        self.assertEqual(list(result), [[1, 1.0]])

    def test_aggregate_works_when_nothing_is_queried(self):
        self.save_history_line('localhost', uid=0, _time=0)
        self.save_history_line('localhost', uid=1, _time=0, cpu=100)
        HistoryShow(self.engine,
                    period_from=dateutil.unixtimestamp_to_gm_datetime(0),
                    period_to=dateutil.unixtimestamp_to_gm_datetime(6),
                    time_unit=3).proceed()

    def test_use_heartbeat_for_number_of_users_and_periods(self):
        self.save_history_line('localhost', uid=0, _time=0)
        self.save_history_line('localhost', uid=0, _time=1)
        self.save_history_line('localhost', uid=0, _time=2)
        self.save_history_line('localhost', uid=0, _time=3)
        self.save_history_line('localhost', uid=0, _time=4)
        self.save_history_line('localhost', uid=0, _time=5)

        self.save_history_line('localhost', uid=1, _time=0, cpu=100)
        self.save_history_line('localhost', uid=2, _time=1, cpu=100)
        self.save_history_line('localhost', uid=3, _time=2, cpu=100)
        self.save_history_line('localhost', uid=4, _time=3, cpu=100)
        self.save_history_line('localhost', uid=5, _time=4, cpu=100)
        self.save_history_line('localhost', uid=6, _time=5, cpu=100)

        result = HistoryShow(
            self.engine,
            period_from=dateutil.unixtimestamp_to_gm_datetime(0),
            period_to=dateutil.unixtimestamp_to_gm_datetime(6),
            show_columns=('From', 'To', 'ID', 'aCPU')).proceed()
        expected_cpu_usage1 = round(1.0 / 6, 5)  # pylint: disable=round-builtin
        self.assertEqual([[0, 6, 1, expected_cpu_usage1],
                          [0, 6, 2, expected_cpu_usage1],
                          [0, 6, 3, expected_cpu_usage1],
                          [0, 6, 4, expected_cpu_usage1],
                          [0, 6, 5, expected_cpu_usage1],
                          [0, 6, 6, expected_cpu_usage1]], list(result))

    def test_ignore_heartbeat_if_there_are_data(self):
        # For id=1, there are only 3 records, all have 100% cpu,
        # so avg should be (100 + 100 + 100) / 3
        # heartbeat
        self.save_history_line('localhost', uid=0, _time=3, cpu=100)
        self.save_history_line('localhost', uid=0, _time=4, cpu=100)
        self.save_history_line('localhost', uid=0, _time=5, cpu=100)
        # user data
        self.save_history_line('localhost', uid=1, _time=3, cpu=100)
        self.save_history_line('localhost', uid=1, _time=4, cpu=100)
        self.save_history_line('localhost', uid=1, _time=5, cpu=100)

        result = HistoryShow(
            self.engine,
            period_from=dateutil.unixtimestamp_to_gm_datetime(0),
            period_to=dateutil.unixtimestamp_to_gm_datetime(6),
            show_columns=('From', 'To', 'ID', 'aCPU'),
            time_unit=6).proceed()
        # we divide by 100 here because show_history has 0..1 range for aCPU
        self.assertListEqual([[0, 6, 1, 300 / 3.0 / 100]], list(result))

    def test_get_ID(self):
        # For id=1, there are only 3 records, all have 100% cpu,
        # so avg should be (100 + 100 + 100) / 3
        # heartbeat
        self.save_history_line('localhost', uid=0, _time=1, cpu=100)
        self.save_history_line('localhost', uid=0, _time=2, cpu=100)
        self.save_history_line('localhost', uid=0, _time=3, cpu=100)
        # user 1 data
        self.save_history_line('localhost', uid=1, _time=1, cpu=100)
        self.save_history_line('localhost', uid=1, _time=2, cpu=100)
        self.save_history_line('localhost', uid=1, _time=3, cpu=100)
        # user 2 data
        self.save_history_line('localhost', uid=2, _time=1, cpu=100)
        self.save_history_line('localhost', uid=2, _time=2, cpu=100)
        self.save_history_line('localhost', uid=2, _time=3, cpu=100)
        # user 3 data
        self.save_history_line('localhost', uid=3, _time=1, cpu=100)
        self.save_history_line('localhost', uid=3, _time=2, cpu=100)
        self.save_history_line('localhost', uid=3, _time=3, cpu=100)

        tested_user_id = 2
        result = HistoryShow(
            self.engine,
            period_from=dateutil.unixtimestamp_to_gm_datetime(0),
            period_to=dateutil.unixtimestamp_to_gm_datetime(4),
            show_columns=('From', 'To', 'ID', 'aCPU'),
            uid=tested_user_id,
            time_unit=3).proceed()
        # we divide by 100 here because show_history has 0..1 range for aCPU
        self.assertListEqual([[0, 3, tested_user_id, 1.0], [3, 6, tested_user_id, 1.0]], list(result))

    @ddt.unpack
    @ddt.data(([1, 2],                [1, 2]),
              ((2, 3, 4),             [2, 3]),
              ([1],                   [1]),
              ([4],                   []),
              ([4, 5],                []),
              ([],                    []),
              (tuple(),               []),
              (list(range(1, 850)),   [1, 2, 3]),
              # test big amount of id's here
              # because some databases have limit variables
              (list(range(1, 2501)),  [1, 2, 3, 2500]),
              (list(range(2, 10000)), [2, 3, 2500, 5000])
              )
    def test_multiple_uids(self, uids, uids_output):
        self.save_history_line('localhost', uid=0, _time=1, cpu=100)
        self.save_history_line('localhost', uid=1, _time=1, cpu=100)

        self.save_history_line('localhost', uid=0, _time=3, cpu=100)
        self.save_history_line('localhost', uid=2, _time=3, cpu=100)
        self.save_history_line('localhost', uid=3, _time=3, cpu=100)
        self.save_history_line('localhost', uid=2500, _time=3, cpu=100)
        self.save_history_line('localhost', uid=5000, _time=3, cpu=100)

        valid_results = HistoryShow(
            self.engine,
            period_from=0,
            period_to=5,
            uid=None).proceed()
        interesting_results = [item for item in valid_results if item[0] in uids]

        result = list(HistoryShow(
            self.engine,
            period_from=0,
            period_to=5,
            uid=uids).proceed())
        self.assertEqual(interesting_results, result)
        result_uids = [line[0] for line in result]
        self.assertListEqual(uids_output, result_uids)

    def test_get_aCPU(self):

        self.save_history_line('localhost', uid=0, _time=1, cpu=100)
        self.save_history_line('localhost', uid=0, _time=2, cpu=100)
        self.save_history_line('localhost', uid=0, _time=3, cpu=100)
        # user 1 data
        self.save_history_line('localhost', uid=1, _time=1, cpu=100)
        self.save_history_line('localhost', uid=1, _time=2, cpu=100)
        self.save_history_line('localhost', uid=1, _time=3, cpu=100)
        # user 2 data
        # For id=2, there are only 3 records
        # first two are in interval 0-3, last one in 3-6
        # as they match heartbeat average will be (cpu1+cpu2)/2 for first one, and cpu3 for second one.
        # average will
        cpu1, cpu2, cpu3 = 110, 120, 130
        self.save_history_line('localhost', uid=2, _time=1, cpu=cpu1)
        self.save_history_line('localhost', uid=2, _time=2, cpu=cpu2)
        self.save_history_line('localhost', uid=2, _time=3, cpu=cpu3)
        # user 3 data
        self.save_history_line('localhost', uid=3, _time=1, cpu=100)
        self.save_history_line('localhost', uid=3, _time=2, cpu=100)
        self.save_history_line('localhost', uid=3, _time=3, cpu=100)

        result = HistoryShow(
            self.engine,
            period_from=dateutil.unixtimestamp_to_gm_datetime(0),
            period_to=dateutil.unixtimestamp_to_gm_datetime(4),
            show_columns=('From', 'To', 'ID', 'aCPU'),
            uid=2,
            time_unit=3).proceed()
        self.assertListEqual([[0, 3, 2, (cpu1+cpu2)/2.0/100], [3, 6, 2, cpu3/100.]], list(result))

    def test_show_all(self):
        self.save_history_line('localhost', uid=0, _time=1, cpu=100)
        self.save_history_line(
            'localhost', uid=503, _time=1, lnproc=9, iops=8, io=6,
            cpu_limit=17, io_limit=19, io_fault=15, memphy_fault=11, memphy=5,
            mep_limit=18, mep_fault=12, lmemphy=21, nproc_fault=13, mem=3,
            cpu_fault=14, mem_fault=10, mep=4, nproc=7, created=0, liops=22,
            mem_limit=20, iops_fault=16, cpu=2)

        fields = ['ID', 'CPUf', 'lCPU', 'lVMem', 'VMemF', 'lPMem', 'PMemF',
                  'EPf', 'lEP']
        by_fault_arg = ['CPUf', 'VMemF', 'PMemF', 'EPf', 'NprocF', 'IOf', 'IOPSf']
        date_from = dateutil.unixtimestamp_to_gm_datetime(0)
        date_to = dateutil.unixtimestamp_to_gm_datetime(4)
        aggregation_period = 3
        notify_min_faults = 1
        users_fault_data_no_show_all = HistoryShow(
            self.engine, date_from, date_to, show_columns=fields,
            time_unit=aggregation_period, order_by='anyF',
            by_fault=by_fault_arg, threshold=notify_min_faults,
        ).proceed()
        # Implement show_all by not specifying fields
        users_fault_data_show_all = HistoryShow(
            self.engine,
            date_from,
            date_to,
            time_unit=aggregation_period,
            order_by='anyF',
            by_fault=by_fault_arg,
            threshold=notify_min_faults,
        ).proceed()

        # without show all key we must take only asked fields
        self.assertTrue(len(list(users_fault_data_no_show_all)[0]) == len(fields))
        # with show all we must get all fields
        self.assertTrue(len(list(users_fault_data_show_all)[0]) > len(fields))

    @ddt.data(True, False)
    def test_skip_maxuid_lve(self, skip_maxuid_lve):
        maxuid_lve = 60001
        reseller_id = 1000000503
        aggregation_period = 3
        self.save_history_line('localhost', uid=0, _time=1, cpu=100)
        self.save_history_line(
            'localhost', uid=503, _time=1, lnproc=9, iops=8, io=6,
            cpu_limit=17, io_limit=19, io_fault=15, memphy_fault=11, memphy=5,
            mep_limit=18, mep_fault=12, lmemphy=21, nproc_fault=13, mem=3,
            cpu_fault=14, mem_fault=10, mep=4, nproc=7, created=0, liops=22,
            mem_limit=20, iops_fault=16, cpu=2)
        self.save_history_line(
            'localhost', uid=maxuid_lve, _time=1, lnproc=9, iops=8, io=6,
            cpu_limit=17, io_limit=19, io_fault=15, memphy_fault=11, memphy=5,
            mep_limit=18, mep_fault=12, lmemphy=21, nproc_fault=13, mem=3,
            cpu_fault=14, mem_fault=10, mep=4, nproc=7, created=0, liops=22,
            mem_limit=20, iops_fault=16, cpu=2)
        self.save_history_line(
            'localhost', uid=reseller_id, _time=1, lnproc=9, iops=8, io=6,
            cpu_limit=17, io_limit=19, io_fault=15, memphy_fault=11, memphy=5,
            mep_limit=18, mep_fault=12, lmemphy=21, nproc_fault=13, mem=3,
            cpu_fault=14, mem_fault=10, mep=4, nproc=7, created=0, liops=22,
            mem_limit=20, iops_fault=16, cpu=2)

        date_from = dateutil.unixtimestamp_to_gm_datetime(0)
        date_to = dateutil.unixtimestamp_to_gm_datetime(4)
        with mock.patch('lvestats.lib.lveinfolib.str_to_bool', return_value=skip_maxuid_lve):
            usage = HistoryShow(
                self.engine,
                date_from,
                date_to,
                time_unit=aggregation_period
            ).proceed()

        ids = [row[0] for row in usage]
        self.assertEqual(maxuid_lve not in ids, skip_maxuid_lve)
        self.assertIn(reseller_id, ids)

    def test_order_by(self):
        # heartbeat
        # user 1 data
        self.save_history_line("localhost", uid=0, _time=1, cpu=0)
        self.save_history_line("localhost", uid=0, _time=2, cpu=0)
        self.save_history_line("localhost", uid=0, _time=3, cpu=0)
        self.save_history_line("localhost", uid=0, _time=5, cpu=0)
        self.save_history_line("localhost", uid=0, _time=7, cpu=0)
        # user 1 data
        self.save_history_line("localhost", uid=1, _time=1, cpu=100)
        self.save_history_line("localhost", uid=1, _time=2, cpu=110)
        self.save_history_line("localhost", uid=1, _time=3, cpu=120)
        # user 2 data
        self.save_history_line("localhost", uid=2, _time=1, cpu=50)
        self.save_history_line("localhost", uid=2, _time=2, cpu=80)
        self.save_history_line("localhost", uid=2, _time=3, cpu=70)
        # user 3 data
        self.save_history_line("localhost", uid=3, _time=1, cpu=30)
        self.save_history_line("localhost", uid=3, _time=3, cpu=20)
        self.save_history_line("localhost", uid=3, _time=5, cpu=100)
        self.save_history_line("localhost", uid=3, _time=7, cpu=30)

        fields = ["ID", "aCPU"]
        date_from = dateutil.unixtimestamp_to_gm_datetime(0)
        date_to = dateutil.unixtimestamp_to_gm_datetime(10)
        order_by_cpu_all_users = HistoryShow(
            self.engine, date_from, date_to, show_columns=fields,
            time_unit=3, order_by="aCPU",
        ).proceed()
        order_by_cpu_for_user = HistoryShow(
            self.engine, date_from, date_to, show_columns=fields, uid=3,
            time_unit=1, order_by="aCPU",
        ).proceed()

        self.assertListEqual([[1, 0.66], [2, 0.4], [3, 0.36]],
                             list(order_by_cpu_all_users))
        self.assertListEqual([[3, 1.0], [3, 0.3], [3, 0.3], [3, 0.2]],
                             list(order_by_cpu_for_user))
        # self.assertListEqual([[3, 0.2], [3, 0.3], [3, 0.3], [3, 1.0]],
        #                      order_by_cpu_for_user)

        for data in (list(order_by_cpu_all_users), list(order_by_cpu_for_user)):
            val = 0
            for row in reversed(data):
                self.assertTrue(val <= row[1])
                val = row[1]

    def test_by_usage(self):
        # heartbeat
        self.save_history_line("localhost", uid=0, _time=1)
        self.save_history_line("localhost", uid=0, _time=3)
        self.save_history_line("localhost", uid=0, _time=5)
        self.save_history_line("localhost", uid=0, _time=7)
        self.save_history_line("localhost", uid=0, _time=9)
        self.save_history_line("localhost", uid=0, _time=11)

        # user 1 data
        self.save_history_line("localhost", uid=1, _time=1, cpu=85, cpu_limit=100,
                               mem=60863, mem_limit=100000, nproc=52, lnproc=100)
        self.save_history_line("localhost", uid=1, _time=3, cpu=95, cpu_limit=100,
                               mem=90863, mem_limit=100000, nproc=72, lnproc=100)
        self.save_history_line("localhost", uid=1, _time=5, cpu=75, cpu_limit=100,
                               mem=80863, mem_limit=100000, nproc=94, lnproc=100)
        self.save_history_line("localhost", uid=1, _time=7, cpu=55, cpu_limit=100,
                               mem=50363, mem_limit=100000, nproc=24, lnproc=100)
        self.save_history_line("localhost", uid=1, _time=9, cpu=69, cpu_limit=100,
                               mem=80822, mem_limit=100000, nproc=5, lnproc=100)
        self.save_history_line("localhost", uid=1, _time=11, cpu=98, cpu_limit=100,
                               mem=90, mem_limit=100000, nproc=100, lnproc=100)

        fields = ["ID", "aCPU", "lCPU", "aVMem", "lVMem", "aNproc", "lNproc"]
        aggregation_period = 1
        date_from = dateutil.unixtimestamp_to_gm_datetime(0)
        date_to = dateutil.unixtimestamp_to_gm_datetime(20)
        default_params = {
            "dbengine": self.engine,
            "period_from": date_from,
            "period_to": date_to,
            "show_columns": fields,
            "uid": 1,
            "time_unit": aggregation_period
        }
        by_usage_cpu = HistoryShow(
            by_usage=["aCPU"],
            by_usage_percentage=0.73,
            **default_params
        ).proceed()
        self.assertEqual(len(list(by_usage_cpu)), 4)

        by_usage_mem = HistoryShow(
            by_usage=["aVMem"],
            by_usage_percentage=0.75,
            **default_params
        ).proceed()
        self.assertTrue(len(list(by_usage_mem)) == 3)

        by_usage_nproc = HistoryShow(
            by_usage=["aNproc"],
            by_usage_percentage=0,
            **default_params
        ).proceed()
        self.assertTrue(len(list(by_usage_nproc)) == 6)

    def test_by_fault(self):
        # user data
        self.save_history_line("localhost", uid=1, _time=1, cpu_fault=5, mem_fault=19, nproc_fault=1)
        self.save_history_line("localhost", uid=1, _time=3, cpu_fault=11, mem_fault=63, nproc_fault=14)
        self.save_history_line("localhost", uid=1, _time=5, cpu_fault=4, mem_fault=10, nproc_fault=7)
        self.save_history_line("localhost", uid=1, _time=7, cpu_fault=2, mem_fault=20, nproc_fault=8)
        self.save_history_line("localhost", uid=1, _time=9, cpu_fault=8, mem_fault=10, nproc_fault=20)
        self.save_history_line("localhost", uid=1, _time=11, cpu_fault=23, mem_fault=15, nproc_fault=12)

        fields = ["ID", "CPUf", "VMemf"]
        aggregation_period = 1
        date_from = dateutil.unixtimestamp_to_gm_datetime(0)
        date_to = dateutil.unixtimestamp_to_gm_datetime(20)
        default_params = {
            "dbengine": self.engine,
            "period_from": date_from,
            "period_to": date_to,
            "show_columns": fields,
            "uid": 1,
            "time_unit": aggregation_period
        }

        by_fault_cpu = HistoryShow(
            by_fault=["CPUf"],
            threshold=10,
            **default_params
        ).proceed().rows
        self.assertEqual(len(by_fault_cpu), 2)

        by_fault_mem = HistoryShow(
            by_fault=["VMemf"],
            threshold=11,
            **default_params
        ).proceed().rows
        self.assertEqual(len(by_fault_mem), 4)

        by_fault_nproc = HistoryShow(
            by_fault=["Nprocf"],
            threshold=10,
            **default_params
        ).proceed().rows
        self.assertEqual(len(by_fault_nproc), 3)

        by_fault_cpu_or_mem = HistoryShow(
            by_fault=["CPUf", "VMemf"],
            threshold=19,
            **default_params
        ).proceed().rows
        self.assertEqual(len(by_fault_cpu_or_mem), 4)

    def test_time_unit_eq_date_period_order_by(self):
        """
        Aggregation period and select data period is same
        """
        # heartbeat
        self.save_history_line("localhost", uid=0, _time=0)
        self.save_history_line("localhost", uid=0, _time=1)
        self.save_history_line("localhost", uid=0, _time=2)

        # user 1 data
        self.save_history_line("localhost", uid=1, _time=0, cpu=35, cpu_limit=150, mem=6086, mem_limit=10000, nproc=52,
                               lnproc=100)
        self.save_history_line("localhost", uid=1, _time=1, cpu=95, cpu_limit=150, mem=9086, mem_limit=10000, nproc=72,
                               lnproc=100)
        self.save_history_line("localhost", uid=1, _time=2, cpu=115, cpu_limit=150, mem=8286, mem_limit=10000, nproc=94,
                               lnproc=100)

        # user 2 data
        self.save_history_line("localhost", uid=2, _time=0, cpu=55, cpu_limit=100, mem=5036, mem_limit=10000, nproc=32,
                               lnproc=50)
        self.save_history_line("localhost", uid=2, _time=1, cpu=69, cpu_limit=100, mem=8082, mem_limit=10000, nproc=23,
                               lnproc=50)
        self.save_history_line("localhost", uid=2, _time=2, cpu=98, cpu_limit=100, mem=90, mem_limit=10000, nproc=1,
                               lnproc=50)

        # user 3 data
        self.save_history_line("localhost", uid=3, _time=0, cpu=15, cpu_limit=50, mem=506, mem_limit=1000, nproc=2,
                               lnproc=20)
        self.save_history_line("localhost", uid=3, _time=1, cpu=29, cpu_limit=50, mem=802, mem_limit=1000, nproc=5,
                               lnproc=20)
        self.save_history_line("localhost", uid=3, _time=2, cpu=38, cpu_limit=50, mem=902, mem_limit=1000, nproc=10,
                               lnproc=20)

        date_from = dateutil.unixtimestamp_to_gm_datetime(0)
        date_to = dateutil.unixtimestamp_to_gm_datetime(10)
        fields = ["ID", "From", "To", "aCPU", "lCPU", "aVMem", "lVMem",
                  "aNproc", "lNproc"]
        params = {
            "dbengine": self.engine,
            "period_from": date_from,
            "period_to": date_to,
            "uid": None,
            "time_unit": 10,
            "show_columns": fields,
            "by_usage": None,
            "by_usage_percentage": None,
            "by_fault": None,
            "threshold": None
        }

        rows = HistoryShow(order_by="aCPU", **params).proceed()
        self.assertListEqual(list(rows), [[1, 0, 10, 0.81667, 1.5, 7819.333, 10000.0, 72.667, 100],
                                          [2, 0, 10, 0.74, 1.0, 4402.667, 10000.0, 18.667, 50],
                                          [3, 0, 10, 0.27333, 0.5, 736.667, 1000.0, 5.667, 20]])

    def test_limit_change_in_aggregation_group(self):
        """
        LVES-273: show correct limits when it`s change in aggregation period
        """
        # heartbeat
        self.save_history_line("localhost", uid=0, _time=1, cpu=0, cpu_limit=10000)
        self.save_history_line("localhost", uid=0, _time=2, cpu=0, cpu_limit=12000)
        self.save_history_line("localhost", uid=0, _time=3, cpu=0, cpu_limit=8000)
        self.save_history_line("localhost", uid=0, _time=4, cpu=0, cpu_limit=18000)
        self.save_history_line("localhost", uid=0, _time=5, cpu=0, cpu_limit=1000)
        self.save_history_line("localhost", uid=0, _time=6, cpu=0, cpu_limit=9000)
        self.save_history_line("localhost", uid=0, _time=7, cpu=0, cpu_limit=9000)
        # user data

        self.save_history_line("localhost", uid=1, _time=1, cpu=1500, cpu_limit=10000)
        self.save_history_line("localhost", uid=1, _time=2, cpu=11500, cpu_limit=12000)
        self.save_history_line("localhost", uid=1, _time=3, cpu=7500, cpu_limit=8000)
        self.save_history_line("localhost", uid=1, _time=4, cpu=15500, cpu_limit=18000)
        self.save_history_line("localhost", uid=1, _time=5, cpu=1000, cpu_limit=1000)
        self.save_history_line("localhost", uid=1, _time=6, cpu=8000, cpu_limit=9000)
        self.save_history_line("localhost", uid=1, _time=7, cpu=8000, cpu_limit=9000)

        date_from = dateutil.unixtimestamp_to_gm_datetime(0)
        date_to = dateutil.unixtimestamp_to_gm_datetime(10)
        fields = ["ID", "From", "To", "aCPU", "lCPU"]
        params = {
            "dbengine": self.engine,
            "period_from": date_from,
            "period_to": date_to,
            "uid": 1,
            "show_columns": fields}

        rows = HistoryShow(time_unit=2, **params).proceed()
        self.assertListEqual(list(rows), [[1, 0, 2, 15.0, 100.0],
                                          [1, 2, 4, 95.0, 120.0],
                                          [1, 4, 6, 82.50, 180.0],
                                          [1, 6, 8, 80.0, 90.0]])

        rows = HistoryShow(time_unit=3, **params).proceed()
        self.assertListEqual(list(rows), [[1, 0, 3, 65.0, 120.0],
                                          [1, 3, 6, 80.0, 180.0],
                                          [1, 6, 9, 80, 90.0]])

    def test_limit_update(self):
        self.save_history_line('localhost', uid=1, _time=0, cpu=10, cpu_limit=50)
        self.save_history_line('localhost', uid=1, _time=1, cpu=100, cpu_limit=100)
        self.save_history_line('localhost', uid=2, _time=0, cpu=100, cpu_limit=100)
        self.save_history_line('localhost', uid=2, _time=1, cpu=100, cpu_limit=100)
        date_from = dateutil.unixtimestamp_to_gm_datetime(0)
        date_to = dateutil.unixtimestamp_to_gm_datetime(2)
        fields = ["lCPU"]
        params = {
            "dbengine": self.engine,
            "period_from": date_from,
            "period_to": date_to,
            "server_id": None,
            "time_unit": 1,
            "limit": 0,
            "show_idle": False,
            "show_columns": fields,
            "by_usage": None,
            "by_usage_percentage": None,
            "by_fault": None,
            "threshold": None}

        user1_data = HistoryShow(uid=1, **params).proceed().rows
        user2_data = HistoryShow(uid=2, **params).proceed().rows
        self.assertNotEqual(user1_data[0], user1_data[1])
        self.assertEqual(user2_data[0], user2_data[1])

    # FIXME - figure out what to do with idle
    def TODO_test_append_missed_groups(self):
        """
        Test for show_idle parameter
        """
        # `ping` data
        self.save_history_line("localhost", uid=0, _time=1, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=2, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=3, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=7, cpu=0, cpu_limit=5000)

        # user1 data
        self.save_history_line("localhost", uid=1, _time=1, cpu=1500, cpu_limit=10000)
        self.save_history_line("localhost", uid=1, _time=2, cpu=11500, cpu_limit=12000)
        self.save_history_line("localhost", uid=1, _time=3, cpu=7500, cpu_limit=8000)
        self.save_history_line("localhost", uid=1, _time=7, cpu=8000, cpu_limit=9000)

        date_from = dateutil.unixtimestamp_to_gm_datetime(0)
        date_to = dateutil.unixtimestamp_to_gm_datetime(10)
        fields = ["ID", "From", "To", "aCPU", "lCPU"]
        params = {
            "dbengine": self.engine,
            "period_from": date_from,
            "period_to": date_to,
            "uid": 1,
            "show_columns": fields,
            "show_idle": True}

        rows = HistoryShow(time_unit=1, **params).proceed()
        self.assertListEqual(list(rows), [[1, 1, 2, 15.0, 100.0],
                                          [1, 2, 3, 115.0, 120.0],
                                          [1, 3, 4, 75.0, 80.0],
                                          [1, 4, 5, 0.0, 0],
                                          [1, 6, 7, 0.0, 0],
                                          [1, 7, 8, 80.0, 90.0]])

    # FIXME - Figure out what to do with idle
    def TODO_test_append_idle_user(self):
        """
        """
        # `ping` data
        self.save_history_line("localhost", uid=0, _time=1, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=2, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=3, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=4, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=5, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=6, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=7, cpu=0, cpu_limit=5000)

        # user1 data
        self.save_history_line("localhost", uid=1, _time=1, cpu=1500, cpu_limit=10000)
        self.save_history_line("localhost", uid=1, _time=2, cpu=11500, cpu_limit=12000)
        self.save_history_line("localhost", uid=1, _time=3, cpu=7500, cpu_limit=8000)
        self.save_history_line("localhost", uid=1, _time=7, cpu=8000, cpu_limit=9000)

        date_from = dateutil.unixtimestamp_to_gm_datetime(0)
        date_to = dateutil.unixtimestamp_to_gm_datetime(10)
        fields = ["ID", "From", "To", "aCPU", "lCPU"]

        result = HistoryShow(dbengine=self.engine, period_from=date_from, period_to=date_to, uid=1, show_columns=fields,
                             time_unit=1).proceed()
        self.assertListEqual(list(result), [[1, 1, 1, 15.0, 100.0],
                                            [1, 2, 2, 115.0, 120.0],
                                            [1, 3, 3, 75.0, 80.0],
                                            [1, 4, 4, 0.0, 80.0],
                                            [1, 5, 5, 0.0, 80.0],
                                            [1, 6, 6, 0.0, 80.0],
                                            [1, 7, 7, 80.0, 90.0]])

    # FIXME - figure out what to do with idle
    def TODO_test_server_show_idle(self):
        self.save_history_line('localhost', uid=1, _time=0, cpu=100)
        self.save_history_line('localhost', uid=1, _time=10, cpu=100)
        self.save_history_line('localhost', uid=2, _time=0, cpu=100)
        self.save_history_line('localhost', uid=2, _time=1, cpu=100)
        date_from = dateutil.unixtimestamp_to_gm_datetime(0)
        date_to = dateutil.unixtimestamp_to_gm_datetime(10)
        fields = ["ID", "From", "To", "aCPU", "lCPU", "aVMem", "lVMem",
                  "aNproc", "lNproc"]
        params = {
            "dbengine": self.engine,
            "period_from": date_from,
            "period_to": date_to,
            "server_id": None,
            "time_unit": 1,
            "show_columns": fields,
            "by_usage": None,
            "by_usage_percentage": None,
            "by_fault": None,
            "threshold": None}
        user1_data = HistoryShow(uid=1, show_idle=False, **params).proceed()
        user2_data = HistoryShow(uid=2, show_idle=False, **params).proceed()
        self.assertEqual(len(user1_data.rows), 2)
        self.assertEqual(len(user2_data.rows), 2)
        user1_data = HistoryShow(uid=1, show_idle=True, **params).proceed()
        user2_data = HistoryShow(uid=2, show_idle=True, **params).proceed()
        self.assertEqual(len(user1_data.rows), 4)
        self.assertEqual(len(user2_data.rows), 2)

    def test_stats_history_limit(self):
        for t in range(50):
            c = t * 2
            self.save_history_line('localhost', uid=0, _time=t)
            self.save_history_line('localhost', uid=1, _time=t, cpu=c)
        date_from = dateutil.unixtimestamp_to_gm_datetime(0)
        date_to = dateutil.unixtimestamp_to_gm_datetime(50)
        fields = ["From", "aCPU"]
        params = {"dbengine": self.engine, "uid": 1, "period_from": date_from, "period_to": date_to,
                  "server_id": None, "time_unit": 1,
                  "show_columns": fields, "by_usage": None, "show_idle": False,
                  "by_usage_percentage": None, "by_fault": None, "threshold": None}
        self.assertEqual(len(HistoryShow(limit=1, **params).proceed().rows), 1)
        self.assertEqual(len(HistoryShow(limit=2, **params).proceed().rows), 2)
        self.assertEqual(len(HistoryShow(limit=49, **params).proceed().rows), 49)
        self.assertEqual(len(HistoryShow(limit=0, **params).proceed().rows), 50)
        data = list(HistoryShow(limit=2000, order_by="aCPU", **params).proceed())
        self.assertEqual(len(data), 50)
        data = list(HistoryShow(limit=1, **params).proceed())
        self.assertAlmostEqual(data[0][1], 0.0)
        data = list(HistoryShow(limit=1, order_by="aCPU", **params).proceed())
        self.assertAlmostEqual(data[0][1], 0.98)
        self.assertEqual(data[0][0], 49)

    def test_limit_lower_than_default(self):
        """
        LVES-299
        test limits lower than default.
        """
        # `ping` data
        self.save_history_line("localhost", uid=0, _time=1, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=2, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=3, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=4, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=5, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=6, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=7, cpu=0, cpu_limit=5000)
        # user1 data
        self.save_history_line("localhost", uid=1, _time=1, cpu=150, cpu_limit=1000)
        self.save_history_line("localhost", uid=1, _time=2, cpu=1150, cpu_limit=1200)
        self.save_history_line("localhost", uid=1, _time=3, cpu=750, cpu_limit=800)
        self.save_history_line("localhost", uid=1, _time=7, cpu=800, cpu_limit=900)

        date_from = dateutil.unixtimestamp_to_gm_datetime(0)
        date_to = dateutil.unixtimestamp_to_gm_datetime(10)
        fields = ["ID", "From", "To", "lCPU"]
        user_data = HistoryShow(uid=1, show_idle=True, dbengine=self.engine, period_from=date_from,
                                period_to=date_to,  time_unit=1,
                                show_columns=fields).proceed()
        self.assertEqual(list(user_data), [[1, 1, 2, 10.0],
                                           [1, 2, 3, 12.0],
                                           [1, 3, 4, 8.0],
                                           [1, 7, 8, 9.0]])

    def test_filter_users(self):
        """
        Get data only for given users.
        """
        # `ping` data
        self.save_history_line("localhost", uid=0, _time=1, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=2, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=3, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=4, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=5, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=6, cpu=0, cpu_limit=5000)
        self.save_history_line("localhost", uid=0, _time=7, cpu=0, cpu_limit=5000)
        # user1 data
        self.save_history_line("localhost", uid=1, _time=1, cpu=150, cpu_limit=1000)
        self.save_history_line("localhost", uid=1, _time=2, cpu=1150, cpu_limit=1200)
        self.save_history_line("localhost", uid=1, _time=3, cpu=750, cpu_limit=800)
        self.save_history_line("localhost", uid=1, _time=7, cpu=800, cpu_limit=900)
        # user2 data
        self.save_history_line("localhost", uid=2, _time=1, cpu=150, cpu_limit=1000)
        self.save_history_line("localhost", uid=2, _time=2, cpu=1150, cpu_limit=1200)
        self.save_history_line("localhost", uid=2, _time=3, cpu=750, cpu_limit=800)
        self.save_history_line("localhost", uid=2, _time=7, cpu=800, cpu_limit=900)
        # user3 data
        self.save_history_line("localhost", uid=3, _time=1, cpu=150, cpu_limit=1000)
        self.save_history_line("localhost", uid=3, _time=2, cpu=1150, cpu_limit=1200)
        self.save_history_line("localhost", uid=3, _time=3, cpu=750, cpu_limit=800)
        self.save_history_line("localhost", uid=3, _time=7, cpu=800, cpu_limit=900)

        date_from = dateutil.unixtimestamp_to_gm_datetime(0)
        date_to = dateutil.unixtimestamp_to_gm_datetime(10)
        fields = ["ID", "From", "To", "lCPU"]
        filter_id = [1, 3]
        user_data = HistoryShow(
            uid=filter_id,
            show_idle=True,
            dbengine=self.engine,
            period_from=date_from,
            period_to=date_to,
            time_unit=1,
            show_columns=fields).proceed()

        for user in user_data:
            self.assertIn(user[0], filter_id)

if __name__ == '__main__':
    unittest.main()
