# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import json
import os
import unittest
import datetime

import mock
import warnings
from ddt import ddt, data, unpack
from testfixtures import Replace, OutputCapture, test_datetime
import csv
import io

from test.base.base_db_test import BaseDbTest
from test.generate_lvestats2db import DatabaseGenerator, DataCase, YEAR
from lvestats.plugins.generic.dbsaver_x60 import main as dbsaver_x60_main

TEST_CASES = list()
# test cases for calculate average cpu
TEST_CASES.append(({'aCPU': 1,   'lCPU': 100},                                  {'aCPU': 1}))
TEST_CASES.append(({'aCPU': 20,  'lCPU': 100},                                  {'aCPU': 20}))
TEST_CASES.append(({'aCPU': 100, 'lCPU': 100, 'CPUf': 100},                     {'aCPU': 100}))

# test cases for calculate maximum cpu
TEST_CASES.append(({'aCPU': 20, 'mCPU': 100,   'lCPU': 100, 'CPUf': 2000},                     {'mCPU': 100}))
TEST_CASES.append(({'aCPU': 20, 'mCPU': 100,   'lCPU': 100, 'CPUf': 1},                        {'mCPU': 100}))
TEST_CASES.append(({'aCPU': 20, 'mCPU': 80,    'lCPU': 100, 'CPUf': 0},                        {'mCPU':  80}))

# test cases for calculate cpu faults
TEST_CASES.append(({'aCPU': 1,   'lCPU': 100, 'CPUf': 0},                        {'CPUf': 0}))
TEST_CASES.append(({'aCPU': 20,  'lCPU': 100, 'CPUf': 1},                        {'CPUf': 1}))
TEST_CASES.append(({'aCPU': 100, 'lCPU': 100, 'CPUf': 234},                      {'CPUf': 234}))

# test cases for calculate average ep
TEST_CASES.append(({'aEP': 1,  'lEP': 100},                                       {'aEP': 1}))
TEST_CASES.append(({'aEP': 20, 'lEP': 100},                                       {'aEP': 20}))
TEST_CASES.append(({'aEP': 100,  'lEP': 100, 'EPf': 100},                         {'aEP': 100}))

# test cases for calculate maximum ep
TEST_CASES.append(({'aEP': 20, 'mEP': 100,   'lEP': 100, 'EPf': 2000},                         {'mEP': 100}))
TEST_CASES.append(({'aEP': 20, 'mEP': 100,   'lEP': 100, 'EPf': 1},                            {'mEP': 100}))
TEST_CASES.append(({'aEP': 20, 'mEP': 80,  'lEP': 100, 'EPf': 0},                              {'mEP':  80}))

# test cases for calculate ep faults
TEST_CASES.append(({'aEP': 1,   'lEP': 100, 'EPf': 0},                             {'EPf': 0}))
TEST_CASES.append(({'aEP': 20,   'lEP': 100, 'EPf': 1},                            {'EPf': 1}))
TEST_CASES.append(({'aEP': 100,   'lEP': 100, 'EPf': 234},                         {'EPf': 234}))

# test cases for calculate average virtual memory
TEST_CASES.append(({'aVMem': 4096,   'lVMem': 409600},                             {'aVMem': 4096}))
TEST_CASES.append(({'aVMem': 81920,  'lVMem': 409600},                             {'aVMem': 81920}))
TEST_CASES.append(({'aVMem': 409600, 'lVMem': 409600, 'VMemF': 100},               {'aVMem': 409600}))

# test cases for calculate maximum virtual memory
TEST_CASES.append(({'aVMem': 81920, 'mVMem': 409600,   'lVMem': 409600, 'VMemF': 2000},        {'mVMem': 409600}))
TEST_CASES.append(({'aVMem': 81920, 'mVMem': 409600,   'lVMem': 409600, 'VMemF': 1},           {'mVMem': 409600}))
TEST_CASES.append(({'aVMem': 81920, 'mVMem': 126976,  'lVMem': 409600, 'VMemF': 0},            {'mVMem':  126976}))

# test cases for calculate virtual memory faults
TEST_CASES.append(({'aVMem': 4096,  'lVMem': 409600, 'VMemF': 0},            {'VMemF': 0}))
TEST_CASES.append(({'aVMem': 81920,  'lVMem': 409600, 'VMemF': 1},           {'VMemF': 1}))
TEST_CASES.append(({'aVMem': 409600,  'lVMem': 409600, 'VMemF': 234},        {'VMemF': 234}))

# test cases for calculate average physical memory
TEST_CASES.append(({'aPMem': 4096,    'lPMem': 409600},                       {'aPMem': 4096}))
TEST_CASES.append(({'aPMem': 81920,   'lPMem': 409600},                       {'aPMem': 81920}))
TEST_CASES.append(({'aPMem': 409600,  'lPMem': 409600, 'PMemF': 100},         {'aPMem': 409600}))

# test cases for calculate maximum physical memory
TEST_CASES.append(({'aPMem': 81920, 'mPMem': 409600,   'lPMem': 409600, 'PMemF': 2000},        {'mPMem': 409600}))
TEST_CASES.append(({'aPMem': 81920, 'mPMem': 409600,   'lPMem': 409600, 'PMemF': 1},           {'mPMem': 409600}))
TEST_CASES.append(({'aPMem': 81920, 'mPMem': 126976,  'lPMem': 409600, 'PMemF': 0},            {'mPMem':  126976}))

# test cases for calculate  physical memory faults
TEST_CASES.append(({'aPMem': 4096,  'lPMem': 409600, 'PMemF': 0},            {'PMemF': 0}))
TEST_CASES.append(({'aPMem': 81920,  'lPMem': 409600, 'PMemF': 1},           {'PMemF': 1}))
TEST_CASES.append(({'aPMem': 409600,  'lPMem': 409600, 'PMemF': 234},        {'PMemF': 234}))

# test cases for calculate average process numbers
TEST_CASES.append(({'aNproc': 1,    'lNproc': 100},                              {'aNproc': 1}))
TEST_CASES.append(({'aNproc': 20,   'lNproc': 100},                              {'aNproc': 20}))
TEST_CASES.append(({'aNproc': 100,    'lNproc': 100, 'NprocF': 100},             {'aNproc': 100}))

# test cases for calculate maximum process numbers
TEST_CASES.append(({'aNproc': 20, 'mNproc': 100,   'lNproc': 100, 'NprocF': 2000},             {'mNproc': 100}))
TEST_CASES.append(({'aNproc': 20, 'mNproc': 100,   'lNproc': 100, 'NprocF': 1},                {'mNproc': 100}))
TEST_CASES.append(({'aNproc': 20, 'mNproc': 80,  'lNproc': 100, 'NprocF': 0},                  {'mNproc':  80}))

# test cases for calculate process numbers faults
TEST_CASES.append(({'aNproc': 1,     'lNproc': 100, 'NprocF': 0},                 {'NprocF': 0}))
TEST_CASES.append(({'aNproc': 20,    'lNproc': 100, 'NprocF': 1},                 {'NprocF': 1}))
TEST_CASES.append(({'aNproc': 100,   'lNproc': 100, 'NprocF': 234},               {'NprocF': 234}))


# test cases for calculate average IO
TEST_CASES.append(({'aIO': 1,     'lIO': 1024},                                      {'aIO': 1}))
TEST_CASES.append(({'aIO': 256,   'lIO': 1024},                                      {'aIO': 256}))
TEST_CASES.append(({'aIO': 1024,  'lIO': 1024, 'IOf': 1024},                         {'aIO': 1024}))

# test cases for calculate maximum IO
TEST_CASES.append(({'aIO': 256, 'mIO': 1024,   'lIO': 1024, 'IOf': 240},                       {'mIO': 1024}))
TEST_CASES.append(({'aIO': 256, 'mIO': 1024,   'lIO': 1024, 'IOf': 1},                         {'mIO': 1024}))
TEST_CASES.append(({'aIO': 256, 'mIO': 768,  'lIO': 1024, 'IOf': 0},                           {'mIO':  768}))

# test cases for calculate IO faults
TEST_CASES.append(({'aIO': 1,    'lIO': 1024, 'IOf': 0},                            {'IOf': 0}))
TEST_CASES.append(({'aIO': 256,  'lIO': 1024, 'IOf': 1},                            {'IOf': 1}))
TEST_CASES.append(({'aIO': 1024, 'lIO': 1024, 'IOf': 234},                          {'IOf': 234}))

# test cases for calculate average IO
TEST_CASES.append(({'aIOPS': 1,     'lIOPS': 1024},                                {'aIOPS': 1}))
TEST_CASES.append(({'aIOPS': 256,   'lIOPS': 1024},                                {'aIOPS': 256}))
TEST_CASES.append(({'aIOPS': 1024,  'lIOPS': 1024, 'IOPSf': 1024},                 {'aIOPS': 1024}))

# test cases for calculate maximum IOPS
TEST_CASES.append(({'aIOPS': 256, 'mIOPS': 1024,   'lIOPS': 1024, 'IOPSf': 240},               {'mIOPS': 1024}))
TEST_CASES.append(({'aIOPS': 256, 'mIOPS': 1024,   'lIOPS': 1024, 'IOPSf': 1},                 {'mIOPS': 1024}))
TEST_CASES.append(({'aIOPS': 256, 'mIOPS': 768,  'lIOPS': 1024, 'IOPSf': 0},                   {'mIOPS':  768}))

# test cases for calculate IOPS faults
TEST_CASES.append(({'aIOPS': 1,    'lIOPS': 1024, 'IOPSf': 0},                    {'IOPSf': 0}))
TEST_CASES.append(({'aIOPS': 256,  'lIOPS': 1024, 'IOPSf': 1},                    {'IOPSf': 1}))
TEST_CASES.append(({'aIOPS': 1024, 'lIOPS': 1024, 'IOPSf': 234},                  {'IOPSf': 234}))


@ddt
class TestFunctionalLveinfoMain(BaseDbTest):

    def setUp(self):
        BaseDbTest.setUp(self)
        self.db_generator = DatabaseGenerator(db_uri=self.engine.url, server_id=self.cfg['server_id'], lve_version=8)
        self.db_generator.dbengine = self.engine
        self.default_data_case = {
            'ID':     500,   'From':   '2015-09-17 00:00', 'To': '2015-09-17 04:00',
            'aCPU':   20,    'lCPU':   100,
            'aEP':    20,    'lEP':    100,
            'aVMem':  81920, 'lVMem':  409600,
            'aPMem':  81920, 'lPMem':  409600,
            'aNproc': 20,    'lNproc': 100,
            'aIO':    256,   'lIO':    1024,
            'aIOPS':  256,   'lIOPS':  1024,
        }
        # generate input arguments
        self.argv_ = [
            '--from', self.default_data_case['From'],
            '--to', self.default_data_case['To'],
            '--server_id', self.cfg['server_id'],
            '-mv2',
            '--json']

    def _generate_database(self, input_):
        # prepare data for generating database
        data_case_dict = dict(self.default_data_case)
        data_case_dict.update(input_)
        # load data from dict instead of csv
        def mock_generate_history(conn_):
            self.db_generator._generate_history(conn_, [data_case_dict], DataCase)

        with mock.patch.object(self.db_generator, 'generate_history',
                               side_effect=mock_generate_history):
            self.db_generator.generate_database()  # generate database

    @data(*TEST_CASES)
    @unpack
    def test_show_history(self, input_, output_):
        from lvestats.lib.info.lveinfomain import main_ as lveinfo_main
        self._generate_database(input_)
        json_string = lveinfo_main(self.cfg, dbengine=self.engine,
                                   argv_=self.argv_ + ['--show-columns', 'ID', 'From', 'To'] + list(output_.keys()))
        self.assertDictContainsSubset(output_, json.loads(json_string)['data'][0])

    @data(*TEST_CASES)
    @unpack
    def test_show_history_x60(self, input_, output_):
        from lvestats.lib.info.lveinfomain import main_ as lveinfo_main
        self._generate_database(input_)
        # generate additional x60 table for testing
        with mock.patch('lvestats.plugins.generic.dbsaver_x60.reboot_lock'):
            dbsaver_x60_main(argv_=['aggregate'], config_=self.cfg, db_engine=self.engine)
        json_string = lveinfo_main(self.cfg, dbengine=self.engine,
                                   argv_=self.argv_ + ['--show-columns', 'ID', 'From', 'To'] + list(output_.keys()))
        self.assertDictContainsSubset(output_, json.loads(json_string)['data'][0])


TEST_CASES_BY_USAGE = (
    ('--by-usage aCPU  -m v1',  [19696]),
    ('--by-usage aCPU  -m v2',  [19696]),
    ('--by-usage mCPU  -m v1',  [18096, 19696]),
    ('--by-usage mCPU  -m v2',  [18096, 19696]),

    ('--by-usage aEP   -m v1',  [19696]),
    ('--by-usage aEP   -m v2',  [19696]),
    ('--by-usage mEP   -m v1',  [18096, 19696]),
    ('--by-usage mEP   -m v2',  [18096, 19696]),

    ('--by-usage aVMem -m v1',  [19696]),
    ('--by-usage aVMem -m v2',  [19696]),
    ('--by-usage mVMem -m v1',  [18096, 19696]),
    ('--by-usage mVMem -m v2',  [18096, 19696]),


    ('--by-usage aPMem -m v1',  [19696]),
    ('--by-usage aPMem -m v2',  [19696]),
    ('--by-usage mPMem -m v1',  [18096, 19696]),
    ('--by-usage mPMem -m v2',  [18096, 19696]),

    ('--by-usage aNproc -m v1', [19696]),
    ('--by-usage aNproc -m v2', [19696]),
    ('--by-usage mNproc -m v1', [18096, 19696]),
    ('--by-usage mNproc -m v2', [18096, 19696]),

    ('--by-usage aIO    -m v1', [19696]),
    ('--by-usage aIO    -m v2', [19696]),
    ('--by-usage mIO    -m v1', [18096, 19696]),
    ('--by-usage mIO    -m v2', [18096, 19696]),

    ('--by-usage aIOPS  -m v1', [19696]),
    ('--by-usage aIOPS  -m v2', [19696]),
    ('--by-usage mIOPS  -m v1', [18096, 19696]),
    ('--by-usage mIOPS  -m v2', [18096, 19696]),


    ('--by-usage aCPU  -m v1 --percentage 50',  [18096, 19696]),
    ('--by-usage aCPU  -m v2 --percentage 50',  [18096, 19696]),
    ('--by-usage mCPU  -m v1 --percentage 50',  [12080, 18096, 19696]),
    ('--by-usage mCPU  -m v2 --percentage 50',  [12080, 18096, 19696]),

    ('--by-usage aEP   -m v1 --percentage 50',  [18096, 19696]),
    ('--by-usage aEP   -m v2 --percentage 50',  [18096, 19696]),
    ('--by-usage mEP   -m v1 --percentage 50',  [12080, 18096, 19696]),
    ('--by-usage mEP   -m v2 --percentage 50',  [12080, 18096, 19696]),

    ('--by-usage aVMem -m v1 --percentage 50',  [18096, 19696]),
    ('--by-usage aVMem -m v2 --percentage 50',  [18096, 19696]),
    ('--by-usage mVMem -m v1 --percentage 50',  [12080, 18096, 19696]),
    ('--by-usage mVMem -m v2 --percentage 50',  [12080, 18096, 19696]),

    ('--by-usage aPMem -m v1 --percentage 50',  [18096, 19696]),
    ('--by-usage aPMem -m v2 --percentage 50',  [18096, 19696]),
    ('--by-usage mPMem -m v1 --percentage 50',  [12080, 18096, 19696]),
    ('--by-usage mPMem -m v2 --percentage 50',  [12080, 18096, 19696]),

    ('--by-usage aNproc -m v1  --percentage 50', [18096, 19696]),
    ('--by-usage aNproc -m v2 --percentage 50', [18096, 19696]),
    ('--by-usage mNproc -m v1 --percentage 50', [12080, 18096, 19696]),
    ('--by-usage mNproc -m v2 --percentage 50', [12080, 18096, 19696]),

    ('--by-usage aIO    -m v1 --percentage 50', [18096, 19696]),
    ('--by-usage aIO    -m v2 --percentage 50', [18096, 19696]),
    ('--by-usage mIO    -m v1 --percentage 50', [12080, 18096, 19696]),
    ('--by-usage mIO    -m v2 --percentage 50', [12080, 18096, 19696]),

    ('--by-usage aIOPS  -m v1 --percentage 50', [18096, 19696]),
    ('--by-usage aIOPS  -m v2 --percentage 50', [18096, 19696]),
    ('--by-usage mIOPS  -m v1 --percentage 50', [12080, 18096, 19696]),
    ('--by-usage mIOPS  -m v2 --percentage 50', [12080, 18096, 19696]),



    ('--by-usage aCPU  -m v1 --percentage 10',  [12080, 18096, 19696]),
    ('--by-usage aCPU  -m v2 --percentage 10',  [12080, 18096, 19696]),
    ('--by-usage mCPU  -m v1 --percentage 10',  [12080, 18096, 19696]),
    ('--by-usage mCPU  -m v2 --percentage 10',  [12080, 18096, 19696]),

    ('--by-usage aEP   -m v1 --percentage 10',  [12080, 18096, 19696]),
    ('--by-usage aEP   -m v2 --percentage 10',  [12080, 18096, 19696]),
    ('--by-usage mEP   -m v1 --percentage 10',  [12080, 18096, 19696]),
    ('--by-usage mEP   -m v2 --percentage 10',  [12080, 18096, 19696]),

    ('--by-usage aVMem -m v1 --percentage 10',  [12080, 18096, 19696]),
    ('--by-usage aVMem -m v2 --percentage 10',  [12080, 18096, 19696]),
    ('--by-usage mVMem -m v1 --percentage 10',  [12080, 18096, 19696]),
    ('--by-usage mVMem -m v2 --percentage 10',  [12080, 18096, 19696]),


    ('--by-usage aPMem -m v1 --percentage 10',  [12080, 18096, 19696]),
    ('--by-usage aPMem -m v2 --percentage 10',  [12080, 18096, 19696]),
    ('--by-usage mPMem -m v1 --percentage 10',  [12080, 18096, 19696]),
    ('--by-usage mPMem -m v2 --percentage 10',  [12080, 18096, 19696]),

    ('--by-usage aNproc -m v1 --percentage 10', [12080, 18096, 19696]),
    ('--by-usage aNproc -m v2 --percentage 10', [12080, 18096, 19696]),
    ('--by-usage mNproc -m v1 --percentage 10', [12080, 18096, 19696]),
    ('--by-usage mNproc -m v2 --percentage 10', [12080, 18096, 19696]),

    ('--by-usage aIO    -m v1 --percentage 10', [12080, 18096, 19696]),
    ('--by-usage aIO    -m v2 --percentage 10', [12080, 18096, 19696]),
    ('--by-usage mIO    -m v1 --percentage 10', [12080, 18096, 19696]),
    ('--by-usage mIO    -m v2 --percentage 10', [12080, 18096, 19696]),

    ('--by-usage aIOPS  -m v1 --percentage 10', [12080, 18096, 19696]),
    ('--by-usage aIOPS  -m v2 --percentage 10', [12080, 18096, 19696]),
    ('--by-usage mIOPS  -m v1 --percentage 10', [12080, 18096, 19696]),
    ('--by-usage mIOPS  -m v2 --percentage 10', [12080, 18096, 19696]),
)


class TestFunctionalX60(BaseDbTest):
    def setUp(self):
        BaseDbTest.setUp(self)
        self.db_generator = DatabaseGenerator(
            db_uri=self.engine.url, server_id=self.cfg['server_id'],
            lve_data_path=os.path.join(os.path.dirname(__file__), 'data', 'lveinfo_by-usage.csv'))
        self.db_generator.dbengine = self.engine
        self.db_generator.generate_database()

    def test_x60_show_all(self):
        from lvestats.lib.info.lveinfomain import main_ as lveinfo_main
        args_all = '--from {year}-09-23 00:00 --to {year}-09-24 00:00 ' \
                   '--show-columns all ' \
                   '--server_id {server_id} --json '.format(year=YEAR, server_id=self.cfg['server_id'])
        json_string_without_x60 = lveinfo_main(
            self.cfg,
            dbengine=self.engine,
            argv_=args_all.split())
        with mock.patch('lvestats.plugins.generic.dbsaver_x60.reboot_lock'):
            dbsaver_x60_main(argv_=['aggregate'], config_=self.cfg, db_engine=self.engine)
        json_string_with_x60 = lveinfo_main(
            self.cfg,
            dbengine=self.engine,
            argv_=args_all.split())
        self.assertEqual(json_string_with_x60, json_string_without_x60)


@ddt
class TestFunctionalLveinfoMain2(BaseDbTest):

    @classmethod
    def setUpClass(cls):
        BaseDbTest.setUpClass()  # create tables
        cls.need_clear_rows = False  # prevent clear rows for every unittest
        cls.db_generator = DatabaseGenerator(
            db_uri=cls.engine.url, server_id=cls.cfg['server_id'],
            lve_data_path=os.path.join(os.path.dirname(__file__), 'data', 'lveinfo_by-usage.csv'))
        cls.db_generator.dbengine = cls.engine
        cls.db_generator.generate_database()
        # generate input arguments
        cls.args_base = '--from {year}-09-23 00:00 --to {year}-09-24 00:00 --show-columns ID ' \
            '--server_id {server_id} --json '.format(year=YEAR, server_id=cls.cfg['server_id'])

    @data(*TEST_CASES_BY_USAGE)
    @unpack
    def test_by_usage(self, args, uid_list):
        from lvestats.lib.info.lveinfomain import main_ as lveinfo_main
        argv_ = self.args_base.split() + args.split()
        json_string = lveinfo_main(self.cfg, dbengine=self.engine, argv_=argv_)
        uid_list_ = [dict_['ID'] for dict_ in json.loads(json_string)['data']]
        self.assertEqual(uid_list, uid_list_)

    def test_user_style(self):
        """
            Just check if there are no exceptions
        """
        from lvestats.lib.info.lveinfomain import main_ as lveinfo_main
        argv_ = '--from {year}-09-23 00:00 --to {year}-09-24 00:00 --show-all ' \
            '--id=12080 --server_id {server_id} --json --style=user'.format(
                year=YEAR, server_id=self.cfg['server_id']).split()
        lveinfo_main(self.cfg, dbengine=self.engine, argv_=argv_)


TEST_CASES_ORDER_BY = (
    ("--order-by aCPU", (20, 19, 18, 17, 16, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1)),
    ("--order-by mCPU", (20, 19, 18, 17, 16, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1)),
    ("--order-by CPUf", [[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]]),
    ("--order-by aVMem", (15, 6, 19, 2, 12, 16, 9, 14, 17, 11, 13, 18, 5, 20, 10, 8, 7, 3, 4, 1)),
    ("--order-by mVMem", ((2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20), 1)),
    ("--order-by VMemF", (20, (13, 15, 16, 17, 18, 19), (14, 10, 9, 8, 7), (2, 3, 4, 5, 6, 11, 12), 1)),
    ("--order-by aEP", (20, 18, 16, 14, 12, 10, 8, 6, 4, 2, (19, 17, 15, 13, 11, 9, 7, 5, 3, 1))),
    ("--order-by mEP", ((20, 18, 16, 14, 12, 10, 8, 6, 4, 2), (19, 17, 15, 13, 11, 9, 7, 5, 3, 1))),
    ("--order-by EPf", (2, 4, 6, 8, 10, 12, 14, 16, 18, 20, (19, 17, 15, 13, 11, 9, 7, 5, 3, 1))),
    ("--order-by aPMem", ((12, 13, 14, 15, 16, 17, 18, 19, 20), 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 1)),
    ("--order-by mPMem", ((12, 13, 14, 15, 16, 17, 18, 19, 20), (2, 3, 4, 5, 6, 7, 8, 9, 10, 11), 1)),
    ("--order-by PMemF", (15, 17, 18, 16, 19, 20, 13, 14, 12, (1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11))),
    ("--order-by aNproc", ((20, 19, 18, 17, 16), 15, 14, 13, 12, (11, 10, 9, 8, 7, 6), 5, 4, 3, 2, 1)),
    ("--order-by mNproc", ((20, 19, 18, 17, 16), 15, 14, 13, 12, (11, 10, 9, 8, 7, 6), 5, 4, 3, 2, 1)),
    ("--order-by NprocF", ((17, 18, 20), (19, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16))),
    ("--order-by aIO", ((4, 7, 10, 13, 16, 19), (3, 6, 9, 12, 15, 18), (2, 5, 8, 11, 14, 17, 20), 1)),
    ("--order-by mIO", ((6, 11, 16), (5, 10, 15, 20), (4, 9, 14, 19), (3, 8, 13, 18), (2, 7, 12, 17), 1)),
    ("--order-by IOf", [[20, 19, 18, 17, 16, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1]]),
    ("--order-by aIOPS", ((2, 10), 12, (6, 8, 11, 18), (14, 16, 20), (4, 13, 17, 19), 7, (3, 5, 9, 15), 1)),
    ("--order-by mIOPS", ((2, 7, 10, 12, 13, 18), (3, 4, 5, 6, 8, 9, 11), (17, 19), (16, 20), (14, 15), 1)),
    ("--order-by IOPSf", (13, (7, 10, 18), (1, 2, 3, 4, 5, 6, 8, 9, 11, 12, 14, 15, 16, 17, 19, 20))),
    ("--order-by anyF", (17, 2, 15, 18, 4, 16, 6, 8, 12, 20, (10, 14), 19, 13, 7, 9, (3, 5, 11), 1)),
    ("--by-fault cpu", [[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]]),
    ("--by-fault ep", (2, 4, 6, 8, 10, 12, 14, 16, 18, 20, (19, 17, 15, 13, 11, 9, 7, 5, 3, 1))),
    ("--by-fault io", [[20, 19, 18, 17, 16, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1]]),
    ("--by-fault iops", (13, (7, 10, 18), (1, 2, 3, 4, 5, 6, 8, 9, 11, 12, 14, 15, 16, 17, 19, 20)))
)


@ddt
class TestFunctionalLveinfoMainOrderBy(BaseDbTest):

    @classmethod
    def setUpClass(cls):
        BaseDbTest.setUpClass()

        cls.server_id = 'localhost'
        cls.db_generator = DatabaseGenerator(
            db_uri=cls.engine.url, server_id=cls.server_id,
            lve_data_path=os.path.join(os.path.dirname(__file__), 'data', 'lveinfo_order-by.csv'))
        cls.db_generator.dbengine = cls.engine
        cls.db_generator.generate_database()
        # generate input arguments
        cls.args_base = '--from {year}-09-23 00:00 --to {year}-09-24 00:00 --show-columns ID ' \
            '--server_id {server_id} --json '.format(year=YEAR, server_id=cls.server_id)

    @staticmethod
    def get_length(value):
        if isinstance(value, int):
            return 1
        else:
            return len(value)

    def get_sublist_by_index(self, data_, index):
        length = 0
        index += 1
        for d in data_:
            length += self.get_length(d)
            if index <= length:
                return d
        return []

    def assertCorrectOrder(self, result, test_data):
        """
        If some values are equal we dont care about their order in result (sorting function may change),
        but this values must be in correct place
        """
        for i, r in enumerate(result):
            test_value = self.get_sublist_by_index(test_data, i)
            if isinstance(test_value, int):
                assert_result = r == test_value
            else:
                assert_result = r in test_value
            self.assertTrue(
                assert_result,
                "Data is not in correct order: \nData: {}\nOrder: {}\nElement {} (index={}) is not in {}".format(
                    result, test_data, r, i, test_value
                )
            )

    @data(*TEST_CASES_ORDER_BY)
    @unpack
    def test_order_by(self, args, uid_list):
        from lvestats.lib.info.lveinfomain import main_ as lveinfo_main
        argv_ = self.args_base.split() + args.split()
        json_string = lveinfo_main(self.cfg, dbengine=self.engine, argv_=argv_)
        uid_list_ = [dict_['ID'] for dict_ in json.loads(json_string)['data']]
        self.assertCorrectOrder(uid_list_, uid_list)


TEST_CASES_BLANK_VALUE = (
    #     command line ------------------------ blanked-items
    ('--blank-value --server_id=lve-ver-8',
     {'aPMem': 0,  'uPMem': 0,  'mPMem': 0,  'lPMem': 102400, 'PMemF': 0,
      'aNproc': 0, 'uNproc': 0, 'mNproc': 0, 'lNproc': 15,    'NprocF': 0,
      'aIO': 0,    'uIO': 0,    'mIO': 0,    'lIO': 1024,     'IOf': 0,
      'aIOPS': 0,  'uIOPS': 0,  'mIOPS': 0,  'lIOPS': 1024,   'IOPSf': 0}),

    ('--blank-value --server_id=lve-ver-6',
     {'aPMem': 0,   'uPMem': 0,   'mPMem': 0,   'lPMem': 102400, 'PMemF': 0,
      'aNproc': 0,  'uNproc': 0,  'mNproc': 0,  'lNproc': 15,    'NprocF': 0,
      'aIO': 0,     'uIO': 0,     'mIO': 0,     'lIO': 1024,     'IOf': 0,
      'aIOPS': '-', 'uIOPS': '-', 'mIOPS': '-', 'lIOPS': '-',    'IOPSf': '-'}),

    ('--blank-value N/A --server_id=lve-ver-8',
     {'aPMem': 0,  'uPMem': 0,  'mPMem': 0,  'lPMem': 102400, 'PMemF': 0,
      'aNproc': 0, 'uNproc': 0, 'mNproc': 0, 'lNproc': 15,    'NprocF': 0,
      'aIO': 0,    'uIO': 0,    'mIO': 0,    'lIO': 1024,     'IOf': 0,
      'aIOPS': 0,  'uIOPS': 0,  'mIOPS': 0,  'lIOPS': 1024,   'IOPSf': 0}),

    ('--blank-value N/A --server_id=lve-ver-6 ',
     {'aPMem': 0,   'uPMem': 0,   'mPMem': 0,   'lPMem': 102400, 'PMemF': 0,
      'aNproc': 0,  'uNproc': 0,  'mNproc': 0,  'lNproc': 15,    'NprocF': 0,
      'aIO': 0,     'uIO': 0,     'mIO': 0,     'lIO': 1024,     'IOf': 0,
      'aIOPS': 'N/A', 'uIOPS': 'N/A', 'mIOPS': 'N/A', 'lIOPS': 'N/A',    'IOPSf': 'N/A'}),
)


@ddt
class TestFunctionalLveinfoMainBlankValue(BaseDbTest):

    @classmethod
    def setUpClass(cls):
        BaseDbTest.setUpClass()

        cls.server_id_template = 'lve-ver-{}'
        cls.need_clear_rows = False
        # generate test database for different servers with different lve versions
        for lve_version in (8, 6):
            server_id = cls.server_id_template.format(lve_version)
            db_generator = DatabaseGenerator(
                db_uri=cls.engine.url, lve_version=lve_version, server_id=server_id,
                lve_data_path=os.path.join(os.path.dirname(__file__), 'data', 'lveinfo_by-usage.csv'))
            db_generator.dbengine = cls.engine
            db_generator.generate_database()

        cls.args_base = '--from {year}-09-23 00:00 --to {year}-09-24 00:00 ' \
                        '--show-all -mv2 --json '.format(year=YEAR)

    @data(*TEST_CASES_BLANK_VALUE)
    @unpack
    def test_value_blank(self, input_, output_):
        from lvestats.lib.info.lveinfomain import main_ as lveinfo_main

        json_string = lveinfo_main(self.cfg, dbengine=self.engine,
                                   argv_=str.split(self.args_base + input_))
        self.assertDictContainsSubset(output_, json.loads(json_string)['data'][0])


TEST_CASES_TIME_UNIT_ = [
    [501, '--time-unit=1m --id=501 --from {year}-01-01 00:00 --to {year}-01-01 00:20'],
    [502, '--time-unit=5m --id=502 --from {year}-01-01 00:00 --to {year}-01-01 01:40'],
    [503, '--time-unit=1h --id=503 --from {year}-01-01 00:00 --to {year}-01-01 20:00']
]

# make combinations with difference columns names
TEST_CASES_TIME_UNIT = list()
for id_, _ in TEST_CASES_TIME_UNIT_:
    for column_name in ('aCPU', 'lCPU', 'CPUf', 'aEP', 'lEP', 'EPf', 'aVMem', 'lVMem', 'VMemF', 'aPMem', 'lPMem',
                        'PMemF', 'aNproc', 'lNproc', 'NprocF', 'aIO', 'lIO', 'IOf', 'aIOPS', 'lIOPS', 'IOPSf'):
        TEST_CASES_TIME_UNIT.append([column_name, id_])


@ddt
class TestFunctionalLveinfoMainTimeUnit(BaseDbTest):

    @classmethod
    def setUpClass(cls):
        BaseDbTest.setUpClass()
        from lvestats.lib.info.lveinfomain import main_ as lveinfo_main
        cls.server_id = cls.cfg['server_id']
        cls.csv_path = os.path.join(os.path.dirname(__file__), 'data', 'lveinfo_time_unit.csv')

        db_generator = DatabaseGenerator(
            db_uri=cls.engine.url, server_id=cls.server_id,
            lve_data_path=cls.csv_path)
        db_generator.dbengine = cls.engine
        db_generator.generate_database()
        cls.csv_test_data = cls.load_csv()
        cls.columns = list(cls.csv_test_data[0].keys())  # obtain columns list from csv

        cls.args_base = '--show-columns {show_columns} -mv2 --json '.format(show_columns=','.join(cls.columns))
        cls.output_cache = dict()
        for id_, case_ in TEST_CASES_TIME_UNIT_:
            cls.output_cache[id_] = lveinfo_main(
                cls.cfg,
                dbengine=cls.engine,
                argv_=str.split(cls.args_base + case_.format(year=YEAR)))

    @classmethod
    def load_csv(cls):
        """
        Load csv and return as dicts list
        """
        data_list = list()
        with open(cls.csv_path) as csv_stream:
            for csv_dict in csv.DictReader(csv_stream):
                dict_ = dict()
                for key, val in list(csv_dict.items()):
                    if key == 'ID':
                        val = int(val)
                    elif key not in ('From', 'To'):
                        val = float(val)
                    dict_[key] = val
                data_list.append(dict_)
        return data_list

    @classmethod
    def get_column(cls, id_, column_name):
        return [dict_[column_name] for dict_ in cls.csv_test_data if int(id_) == int(dict_['ID'])]

    @data(*TEST_CASES_TIME_UNIT)
    @unpack
    def test_time_unit(self, column_name, id_):

        # extract column values from lveinfo output
        column_val_ = [dict_[column_name] for dict_ in json.loads(self.output_cache[id_])['data']]
        # extract column values from csv test data
        column_val = self.get_column(id_, column_name)
        self.assertListEqual(column_val, column_val_)


@ddt
class TestFunctionalDynTimeUnit(BaseDbTest):

    @classmethod
    def setUpClass(cls):
        BaseDbTest.setUpClass()
        cls.server_id = cls.cfg['server_id']
        cls.csv_path = os.path.join(os.path.dirname(__file__), 'data', 'lveinfo_time_unit_dyn.csv')

        db_generator = DatabaseGenerator(
            db_uri=cls.engine.url, server_id=cls.server_id,
            lve_data_path=cls.csv_path)
        db_generator.dbengine = cls.engine
        db_generator.generate_database()
        cls.need_clear_rows = False  # prevent clear generated data in tearDown
        cls.base_args = ['--id=7777', '--csv', '--compat=v1', '--style=user',
                         '--show-all']

    @staticmethod
    def _csv_str_to_dt(time_str):
        return datetime.datetime.strptime(time_str, '%m-%d %H:%M').replace(
            year=2000)

    def parse_csv_from_to(self, csv_str):
        csv_file = io.StringIO(csv_str)
        csv_dict_reader = csv.DictReader(csv_file, delimiter=',')
        for dict_ in csv_dict_reader:
            yield self._csv_str_to_dt(dict_['From']), \
                  self._csv_str_to_dt(dict_['To'])

    @data('30d', '7d', '1d', '4h', '1h', '30m', '10m')
    def test_check_from_to_structure(self, period):
        from lvestats.lib.info.lveinfomain import main_ as lveinfo_main
        # change datetime.now()
        with Replace('lvestats.lib.parsers.lveinfoargparse.datetime.datetime',
                     test_datetime(2000, 5, 30, 5, 0, 0)):
            with OutputCapture() as output_csv:  # capture all output to stdout
                # TODO: investigate why mysql on cl6 i686 prints warnings
                # /opt/alt/python37/lib/python3.7/site-packages/sqlalchemy/dialects/mysql/base.py:877:
                # SAWarning: Datatype FLOAT does not support CAST on MySQL; the CAST will be skipped.
                # self.dialect.type_compiler.process(cast.typeclause.type))
                with warnings.catch_warnings(record=True):
                    lveinfo_main(
                            self.cfg, dbengine=self.engine,
                            argv_=self.base_args + ['--period', period])
        # prepare parsed csv data for checking
        from_to_list = list(self.parse_csv_from_to(output_csv.captured))
        from_column = [_[0] for _ in from_to_list]
        to_column = [_[1] for _ in from_to_list]
        time_unit_list = [_to - _from for _from, _to in from_to_list]

        # check that rows present
        self.assertGreaterEqual(
            len(from_to_list), 10,
            msg='The number of lines must be greater than %s\n' + output_csv.captured)
        # check that rows no more than 100
        self.assertLessEqual(
            len(from_to_list), 100,
            msg='The number of rows to be less than %s\n' + output_csv.captured)
        # check order by time
        self.assertEqual(
            sorted(from_column), from_column,
            msg='"From" column must be sorted in increasing\n' + output_csv.captured)
        self.assertEqual(
            sorted(to_column), to_column,
            msg='"To" column must be sorted in increasing\n' + output_csv.captured)
        # check that From and To date changes
        self.assertEqual(len(set(from_column)), len(from_column),
            msg='All values in the column "From" must be unique\n' + output_csv.captured)
        self.assertEqual(len(set(to_column)), len(to_column),
            msg='All values in the column "To" must be unique\n' + output_csv.captured)
        # check time-unit order
        self.assertEqual(sorted(time_unit_list, reverse=True), time_unit_list,
            msg='time-unit (time-unit = To - From) must be dynamically reduced\n' + output_csv.captured)
        # check 1min time-unit present
        self.assertIn(datetime.timedelta(minutes=1), time_unit_list,
            msg='The value of a time-unit = 10 minutes must be present\n' + output_csv.captured)


if __name__ == '__main__':
    unittest.main()
