# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import datetime
import unittest

from ddt import ddt, data, unpack
from mock import patch


TEST_CASES_PERIOD = (
    # arguments                       From                                   To
    ('--period 1d',                   datetime.datetime(1999, 12, 31, 0, 0), datetime.datetime(2000, 1, 1, 0, 0)),
    ('--period 1h',                   datetime.datetime(1999, 12, 31, 23, 0), datetime.datetime(2000, 1, 1, 0, 0)),
    ('--period 1m',                   datetime.datetime(1999, 12, 31, 23, 59), datetime.datetime(2000, 1, 1, 0, 0)),
    ('--period 25',                   datetime.datetime(1999, 12, 31, 23, 35), datetime.datetime(2000, 1, 1, 0, 0)),
    ('--period 1d3h13m',              datetime.datetime(1999, 12, 30, 20, 47), datetime.datetime(2000, 1, 1, 0, 0)),
    ('--from 2014-01-01 13:25 '
     '--to 2014-02-13 17:13',         datetime.datetime(2014, 1, 1, 13, 25), datetime.datetime(2014, 2, 13, 17, 13)),
    ('--from 14-01-01 --to 14-02-13', datetime.datetime(2014, 1, 1, 0, 0),   datetime.datetime(2014, 2, 13, 0, 0)),
)


TEST_CASES_ORDER_BY = (
    # input data
    ('--order-by=cpu_avg',              'aCPU'),
    ('--order-by "aCPU"',               'aCPU'),
    ('--order-by aCpu',                 'aCPU'),
    ('-o total_pmem_faults',            'PMemF'),
    ('-o=PMEMF',                        'PMemF'),
    ('-o="pmemf"',                      'PMemF'),
    ('--order-by any_faults',           'anyF'),
    ('--order-by anyF',                 'anyF'),
    ('--order-by anyf',                 'anyF'),
)

TEST_CASES_BY_FAULT = (
    ('--by-fault mem,vmEm,VMemF',        ['VMemF', 'VMemF', 'VMemF']),
    ('--by-fault=any_fauLts,iops,IOf',   ['anyF', 'IOPSf', 'IOf']),
    ('--by-fault mCpU cpu cPUf',         ['CPUf', 'CPUf', 'CPUf']),
    ('--by-fault mcpu cpu CPUf',         ['CPUf', 'CPUf', 'CPUf']),
)

TEST_CASES_SHOW_COLUMNS = (
    # input data
    ('--show-columns From To ID',       ['From', 'To', 'ID']),
    ('--show-columns aCPU,aCPU,aCPU',   ['aCPU', 'aCPU', 'aCPU']),
    ('--show-columns=From,To,ID,IOPSf', ['From', 'To', 'ID', 'IOPSf']),
    ('--show-columns id aiops liops',   ['ID', 'aIOPS', 'lIOPS']),
    ('--show-all -m=v2',                 ['From',   'To',     'ID',
                                          'aCPU',   'uCPU',   'mCPU',   'lCPU',   'CPUf',
                                          'aEP',    'uEP',    'mEP',    'lEP',    'EPf',
                                          'aVMem',  'uVMem',  'mVMem',  'lVMem',  'VMemF',
                                          'aPMem',  'uPMem',  'mPMem',  'lPMem',  'PMemF',
                                          'aNproc', 'uNproc', 'mNproc', 'lNproc', 'NprocF',
                                          'aIO',    'uIO',    'mIO',    'lIO',    'IOf',
                                          'aIOPS',  'mIOPS',  'uIOPS',  'lIOPS',  'IOPSf']),
    ('--show-columns all -m v2',         ['From',   'To',     'ID',
                                          'aCPU',   'uCPU',   'mCPU',   'lCPU',   'CPUf',
                                          'aEP',    'uEP',    'mEP',    'lEP',    'EPf',
                                          'aVMem',  'uVMem',  'mVMem',  'lVMem',  'VMemF',
                                          'aPMem',  'uPMem',  'mPMem',  'lPMem',  'PMemF',
                                          'aNproc', 'uNproc', 'mNproc', 'lNproc', 'NprocF',
                                          'aIO',    'uIO',    'mIO',    'lIO',    'IOf',
                                          'aIOPS',  'mIOPS',  'uIOPS',  'lIOPS',  'IOPSf']),

    # test case for backward compatibility (lveinfo 0.1.x)
    ('--show-all -m v1',                ['ID', 'aCPU', 'mCPU', 'lCPU', 'aEP', 'mEP', 'lEP', 'aVMem', 'mVMem', 'lVMem',
                                         'VMemF', 'EPf', 'aPMem', 'mPMem', 'lPMem', 'aNproc', 'mNproc', 'lNproc',
                                         'PMemF', 'NprocF', 'aIO', 'mIO', 'lIO', 'aIOPS', 'mIOPS', 'lIOPS']),
    ('--show-columns all -m v1',        ['ID', 'aCPU', 'mCPU', 'lCPU', 'aEP', 'mEP', 'lEP', 'aVMem', 'mVMem', 'lVMem',
                                         'VMemF', 'EPf', 'aPMem', 'mPMem', 'lPMem', 'aNproc', 'mNproc', 'lNproc',
                                         'PMemF', 'NprocF', 'aIO', 'mIO', 'lIO', 'aIOPS', 'mIOPS', 'lIOPS']),
)

TEST_CASES_LIMIT = (
    ('',                  10),
    ('-m v1',             10),
    ('-m v2',             10),
    ('--limit 0',          0),
    ('--limit 10',        10),
    ('--limit 100',       100),

    ('--csv',               0),
    ('--csv --limit  10',  10),
    ('--csv --limit   2',   2),
    ('--csv --limit   0',   0),
    ('--csv --limit 100', 100),

    ('--json',               0),
    ('--json --limit 10',   10),
    ('--json --limit  2',    2),
    ('--json --limit  0',    0),
    ('--json --limit 100', 100),

    ('--id 0',               0),
    ('--id 0 --limit=10',   10),
)

TEST_CASES_RESELLER = (
    ('--reseller-name resel1', 1000001241),
    ('--reseller-name resel2', 1000001242),
    ('--reseller-id 1241',     1000001241),
    ('--reseller-id 1242',     1000001242),
)


@ddt
class TestLveinfoArgParse(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        from lvestats.lib.parsers.lveinfoargparse import lveinfo_parser
        datetime_now = datetime.datetime(2000, 1, 1, 0, 0)  # Millennium time
        cls.command_line_parser = lveinfo_parser(
                name='lveinfo',
                ver='default',
                config={'mode': 'v1'},
                _datetime_now=datetime_now)
        cls.command_line_parser.lve_version = 8

    @data(*TEST_CASES_PERIOD)
    @unpack
    def test_period(self, args_, from_, to_):
        namespace = self.command_line_parser.parse_args(args_.split())
        self.assertEqual(from_, getattr(namespace, 'from'))
        self.assertEqual(to_, namespace.to)

    @data(*TEST_CASES_ORDER_BY)
    @unpack
    def test_order_by(self, args_, output_):
        namespace = self.command_line_parser.parse_args(args_.split())

        self.assertEqual(output_, namespace.order_by)

    @data(*TEST_CASES_BY_FAULT)
    @unpack
    def test_by_fault(self, args_, output_):
        namespace = self.command_line_parser.parse_args(args_.split())
        self.assertEqual(output_, namespace.by_fault)

    @data(*TEST_CASES_SHOW_COLUMNS)
    @unpack
    def test_show_columns(self, args_, output_):
        namespace = self.command_line_parser.parse_args(args_.split())
        self.assertEqual(namespace.show_columns, output_)

    @data(*TEST_CASES_LIMIT)
    @unpack
    def test_limit(self, args_, output_):
        namespace = self.command_line_parser.parse_args(args_.split())
        self.assertEqual(namespace.limit, output_)

    @staticmethod
    def _mock_get_lvp_id(name):
        resellers_dict = {
            'resel1': 1000001241,
            'resel2': 1000001242,
            1241: 1000001241,
            1242: 1000001242,
        }
        return resellers_dict.get(name)

    @data(*TEST_CASES_RESELLER)
    @unpack
    def test_reseller(self, args_, output_):
        with patch(
                'lvestats.lib.parsers.lveinfoargparse.uidconverter.convert_id_to_lvp_id',
                self._mock_get_lvp_id):
            with patch(
                    'lvestats.lib.parsers.lveinfoargparse.uidconverter.convert_name_to_lvp_id',
                    self._mock_get_lvp_id):
                namespace = self.command_line_parser.parse_args(args_.split())
                self.assertEqual(namespace.id, output_)


class TestArgparseDBGov(unittest.TestCase):
    datetime_now = datetime.datetime(2000, 1, 1, 0, 0)  # Millennium time

    @classmethod
    def setUpClass(cls):
        from lvestats.lib.parsers.lveinfoargparse import dbgov_parser
        cls.parser = dbgov_parser(config={'mode': 'v1', 'server_id': 'localhost'}, _datetime_now=cls.datetime_now)

    def test_010_order_by(self):
        self.assertDictContainsSubset(
            {'order_by': 'con'},
            vars(self.parser.parse_args('--order-by=con'.split())))

    def test_020_order_by(self):
        self.assertDictContainsSubset(
            {'order_by': 'cpu'},
            vars(self.parser.parse_args('-o cpu'.split())))

    def test_030_format(self):
        self.assertDictContainsSubset(
            {'format': ['TS', 'USER', 'CPU', 'READ', 'WRITE', 'CON']},
            vars(self.parser.parse_args('--format TS USER CPU READ WRITE CON'.split())))

    def test_040_format(self):
        self.assertDictContainsSubset(
            {'format': ['TS', 'USER', 'CPU', 'READ', 'WRITE', 'CON']},
            vars(self.parser.parse_args('-b=ts,username,cpu,read,write,con'.split())))


if __name__ == '__main__':
    unittest.main()
