# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import os
import unittest
import time

from lvestats.lib.lveinfolib import OutputFormatter


class TestOutputFormatter(unittest.TestCase):
    def setUp(self):
        self.rows = [[1, 2, 3, 4],
                     [1, 2, 3, 4],
                     [1, 2, 3, 4]]
        self.fields = ["1", "2", "3", "4"]

        os.environ["TZ"] = "US/Eastern"
        time.tzset()

    def test_01_hide_fields(self):
        out_f = OutputFormatter(rows=self.rows, fields=self.fields)
        out_f.hide_fields(["2"])
        self.assertEqual(out_f.get_fields(), ["1", "3", "4"])

    def test_02_hide_fields(self):
        out_f = OutputFormatter(rows=self.rows, fields=self.fields)
        out_f.hide_fields(["2"])
        self.assertEqual(list(out_f), [[1, 3, 4], [1, 3, 4], [1, 3, 4]])
        self.assertEqual(out_f.get_corrected_list(), [[1, 3, 4], [1, 3, 4], [1, 3, 4]])

    def test_03_hide_fields(self):
        out_f = OutputFormatter(rows=self.rows, fields=self.fields)
        out_f.hide_fields(["2"])
        out_f.hide_fields(["2", "3"])
        out_f.add_order(["3", "4"], "percentage")
        self.assertEqual(list(out_f), [[1, "400%"],
                                       [1, "400%"],
                                       [1, "400%"]])
        self.assertEqual(out_f.get_fields(), ["1", "4"])
        self.assertEqual(out_f.get_corrected_list(), [[1, "400%"],
                                                      [1, "400%"],
                                                      [1, "400%"]])

    def test_04_hide_fields(self):
        out_f = OutputFormatter(fields=["1", "2", "3", "del_me", "4"], )
        out_f.set_rows(rows=[[1, 2, 3, "del_me", 4],
                             [1, 2, 3, "del_me", 4],
                             [1, 2, 3, "del_me", 4]])
        out_f.hide_fields(["del_me"])  # delete only fields without rows
        out_f.hide_fields(["2"])  # delete first column (item with rows)
        self.assertEqual(list(out_f), [[1, 3, 4],
                                       [1, 3, 4],
                                       [1, 3, 4]])
        self.assertEqual(out_f.get_corrected_list(), [[1, 3, 4],
                                                      [1, 3, 4],
                                                      [1, 3, 4]])

    def test_05_add_order(self):
        out_f = OutputFormatter(fields=["From", "To"], orders=[
            (["From", "To"], "datetime"),
            (["From", "To"], "strftime")
        ])
        out_f.set_rows(rows=[[1414059360, 1414059420], [1414059360, 1414059420]])
        self.assertEqual(list(out_f), [["10-23 06:16", "10-23 06:17"],
                                       ["10-23 06:16", "10-23 06:17"]])
        self.assertEqual(out_f.get_corrected_list(),
                         [["10-23 06:16", "10-23 06:17"],
                          ["10-23 06:16", "10-23 06:17"]])

    def test_06_blank_value_percentage(self):
        rows = [[1, 2, "-", 4],
                [1, 2, None, 4],
                [1, 2, "*", 4],
                [1, 2, "n/a", 4],
                [1, 2, 3, 4]]
        fields = ["1", "2", "3", "4"]
        out_f = OutputFormatter(rows=rows, fields=fields)
        out_f.add_order(["3"], "percentage")
        self.assertEqual(list(out_f), [[1, 2, "-", 4],
                                       [1, 2, "-", 4],
                                       [1, 2, "*", 4],
                                       [1, 2, "n/a", 4],
                                       [1, 2, "300%", 4]])
        self.assertEqual(out_f.get_corrected_list(),
                         [[1, 2, "-", 4],
                          [1, 2, "-", 4],
                          [1, 2, "*", 4],
                          [1, 2, "n/a", 4],
                          [1, 2, "300%", 4]])

    def test_06_blank_value_powers_of_1000_and_bytes(self):
        rows = [
            [None, 2, "-", 4],
            [1, 2048, None, 4],
            [2000, None, "*", 4],
            [1100000, 1100000, "n/a", 4],
            [6400000000, 6400000000, 3, 4],
            [999, 999, 0.111, 0],
            [5090, 6666, 0, 1],
            [7008000, 9226, 0, 1],
            [111111, 111111, 0, 1],
            [11111, 11111, 0, 1],
            [1111111, 1111111, 0, 1],
        ]
        fields = ["1", "2", "3", "4"]
        out_f = OutputFormatter(rows=rows, fields=fields)
        out_f.add_order(["1"], "powers_of_1000")
        out_f.add_order(["2"], "bytes")
        out_f.add_order(["3"], "percentage")
        self.assertEqual(
            list(out_f),
            [
                ["-", "2B", "-", 4],
                ["1", "2.0KB", "-", 4],
                ["2.0K", "-", "*", 4],
                ["1.1M", "1.0MB", "n/a", 4],
                ["6.4G", "6.0GB", "300%", 4],
                ["999", "999B", "11.1%", 0],
                ["5.1K", "6.5KB", "0%", 1],
                ["7.0M", "9.0KB", "0%", 1],
                ["111K", "108KB", "0%", 1],
                ["11.1K", "10.9KB", "0%", 1],
                ["1.1M", "1.1MB", "0%", 1]
            ])
        self.assertEqual(
            out_f.get_corrected_list(),
            [
                ["-", "2B", "-", 4],
                ["1", "2.0KB", "-", 4],
                ["2.0K", "-", "*", 4],
                ["1.1M", "1.0MB", "n/a", 4],
                ["6.4G", "6.0GB", "300%", 4],
                ["999", "999B", "11.1%", 0],
                ["5.1K", "6.5KB", "0%", 1],
                ["7.0M", "9.0KB", "0%", 1],
                ["111K", "108KB", "0%", 1],
                ["11.1K", "10.9KB", "0%", 1],
                ["1.1M", "1.1MB", "0%", 1]
            ]

        )

    def test_get_correct_list(self):
        rows = [[1, 2, "-", 4],
                [1, 2, None, 4],
                [1, 2, "*", 4],
                [1, 2, "n/a", 4],
                [1, 2, 3, 4]]
        fields = ["1", "2", "3", "4"]
        out_f = OutputFormatter(rows=rows, fields=fields)
        self.assertEqual(out_f.get_corrected_list(), [[1, 2, "-", 4],
                                                      [1, 2, None, 4],
                                                      [1, 2, "*", 4],
                                                      [1, 2, "n/a", 4],
                                                      [1, 2, 3, 4]])


if __name__ == "__main__":
    unittest.main()
