# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import io
import csv
import os
import unittest
import warnings

import ddt
from testfixtures import OutputCapture

from test.base.base_db_test import BaseDbTest
from test.generate_lvestats2db import DatabaseGenerator
from lvestats.lib.info.lveinfomain import is_cpu_output_normalizable, normalize_cpu
from lvestats.lib.info.lveinfomain import main_ as lveinfo_main


class Namespace(object):
    def __init__(self, l):
        self.id = l[0]
        self.json = l[1] == 'json'
        self.csv = l[1] == 'csv'
        self.style = l[2]


input_rows = [[1000.,  2000.,  5000.],
              [10000., 10000., 10000.],
              [10000., 10000., 20000.],
              [0.,     2000.,  2000.]]


@ddt.ddt
class TestNormalizer(unittest.TestCase):
    DDT_PARAM_SHOULD_NORMALIZE_TRUE = [(10, 'csv', 'user'),
                                       (10, 'json', 'user')]
    DDT_PARAM_SHOULD_NORMALIZE_FALSE = [(10, 'csv', 'admin'),
                                        (10, 'json', 'admin'),
                                        (10, 'other', 'user'),
                                        (None, 'csv', 'user'),
                                        (None, 'json', 'user'),
                                        (None, 'other', 'user'),
                                        (10, 'csv', 'admin'),
                                        (10, 'json', 'admin')]
    DDT_PARAM_DATA = [(['aCPU', 'mCPU', 'lCPU'],  [[20.0, 40.0, 100.0],
                                                   [100.0, 100.0, 100.0],
                                                   [50.0, 50.0, 100.0],
                                                   [0.0, 100.0, 100.0]]),
                      (['other', 'mCPU', 'lCPU'], [[1000.,  40.0, 100.0],
                                                   [10000.0, 100.0, 100.0],
                                                   [10000.0, 50.0, 100.0],
                                                   [0.,     100.,  100.]]),
                      (['aCPU', 'other', 'lCPU'], [[20.,  2000.,   100.],
                                                   [100., 10000.,  100.],
                                                   [50.,  10000.,  100.],
                                                   [0.,     2000.,   100.]]),
                      (['aCPU', 'mCPU', 'other'], input_rows)
                      ]

    @ddt.data(*DDT_PARAM_SHOULD_NORMALIZE_TRUE)
    def test_is_cpu_output_normalizable_true(self, value):
        self.assertTrue(is_cpu_output_normalizable(Namespace(value)))

    @ddt.data(*DDT_PARAM_SHOULD_NORMALIZE_FALSE)
    def test_is_cpu_output_normalizable_false(self, value):
        self.assertFalse(is_cpu_output_normalizable(Namespace(value)))

    @ddt.data(*DDT_PARAM_DATA)
    @ddt.unpack
    def test_normalize_cpu(self, fields, rows):
        self.assertEqual(normalize_cpu(input_rows, fields), rows)


@ddt.ddt
class TestFunctionalUserStyle(BaseDbTest):

    @classmethod
    def setUpClass(cls):
        BaseDbTest.setUpClass()  # create tables
        cls.need_clear_rows = False  # prevent clear rows for every unittest
        cls.db_generator = DatabaseGenerator(
            db_uri=cls.engine.url, server_id=cls.cfg['server_id'],
            lve_data_path=os.path.join(os.path.dirname(__file__), 'data', 'lveinfo_style_user.csv'))
        cls.db_generator.dbengine = cls.engine
        cls.db_generator.generate_database()
        # generate input arguments

    @staticmethod
    def parse_csv_output(csv_str):
        csv_file = io.StringIO(csv_str)
        csv_dict_reader = csv.DictReader(csv_file, delimiter=',')
        return list(csv_dict_reader)

    @ddt.data(10001, 10002, 10003)
    def test_style_user(self, uid):
        argv_ = ['--id', str(uid), '--style=user',
                 '--from', '2000-09-17', '01:00',
                 '--to', '2000-09-17', '02:00',
                 '-mv1', '--show-all', '--csv', '--time-unit=1h']
        with OutputCapture() as output_csv:  # capture all output to stdout
            # TODO: investigate why mysql on cl6 i686 prints warnings
            # /opt/alt/python37/lib/python3.7/site-packages/sqlalchemy/dialects/mysql/base.py:877:
            # SAWarning: Datatype FLOAT does not support CAST on MySQL; the CAST will be skipped.
            # self.dialect.type_compiler.process(cast.typeclause.type))
            with warnings.catch_warnings(record=True):
                lveinfo_main(self.cfg, dbengine=self.engine, argv_=argv_)
        # filter needed columns
        output = self.parse_csv_output(output_csv.captured)[0]
        self.assertDictContainsSubset(
            {'aCPU': '20', 'mCPU': '50', 'lCPU': '100'}, output)

if __name__ == "__main__":
    unittest.main()
