# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import time
import os
import unittest
from uuid import uuid4

import ddt
import mock
from clcommon.clproc import LIMIT_LVP_ID

import test.helpers as helpers
import lvestats.lib.config
import test.test_func
from test.base.base_db_test import BaseDbTest
from test.generate_lvestats2db import DatabaseGenerator, YEAR
from lvestats.lib.chart import Renderer, SvgChart
from lvestats.lib.chart.lvechartmain import LveChart
from lvestats.lib.commons import dateutil
from lvestats.lib.commons.func import serialize_lve_id
import lvestats.lib.commons.func

__author__ = 'alukyanov'


_USERMAP = {503: 'user503', 501: 'user501'}
_INV_USERMAP = {v: k for k, v in _USERMAP.items()}


class BaseLVEMainChartTest(BaseDbTest):
    svg = '/tmp/{0}.svg'.format(uuid4())

    @classmethod
    def setUpClass(cls):
        BaseDbTest.setUpClass()
        cls.db_generator = DatabaseGenerator(
            db_uri=cls.engine.url, server_id=cls.cfg['server_id'],
            lve_data_path=os.path.join(os.path.dirname(__file__), 'data', 'lve_chart_base.csv'))
        cls.db_generator.dbengine = cls.engine
        cls.db_generator.generate_database()
        cls.need_clear_rows = False


@ddt.ddt
class TestLVEMainChart(BaseLVEMainChartTest):
    def setUp(self):
        patcher1 = mock.patch('lvestats.lib.config.read_config', return_value={})
        patcher1.start()
        self.addCleanup(patcher1.stop)

    def test_lvemainchart(self):
        LveChart(self.cfg).main(
            args='--from {year}-09-23 00:00 --to {year}-09-23 01:00 --id 100000 --show-all'
            ' --format svg --output {svg}'.format(
            year=YEAR, svg=self.svg).split(), debug_engine=self.engine)

    def test_lvemainchart_normalize_user_cpu(self):
        chart = LveChart(self.cfg)
        chart.governor_mode = "none"
        chart.is_normalized_user_cpu = True
        chart.main(
            args='--from {year}-09-23 00:00 --to {year}-09-23 01:00 --id 100000 --show-all'
                 ' --style user --format svg --output {svg}'.format(
                year=YEAR, svg=self.svg).split(),
            debug_engine=self.engine)

    @unittest.skipIf(os.getuid() == 0, "alt-imageMagic package not exists for ubuntu")
    def test_unknown_user(self):
        LveChart(self.cfg).main(
            args=['--user', 'NotKnownOne', '--output', self.svg],
            debug_engine=self.engine)

    def test_wrong_db_type(self):
        chart = LveChart({
            'db_type': 'sqliteqw',
            'server_id': 'localhost',
        })
        with helpers.capture_all(chart.main, []) as output:
            self.assertEqual(1, output["result"])
            self.assertIn('sqliteqw', output["stderr"])

    @ddt.data(4, 6, 8, 9)
    @mock.patch('os.getuid', return_value=0)
    @mock.patch("lvestats.lib.lveinfolib.get_lve_version")
    def test_no_faults_message(self, ddt_lve_version, mock_get_lve_version, os_getuid):
        global expected_message_appeared
        expected_message_appeared = False
        mock_get_lve_version.return_value = ddt_lve_version
        old_add_graph = SvgChart.add_graph

        def add_graph(*args, **kwargs):
            global expected_message_appeared
            if kwargs['message'] == 'No Faults':
                expected_message_appeared = True
            old_add_graph(*args, **kwargs)

        SvgChart.add_graph = add_graph

        LveChart(self.cfg).main(
            args='--from {year}-09-23 00:00 --to {year}-09-23 01:00 --id 100000 --show-all'
                 ' --style user --format svg --output {svg}'.format(
            year=YEAR,
            svg="/tmp/test_chart_lve{}_with_faults.svg".format(ddt_lve_version)).split(),
            debug_engine=self.engine)
        self.assertFalse(expected_message_appeared, '"No Faults" message was shown')

        LveChart(self.cfg).main(
            args='--from {year}-09-23 00:00 --to {year}-09-23 01:00 --id 200000 --show-all'
                 ' --style user --format svg --output {svg}'.format(
            year=YEAR,
            svg="/tmp/test_chart_lve{}_no_faults.svg".format(ddt_lve_version)).split(),
            debug_engine=self.engine)
        self.assertTrue(expected_message_appeared, '"No Faults" message was not shown')

        SvgChart.add_graph = old_add_graph

    # LVESTATS-97
    @ddt.data(
        (lambda: 1000, 'reseller1', ['user2'], lambda: None),
        (lambda: 1000, 'reseller1', ['user1', 'user2'], lambda: 1000),
        (lambda: 1000, 'reseller1', ['user4', 'user3'], lambda: None),
        (lambda: serialize_lve_id(LIMIT_LVP_ID, 10001), 'reseller1', [], lambda: None),
        (lambda: serialize_lve_id(LIMIT_LVP_ID, 10001), 'reseller2', [], lambda: serialize_lve_id(LIMIT_LVP_ID, 10001))
    )
    @ddt.unpack
    def test_chartmain_main_admin(self, get_uid, reseller_name, reseller_users, get_result):
        uid = get_uid()
        args = [
            '--id={}'.format(uid),
            '--for-reseller={}'.format(reseller_name),
            '--output=/tmp/test.svg']

        def mock_reseller_to_uid(reseller):
            return {
                'reseller1': serialize_lve_id(LIMIT_LVP_ID, 10000),
                'reseller2': serialize_lve_id(LIMIT_LVP_ID, 10001)
            }.get(reseller, None)

        def mock_username_to_uid(username):
            return {
                'user1': 1000,
                'user2': 1001
            }.get(username, None)
        with \
                mock.patch('os.getuid', return_value=0),\
                mock.patch('lvestats.lib.cloudlinux_statistics.get_users_for_reseller', return_value=reseller_users), \
                mock.patch('lvestats.lib.cloudlinux_statistics.username_to_uid_local', side_effect=mock_username_to_uid), \
                mock.patch('lvestats.lib.cloudlinux_statistics.reseller_to_uid', side_effect=mock_reseller_to_uid), \
                mock.patch('lvestats.lib.chart.ChartMain._render', return_value=None), \
                mock.patch('lvestats.lib.chart.ChartMain._output'), \
                mock.patch('lvestats.lib.chart.lvechartmain.LveChart.get_chart_data', return_value=([], [], [])) as mock_get_chart_data:
            chart = LveChart(self.cfg)
            with helpers.capture_all(lambda: chart.main(args, self.engine)) as output:
                result = get_result()
                if result is None:
                    self.assertEqual(len(mock_get_chart_data.call_args_list), 0)
                    self.assertEqual(output['result'], 1)
                    self.assertNotEqual(output['stderr'].strip(), '')
                    self.assertEqual(output['stdout'].strip(), '')
                else:
                    self.assertEqual(len(mock_get_chart_data.call_args_list), 1)
                    self.assertEqual(mock_get_chart_data.call_args_list[0][1]['user_id'], result)
                    self.assertEqual(output['result'], 0)
                    self.assertEqual(output['stderr'].strip(), '')
                    self.assertEqual(output['stdout'].strip(), '')

    @ddt.data(
        (2222, 'reseller1', 2222),
        (1111, 'reseller1', None),
        (1111, None, None),
        (None, 'reseller1', 2222),
        (None, None, 2222),
    )
    @ddt.unpack
    def test_chartmain_main_user(self, uid, reseller_name, result):
        TEST_USER = 'user1'
        args = []
        if uid:
            args.append('--id={}'.format(uid))
        if reseller_name:
            args.append('--for-reseller={}'.format(reseller_name))
        args.append('--output=/tmp/test.svg')

        def mock_username_to_uid(username):
            return {
                TEST_USER: 2222,
                'user2': 1001
            }.get(username, None)
        with \
                mock.patch('os.getuid', return_value=2222), \
                mock.patch('pwd.getpwuid', return_value=[TEST_USER]), \
                mock.patch('lvestats.lib.chart.lvechartmain.getuser', return_value='user'), \
                mock.patch('lvestats.lib.cloudlinux_statistics.getuser', return_value='user'), \
                mock.patch('lvestats.lib.cloudlinux_statistics.username_to_uid_local', side_effect=mock_username_to_uid), \
                mock.patch('lvestats.lib.chart.ChartMain._render', return_value=None), \
                mock.patch('lvestats.lib.chart.ChartMain._output'), \
                mock.patch('lvestats.lib.chart.lvechartmain.LveChart.get_chart_data', return_value=([], [], [])) as mock_get_chart_data:
            chart = LveChart(self.cfg)
            with helpers.capture_all(lambda: chart.main(args, self.engine)) as output:
                if result is None:
                    self.assertEqual(len(mock_get_chart_data.call_args_list), 0)
                    self.assertEqual(output['result'], 1)
                    self.assertEqual(output['stderr'].strip(), 'Permission denied')
                    self.assertEqual(output['stdout'].strip(), '')
                else:
                    self.assertEqual(len(mock_get_chart_data.call_args_list), 1)
                    self.assertEqual(mock_get_chart_data.call_args_list[0][1]['user_id'], result)
                    self.assertEqual(output['result'], 0)
                    self.assertEqual(output['stderr'].strip(), '')
                    self.assertEqual(output['stdout'].strip(), '')


@ddt.ddt
class TestLoadingData(BaseDbTest):
    def setUp(self):
        BaseDbTest.setUp(self)
        os.environ['TZ'] = 'UTC'  # set the zero time zone
        time.tzset()
        self.cfg["aggregation_period"] = 100
        self.chart = LveChart(config=self.cfg)
        self.date_from = dateutil.gm_to_local(dateutil.unixtimestamp_to_gm_datetime(100))
        self.date_to = dateutil.gm_to_local(dateutil.unixtimestamp_to_gm_datetime(1005))
        self.clear_rows = True
        self.server = 'localhost'

    def tearDown(self):
        del os.environ['TZ']
        time.tzset()
        BaseDbTest.tearDown(self)

    @ddt.data(["2016-09-23 00:00", "2016-09-23 00:01", 60, 0],
              ["2016-09-23 00:00", "2016-10-23 00:00", 30 * 24 * 60 * 60, 100],
              # assumes max 500 points per on a graph
              ["2016-12-31 10:00", "2017-01-05 11:00", 5 * 24 * 60 * 60 + 60 * 60, 100])
    @ddt.unpack
    def test_time_step(self, from_, to_, seconds, step):
        if step == 0:
            step = self.chart.MINIMAL_TIME_STEP
        result = self.chart.get_time_step(dateutil.parse_date(from_), dateutil.parse_date(to_))
        self.assertEqual((seconds, step), result)

    def generate(self, from_, to_, uid, val, luck):
        time_x = (from_ // 100) * 100
        connection, trans = self.bulk_history_start()
        import random
        while time_x <= to_:
            if random.randint(0, 100) <= luck:  # generate some missing entries
                self.bulk_history_add(connection, self.server, uid=uid, _time=time_x, cpu=val, cpu_limit=val, io=val, io_limit=val)
            time_x += 100
        trans.commit()

    @ddt.data([100, 7000, 10], [1000, 133303, 20], [239839, 270001, 30])
    @ddt.unpack
    def test_load_data(self, from_, to_, luck):
        """
        :param from_: sec to use as from
        :param to_: sec to use as to
        :param luck: % of records for the user to create
        :return:
        """
        from datetime import datetime
        from_ts, to_ts = datetime.fromtimestamp(from_), datetime.fromtimestamp(to_)
        self.generate(from_, to_, 0, 0, 99)
        self.generate(from_, to_, 1, 3000, luck)
        _, time_step = self.chart.get_time_step(from_ts, to_ts)
        result = self.chart.load_data(engine=self.engine, server="localhost",
                                      time_step=time_step, user_id=1, utc_from=from_ts, utc_to=to_ts)
        for i in range(1, len(result)):
            self.assertEqual(time_step, result[i][0] - result[i - 1][0])


class TestGetChartData(BaseDbTest):
    def setUp(self):
        mb = 1024.0 * 1024  # bytes in one megabyte
        BaseDbTest.setUp(self)
        os.environ['TZ'] = 'UTC'  # set the zero time zone
        time.tzset()
        self.save_history_line("localhost", uid=0, _time=100,  cpu=0, cpu_limit=0, io=0, io_limit=0)
        self.save_history_line("localhost", uid=0, _time=200,  cpu=0, cpu_limit=0, io=0, io_limit=0)
        self.save_history_line("localhost", uid=0, _time=300,  cpu=0, cpu_limit=0, io=0, io_limit=0)
        self.save_history_line("localhost", uid=0, _time=400,  cpu=0, cpu_limit=0, io=0, io_limit=0)
        self.save_history_line("localhost", uid=0, _time=500,  cpu=0, cpu_limit=0, io=0, io_limit=0)
        self.save_history_line("localhost", uid=0, _time=600,  cpu=0, cpu_limit=0, io=0, io_limit=0)
        self.save_history_line("localhost", uid=0, _time=700,  cpu=0, cpu_limit=0, io=0, io_limit=0)
        self.save_history_line("localhost", uid=0, _time=800,  cpu=0, cpu_limit=0, io=0, io_limit=0)
        self.save_history_line("localhost", uid=0, _time=900,  cpu=0, cpu_limit=0, io=0, io_limit=0)
        self.save_history_line("localhost", uid=0, _time=1000, cpu=0, cpu_limit=0, io=0, io_limit=0)

        self.save_history_line("localhost", uid=1, _time=100, cpu=250,   cpu_limit=5000, io=0.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=1, _time=200, cpu=0,     cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=1, _time=300, cpu=500,   cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=1, _time=400, cpu=1000,  cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=1, _time=500, cpu=750,   cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=1, _time=600, cpu=1500,  cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=1, _time=700, cpu=100,   cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=1, _time=800, cpu=800,   cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=1, _time=900, cpu=5000,  cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=1, _time=1000, cpu=1000, cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)

        self.save_history_line("localhost", uid=3, _time=100, cpu=250,   cpu_limit=5000, io=0.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=3, _time=200, cpu=0,     cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=3, _time=300, cpu=500,   cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=3, _time=400, cpu=1000,  cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=3, _time=500, cpu=750,   cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=3, _time=600, cpu=1500,  cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=3, _time=700, cpu=100,   cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=3, _time=800, cpu=800,   cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=3, _time=900, cpu=5000,  cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=3, _time=1000, cpu=1000, cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)

        self.save_history_line("localhost", uid=6, _time=100, cpu=250,   cpu_limit=5000, io=0.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=6, _time=200, cpu=0,     cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=6, _time=300, cpu=500,   cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=6, _time=400, cpu=1000,  cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=6, _time=500, cpu=750,   cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=6, _time=600, cpu=1500,  cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=6, _time=700, cpu=100,   cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=6, _time=800, cpu=800,   cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=6, _time=900, cpu=5000,  cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=6, _time=1000, cpu=1000, cpu_limit=5000, io=2.5*mb, io_limit=12.5*mb)

        self.save_history_line("localhost", uid=7, _time=100, cpu=250,   cpu_limit=0)
        self.save_history_line("localhost", uid=7, _time=200, cpu=0,     cpu_limit=0)
        self.save_history_line("localhost", uid=7, _time=300, cpu=500,   cpu_limit=0)
        self.save_history_line("localhost", uid=7, _time=400, cpu=1000,  cpu_limit=0)
        self.save_history_line("localhost", uid=7, _time=500, cpu=750,   cpu_limit=0)
        self.save_history_line("localhost", uid=7, _time=600, cpu=1500,  cpu_limit=0)
        self.save_history_line("localhost", uid=7, _time=700, cpu=100,   cpu_limit=0)
        self.save_history_line("localhost", uid=7, _time=800, cpu=800,   cpu_limit=0)
        self.save_history_line("localhost", uid=7, _time=900, cpu=5000,  cpu_limit=0)
        self.save_history_line("localhost", uid=7, _time=1000, cpu=1000, cpu_limit=0)

        self.save_history_line("localhost", uid=2, _time=100, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=2, _time=200, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=2, _time=300, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=2, _time=400, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=2, _time=500, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=2, _time=600, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=2, _time=700, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=2, _time=800, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=2, _time=900, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=2, _time=1000, cpu=5000, cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)

        self.save_history_line("localhost", uid=4, _time=100, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=4, _time=200, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=4, _time=300, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=4, _time=400, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=4, _time=500, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=4, _time=600, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=4, _time=700, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=4, _time=800, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=4, _time=900, cpu=5000,  cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)
        self.save_history_line("localhost", uid=4, _time=1000, cpu=5000, cpu_limit=5000, io=12.5*mb, io_limit=12.5*mb)

        self.save_dbgov_line('localhost', uid=1,  _time=0,    sum_cpu=1, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=1,  _time=99,   sum_cpu=2, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=1,  _time=100,  sum_cpu=3, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=1,  _time=101,  sum_cpu=4, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=1,  _time=500,  sum_cpu=5, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=1,  _time=700,  sum_cpu=6, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=1,  _time=750,  sum_cpu=7, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=1,  _time=800,  sum_cpu=8, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=1,  _time=1000, sum_cpu=9, sum_write=1, sum_read=1)

        self.save_dbgov_line('localhost', uid=4,  _time=0,    sum_cpu=1, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=4,  _time=99,   sum_cpu=2, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=4,  _time=100,  sum_cpu=3, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=4,  _time=101,  sum_cpu=4, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=4,  _time=500,  sum_cpu=5, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=4,  _time=700,  sum_cpu=6, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=4,  _time=750,  sum_cpu=7, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=4,  _time=800,  sum_cpu=8, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=4,  _time=1000, sum_cpu=9, sum_write=1, sum_read=1)

        self.save_dbgov_line('localhost', uid=2, _time=100,  sum_cpu=3, sum_write=1)
        self.save_dbgov_line('localhost', uid=2, _time=101,  sum_cpu=4, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=2, _time=500,  sum_cpu=5, limit_cpu_on_period_end=10)
        self.save_dbgov_line('localhost', uid=2, _time=700,  sum_cpu=6, sum_write=8, sum_read=1)
        self.save_dbgov_line('localhost', uid=2, _time=750,  sum_cpu=7, limit_cpu_on_period_end=10)
        self.save_dbgov_line('localhost', uid=2, _time=800,  sum_cpu=8, sum_read=1, limit_cpu_on_period_end=10)
        self.save_dbgov_line('localhost', uid=2, _time=1100, sum_cpu=9, sum_write=1, sum_read=1)

        self.save_dbgov_line('localhost', uid=3, _time=100,  sum_cpu=3, sum_write=1)
        self.save_dbgov_line('localhost', uid=3, _time=101,  sum_cpu=4, sum_write=1, sum_read=1)
        self.save_dbgov_line('localhost', uid=3, _time=500,  sum_cpu=5, limit_cpu_on_period_end=10)
        self.save_dbgov_line('localhost', uid=3, _time=700,  sum_cpu=6, sum_write=8, sum_read=1)
        self.save_dbgov_line('localhost', uid=3, _time=750,  sum_cpu=7, limit_cpu_on_period_end=10)
        self.save_dbgov_line('localhost', uid=3, _time=800,  sum_cpu=8, sum_read=1, limit_cpu_on_period_end=10)
        self.save_dbgov_line('localhost', uid=3, _time=1100, sum_cpu=9, sum_write=1, sum_read=1)

        self.cfg["aggregation_period"] = 100
        self.chart = LveChart(self.cfg)
        self.date_from = dateutil.gm_to_local(dateutil.unixtimestamp_to_gm_datetime(100))
        self.date_to = dateutil.gm_to_local(dateutil.unixtimestamp_to_gm_datetime(1005))
        self.svg = '/tmp/{0}.svg'.format(uuid4())

        patcher1 = mock.patch('lvestats.lib.config.read_config', return_value={})
        patcher1.start()
        self.addCleanup(patcher1.stop)

    @mock.patch.object(lvestats.lib.config, "is_normalized_user_cpu", return_value=True)
    def test_normalize_unlimited(self, is_normalized):
        self.chart.main(
            args="--from 1970-01-01 00:00 --to 1970-01-01 00:10 --id 7 --show-all --style user --format svg --output /dev/null".split(),
            debug_engine=self.engine)

    def test_get_chart_data_without_dbgov_data_govmode_all(self):
        self.chart.governor_mode = "all"
        data_collected, times, _ = self.chart.get_chart_data(
            engine=self.engine,
            from_ts=self.date_from,
            to_ts=self.date_to,
            server="localhost",
            user_id=6)
        self.assertEqual(times,                   [100, 200, 300, 400,  500, 600,  700, 800, 900,  1000])
        self.assertEqual(data_collected['dbcpu'], [0,   0,   0,   0,    0,   0,    0,   0,   0,    0])
        self.assertEqual(data_collected['dbio'],  [0,   0,   0,   0,    0,   0,    0,   0,   0,    0])
        self.assertEqual(data_collected['acpu'],  [2.5, 0.0, 5.0, 10.0, 7.5, 15.0, 1.0, 8.0, 50.0, 10.0])

    def test_get_chart_data_without_dbgov_data_govmode_abusers(self):
        self.chart.governor_mode = "abusers"
        data_collected, times, _ = self.chart.get_chart_data(
            engine=self.engine,
            from_ts=self.date_from,
            to_ts=self.date_to,
            server="localhost",
            user_id=6)
        self.assertEqual(times,                   [100, 200, 300, 400,  500, 600,  700, 800, 900,  1000])
        self.assertEqual(data_collected['dbcpu'], [0,   0,   0,   0,    0,   0,    0,   0,   0,    0])
        self.assertEqual(data_collected['dbio'],  [0,   0,   0,   0,    0,   0,    0,   0,   0,    0])
        self.assertEqual(data_collected['acpu'],  [2.5, 0.0, 5.0, 10.0, 7.5, 15.0, 1.0, 8.0, 50.0, 10.0])

    def test_get_chart_data_without_dbgov_data_govmode_no_config(self):
        self.chart.governor_mode = "none"
        data_collected, times, _ = self.chart.get_chart_data(
            engine=self.engine,
            from_ts=self.date_from,
            to_ts=self.date_to,
            server="localhost",
            user_id=1)
        self.assertEqual(times, [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000])
        self.assertEqual(data_collected['dbcpu'], [])
        self.assertEqual(data_collected['dbio'], [])
        self.assertEqual(data_collected['acpu'], [2.5, 0.0, 5.0, 10.0, 7.5, 15.0, 1.0, 8.0, 50.0, 10.0])

    def test_get_chart_data_with_dbgov_data_less_than_lve_stats1(self):
        self.chart.governor_mode = "all"
        data_collected, times, _ = self.chart.get_chart_data(
            engine=self.engine,
            from_ts=self.date_from,
            to_ts=self.date_to,
            server="localhost",
            user_id=4)
        self.assertEqual(data_collected['dbcpu'], [3.5, 0, 0, 0, 5.0, 0, 6.0, 7.5, 0, 9.0])
        self.assertEqual(data_collected['dbio'], [2097152.0, 0, 0, 0, 2097152.0, 0, 2097152.0, 2097152.0, 0, 2097152.0])
        self.assertEqual(times, [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000])

    def test_get_chart_data_with_dbgov_data_bigger_than_lve_stats1_govmode_abusers(self):
        self.chart.governor_mode = "abusers"
        data_collected, times, _ = self.chart.get_chart_data(
            engine=self.engine,
            from_ts=self.date_from,
            to_ts=self.date_to,
            server="localhost",
            user_id=1)
        self.assertEqual(data_collected['acpu'],  [2.5,       0.0, 5.0, 10.0, 7.5,       15.0, 1.0,       8.0,       50.0, 10.0])
        self.assertEqual(data_collected['dbcpu'], [3.5,       0,   0,   0,    5.0,       0,    6.0,       7.5,       0,    9.0])
        self.assertEqual(data_collected['dbio'],  [2097152.0, 0,   0,   0,    2097152.0, 0,    2097152.0, 2097152.0, 0,    2097152.0])
        self.assertEqual(times,                   [100,       200, 300, 400,  500,       600,  700,       800,       900,  1000])

    @mock.patch("lvestats.lib.commons.func.get_governor_config_path", return_value="/var/run/mysql-governor-config.xml")
    @mock.patch('xml.dom.minidom.parse', side_effect=test.test_func.mock_dbgov_config)
    def test_get_chart__ignore_unknown_user(self, mock_parse_config_file, governor_config_path):
        self.chart.governor_mode = "abusers"
        with mock.patch('lvestats.lib.chart.lvechartmain.uid_to_username', return_value="unknown_user"):
            data_collected, times, _ = self.chart.get_chart_data(
                engine=self.engine,
                from_ts=self.date_from,
                to_ts=self.date_to,
                server="localhost",
                user_id=1)
            self.assertEqual(data_collected['acpu'],  [2.5,       0.0, 5.0, 10.0, 7.5,       15.0, 1.0,       8.0,       50.0, 10.0])
            self.assertEqual(data_collected['dbcpu'], [3.5, 0, 0, 0, 5.0, 0, 6.0, 7.5, 0, 9.0])
            self.assertEqual(data_collected['dbio'],  [2097152.0, 0, 0, 0, 2097152.0, 0, 2097152.0, 2097152.0, 0, 2097152.0])
            self.assertEqual(times,                   [100,       200, 300, 400,  500,       600,  700,       800,       900,  1000])

    @mock.patch("lvestats.lib.commons.func.get_governor_config_path", return_value="/var/run/mysql-governor-config.xml")
    @mock.patch('xml.dom.minidom.parse', side_effect=test.test_func.mock_dbgov_config)
    def test_get_chart__ignore__user1(self, mock_parse_config_file, governor_config_path):
        self.chart.governor_mode = "abusers"
        with mock.patch('lvestats.lib.chart.lvechartmain.uid_to_username', return_value="user1"):
            data_collected, times, _ = self.chart.get_chart_data(
                engine=self.engine,
                from_ts=self.date_from,
                to_ts=self.date_to,
                server="localhost",
                user_id=1)
            self.assertEqual(data_collected['acpu'],  [2.5,       0.0, 5.0, 10.0, 7.5,       15.0, 1.0,       8.0,       50.0, 10.0])
            self.assertEqual(data_collected['dbcpu'], [])
            self.assertEqual(data_collected['dbio'],  [])
            self.assertEqual(times,                   [100,       200, 300, 400,  500,       600,  700,       800,       900,  1000])

    @mock.patch("lvestats.lib.commons.func.get_governor_config_path", return_value="/var/run/mysql-governor-config.xml")
    @mock.patch('xml.dom.minidom.parse', side_effect=test.test_func.mock_dbgov_config)
    def test_get_chart__ignore_unknown_user2(self, mock_parse_config_file, governor_config_path):
        self.chart.governor_mode = "abusers"
        with mock.patch('lvestats.lib.chart.lvechartmain.uid_to_username', return_value="user2"):
            data_collected, times, _ = self.chart.get_chart_data(
                engine=self.engine,
                from_ts=self.date_from,
                to_ts=self.date_to,
                server="localhost",
                user_id=1)
            self.assertEqual(data_collected['acpu'],  [2.5,       0.0, 5.0, 10.0, 7.5,       15.0, 1.0,       8.0,       50.0, 10.0])
            self.assertEqual(data_collected['dbcpu'], [3.5, 0, 0, 0, 5.0, 0, 6.0, 7.5, 0, 9.0])
            self.assertEqual(data_collected['dbio'],  [2097152.0, 0, 0, 0, 2097152.0, 0, 2097152.0, 2097152.0, 0, 2097152.0])
            self.assertEqual(times,                   [100,       200, 300, 400,  500,       600,  700,       800,       900,  1000])

    @mock.patch("lvestats.lib.commons.func.get_governor_config_path", return_value="/var/run/mysql-governor-config.xml")
    @mock.patch('xml.dom.minidom.parse', side_effect=test.test_func.mock_dbgov_config)
    def test_get_chart__ignore_unknown_user3(self, mock_parse_config_file, governor_config_path):
        self.chart.governor_mode = "abusers"
        with mock.patch('lvestats.lib.chart.lvechartmain.uid_to_username', return_value="user3"):
            data_collected, times, _ = self.chart.get_chart_data(
                engine=self.engine,
                from_ts=self.date_from,
                to_ts=self.date_to,
                server="localhost",
                user_id=1)
            self.assertEqual(data_collected['acpu'],  [2.5,       0.0, 5.0, 10.0, 7.5,       15.0, 1.0,       8.0,       50.0, 10.0])
            self.assertEqual(data_collected['dbcpu'], [3.5, 0, 0, 0, 5.0, 0, 6.0, 7.5, 0, 9.0])
            self.assertEqual(data_collected['dbio'],  [2097152.0, 0, 0, 0, 2097152.0, 0, 2097152.0, 2097152.0, 0, 2097152.0])
            self.assertEqual(times,                   [100,       200, 300, 400,  500,       600,  700,       800,       900,  1000])

    def test_get_chart_data_with_dbgov_data_bigger_than_lve_stats1_govmode_no_config(self):
        self.chart.governor_mode = "none"
        data_collected, times, _ = self.chart.get_chart_data(
            engine=self.engine,
            from_ts=self.date_from,
            to_ts=self.date_to,
            server="localhost",
            user_id=1)
        self.assertEqual(data_collected['acpu'], [2.5, 0.0, 5.0, 10.0, 7.5, 15.0, 1.0, 8.0, 50.0, 10.0])
        self.assertEqual(data_collected['dbcpu'], [])
        self.assertEqual(data_collected['dbio'], [])
        self.assertEqual(times, [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000])

    def test_get_chart_data_with_dbgov_data_less_than_lve_stats2(self):
        self.chart.governor_mode = "all"
        data_collected, times, _ = self.chart.get_chart_data(
            engine=self.engine,
            from_ts=self.date_from,
            to_ts=self.date_to,
            server="localhost",
            user_id=2)
        self.assertEqual(data_collected['dbcpu'], [3.5, 0, 0, 0, 5.0, 0, 6.0, 7.5, 0, 0])
        self.assertEqual(data_collected['dbio'], [1572864.0, 0, 0, 0, 0.0, 0, 9437184.0, 524288.0, 0, 0])
        self.assertEqual(times, [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000])

    def test_get_chart_data_with_dbgov_data_bigger_than_lve_stats2_govmode_abusers(self):
        self.chart.governor_mode = "abusers"
        data_collected, times, _ = self.chart.get_chart_data(
            engine=self.engine,
            from_ts=self.date_from,
            to_ts=self.date_to,
            server="localhost",
            user_id=3)
        self.assertEqual(data_collected['dbcpu'], [3.5,       0,   0,   0,   5.0, 0,   6.0,       7.5,      0,   0])
        self.assertEqual(data_collected['dbio'],  [1572864.0, 0,   0,   0,   0.0, 0,   9437184.0, 524288.0, 0,   0])
        self.assertEqual(times,                   [100,       200, 300, 400, 500, 600, 700,       800,      900, 1000])

    def test_get_chart_data_with_dbgov_data_bigger_than_lve_stats1(self):
        """
        In case where dbgov point is bigger than lvestatshistory we get dbcpu and dbio from lvestatshistory
        :return:
        """
        self.chart.governor_mode = "all"
        data_collected, times, _ = self.chart.get_chart_data(
            engine=self.engine,
            from_ts=self.date_from,
            to_ts=self.date_to,
            server="localhost",
            user_id=1)
        self.assertEqual(times,                   [100,      200, 300, 400, 500,       600, 700,       800,       900, 1000])
        self.assertEqual(data_collected['dbcpu'], [2.5,      0,   0,   0,   5.0,       0,   1.0,       7.5,       0,   9.0])
        self.assertEqual(data_collected['dbio'],  [524288.0, 0,   0,   0,   2097152.0, 0,   2097152.0, 2097152.0, 0,   2097152.0])

    def test_get_chart_data_with_dbgov_data_bigger_than_lve_stats2(self):
        """
        In case where dbgov point is bigger than lvestatshistory we get dbcpu and dbio from lvestatshistory
        :return:
        """
        self.chart.governor_mode = "all"
        data_collected, times, _ = self.chart.get_chart_data(
            engine=self.engine,
            from_ts=self.date_from,
            to_ts=self.date_to,
            server="localhost",
            user_id=3)
        self.assertEqual(data_collected['dbcpu'], [2.5, 0, 0, 0, 5.0, 0, 1.0, 7.5, 0, 0])
        self.assertEqual(data_collected['dbio'], [524288.0, 0, 0, 0, 0.0, 0, 2621440.0, 524288.0, 0, 0])
        self.assertEqual(times, [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000])


input_avg = list()
output_avg = list()

input_avg.append([                                                                                                  ])
output_avg.append([[60, 0], [120, 0], [180, 0], [240, 0], [300, 0], [360, 0], [420, 0], [480, 0], [540, 0], [600, 0]])

input_avg.append( [                   [180, 1], [240, 1]                                                            ])
output_avg.append([[60, 0], [120, 0], [180, 1], [240, 1], [300, 0], [360, 0], [420, 0], [480, 0], [540, 0], [600, 0]])

input_avg.append( [[60, 1], [120, 1], [180, 1], [240, 1]                                                            ])
output_avg.append([[60, 1], [120, 1], [180, 1], [240, 1], [300, 0], [360, 0], [420, 0], [480, 0], [540, 0], [600, 0]])

input_avg.append( [                                                                     [480, 1], [540, 1], [600, 1]])
output_avg.append([[60, 0], [120, 0], [180, 0], [240, 0], [300, 0], [360, 0], [420, 0], [480, 1], [540, 1], [600, 1]])

input_avg.append( [                   [180, 1], [240, 1],                               [480, 1], [540, 1]          ])
output_avg.append([[60, 0], [120, 0], [180, 1], [240, 1], [300, 0], [360, 0], [420, 0], [480, 1], [540, 1], [600, 0]])

input_avg.append( [                   [180, 1], [240, 1],                               [480, 1], [540, 1], [600, 1]])
output_avg.append([[60, 0], [120, 0], [180, 1], [240, 1], [300, 0], [360, 0], [420, 0], [480, 1], [540, 1], [600, 1]])

input_avg.append( [[60, 1], [120, 1], [180, 1], [240, 1],                               [480, 1], [540, 1], [600, 1]])
output_avg.append([[60, 1], [120, 1], [180, 1], [240, 1], [300, 0], [360, 0], [420, 0], [480, 1], [540, 1], [600, 1]])

input_avg.append( [[60, 1], [120, 1], [180, 1], [240, 1], [300, 1], [360, 1], [420, 1], [480, 1], [540, 1], [600, 1]])
output_avg.append([[60, 1], [120, 1], [180, 1], [240, 1], [300, 1], [360, 1], [420, 1], [480, 1], [540, 1], [600, 1]])

input_avg.append( [[60, 0], [120, 0], [180, 0], [240, 0], [300, 0], [360, 0], [420, 0], [480, 0], [540, 0], [600, 0]])
output_avg.append([[60, 0], [120, 0], [180, 0], [240, 0], [300, 0], [360, 0], [420, 0], [480, 0], [540, 0], [600, 0]])

input_limit = list()
output_limit = list()

input_limit.append( [                                                                                                ])
output_limit.append([[60, 0], [120, 0], [180, 0], [240, 0], [300, 0], [360, 0], [420, 0], [480, 0], [540, 0], [600, 0]])

input_limit.append( [                   [180, 2], [240, 1]                                                            ])
output_limit.append([[60, 2], [120, 2], [180, 2], [240, 1], [300, 1], [360, 1], [420, 1], [480, 1], [540, 1], [600, 1]])

input_limit.append( [                             [240, 3]                                                            ])
output_limit.append([[60, 3], [120, 3], [180, 3], [240, 3], [300, 3], [360, 3], [420, 3], [480, 3], [540, 3], [600, 3]])

input_limit.append( [                             [240, 3]                                                            ])
output_limit.append([[60, 3], [120, 3], [180, 3], [240, 3], [300, 3], [360, 3], [420, 3], [480, 3], [540, 3], [600, 3]])

input_limit.append( [         [120, 3],           [240, 4],                     [420, 5],                     [600, 6]])
output_limit.append([[60, 3], [120, 3], [180, 3], [240, 4], [300, 4], [360, 4], [420, 5], [480, 5], [540, 5], [600, 6]])


@ddt.ddt
class TestGetChartData2(unittest.TestCase):
    def setUp(self):
        self.time_unit = 60
        self.period_from, self.period_to = (60, 660)

    @ddt.data(*list(zip(output_avg, input_avg)))
    @ddt.unpack
    def test_make_lvedata_with_idle_avg(self, lvedata_with_idle_, lvedata):
        show_columns = ('From', 'aCPU')
        lvedata_with_idle = LveChart.make_lvedata_with_idle(
            lvedata=lvedata, show_columns=show_columns,
            period_from=self.period_from, period_to=self.period_to,
            time_step=self.time_unit)
        self.assertListEqual(lvedata_with_idle, lvedata_with_idle_)

    @ddt.data(*list(zip(output_limit, input_limit)))
    @ddt.unpack
    def test_make_lvedata_with_idle_limit(self, lvedata_with_idle_, lvedata):
        show_columns = ('From', 'lCPU')
        lvedata_with_idle = LveChart.make_lvedata_with_idle(
            lvedata=lvedata, show_columns=show_columns,
            period_from=self.period_from, period_to=self.period_to,
            time_step=self.time_unit)
        self.assertListEqual(lvedata_with_idle, lvedata_with_idle_)


class TestFaultsOnCharts(BaseDbTest):
    def setUp(self):
        BaseDbTest.setUp(self)
        os.environ['TZ'] = 'UTC'  # set the zero time zone
        time.tzset()
        self.save_history_line("localhost", uid=1, _time=0, cpu=250, cpu_limit=5000, cpu_fault=1, keepalive=True)
        self.save_history_line("localhost", uid=1, _time=60, cpu=150, cpu_limit=5000, keepalive=True)
        self.save_history_line("localhost", uid=1, _time=120, cpu=350, cpu_limit=3000, cpu_fault=10, keepalive=True)
        self.save_history_line("localhost", uid=1, _time=180, cpu=350, cpu_limit=1000, cpu_fault=2, keepalive=True)
        self.save_history_line("localhost", uid=1, _time=240, cpu=350, cpu_limit=1000, keepalive=True)

        self.save_history_line("localhost", uid=2, _time=0, cpu=250, cpu_limit=5000, cpu_fault=1)
        self.save_history_line("localhost", uid=2, _time=60, cpu=150, cpu_limit=5000)
        self.save_history_line("localhost", uid=2, _time=120, cpu=2350, cpu_limit=3000, cpu_fault=10)
        self.save_history_line("localhost", uid=2, _time=180, cpu=350, cpu_limit=1000, cpu_fault=2)
        self.save_history_line("localhost", uid=2, _time=240, cpu=350, cpu_limit=1000)

        self.save_history_line("localhost", uid=3, _time=0, cpu=5000, cpu_limit=5000, cpu_fault=1)
        self.save_history_line("localhost", uid=3, _time=60, cpu=150, cpu_limit=5000)
        self.save_history_line("localhost", uid=3, _time=120, cpu=2350, cpu_limit=3000, cpu_fault=10)
        self.save_history_line("localhost", uid=3, _time=180, cpu=1000, cpu_limit=1000, cpu_fault=2)
        self.save_history_line("localhost", uid=3, _time=240, cpu=999, cpu_limit=1000)

        self.cfg["aggregation_period"] = 60
        self.chart = LveChart(self.cfg)
        self.date_from = dateutil.gm_to_local(dateutil.unixtimestamp_to_gm_datetime(0))
        self.date_to = dateutil.gm_to_local(dateutil.unixtimestamp_to_gm_datetime(250))
        self.svg = '/tmp/{0}.svg'.format(uuid4())
        self.renderer = Renderer(max_points_on_graph=4)

        patcher1 = mock.patch('lvestats.lib.config.read_config', return_value={})
        patcher1.start()
        self.addCleanup(patcher1.stop)

    def test_faults1(self):
        data_collected, times, _ = self.chart.get_chart_data(
            engine=self.engine,
            from_ts=self.date_from,
            to_ts=self.date_to,
            server="localhost",
            user_id=1)
        _, _, fault_lines, _ = self.renderer._add_graph(
            data_collected,
            ('CPUf', 'aCPU', 'lCPU'),
            {'aCPU': ('average', 'green'), 'lCPU': ('limit', 'red')},
            times)
        self.assertEqual(fault_lines, [((0.0, 2.5), (0.0, 50.0)), ((120.0, 3.0), (120.0, 30.0)), ((180.0, 3.25), (180.0, 10.0))])
        for fault_line in fault_lines:
            fault_coords = self.renderer.svg_chart.convert_coordinates(fault_line, 240.0, 50.0, 0.0, 1.5, 6.0)
            self.assertGreater(fault_coords[0][0], 0)
            self.assertGreater(fault_coords[0][1], 0)
            self.assertGreater(fault_coords[1][0], 0)
            self.assertGreater(fault_coords[1][1], 0)

    def test_faults2(self):
        data_collected, times, _ = self.chart.get_chart_data(
            engine=self.engine,
            from_ts=self.date_from,
            to_ts=self.date_to,
            server="localhost",
            user_id=2)
        _, _, fault_lines, _ = self.renderer._add_graph(
            data_collected,
            ('CPUf', 'aCPU', 'lCPU'),
            {'aCPU': ('average', 'green'), 'lCPU': ('limit', 'red')},
            times)
        self.assertEqual(fault_lines, [((0.0, 2.5), (0.0, 50.0)), ((120.0, 23.5), (120.0, 30.0))])
        for fault_line in fault_lines:
            fault_coords = self.renderer.svg_chart.convert_coordinates(fault_line, 240.0, 50.0, 0.0, 1.5, 6.0)
            self.assertGreater(fault_coords[0][0], 0)
            self.assertGreater(fault_coords[0][1], 0)
            self.assertGreater(fault_coords[1][0], 0)
            self.assertGreater(fault_coords[1][1], 0)

    def test_faults3(self):
        data_collected, times, _ = self.chart.get_chart_data(
            engine=self.engine,
            from_ts=self.date_from,
            to_ts=self.date_to,
            server="localhost",
            user_id=3)
        _, _, fault_lines, _ = self.renderer._add_graph(
            data_collected,
            ('CPUf', 'aCPU', 'lCPU'),
            {'aCPU': ('average', 'green'), 'lCPU': ('limit', 'red')},
            times)
        self.assertEqual(fault_lines, [((120.0, 23.5), (120.0, 30.0))])
        for fault_line in fault_lines:
            fault_coords = self.renderer.svg_chart.convert_coordinates(fault_line, 240.0, 50.0, 0.0, 1.5, 6.0)
            self.assertGreater(fault_coords[0][0], 0)
            self.assertGreater(fault_coords[0][1], 0)
            self.assertGreater(fault_coords[1][0], 0)
            self.assertGreater(fault_coords[1][1], 0)

if __name__ == '__main__':
    unittest.main()
