# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import logging
import multiprocessing
import time
import unittest

import mock
from sqlalchemy.exc import NoSuchTableError

from test.base.base_db_test import BaseDbTest
from lvestats.eventloop.main_loop import MainLoop
import lvestats.eventloop.plugin_executors
from lvestats.lib.commons import func

__author__ = 'shaman'

manager = multiprocessing.Manager()
ns = manager.Namespace()
get_lve_version_backup = None
old_KILL_PROCESS_TIMEOUT = lvestats.eventloop.plugin_executors.KILL_PROCESS_TIMEOUT


class HangingPlugin(object):
    __is_user_plugin__ = True

    def execute(self, lve_data):
        ns.broken = False
        for _ in range(10000):
            time.sleep(1)
        ns.broken = True


class ExceptionPlugin(object):
    def execute(self, lve_data):
        lve_data['broken'] = True
        raise Exception('I am broken, sorry')


class NormalPlugin(object):
    def execute(self, lve_data):
        # print 'Side effect'
        ns.ok = 'broken' not in lve_data


class NowCheckingPlugin(object):
    def execute(self, lve_data):
        if not hasattr(self, 'now'):
            raise Exception('No such attribute "now"!')

        if not self.now:
            raise Exception('Now is not set!')

        ns.now_is_set = True


class DbErrorPlugin(object):
    def execute(self, lve_data):
        raise NoSuchTableError("SELECT 1;", [], None, False)


class PeriodPlugin(object):
    period = 3

    def execute(self, lve_data):
        lve_data['count'] = lve_data.get('count', 0) + 1
        ns.count = lve_data['count']


class CountPlugin(object):
    def execute(self, lve_data):
        lve_data['count'] = lve_data.get('count', 0) + 1
        ns.count = lve_data['count']


class PeriodInstancePlugin(object):
    def __init__(self):
        super(PeriodInstancePlugin, self).__init__()
        self.period = 3

    def execute(self, lve_data):
        lve_data['count'] = lve_data.get('count', 0) + 1
        ns.count = lve_data['count']


class NotInitializingPlugin(object):
    def __init__(self):
        super(NotInitializingPlugin, self).__init__()
        raise Exception('Bla-bla')

    def execute(self, lve_data):
        pass


def base_exception():
    raise BaseException()


class TestMainLoop(BaseDbTest):
    @classmethod
    def setUpClass(cls):
        BaseDbTest.setUpClass()
        lvestats.eventloop.plugin_executors.KILL_PROCESS_TIMEOUT = 2
        global get_lve_version_backup
        get_lve_version_backup = func.get_lve_version

        def mock_lve_version(f=None):
            return 8

        func.get_lve_version = mock_lve_version

    @classmethod
    def tearDownClass(cls):
        BaseDbTest.tearDownClass()
        func.get_lve_version = get_lve_version_backup
        lvestats.eventloop.plugin_executors.KILL_PROCESS_TIMEOUT = old_KILL_PROCESS_TIMEOUT

    def setUp(self):
        BaseDbTest.setUp(self)
        logging.basicConfig(level=logging.DEBUG)

    def test_not_able_to_init_plugin_should_not_ruin_the_system(self):
        main_loop = MainLoop(
            self.cfg,
            interval=0.2,
            plugins=[NotInitializingPlugin, NormalPlugin],
            plugin_timeout=0.1)
        main_loop.run(times=1)
        self.assertTrue(ns.ok)

    def test_exit_invocation_should_stop_the_process_and_loop(self):
        main_loop = MainLoop(
            self.cfg,
            interval=0.2,
            plugins=[NormalPlugin],
            plugin_timeout=0.1)
        main_loop.exit()
        main_loop.run()

    def test_hanging_plugin_should_be_killed_and_not_break_the_data(self):
        main_loop = MainLoop(
            self.cfg,
            interval=0.2,
            plugins=[HangingPlugin, NormalPlugin],
            plugin_timeout=0.1)
        main_loop.run(times=1)
        self.assertTrue(ns.ok)
        self.assertFalse(ns.broken)

    def test_main_loop_should_get_stacktrace_for_hanging_plugin(self):
        main_loop = MainLoop(
            self.cfg,
            interval=0.2,
            plugins=[HangingPlugin],
            plugin_timeout=0.1)
        main_loop.run(times=1)

    def test_exception_in_plugin_should_not_prevent_other_plugins_to_execute_and_not_break_the_data(self):
        ns.ok = False

        main_loop = MainLoop(
            self.cfg,
            interval=0.2,
            plugins=[ExceptionPlugin, NormalPlugin],
            plugin_timeout=0.1)
        main_loop.run(times=1)
        self.assertTrue(ns.ok)

    def test_period_between_executions(self):
        for plugin_class in (PeriodPlugin, PeriodInstancePlugin):
            main_loop = MainLoop(
                self.cfg,
                interval=0.1,
                plugins=[plugin_class],
                plugin_timeout=0.1)
            main_loop.run(times=5)
            cnt = ns.count
            # during 5 seconds plugin with period = 3 should be called only 1 or 2 times
            self.assertTrue(cnt in [1, 2],
                            msg='%s.execute should be called [1, 2] times but is called %d times '
                                'instead' % (str(plugin_class), cnt))

    def test_now_is_set(self):
        ns.now_is_set = False

        main_loop = MainLoop(
            self.cfg,
            interval=0.1,
            plugins=[NowCheckingPlugin],
            plugin_timeout=0.1)
        main_loop.run(times=1)
        self.assertTrue(ns.now_is_set)

    def test_recover_db(self):
        ns.ok = False
        main_loop = MainLoop(
            self.cfg,
            interval=0.1,
            plugins=[DbErrorPlugin, NormalPlugin],
            plugin_timeout=0.2)
        main_loop.run(times=1)
        self.assertTrue(ns.ok)

    # mock for imitation fail recover database
    @mock.patch.object(lvestats.eventloop.plugin_executors.PluginContext, 'recover_db', base_exception)
    def test_exit_if_recover_db_fails(self):
        ns.ok = False
        main_loop = MainLoop(
            self.cfg,
            interval=0.1,
            plugins=[CountPlugin, DbErrorPlugin, NormalPlugin],
            plugin_timeout=0.2)
        main_loop.run(times=3)
        self.assertFalse(ns.ok)
        self.assertEqual(ns.count, 1)


if __name__ == '__main__':
    unittest.main()
