#!/usr/bin/python
# coding=utf-8
import os
import pickle
import unittest
from unittest import TestCase

import time
from ddt import ddt, data, unpack
from mock import patch

from lvestats.lib.notifications_helper import NotificationsHelper

__author__ = "Aleksandr Shyshatsky"


@ddt
class TestNotificationsHelper(TestCase):
    def setUp(self):
        self._notifications = NotificationsHelper()
        self._notifications.STATSNOTIFIER_LAST_TS = \
            os.path.join(os.path.dirname(__file__), 'data', 'statsnotifier_last.ts')
        self._notifications.RESELLERS_NOTIFICATIONS_STORAGE = \
            os.path.join(os.path.dirname(__file__), 'data', 'statsnotifier_timestamps.bin')

    # def test_load_from_persistent_storage(self):
    #     self.fail()
    #
    @data([{'res1': 1234, 'res2': 345}, {'res1': 4321, 'res4': 123}],
          [{'res6': -1, 'res2': 345},   {'res9': 123, 'res6': 123}])
    @unpack
    def test_persistent_storage(self, users_notification_info, resellers_notification_info):
        self._notifications._users_notification_info = users_notification_info
        self._notifications._resellers_notification_info = resellers_notification_info

        self._notifications.save_to_persistent_storage()

        # check if it is saved as should be
        with open(self._notifications.RESELLERS_NOTIFICATIONS_STORAGE, 'rb') as f:
            self.assertEqual((users_notification_info, resellers_notification_info), pickle.load(f))

        # clean data
        self._notifications._users_notification_info = None
        self._notifications._resellers_notification_info = None

        # load data from file
        self._notifications._load_from_persistent_storage()

        # check if it is loaded as should be
        self.assertEqual(self._notifications._users_notification_info, users_notification_info)
        self.assertEqual(self._notifications._resellers_notification_info, resellers_notification_info)

    @data([1234, 1234],
          ["qwert", -1])
    @unpack
    def test_save_ts_to_file(self, save, result):
        self._notifications._save_ts_to_file(save)
        self.assertEqual(self._notifications._read_ts_from_file(), result)

    def test_mark_resellers_notified(self):
        self._notifications.mark_resellers_notified(['res1', 'res3'])
        self.assertIn('res1', self._notifications._resellers_notification_info)
        self.assertIn('res3', self._notifications._resellers_notification_info)

    def test_mark_users_notified(self):
        self._notifications.mark_users_notified(['res3', 'res4'])
        self.assertIn('res3', self._notifications._users_notification_info)
        self.assertIn('res4', self._notifications._users_notification_info)

    def test_mark_admin_notified(self):
        self._notifications.mark_admin_notified()
        self.assertAlmostEqual(self._notifications._admin_notify_time, time.time(), 1)

    def test_users_need_notification(self):
        self._notifications._users_notification_info['reseller1'] = 120

        # users were just notified
        with patch('time.time', return_value=120):
            self.assertFalse(self._notifications.users_need_notification('reseller1', 120))

        # only 119 seconds elapsed
        with patch('time.time', return_value=240):
            self.assertFalse(self._notifications.users_need_notification('reseller1', 120))

        with patch('time.time', return_value=241):
            self.assertTrue(self._notifications.users_need_notification('reseller1', 120))

        # new users always need notification
        self.assertTrue(self._notifications.users_need_notification('reseller2', 120))

    def test_reseller_need_notification(self):
        self._notifications._resellers_notification_info['reseller1'] = 120

        # resellers was just notified
        with patch('time.time', return_value=120):
            self.assertFalse(self._notifications.reseller_need_notification('reseller1', 120))

        # only 119 seconds elapsed
        with patch('time.time', return_value=240):
            self.assertFalse(self._notifications.reseller_need_notification('reseller1', 120))

        with patch('time.time', return_value=241):
            self.assertTrue(self._notifications.reseller_need_notification('reseller1', 120))

        # new resellers always need notification
        self.assertTrue(self._notifications.reseller_need_notification('reseller2', 120))

    def test_admin_need_notification(self):
        self._notifications._admin_notify_time = 120

        # admin was just notified
        with patch('time.time', return_value=120):
            self.assertFalse(self._notifications.admin_need_notification(120))

        # only 119 seconds elapsed
        with patch('time.time', return_value=240):
            self.assertFalse(self._notifications.admin_need_notification(120))

        with patch('time.time', return_value=241):
            self.assertTrue(self._notifications.admin_need_notification(120))

if __name__ == '__main__':
    unittest.main(verbosity=2)
