# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
from test.base.base_db_test import BaseDbTest

import ddt
import mock

import lvestats.lib.ustate
from lvestats.plugins.generic import aggregators
from lvestats.plugins.generic.snapshot_saver import SnapshotSaver
from test.test_snapshot_saver import LVEUsage

test_case_dict_sum_values = (
    ([{"a": 1, "b": 6, "c": 4}, {}], {"a": 1, "b": 6, "c": 4}),
    ([{}, {}], {}),
    ([{}, {"a": 1, "b": 6, "c": 4}], {"a": 1, "b": 6, "c": 4}),
    ([{"a": 1, "c": 3}, {"a": 1, "b": 2}], {"a": 2, "b": 2, "c": 3}))


class MockAggregatedLveUsage(aggregators.AggregatedLveUsage):
    def __init__(self, **kwargs):
        super(MockAggregatedLveUsage, self).__init__(8)
        for (k, v) in kwargs.items():
            setattr(self, k, v)


@ddt.ddt
class TestLveUsageAggregator5S(BaseDbTest):
    def setUp(self):
        self.lve_data = {"lve_usages": [], "lve_usages5s": []}
        self.plugin5s = aggregators.LveUsageAggregator5S()
        self.plugin5s.now = 100
        self.plugin5s.set_config({})
        self.plugin = aggregators.LveUsageAggregator()
        self.plugin.now = 100

    @ddt.data(*test_case_dict_sum_values)
    @ddt.unpack
    def test_dict_sum_values(self, test_data, output):
        self.assertEqual(self.plugin5s.dict_sum_values(*test_data), output)

    def test_aggregator_dont_add_faults(self):
        self.lve_data = {
            "lve_usages": [
                {1000: LVEUsage(cpu_limit=2500, cpu_usage=2600, cpu_fault=1)},
                {1000: LVEUsage(cpu_limit=2500, cpu_usage=2600, cpu_fault=4)}
            ],
            "lve_usages_5s": [{1000: LVEUsage(cpu_limit=2500, cpu_usage=2600, cpu_fault=1)}]
        }
        self.plugin.execute(self.lve_data)
        self.assertNotIn("faults", self.lve_data)

    def test_aggregator5s_add_faults(self):
        self.lve_data["lve_usages_5s"] = [{1000: LVEUsage(cpu_limit=2500, cpu_usage=2600, cpu_fault=1)}]
        self.plugin5s.execute(self.lve_data)
        self.assertIn("faults", self.lve_data)
        self.assertEqual(self.lve_data["faults"], {1000: {"cpu_fault": 1}})

        self.lve_data["lve_usages_5s"] = [{1000: LVEUsage(cpu_limit=2500, cpu_usage=2600, cpu_fault=4)}]
        self.plugin5s.execute(self.lve_data)
        self.assertEqual(self.lve_data["faults"], {1000: {"cpu_fault": 5}})

        self.lve_data["lve_usages_5s"] = [{1000: LVEUsage(cpu_limit=25, cpu_usage=2600, cpu_fault=5, mem_fault=7)}]
        self.plugin5s.execute(self.lve_data)
        self.assertEqual(self.lve_data["faults"], {1000: {"cpu_fault": 10, "mem_fault": 7}})

        # LveUsageAggregator does not change "faults" data
        self.plugin.execute(self.lve_data)
        self.assertEqual(self.lve_data["faults"], {1000: {"cpu_fault": 10, "mem_fault": 7}})

        self.lve_data["lve_usages_5s"] = [{1001: LVEUsage(cpu_limit=25, cpu_usage=2600, cpu_fault=2, memphy_fault=4)}]
        self.plugin5s.execute(self.lve_data)
        self.assertEqual(self.lve_data["faults"], {
            1000: {"cpu_fault": 10, "mem_fault": 7},
            1001: {"cpu_fault": 2, "memphy_fault": 4}})

    def test_aggregator5s_snapshots_disabled(self):
        self.plugin5s.set_config({"disable_snapshots": "true"})
        self.lve_data["lve_usages_5s"] = [{1000: LVEUsage(cpu_limit=2500, cpu_usage=2600, cpu_fault=1)}]
        self.plugin5s.execute(self.lve_data)
        self.assertNotIn("faults", self.lve_data)

    @mock.patch.object(lvestats.plugins.generic.snapshot_saver, "get_lveps", return_value={})
    def test_faults_cleaned(self, get_lveps):
        snapshot_saver = SnapshotSaver()
        snapshot_saver.now = 100
        snapshot_saver.engine = self.engine
        self.lve_data["lve_usages_5s"] = [{1000: LVEUsage(cpu_limit=2500, cpu_usage=2600, cpu_fault=1)}]
        self.plugin5s.execute(self.lve_data)
        self.assertEqual(self.lve_data["faults"], {1000: {"cpu_fault": 1}})
        with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
            snapshot_saver.execute(self.lve_data)
        self.assertEqual(self.lve_data["faults"], {})

    @ddt.data(
        ({}, {100: MockAggregatedLveUsage()}, {}),
        ({'faults': {1000: {"cpu_fault": 1}}}, {100: MockAggregatedLveUsage()}, {1000: {"cpu_fault": 1}}),
        (
            {'faults': {1000: {"cpu_fault": 1}}},
            {100: MockAggregatedLveUsage(nproc_fault=1)},
            {1000: {"cpu_fault": 1}, 100: {"nproc_fault": 1}}),
        (
            {'faults': {100: {"cpu_fault": 1}}},
            {100: MockAggregatedLveUsage(cpu_fault=1)},
            {100: {"cpu_fault": 2}}),
        (
            {'faults': {100: {"cpu_fault": 1}}},
            {100: MockAggregatedLveUsage(cpu_fault=1), 101: MockAggregatedLveUsage(io_fault=1)},
            {100: {"cpu_fault": 2}, 101: {"io_fault": 1}}),
        (
            {'faults': {101: {"cpu_fault": 1}}},
            {100: MockAggregatedLveUsage(cpu_fault=1), 101: MockAggregatedLveUsage(io_fault=1)},
            {100: {"cpu_fault": 1}, 101: {"cpu_fault": 1, "io_fault": 1}}))
    @ddt.unpack
    def test_aggregate_faults(self, lve_data, lve_usage, result):
        self.plugin5s.aggregate_faults(lve_data, lve_usage)
        self.assertEqual(lve_data['faults'], result)
