# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import itertools
import typing
from random import Random
from types import SimpleNamespace
from collections.abc import Sequence
from datetime import timedelta
from unittest import TestCase
from unittest.mock import MagicMock

import sqlalchemy as sa

from lvestats.orm import BurstingEventType
from lvestats.plugins.generic.burster.storage.base import (
    InBurstingEventRow,
    init_db_schema,
    drop_db_schema,
)
from lvestats.plugins.generic.burster.storage.save import save_events_to_db
from lvestats.plugins.generic.burster.storage.load import load_bursting_enabled_intervals_from_db
from lvestats.plugins.generic.burster.storage.cleanup import cleanup_old_events
from lvestats.plugins.generic.burster.common import (
    BurstingMultipliers,
    Timestamp,
    LveLimits,
    LveStats,
    InvalidStateError,
    LveId,
    LveUsage,
    LveState,
    AdjustStepData,
)
from lvestats.plugins.generic.burster.history import (
    IntervalType,
    LveHistory,
)
from lvestats.plugins.generic.burster.lve_sm import (
    LveStateManager,
)
from lvestats.plugins.generic.burster.lves_tracker import (
    LvesTracker,
)
from lvestats.plugins.generic.burster.adjust import (
    Adjuster,
    StepCalculator,
)
from lvestats.plugins.generic.burster.overload import (
    OverloadChecker,
    OverloadCheckResult,
)
from lvestats.plugins.generic.burster.config import (
    ConfigUpdate,
    PluginConfig,
    MissingKeysInRawConfig,
)


class TestLveHistory(TestCase):
    def test_get_total_duration(self):
        with self.subTest('empty history'):
            history = LveHistory()
            self.assertAlmostEquals(
                history.get_overusing_duration(6.0),
                0.0,
            )

        with self.subTest('first interval is overusing'):
            history = LveHistory(IntervalType.OVERUSING, (1.0, 2.0, 3.0, 4.0)).get_overusing_duration(6.0)
            self.assertAlmostEquals(
                history,
                2.0,
            )
        with self.subTest('first interval is normal'):
            history = LveHistory(IntervalType.NORMAL, (1.0, 2.0, 3.0, 4.0)).get_overusing_duration(6.0)
            self.assertAlmostEquals(
                history,
                3.0,
            )

    def test_trim(self):
        with self.subTest('trim empty history'):
            now = 6.0
            history = LveHistory().trim(now)
            self.assertSequenceEqual(
                (history.first_interval_type, history.get_overusing_intervals(now), history.ongoing_interval_type),
                (None, tuple(), None),
            )
            self.assertAlmostEqual(history.get_overusing_duration(now), 0.0)

        with self.subTest('trim before first timestamp'):
            now = 6.0
            history = LveHistory(IntervalType.OVERUSING, (1.0, 2.0, 3.0, 4.0)).trim(0.5)
            self.assertSequenceEqual(
                (history.first_interval_type, history.get_overusing_intervals(now), history.ongoing_interval_type),
                (IntervalType.OVERUSING, ((1.0, 2.0), (3.0, 4.0),), IntervalType.NORMAL),
            )
            self.assertAlmostEqual(history.get_overusing_duration(now), 2.0)

        with self.subTest('trim in the middle of normal interval'):
            now = 6.0
            history = LveHistory(IntervalType.OVERUSING, (1.0, 2.0, 3.0, 4.0)).trim(2.5)
            self.assertSequenceEqual(
                (history.first_interval_type, history.get_overusing_intervals(now), history.ongoing_interval_type),
                (IntervalType.NORMAL, ((3.0, 4.0),), IntervalType.NORMAL),
            )
            self.assertAlmostEqual(history.get_overusing_duration(now), 1.0)

        with self.subTest('cleanup in the middle of overusing interval'):
            now = 6.0
            history = LveHistory(IntervalType.NORMAL, (1.0, 2.0, 3.0, 4.0)).trim(2.5)
            self.assertSequenceEqual(
                (history.first_interval_type, history.get_overusing_intervals(now), history.ongoing_interval_type),
                (IntervalType.OVERUSING, ((2.5, 3.0), (4.0, now)), IntervalType.OVERUSING),
            )
            self.assertAlmostEqual(history.get_overusing_duration(now), 2.5)

        with self.subTest(f'trim after last timestamp inside normal interval'):
            now = 6.0
            history = LveHistory(IntervalType.NORMAL, (1.0, 2.0, 3.0)).trim(5.0)
            self.assertSequenceEqual(
                (history.first_interval_type, history.get_overusing_intervals(now), history.ongoing_interval_type),
                # NOTE(vlebedev): Number of intervals is odd so last interval has the same type as the first one.
                (IntervalType.NORMAL, tuple(), IntervalType.NORMAL),
            )
            self.assertAlmostEqual(history.get_overusing_duration(now), 0.0)

        with self.subTest('trim after last timestamp inside overusing interval'):
            now = 6.0
            history = LveHistory(IntervalType.OVERUSING, (1.0, 2.0, 3.0)).trim(5.0)
            self.assertSequenceEqual(
                (history.first_interval_type, history.get_overusing_intervals(now), history.ongoing_interval_type),
                # NOTE(vlebedev): Number of intervals is odd so last interval has the same type as the first one.
                (IntervalType.OVERUSING, ((5.0, 6.0),), IntervalType.OVERUSING),
            )
            self.assertAlmostEqual(history.get_overusing_duration(now), 1.0)

        with self.subTest('trim right at a timestamp in the middle when timestamps number is even'):
            now = 6.0
            history = LveHistory(IntervalType.OVERUSING, (1.0, 2.0, 3.0, 4.0)).trim(2.0)
            self.assertSequenceEqual(
                (history.first_interval_type, history.get_overusing_intervals(now), history.ongoing_interval_type),
                (IntervalType.NORMAL, ((3.0, 4.0),), IntervalType.NORMAL),
            )
            self.assertEqual(history.timestamps, (2.0, 3.0, 4.0))
            self.assertAlmostEqual(history.get_overusing_duration(now), 1.0)

        with self.subTest('trim right at the first timestamp'):
            now = 6.0
            history = LveHistory(IntervalType.OVERUSING, (1.0, 2.0, 3.0)).trim(3.0)
            self.assertSequenceEqual(
                (history.first_interval_type, history.get_overusing_intervals(now), history.ongoing_interval_type),
                (IntervalType.OVERUSING, ((3.0, 6.0),), IntervalType.OVERUSING),
            )
            self.assertEqual(history.timestamps, (3.0,))
            self.assertAlmostEqual(history.get_overusing_duration(now), 3.0)

        with self.subTest('trim right at a timestamp in the middle when timestamps number is odd'):
            now = 6.0
            history = LveHistory(IntervalType.OVERUSING, (2.0, 3.0, 4.0)).trim(3.0)
            self.assertEqual(
                (history.first_interval_type, history.get_overusing_intervals(now), history.ongoing_interval_type),
                (IntervalType.NORMAL, ((4.0, 6.0),), IntervalType.OVERUSING),
            )
            self.assertAlmostEqual(history.get_overusing_duration(now), 2.0)

    def test_append(self):
        with self.subTest('append to empty history'):
            now = 6.0
            history = LveHistory().append(1.0, BurstingEventType.STARTED)
            self.assertSequenceEqual(
                (history.first_interval_type, history.get_overusing_intervals(now), history.ongoing_interval_type),
                (IntervalType.OVERUSING, ((1.0, 6.0),), IntervalType.OVERUSING),
            )
            self.assertAlmostEqual(history.get_overusing_duration(now), 5.0)

        with self.subTest('append multiple timestamps - alternations are properly handled'):
            now = 6.0
            history = (
                LveHistory()
                .append(1.0, BurstingEventType.STARTED)
                .append(2.0, BurstingEventType.STOPPED)
                .append(3.0, BurstingEventType.STARTED)
                .append(4.0, BurstingEventType.STOPPED)
            )
            self.assertSequenceEqual(
                (history.first_interval_type, history.get_overusing_intervals(now), history.ongoing_interval_type),
                (IntervalType.OVERUSING, ((1.0, 2.0), (3.0, 4.0)), IntervalType.NORMAL),
            )
            self.assertAlmostEqual(history.get_overusing_duration(now), 2.0)


class TestLveStateManager(TestCase):
    def setUp(self) -> None:
        self.clock = itertools.count(1.0)

    def test_expected_flow(self):
        # NOTE(vlebedev): Start in unbursted state.
        manager = _create_lve_manager(_make_readings_update_request(
            next(self.clock), 2.0, 1.0, 1.0, 1.0, 2.0, 1.0,
        ), lve_id=1)
        state_changes = []
        manager.on_state_changed.register(lambda *ss: state_changes.append(ss))

        manager.step(LveStateManager.Burst(next(self.clock)))

        # NOTE(vlebedev): We can neither issue `Burst`, nor `Unburst` commands while manager is waiting for
        #                 confirmation from new readings that it was applied.
        with self.assertRaises(InvalidStateError):
            manager.step(LveStateManager.Burst(next(self.clock)))

        with self.assertRaises(InvalidStateError):
            manager.step(LveStateManager.Unburst(next(self.clock)))

        # NOTE(vlebedev): Now confirm that limits were switched and LVE is now in bursted, but not overusing state.
        manager.step(_make_readings_update_request(next(self.clock), 2.0, 1.0, 1.5, 1.0, 3.0, 1.0))

        # NOTE(vlebedev): Now manager known that LVE is in bursted state so there is no point in issuing `Unburst`
        with self.assertRaises(InvalidStateError):
            manager.step(LveStateManager.Burst(next(self.clock)))

        # Now switch to overusing state.
        for cpu_usage in [2.1, 2.5, 3.0]:
            manager.step(_make_readings_update_request(next(self.clock), 2.0, 1.0, cpu_usage, 1.0, 3.0, 1.0))

            # NOTE(vlebedev): Now manager known that LVE is in bursted state so there is no point in issuing `Unburst`
            with self.assertRaises(InvalidStateError):
                manager.step(LveStateManager.Burst(next(self.clock)))

        self.assertEqual(state_changes, [
            (LveState.UNBURSTED, LveState.BURSTED),
            (LveState.BURSTED, LveState.OVERUSING),
        ])
        self.assertTrue(manager.history_contains_overusing)

    def test_disappeared_and_reappeared(self):
        # NOTE(vlebedev): Create manager for LVE already in bursted state.
        make_bursted_readings = lambda: _make_readings_update_request(next(self.clock), 2.0, 1.0, 1.0, 1.0, 3.0, 1.0)
        manager = _create_lve_manager(make_bursted_readings(), lve_id=1)
        state_changes = []
        manager.on_state_changed.register(lambda *ss: state_changes.append(ss))

        self.assertFalse(manager.history_contains_overusing)
        for _ in range(2):
            # NOTE(vlebedev): Feed unbursted readings.
            manager.step(make_bursted_readings())

            with self.assertRaises(InvalidStateError):
                manager.step(LveStateManager.Burst(next(self.clock)))

            # NOTE(vlebedev): Feed unbursted readings.
            manager.step(_make_readings_update_request(next(self.clock), 2.0, 1.0, 2.5, 1.0, 3.0, 1.0))

            manager.step(LveStateManager.Disappered(next(self.clock)))

            for cmd_cls in (LveStateManager.Burst, LveStateManager.Unburst):
                now = next(self.clock)
                with self.assertRaises(InvalidStateError):
                    manager.step(cmd_cls(now))

        self.assertEqual(state_changes, [
            (LveState.BURSTED, LveState.OVERUSING),
            (LveState.OVERUSING, LveState.EXISTED),
            (LveState.EXISTED, LveState.BURSTED),
            (LveState.BURSTED, LveState.OVERUSING),
            (LveState.OVERUSING, LveState.EXISTED),
        ])
        self.assertTrue(manager.history_contains_overusing)

    # def _assert_props_equal(self, obj: object, props: dict[str, bool]):
    #     self.assertDictEqual(props, {p: getattr(obj, p) for p in props.keys()})


class TestLvesTracker(TestCase):
    def setUp(self) -> None:
        self.clock = itertools.count(1.0)

    def test_expected_flow(self):
        now = next(self.clock)
        id_to_readings = {
            # unbursted
            1: _make_readings_update_request(now, 2, 1, 1, 1, 2, 1),
            # bursted
            2: _make_readings_update_request(now, 2, 1, 1, 1, 4, 1),
            # overusing
            3: _make_readings_update_request(now, 2, 1, 3, 1, 4, 1),
        }
        added_from_signal = set()
        tracker = LvesTracker(
            create_lve_manager=lambda *args, lve_id, **kwargs: _create_lve_manager(
                lve_id=lve_id,
                initial_readings=id_to_readings[lve_id],
            ),
            deserialize_lve_id=lambda lve_id: lve_id,
        )
        tracker.on_manager_added.register(added_from_signal.add)

        def tracker_update(now, currently_known_ids):
            tracker.update(
                now=now,
                normal_limits_by_id={lid: r.normal_limits for lid, r in id_to_readings.items()},
                stats_by_id={lid: r.stats for lid, r in id_to_readings.items() if lid in currently_known_ids},
                usages_by_id={lid: r.usage for lid, r in id_to_readings.items()},
            )

        with self.subTest('LVEs appearance detected'):
            tracker_update(
                now=next(self.clock),
                currently_known_ids=tuple(id_to_readings.keys()),
            )

            for expected, actual in [
                (tracker.unbursted, {
                    tracker.serialized_id_to_manager[1],
                }),
                (tracker.bursted, {
                    tracker.serialized_id_to_manager[2],
                    tracker.serialized_id_to_manager[3],
                }),
                (tracker.overusing, {
                    tracker.serialized_id_to_manager[3],
                }),
            ]:
                self.assertSetEqual(expected, actual)

            self.assertSetEqual(added_from_signal, set(tracker.serialized_id_to_manager.values()))

        with self.subTest('LVEs disappearance detected'):
            tracker_update(
                now=next(self.clock),
                currently_known_ids=tuple(),
            )
            for state_set in (tracker.unbursted, tracker.bursted, tracker.overusing, tracker.quota_exceeded):
                self.assertSetEqual(state_set, set())

    def test_garbage_collection(self):
        quota, quota_window = timedelta(seconds=1), timedelta(seconds=2)
        lve_id, manager = 42, None

        def create_lve_manager(
            *args,
            lve_id: LveId,
            now: Timestamp,
            normal_limits: LveLimits,
            stats: LveStats,
            usage: LveUsage,
            **kwargs,
        ):
            nonlocal manager
            manager = LveStateManager(
                now=now,
                lve_id=lve_id,
                bursting_multipliers=BurstingMultipliers(cpu=2.0, io=3.0),
                initial_normal_limits=normal_limits,
                initial_stats=stats,
                initial_usage=usage,
                quota=quota,
                quota_window=quota_window,
                apply_lve_settings=lambda *args, **kwargs: None,
            )
            return manager


        tracker = LvesTracker(
            create_lve_manager=create_lve_manager,
            deserialize_lve_id=lambda lve_id: lve_id
        )

        # TODO(vlebedev): Check for case when last state is "overusing" before LVE disappears.
        for update_request in itertools.repeat([
            (2, 1, 3, 1, 4, 1), # overusing
            (2, 1, 1, 1, 4, 1), # bursted
        ], 3):
            for reading in update_request:
                now = next(self.clock)
                update_request = _make_readings_update_request(now, *reading)
                tracker.update(
                    now=now,
                    normal_limits_by_id={lve_id: update_request.normal_limits},
                    stats_by_id={lve_id: update_request.stats},
                    usages_by_id={lve_id: update_request.usage},
                )

        # NOTE(vlebedev): LVE disappeared.
        tracker.update(
            now=next(self.clock),
            normal_limits_by_id=dict(),
            stats_by_id=dict(),
            usages_by_id=dict(),
        )

        self.assertEqual(manager.state, LveState.EXISTED)
        for state_set in (tracker.unbursted, tracker.bursted, tracker.overusing, tracker.quota_exceeded):
            self.assertSetEqual(state_set, set())
        self.assertIn(manager.lve_id, tracker.serialized_id_to_manager)

        # NOTE(vlebedev): Fast forward to ensure that history is fully trimmed.
        try:
            for _ in range(int(quota_window.total_seconds() * 2)):
                tracker.update(
                    now=next(self.clock),
                    normal_limits_by_id=dict(),
                    stats_by_id=dict(),
                    usages_by_id=dict(),
                )
        except BaseException as e:
            raise e

        for state_set in (tracker.unbursted, tracker.bursted, tracker.overusing, tracker.quota_exceeded):
            self.assertSetEqual(state_set, set())
        self.assertNotIn(manager.lve_id, tracker.serialized_id_to_manager)

    # def test_lve_id_with_history_disappeared(self):
    #     pass

    # def test_lve_id_with_no_history_disappeared(self):
    #     pass


class TestStorage(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.engine = sa.create_engine("sqlite+pysqlite:///:memory:", echo=True)

    def setUp(self) -> None:
        self.random = Random(4728)
        self.server_id = self.id()
        init_db_schema(self.engine)

    def test_cutoff_in_the_middle_of_duplicated_events_group(self):
        self.skipTest('Query must be fixed to properly handle this case')

        history = self.save_events_then_load_history(5, [
            (1, BurstingEventType.STARTED),
            (2, BurstingEventType.STOPPED),
            (3, BurstingEventType.STARTED),
            (4, BurstingEventType.STARTED),
            (5, BurstingEventType.STARTED),
            (6, BurstingEventType.STOPPED),
        ])

        self.assertEqual(history, LveHistory(IntervalType.OVERUSING, (3, 6)))

    # def test_events_written(self):
    #     pass

    # def test_well_formatted_events_loaded(self):
    #     pass

    # def test_last_event_is_started(self):
    #     pass

    def test_duplicated_events_normalized_when_loaded(self):
        history = self.save_events_then_load_history(0, [
            (1, BurstingEventType.STARTED),
            (2, BurstingEventType.STARTED),
            (3, BurstingEventType.STOPPED),
            (4, BurstingEventType.STARTED),
            (5, BurstingEventType.STOPPED),
            (6, BurstingEventType.STOPPED),
            (7, BurstingEventType.STOPPED),
            (8, BurstingEventType.STARTED),
            (9, BurstingEventType.STOPPED),
        ])

        self.assertEqual(history, LveHistory(IntervalType.OVERUSING, (1, 3, 4, 5, 8, 9)))

    def test_last_overusing_events_dropped(self):
        history = self.save_events_then_load_history(0, [
            (1, BurstingEventType.STARTED),
        ])

        self.assertEqual(history, LveHistory())

    def test_events_cleanup(self):
        lve_id = 1
        self.save_events(lve_id, [
            (1, BurstingEventType.STARTED),
            (2, BurstingEventType.STARTED),
            (3, BurstingEventType.STOPPED),
            (4, BurstingEventType.STARTED),
            (5, BurstingEventType.STOPPED),
            (6, BurstingEventType.STOPPED),
            (7, BurstingEventType.STOPPED),
            (8, BurstingEventType.STARTED),
            (9, BurstingEventType.STOPPED),
        ])
        cleanup_old_events(self.engine, self.server_id, Timestamp(5))
        history = self.load_events(lve_id, 0)

        self.assertEqual(history, LveHistory(IntervalType.NORMAL, (6, 8, 9)))

    def tearDown(self) -> None:
        drop_db_schema(self.engine)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.engine.dispose()

    def save_events_then_load_history(
        self, cutoff: int, events_data: Sequence[tuple[int, BurstingEventType]]
    ) -> LveHistory:
        lve_id = self.random.randint(1, 1000)
        self.save_events(lve_id, events_data)
        return self.load_events(lve_id, cutoff)

    def save_events(self, lve_id, events_data: Sequence[tuple[int, BurstingEventType]]):
        save_events_to_db(
            engine=self.engine,
            server_id=self.server_id,
            events=[InBurstingEventRow(
                lve_id=lve_id,
                timestamp=Timestamp(ts),
                event_type=event_type,
            ) for ts, event_type in events_data],
        )

    def load_events(self, lve_id, cutoff: int) -> LveHistory:
        results = load_bursting_enabled_intervals_from_db(
            engine=self.engine,
            cutoff=Timestamp(cutoff),
            server_id=self.server_id,
        )
        return results[lve_id]


class TestAdjuster(TestCase):
    def setUp(self) -> None:
        self.clock = itertools.count(1.0)

    def test_expected_flow(self):
        tracker = MagicMock(
            spec_set=LvesTracker,
            unbursted=frozenset(),
            bursted=frozenset(),
            overusing=frozenset(),
            quota_exceeded=frozenset()
        )
        is_overloaded = MagicMock(
            spec_set=OverloadChecker,
            return_value=OverloadCheckResult(True, 0.9),
        )
        step_calculator = MagicMock(
            spec_set=StepCalculator,
            get_enable_step=lambda *args, **kwargs: 5,
            get_disable_step=lambda *args, **kwargs: 2,
        )
        adjuster = Adjuster(
            lves_tracker=tracker,
            get_normal_limits=lambda: dict(),
            step_calculator=step_calculator,
            is_server_overloaded=is_overloaded,
        )

        adjuster.step(AdjustStepData(
            now=next(self.clock),
            lve_active_ids=dict(),
            stats=dict(),
            lve_usages_by_id=dict(),
        ))


class TestConfigUpdate(TestCase):
    def setUp(self):
        self.config = PluginConfig(
            bursting_enabled='true',
            server_id='localhost',
            bursting_quota_sec='5',
            bursting_quota_window_sec='3600',
            bursting_idle_time_threshold='0.5',
            bursting_cpu_multiplier='2',
            bursting_io_multiplier='2',
            bursting_database_dump_period_sec='240',
            bursting_idle_time_samples_num='1',
            bursting_debug_mode='true',
        )
        self.random = Random(2345)

    def test_mandatory_params_missing(self):
        required_keys = tuple(PluginConfig.__required_keys__)
        removed_keys = frozenset(self.random.choices(required_keys, k=int(len(required_keys) / 3)))
        for k in removed_keys:
            del self.config[k]

        with self.assertRaises(MissingKeysInRawConfig) as cm:
            ConfigUpdate.from_plugin_config(self.config)
        self.assertSetEqual(cm.exception.missing_raw_keys, removed_keys)

    def test_unexpected_params(self):
        extra_items = {
            "test": "42",
        }
        assert extra_items.keys() & PluginConfig.__annotations__.keys() == set()
        self.config.update(extra_items)

        ConfigUpdate.from_plugin_config(self.config)

    def test_no_optional_keys(self):
        for k in PluginConfig.__optional_keys__:
            del self.config[k]

        ConfigUpdate.from_plugin_config(self.config)


def _create_lve_manager(
    initial_readings: LveStateManager.UpdateReadings,
    lve_id: int,
    quota: timedelta = timedelta(seconds=5),
    quota_window: timedelta = timedelta(hours=1),
) -> LveStateManager:
    return LveStateManager(
        now=initial_readings.now,
        lve_id=LveId(lve_id),
        bursting_multipliers=BurstingMultipliers(cpu=2.0, io=3.0),
        initial_normal_limits=initial_readings.normal_limits,
        initial_stats=initial_readings.stats,
        initial_usage=initial_readings.usage,
        quota=quota,
        quota_window=quota_window,
        apply_lve_settings=lambda *args, **kwargs: None,
    )


def _make_readings_update_request(
    now: Timestamp,
    normal_cpu_limit: int,
    normal_io_limit: int,
    cpu_usage: int,
    io_usage: int,
    kernel_cpu_limit: int,
    kernel_io_limit: int,
) -> LveStateManager.UpdateReadings:

    assert kernel_cpu_limit >= normal_cpu_limit
    assert kernel_cpu_limit >= cpu_usage
    assert kernel_io_limit >= normal_io_limit
    assert kernel_io_limit >= io_usage

    return LveStateManager.UpdateReadings(
        now=now,
        normal_limits=LveLimits(
            cpu=normal_cpu_limit,
            io=normal_io_limit,
        ),
        stats=MagicMock(
            spec=LveStats,
            cpu=kernel_cpu_limit,
            io=kernel_io_limit,
            cpu_usage=cpu_usage,
            io_usage=io_usage,
            reseller_id=0,
        ),
        usage=typing.cast(LveUsage, SimpleNamespace(
            cpu_usage=cpu_usage,
            io_usage=io_usage,
        ))
    )
