# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import json
import os
import unittest
import mock

import ddt

import lvestats.plugins.generic.analyzers as analyzers
from lvestats.lib.commons.func import serialize_lve_id, LIMIT_LVP_ID, deserialize_lve_id
from lvestats.plugins.generic.persistors import CloudLinuxTopFileSaver, MySQLTopFileSaver


@ddt.ddt
class TestPluginCloudLinuxTopFileSaver(unittest.TestCase):
    usage = analyzers.LVEUsage(8)
    usage2 = analyzers.LVEUsage(8)
    usage2.cpu_usage = 100
    usage2.io = 1
    usage2.lcpu = 100
    usage2.cpu_fault = 4
    usage2.mem_usage = 1000
    with mock.patch(
            'lvestats.lib.config.read_config',
            return_value={}
    ):
        DDT_LIST = [
            {
                'LVE_VERSION': 8,
                'lve_usage_5s': {}},
            {
                'LVE_VERSION': 8,
                'lve_usage_5s': {
                    0: usage,
                    1: usage,
                    2: usage2}},
            {
                'LVE_VERSION': 6,
                'lve_usage_5s': {
                    0: usage,
                    1: usage2}},
            {
                'LVE_VERSION': 6,
                'lve_usage_5s': {
                    0: usage2,
                    1: usage,
                    100: usage2,
                    67: usage}},
            {
                'LVE_VERSION': 10,
                'lve_usage_5s': {
                    0: usage2,
                    100: usage2,
                    serialize_lve_id(LIMIT_LVP_ID, 100): usage2}}]

    def setUp(self):
        self.mock_file = '/tmp/lvestats2_cloudlinux_top.json'
        self.cloudlinux_top_plugin = CloudLinuxTopFileSaver(fname=self.mock_file)

        patcher1 = mock.patch('lvestats.lib.config.read_config', return_value={})
        patcher1.start()
        self.addCleanup(patcher1.stop)

    @ddt.data(*DDT_LIST)
    def test_execute(self, value):
        self.cloudlinux_top_plugin.execute(value)

        with open(self.mock_file, 'r') as json_file:
            try:
                result = json.load(json_file)
            except (IndexError, KeyError, IOError) as e:
                self.fail(e)
            else:
                self.assertTrue("users" in result)
                self.assertTrue(0 not in result["users"])
                for uid in set(value["lve_usage_5s"]) - {0}:
                    _, is_reseller = deserialize_lve_id(uid)
                    if is_reseller:
                        # There should be one reseller in resulted json for every given
                        test_users = [user for user in result["resellers"] if user["id"] == uid]
                        self.assertTrue(len(test_users) == 1)
                    else:
                        # There should be one user in resulted json for every given (except 0)
                        test_users = [user for user in result["users"] if user["id"] == uid]
                        self.assertTrue(len(test_users) == 1)

                    # in lve_usage_5s CPU measurement unit is 100 * % of one CPU core
                    # check if CPU  measurement unit is set as % of one CPU core
                    self.assertEqual(int(test_users[0]['usage']['cpu']['all'] * 100),
                                     value["lve_usage_5s"][uid].cpu_usage)
                    self.assertEqual(int(test_users[0]['limit']['cpu']['all'] * 100),
                                     value["lve_usage_5s"][uid].lcpu)
                a6 = True
                a8 = value["LVE_VERSION"] >= 8
                for user in result["users"]:
                    b6 = "mem" in user["usage"] and "mem" in user["limit"] and \
                         "pno" in user["usage"] and "pno" in user["limit"]
                    b8 = "iops" in user["usage"] and "iops" in user["limit"]
                    self.assertTrue(a6 == b6)  # a6 xnor b6
                    self.assertTrue(a8 == b8)  # a8 xnor b8
        os.unlink(self.mock_file)


@ddt.ddt
class TestMySQLTopFileSaver(unittest.TestCase):

    def setUp(self):
        self.mock_file = '/tmp/lvestats2_cloudlinux_dbtop.json'
        self.cloudlinux_dbtop_plugin = MySQLTopFileSaver(fname=self.mock_file)

    DDT_GET_DBTOP_INFO = [
        [
            ['User            . cpu(%)             .  read(B/s)          .  write(B/s)      . CAUSE  '],
            {}],
        [
            ['User            . cpu(%)             .  read(B/s)          .  write(B/s)      . CAUSE  ',
             'user1              0/0/0                0/0/0                 0/0/0              -'],
            {'user1': {'cause_of_restrict': '-', 'time_of_restrict': 0, 'cpu': 0, 'io': 0}}],
        [
            ['User            . cpu(%)             .  read(B/s)          .  write(B/s)      . CAUSE  ',
             'user1              39/13/3              603K/745K/149K        529K/146K/29219    R/c:cpu/59'],
            {'user1': {'cause_of_restrict': 'cpu', 'time_of_restrict': 59, 'cpu': 39, 'io': 1159168}}],
        [
            ['User            . cpu(%)             .  read(B/s)          .  write(B/s)      . CAUSE  ',
             'user1              0/0/0                0/0/0                 0/0/0              -',
             'user2              69/69/37             183K/81394/220K       1173K/536K/304K    R/c:cpu/60',
             'user3              9/9/7                883K/81394/220K       117M/53M/30M       R/c:write/21'],
            {'user1': {'cause_of_restrict': '-', 'time_of_restrict': 0, 'cpu': 0, 'io': 0},
             'user2': {'cause_of_restrict': 'cpu', 'time_of_restrict': 60, 'cpu': 69, 'io': 1388544},
             'user3': {'cause_of_restrict': 'write', 'time_of_restrict': 21, 'cpu': 9, 'io': 123587584}}]
    ]

    @ddt.data(*DDT_GET_DBTOP_INFO)
    @ddt.unpack
    def test_get_dbtop_info(self, dbtop_output, expected_result):
        self.cloudlinux_dbtop_plugin._dbtop_exist = True
        self.cloudlinux_dbtop_plugin._get_dbtop_output = lambda: dbtop_output
        result = self.cloudlinux_dbtop_plugin.get_dbtop_info()
        self.assertEqual(result, expected_result)
