# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import configparser
import unittest

import mock
from clcommon import clconfpars

from lvestats.lib import config


class TestPluginConfig(unittest.TestCase):
    def setUp(self):
        self.load_conf_backup = clconfpars.load

    def tearDown(self):
        clconfpars.load = self.load_conf_backup

    def return_dict(self, arg=None):
        dict1 = {'a': '1', 'b': '2', 'c': '3'}
        dict2 = {'z': '1', 'x': '2', 'y': '3'}
        if (arg == '/etc/sysconfig/lvestats2') or (arg is None):
            return dict1
        else:
            return dict2

    def return_dict2(self, arg=None):
        dict2 = {'z': '1', 'x': '2', 'y': '3'}
        dict3 = {'w': '1', 'e': '2', 'r': '3'}
        if (arg == '/etc/sysconfig/lvestats2') or (arg is None):
            ie = IOError()
            ie.errno=13
            raise ie
        elif arg == '/etc/sysconfig/lvestats2.readonly':
            return dict3
        else:
            return dict2

    def load_with_parsing_error(self, *args, **kwargs):
        raise configparser.ParsingError('fn')

    def test_read_global(self):
        clconfpars.load = self.return_dict
        global_cfg = config.read_config()
        self.assertIsNotNone(global_cfg)
        self.assertTrue('a' in global_cfg)
        self.assertTrue('b' in global_cfg)
        self.assertTrue('d' not in global_cfg)

    @mock.patch('os.path.isfile')
    def test_read_config_plugin(self, mock_isfile):
        clconfpars.load = self.return_dict
        mock_isfile.return_value = True
        global_cfg = config.read_plugin_config(config.read_config(), 'plugin_name')
        self.assertIsNotNone(global_cfg)
        self.assertTrue('a' in global_cfg)
        self.assertTrue('b' in global_cfg)
        self.assertTrue('d' not in global_cfg)
        self.assertTrue('z' in global_cfg)
        self.assertTrue('x' in global_cfg)
        self.assertTrue('u' not in global_cfg)

    def test_read_config_error_plugin(self):
        clconfpars.load = self.return_dict
        global_cfg = config.read_plugin_config(config.read_config(), 'plugin_name')
        self.assertIsNotNone(global_cfg)
        self.assertTrue('a' in global_cfg)
        self.assertTrue('b' in global_cfg)
        self.assertTrue('d' not in global_cfg)
        self.assertTrue('z' not in global_cfg)
        self.assertTrue('x' not in global_cfg)
        self.assertTrue('u' not in global_cfg)

    def test_read_config_read_only(self):
        clconfpars.load = self.return_dict2
        global_cfg = config.read_config()
        self.assertIsNotNone(global_cfg)
        self.assertTrue('a' not in global_cfg)
        self.assertTrue('b' not in global_cfg)
        self.assertTrue('d' not in global_cfg)
        self.assertTrue('z' not in global_cfg)
        self.assertTrue('x' not in global_cfg)
        self.assertTrue('u' not in global_cfg)
        self.assertTrue('w' in global_cfg)
        self.assertTrue('e' in global_cfg)
        self.assertTrue('r' in global_cfg)

    def test_read_global_config_with_error(self):
        clconfpars.load = self.load_with_parsing_error
        with self.assertRaises(config.ConfigError):
            config.read_config()

    def test_read_global_readonly_plugin_with_error(self):
        exceptions = iter([IOError(13, "Permission denied"), configparser.ParsingError('fn')])

        def load_with_side_effect(self, *args, **kwargs):
            """
                Raises exceptions from list, then works fine.
            """
            config._load_config_count += 1
            raise next(exceptions)

        config._load_config_count = 0
        clconfpars.load = load_with_side_effect
        with self.assertRaises(config.ConfigError):
            config.read_config()
        self.assertEqual(config._load_config_count, 2)

if __name__ == '__main__':
    unittest.main()
