# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import mock
import sqlalchemy.engine.base
import time

from sqlalchemy import select
from test.base.base_db_test import BaseDbTest
from lvestats.orm.history import history_x60
from lvestats.lib.lveinfolib import HistoryShow
from lvestats.plugins.generic.dbsaver_x60 import DBSaverX60


class TestDBSaverX60(BaseDbTest):
    def setUp(self):
        BaseDbTest.setUp(self)
        self.dbsaverx60 = DBSaverX60()
        self.dbsaverx60.set_config(self.cfg)
        self.dbsaverx60.set_db_engine(self.engine)

    def test_insert_big_chunk(self):
        history_show_proceed_result = [
            (i, 0.0, 5000, 0, 0.0, 15, 0, 0.0, 25, 0, 0.0, 25, 0, 0.0, 15, 0, 0.0, 1048576, 0, 0, 1024, 0.0) for i in range(5000)]
        with mock.patch.object(HistoryShow, "proceed", return_value=history_show_proceed_result), \
                mock.patch.object(sqlalchemy.engine.base.Transaction, "rollback") as rollback, \
                mock.patch('lvestats.plugins.generic.dbsaver_x60.reboot_lock'):
            self.dbsaverx60.aggregate_db_data_by_hours(1506099654, 1506185959)
            self.assertFalse(rollback.called)

    def test_aggregate_with_void(self):
        data = [
            dict(server_id='localhost', uid=0, _time=0),
            dict(server_id='localhost', uid=1, _time=0, cpu=100),
            dict(server_id='localhost', uid=2, _time=0, cpu=100),
            dict(server_id='localhost', uid=0, _time=60),
            dict(server_id='localhost', uid=1, _time=60, cpu=100),
            dict(server_id='localhost', uid=2, _time=60, cpu=100),
            dict(server_id='localhost', uid=0, _time=10001),
            dict(server_id='localhost', uid=1, _time=10001, cpu=100),
            dict(server_id='localhost', uid=2, _time=10001, cpu=100),
            dict(server_id='localhost', uid=0, _time=10061),
            dict(server_id='localhost', uid=1, _time=10061, cpu=100),
            dict(server_id='localhost', uid=2, _time=10061, cpu=100),

            # data for next hour, so previous will be aggregated
            dict(server_id='localhost', uid=2, _time=10800, cpu=100)]
        for d in data:
            self.save_history_line(**d)
        self.dbsaverx60.now = time.time()
        with mock.patch('lvestats.plugins.generic.dbsaver_x60.reboot_lock'):
            self.dbsaverx60.execute({})
        history_x60_lines = list(self.engine.execute(select([history_x60.created])))
        self.assertEqual(len(history_x60_lines), 5)

    def test_aggregate_table(self):
        self.dbsaverx60.execute_timeout = 0
        data = [
            # strange thing in aggregator:
            # we pass utc_from=0 to aggregate_db_data_by_hours
            # but then call HistoryShow:
            # period_from=hour_from + 1,
            # so first element is not aggregated, but used
            dict(server_id='localhost', uid=0, _time=0),

            dict(server_id='localhost', uid=0, _time=1),
            dict(server_id='localhost', uid=1, _time=1, cpu=100),
            dict(server_id='localhost', uid=2, _time=1, cpu=100),

            dict(server_id='localhost', uid=0, _time=60),
            dict(server_id='localhost', uid=1, _time=60, cpu=200)]
        for d in data:
            self.save_history_line(**d)

        # first hour is not full, nothing should be aggregated
        self.dbsaverx60.now = 3000
        self.dbsaverx60.execute({})
        history_x60_lines = list(self.engine.execute(select([history_x60.created])))
        self.assertCountEqual(history_x60_lines, [])

        data = [
            dict(server_id='localhost', uid=0, _time=3601),
            dict(server_id='localhost', uid=1, _time=3601, cpu=100),
            dict(server_id='localhost', uid=2, _time=3601, cpu=200),

            dict(server_id='localhost', uid=0, _time=3660),
            dict(server_id='localhost', uid=1, _time=3660, cpu=100),
            dict(server_id='localhost', uid=2, _time=3660, cpu=100)]
        for d in data:
            self.save_history_line(**d)

        # first hour is full, but second is not
        # only first hour should be aggregated
        self.dbsaverx60.now = 4000
        with mock.patch('lvestats.plugins.generic.dbsaver_x60.reboot_lock'):
            self.dbsaverx60.execute({})
        history_x60_lines = list(self.engine.execute(select([history_x60.created, history_x60.id,
                                                             history_x60.time, history_x60.cpu])))
        self.assertCountEqual(history_x60_lines,
                              [(3600, 1, 2, 150), (3600, 2, 2, 50.)])

        data = [
            dict(server_id='localhost', uid=0, _time=7201)]
        for d in data:
            self.save_history_line(**d)

        # second hour is done and should be aggregated
        self.dbsaverx60.now = 10000
        with mock.patch('lvestats.plugins.generic.dbsaver_x60.reboot_lock'):
            self.dbsaverx60.execute({})
        history_x60_lines = list(self.engine.execute(select([history_x60.created, history_x60.id,
                                                             history_x60.time, history_x60.cpu])))
        self.assertCountEqual(history_x60_lines,
                              [(3600, 1, 2, 150), (3600, 2, 2, 50.),
                               (7200, 1, 2, 100), (7200, 2, 2, 150.)])
