import unittest
import mock
from ddt import ddt, data, unpack

from lvestat import LVEStat
from lvestats.plugins.other.lvedestroyer import LVEDestroyer, LVERecord

DEFAULT_LVESTAT_10_PARAMETERS = '\t'.join([
    # '0',        # reseller_id
    # '0',        # id
    '10000',    # cpu
    '100',      # lcpuw
    '1',        # ncpu
    '20',       # lep
    '100',      # lnproc
    '0',        # lmem
    '262144',   # lmemphy
    '1024',     # io
    '1024',     # liops
    '0',        # lNETO    not implemented yet
    '0',        # lNETI    not implemented yet
    '0',        # mep
    '0',        # cpu_usage
    '0',        # mem_usage
    '0',        # io_usage
    '0',        # mem_fault
    '0',        # mep_fault
    '0',        # memphy
    '0',        # memphy_fault
    '0',        # nproc
    '0',        # nproc_fault
    '0',        # iops
    '0',        # NETO     not implemented yet
    '0',        # NETI     not implemented yet
])
DEFAULT_LVESTAT_10_BASESTRING = '0,0\t' + DEFAULT_LVESTAT_10_PARAMETERS
DEFAULT_LVESTAT = LVEStat(DEFAULT_LVESTAT_10_BASESTRING, 10)
CUSTOM_LVESTAT_BASESTRING = (
    '0,4294967295    0       0       0       0       0       0       0       0       0       0       0       0'
    '       511866316351    380455  0       0       0       10007   0       178     0       0       0       0'
)
CUSTOM_LVESTAT = LVEStat(CUSTOM_LVESTAT_BASESTRING, 10)


@ddt
class TestLVEDestroyer(unittest.TestCase):

    def setUp(self) -> None:
        self.plugin_instance = LVEDestroyer()

    @data(
        ({}, False),
        ({'iterations': 0}, False),
        ({'iterations': 'nope'}, False),
        ({'iterations': 1}, True),
    )
    @unpack
    def test_set_config(self, config: dict, enabled_: bool):
        self.plugin_instance.set_config(config)
        self.assertEqual(self.plugin_instance._enabled, enabled_)

    @data(
        (DEFAULT_LVESTAT, False),
        (DEFAULT_LVESTAT, True),
    )
    @unpack
    def test_set_default(self, stat: LVEStat, enabled_: bool):
        with mock.patch.object(self.plugin_instance, '_enabled', enabled_):
            self.plugin_instance.set_default(stat)
            self.assertIs(bool(self.plugin_instance.default_stat), enabled_)

    @data(
        (DEFAULT_LVESTAT, False),
        (CUSTOM_LVESTAT, True),
    )
    @unpack
    def test_is_stat_tracked(self, stat: LVEStat, out_: bool):
        with mock.patch.object(self.plugin_instance, '_enabled', True):
            self.plugin_instance.set_default(DEFAULT_LVESTAT)
            self.assertEqual(self.plugin_instance.is_stat_tracked(stat), out_)

    @data(
        (CUSTOM_LVESTAT, True),
        (CUSTOM_LVESTAT, False),
    )
    @unpack
    def test_check_lve(self, stat: LVEStat, record_check_: bool):
        with mock.patch('lvestats.plugins.other.lvedestroyer.LVERecord.has_changed', return_value=False), \
                mock.patch('lvestats.plugins.other.lvedestroyer.LVERecord.check', return_value=record_check_), \
                mock.patch.object(self.plugin_instance, 'lves', {stat.id: LVERecord(stat.cpu, 5)}):
            self.plugin_instance.check_lve(stat)
            if record_check_:
                self.assertIn(stat.id, self.plugin_instance.to_be_destroyed)
                self.assertNotIn(stat.id, self.plugin_instance.lves)
            else:
                self.assertGreater(self.plugin_instance.lves[stat.id].count, 0)

    @data(
        DEFAULT_LVESTAT_10_BASESTRING,
        '0,4294967295\t' + DEFAULT_LVESTAT_10_PARAMETERS,
        '0,1111111111\t' + DEFAULT_LVESTAT_10_PARAMETERS,
    )
    def test_add_stat(self, stat_basestring: str):
        stat = LVEStat(stat_basestring, 10)
        with mock.patch.object(self.plugin_instance, '_enabled', True), \
                mock.patch.object(self.plugin_instance, 'lves', {4294967295: LVERecord(stat.cpu, 5)}), \
                mock.patch.object(self.plugin_instance, 'default_stat', DEFAULT_LVESTAT), \
                mock.patch.object(self.plugin_instance, 'set_default') as _mock_set_default, \
                mock.patch.object(self.plugin_instance, 'check_lve') as _mock_check_lve, \
                mock.patch.object(self.plugin_instance, 'is_stat_tracked', return_value=False) as _mock_is_stat_tracked:
            self.plugin_instance.add_stat(stat)
            if stat.id == 0:
                _mock_set_default.assert_called()
            elif stat.id == 4294967295:
                _mock_check_lve.assert_called()
            else:
                _mock_is_stat_tracked.assert_called()
                self.assertIn(1111111111, self.plugin_instance.lves)

    @data(
        ([4294967295, 1111111111], b'4294967295 1111111111'),
    )
    @unpack
    def test_process_destroy(self, to_destroy: list, out_: bytes):
        with mock.patch.object(self.plugin_instance, '_enabled', True), \
                mock.patch.object(self.plugin_instance, 'to_be_destroyed', to_destroy), \
                mock.patch('subprocess.Popen') as _mock_popen:
            self.plugin_instance.process_destroy()
            # check that p.communicate() of process_destroy() called with the right argument bytestring
            self.assertIn(mock.call().__enter__().communicate(out_), _mock_popen.mock_calls)

    @data(
        {'stats': {DEFAULT_LVESTAT.id: DEFAULT_LVESTAT, CUSTOM_LVESTAT.id: CUSTOM_LVESTAT}}
    )
    def test_execute(self, lve_data):
        with mock.patch.object(self.plugin_instance, '_enabled', True), \
                mock.patch.object(self.plugin_instance, 'set_default') as _mock_set_default, \
                mock.patch.object(self.plugin_instance, 'add_stat') as _mock_add_stat, \
                mock.patch.object(self.plugin_instance, 'process_destroy') as _mock_process_destroy:
            self.plugin_instance.execute(lve_data)
            _mock_set_default.assert_called_with(DEFAULT_LVESTAT)
            _mock_add_stat.assert_called_with(CUSTOM_LVESTAT)
            _mock_process_destroy.assert_called()


if __name__ == '__main__':
    unittest.main()
