# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import time
import unittest
from contextlib import contextmanager
from random import randint
from itertools import combinations_with_replacement

import mock
from ddt import ddt, data, unpack
from lvestats.plugins.generic.statsnotifier import NotifyFaultsOptions, ResellerSettings, NotifySettings, AdminSettings, StatsNotifier

import test.testcases_data as testcases_data
from test.base.base_db_test import BaseDbTest
from test.testcases_data import NO_SETTINGS

__author__ = 'alukyanov'


@ddt
class TestStatsNotifier(BaseDbTest):
    maxDiff = None

    def setUp(self):
        BaseDbTest.setUp(self)
        # set empty config as default
        _mock_default_config = mock.patch(
            'lvestats.plugins.generic.statsnotifier.get_admin_notification',
            return_value=testcases_data.get_admin_settings())
        _mock_default_config.start()
        self.addCleanup(_mock_default_config.stop)

        self.statsnotifier = StatsNotifier()
        self.statsnotifier.set_config(self.cfg)
        self.statsnotifier.set_db_engine(self.engine)

        patcher1 = mock.patch('lvestats.lib.config.read_config', return_value={})
        patcher1.start()
        self.addCleanup(patcher1.stop)

    @contextmanager
    def mock_admin_settings(self, **kwargs):
        with mock.patch('lvestats.plugins.generic.statsnotifier.get_admin_notification',
                        return_value=testcases_data.get_admin_settings(**kwargs)) as mock_settings:
            self.statsnotifier._load_plugin_settings()
            yield mock_settings

    def _get_notify_faults_types(self, notify_cpu=False, notify_io=False, notify_iops=False,
                                 notify_memory=False, notify_ep=False, notify_nproc=False):
        return NotifyFaultsOptions(
            NOTIFY_CPU=notify_cpu,
            NOTIFY_IO=notify_io,
            NOTIFY_IOPS=notify_iops,
            NOTIFY_MEMORY=notify_memory,
            NOTIFY_EP=notify_ep,
            NOTIFY_NPROC=notify_nproc
        )

    def _generate_data(self, cpu_fault=None):
        ts = int(time.time()) - self.period
        uid = self.uid
        values = [dict(
            _time=ts + (60 * i),
            uid=uid,
            server_id=self.cfg['server_id'],

            cpu=2500,
            cpu_limit=5000,
            cpu_fault=randint(0, 10) if cpu_fault is None else cpu_fault,

            mep=10,
            mep_limit=20,
            mep_fault=randint(0, 10),

            io=524288,
            io_limit=1048576,
            io_fault=randint(0, 10),

            mem=131072,
            mem_limit=262144,
            mem_fault=randint(0, 10),

            lmemphy=262144,
            memphy=20,
            memphy_fault=randint(0, 10),

            lnproc=100,
            nproc=6,
            nproc_fault=randint(0, 10),

            liops=1024,
            iops=512,
            iops_fault=randint(0, 10),
        ) for i in range(3 * 60)]
        return values

    @data([505],
          [505, 506],
          [505, 506, 1000])
    def test_statsnotifier_lvestatshistory(self, users):
        self.period = 3 * 60 * 60

        for i, user_id in enumerate(users):
            self.uid = user_id
            for kwargs in self._generate_data(cpu_fault=1):
                # save only first record with keep alive
                self.save_history_line(keepalive=(i == 0), **kwargs)

        users_faults = list(self.statsnotifier.get_users_fault(self.period, 1, self._get_notify_faults_types(notify_cpu=True)))
        self.assertCountEqual(users, [user['ID'] for user in users_faults])

    @data(*combinations_with_replacement([True, False], 2))
    @unpack
    def test_statsnotifier_get_users_fault_convert_units(self, notify_memory, notify_nproc):
        self.period = 3 * 60 * 60
        self.uid = 505
        for kwargs in self._generate_data():
            self.save_history_line(keepalive=True, **kwargs)  # save record with keep alive

        notify_faults_types = self._get_notify_faults_types(
            notify_cpu=True,
            notify_memory=notify_memory,
            notify_nproc=notify_nproc
        )

        users_faults = list(self.statsnotifier.get_users_fault(self.period, 1, notify_faults_types))
        self.assertEqual(self.uid, users_faults[0]['ID'])

        # this transformations should be done even if notify is disabled
        self.assertEqual(262144 * 4, users_faults[0]["lVMem"])
        self.assertEqual(262144 * 4, users_faults[0]["lPMem"])
        self.assertEqual(1048576 // 1024,  users_faults[0]["lIO"])

        self.assertEqual(131072 * 4, users_faults[0]["aVMem"])
        self.assertEqual(20 * 4, users_faults[0]["aPMem"])
        self.assertEqual(524288 // 1024,  users_faults[0]["aIO"])

        # other units should not be converted
        self.assertEqual(50., users_faults[0]["lCPU"])
        self.assertEqual(100, users_faults[0]["lNproc"])
        self.assertEqual(1024,  users_faults[0]["lIOPS"])

        self.assertEqual(25., users_faults[0]["aCPU"])
        self.assertEqual(6, users_faults[0]["aNproc"])
        self.assertEqual(512,  users_faults[0]["aIOPS"])

    def test_statsnotifier(self):
        self.period = 60 * 60
        period_to = int(time.time())
        test_data = [dict(_time=period_to-10,    cpu_fault=5, cpu=6, cpu_limit=6, mem_fault=1),
                     dict(_time=period_to-30,    cpu_fault=11,                nproc_fault=14),
                     dict(_time=period_to-50,    cpu_fault=4,                 nproc_fault=7),
                     dict(_time=period_to-70,    cpu_fault=2,   mem_fault=20, nproc_fault=8),
                     dict(_time=period_to-90,    cpu_fault=8,   mem_fault=10),
                     dict(_time=period_to-110,   cpu_fault=23,  mem_fault=15, nproc_fault=12),
                     dict(_time=period_to-120),
                     dict(_time=period_to-130,   cpu_fault=1),
                     dict(_time=period_to-140,                  mem_fault=5)]
        for kwargs in test_data:
            self.save_history_line(server_id=self.cfg['server_id'], uid=1, keepalive=True, **kwargs)

        notify_faults_types = self._get_notify_faults_types(
            **{'notify_cpu':     True,
               'notify_memory':  True,
               'notify_nproc':   True})
        users_stat_ret = list(self.statsnotifier.get_users_fault(self.period, 1, notify_faults_types))
        self.assertEqual(54, users_stat_ret[0]["CPUf"])
        self.assertEqual(51, users_stat_ret[0]["VMemF"])
        self.assertEqual(41, users_stat_ret[0]["NprocF"])
        self.assertEqual(0,  users_stat_ret[0]["PMemF"])

    TIME = time.time()

    USE_TIMESTAMP_TEST_DATA = (
        # now, saved timestamp, assertion result
        (TIME, False),
        (5, True),
        (TIME - 49, False),
        (TIME - 51, True),
        (TIME + 50, False),
        (None, True),
        (None, True)
    )

    @data([True,  True,  1,   [1, 2, 3]],
          [True,  True,  11,  [2]],
          [False, True,  11,  []],
          [True,  True,  7,   [2, 3]],
          [False, True,  7,   [3]],
          [False, True,  5,   [2, 3]],
          [True,  False, 4,   [1, 2, 3]],
          [False, False, 100, []])
    @unpack
    def test_notify_min_faults(self, notify_cpu, notify_memory, notify_min_faults, result_uids):
        self.period = 60
        period_to = int(time.time())
        test_data = [
            dict(uid=0, _time=period_to - 10),
            dict(uid=1, _time=period_to - 10, cpu_fault=5,  mem_fault=1),
            dict(uid=2, _time=period_to - 10, cpu_fault=11, mem_fault=5),
            dict(uid=3, _time=period_to - 10, cpu_fault=4,  mem_fault=7)]
        for kwargs in test_data:
            self.save_history_line(server_id=self.cfg['server_id'], **kwargs)

        notify_faults_types = self._get_notify_faults_types(
            **{'notify_cpu': notify_cpu,
               'notify_memory': notify_memory}
        )
        users_stat_ret = self.statsnotifier.get_users_fault(self.period, notify_min_faults, notify_faults_types)
        self.assertCountEqual(result_uids, [user['ID'] for user in users_stat_ret])

    RESELLERS = {
        60001: 'res1',
        60002: 'res2',
        60003: 'res3',
        60004: 'res4',
        60005: None}

    @data(
        dict(with_admin=False,
             expected=[(60001, 'res1_settings'), (60002, 'default_settings'),
                       (60003, 'res3_settings'), (60004, 'default_settings')]),
        dict(with_admin=True,
             expected=[(60001, 'res1_settings'), (60002, 'default_settings'),
                       (60003, 'res3_settings'), (60004, 'default_settings'), (0, 'admin_settings')]))
    @unpack
    def test_get_activated_resellers_settings(self, with_admin, expected):
        def mock_get_reseller_settings(reseller):
            # we do not need real settings here, just make sure that them are returned
            if reseller in {'res4', 'res2'}:
                return None
            return '{reseller}_settings'.format(reseller=reseller)

        with \
                mock.patch.object(
                    self.statsnotifier._proc_lve, 'resellers_supported', return_value=True), \
                mock.patch.object(
                    self.statsnotifier._proc_lve, 'lvp_id_list', return_value=list(self.RESELLERS.keys())), \
                mock.patch.object(
                    self.statsnotifier._name_map, 'get_name', side_effect=lambda i: self.RESELLERS[i]), \
                mock.patch.object(
                    self.statsnotifier, '_default_reseller_settings', 'default_settings'), \
                mock.patch.object(
                    self.statsnotifier, '_admin_settings', 'admin_settings'), \
                mock.patch.object(
                    self.statsnotifier, 'get_reseller_settings', side_effect=mock_get_reseller_settings):
            resellers = self.statsnotifier._get_resellers_settings(with_admin=with_admin)
            self.assertCountEqual(resellers, expected)

    def test_get_default_reseller_settings(self):
        admin_settings = mock.Mock(spec_set=AdminSettings)
        admin_settings.NOTIFY_FAULTS_TYPES = mock.Mock(spec_set=NotifyFaultsOptions)
        admin_settings.NOTIFY_OPTIONS_SELF = mock.Mock(spec_set=NotifySettings)
        admin_settings.NOTIFY_OPTIONS_CUSTOMER = mock.Mock(spec_set=NotifySettings)
        admin_settings.NOTIFY_ADMIN = mock.Mock(spec_set=bool)
        admin_settings.NOTIFY_CUSTOMERS_ON_FAULTS = mock.Mock(spec_set=bool)
        admin_settings.NOTIFY_RESELLER_CUSTOMERS = mock.Mock(spec_set=bool)
        admin_settings.NOTIFY_RESELLER_ON_CUSTOMERS_FAULTS = mock.Mock(spec_set=bool)

        with mock.patch.object(self.statsnotifier, '_admin_settings', admin_settings):
            reseller_settings = self.statsnotifier.get_default_reseller_settings()

        self.assertEqual(reseller_settings.NOTIFY_FAULTS_TYPES, admin_settings.NOTIFY_FAULTS_TYPES)
        self.assertEqual(reseller_settings.NOTIFY_OPTIONS_SELF, admin_settings.NOTIFY_OPTIONS_SELF)
        self.assertEqual(reseller_settings.NOTIFY_OPTIONS_CUSTOMER, admin_settings.NOTIFY_OPTIONS_CUSTOMER)
        self.assertEqual(reseller_settings.NOTIFY_RESELLER_ON_CUSTOMERS_FAULTS, admin_settings.NOTIFY_RESELLER_ON_CUSTOMERS_FAULTS)
        self.assertEqual(reseller_settings.NOTIFY_CUSTOMERS_ON_FAULTS, admin_settings.NOTIFY_RESELLER_CUSTOMERS)
        self.assertEqual(reseller_settings.NOTIFY_RESELLER_ON_TOTAL_FAULTS, admin_settings.NOTIFY_RESELLER_ON_CUSTOMERS_FAULTS)

    def test_get_resellers_options_filtered(self):
        resellers_settings_1 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings_1.NOTIFY_RESELLER_ON_TOTAL_FAULTS = True
        resellers_settings_1.NOTIFY_OPTIONS_SELF.NOTIFY_INTERVAL = 120

        resellers_settings_2 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings_2.NOTIFY_RESELLER_ON_TOTAL_FAULTS = False
        resellers_settings_2.NOTIFY_OPTIONS_SELF.NOTIFY_INTERVAL = 200

        resellers_settings_3 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings_3.NOTIFY_RESELLER_ON_TOTAL_FAULTS = False
        resellers_settings_3.NOTIFY_OPTIONS_SELF.NOTIFY_INTERVAL = 500

        resellers = [
            (1000, resellers_settings_1),
            (1001, resellers_settings_2),
            (1002, resellers_settings_3)]

        resellers_after_filter = [
            (1000, resellers_settings_1)]

        with mock.patch.object(self.statsnotifier, '_get_resellers_settings', return_value=resellers):
            # let's emulate reseller_need_notification and remove resellers
            # with interval more than 200
            with mock.patch.object(
                    self.statsnotifier._notifications, 'reseller_need_notification',
                    side_effect=lambda rid, interval: interval < 200):
                resellers_filtered = list(self.statsnotifier._get_resellers_options_filtered())
                self.assertCountEqual(resellers_after_filter, resellers_filtered)

    def test_get_customers_options_filtered(self):
        resellers_settings_1 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings_1.NOTIFY_CUSTOMERS_ON_FAULTS = True
        resellers_settings_1.NOTIFY_OPTIONS_CUSTOMER.NOTIFY_INTERVAL = 120

        resellers_settings_2 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings_2.NOTIFY_CUSTOMERS_ON_FAULTS = False
        resellers_settings_2.NOTIFY_OPTIONS_CUSTOMER.NOTIFY_INTERVAL = 120

        resellers_settings_3 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings_3.NOTIFY_CUSTOMERS_ON_FAULTS = True
        resellers_settings_3.NOTIFY_OPTIONS_CUSTOMER.NOTIFY_INTERVAL = 500

        resellers = [
            (1000, resellers_settings_1),
            (1001, resellers_settings_2),
            (1002, resellers_settings_3)]

        resellers_after_filter = [(1000, resellers_settings_1)]

        with mock.patch.object(self.statsnotifier, '_get_resellers_settings', return_value=resellers):
            # let's emulate reseller_need_notification and remove resellers
            # with interval more than 200
            with mock.patch.object(
                    self.statsnotifier._notifications, 'users_need_notification',
                    side_effect=lambda rid, interval: interval < 200):
                resellers_filtered = list(self.statsnotifier._get_customers_options_filtered())
                self.assertCountEqual(resellers_after_filter, resellers_filtered)

    def test_get_reseller_customers_faults_options_filtered(self):
        resellers_settings_1 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings_1.NOTIFY_RESELLER_ON_CUSTOMERS_FAULTS = True
        resellers_settings_1.NOTIFY_OPTIONS_SELF.NOTIFY_INTERVAL = 120

        resellers_settings_2 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings_2.NOTIFY_RESELLER_ON_CUSTOMERS_FAULTS = False
        resellers_settings_2.NOTIFY_OPTIONS_SELF.NOTIFY_INTERVAL = 120

        resellers_settings_3 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings_3.NOTIFY_RESELLER_ON_CUSTOMERS_FAULTS = True
        resellers_settings_3.NOTIFY_OPTIONS_SELF.NOTIFY_INTERVAL = 500

        resellers = [
            (1000, resellers_settings_1),
            (1001, resellers_settings_2),
            (1002, resellers_settings_3)]

        resellers_after_filter = [(1000, resellers_settings_1)]

        with mock.patch.object(self.statsnotifier, '_get_resellers_settings', return_value=resellers):
            # let's emulate reseller_need_notification and remove resellers
            # with interval more than 200
            with mock.patch.object(
                    self.statsnotifier._notifications, 'reseller_need_notification',
                    side_effect=lambda rid, interval: interval < 200):
                resellers_filtered = list(self.statsnotifier._get_reseller_customers_faults_options_filtered())
                self.assertCountEqual(resellers_after_filter, resellers_filtered)

    def test_get_resellers_options_grouped(self):
        notify_settings_1 = NotifySettings(
            NOTIFY_MIN_FAULTS=1,
            NOTIFY_INTERVAL=120)

        notify_settings_2 = NotifySettings(
            NOTIFY_MIN_FAULTS=2,
            NOTIFY_INTERVAL=120)

        notify_faults_options_1 = NotifyFaultsOptions(
            NOTIFY_CPU=True,
            NOTIFY_IO=True,
            NOTIFY_IOPS=True,
            NOTIFY_MEMORY=True,
            NOTIFY_EP=True,
            NOTIFY_NPROC=True)

        notify_faults_options_2 = NotifyFaultsOptions(
            NOTIFY_CPU=True,
            NOTIFY_IO=True,
            NOTIFY_IOPS=True,
            NOTIFY_MEMORY=False,
            NOTIFY_EP=True,
            NOTIFY_NPROC=True)

        resellers_settings_1 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings_1.NOTIFY_OPTIONS_SELF = notify_settings_1
        resellers_settings_1.NOTIFY_FAULTS_TYPES = notify_faults_options_1

        resellers_settings_2 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings_2.NOTIFY_OPTIONS_SELF = notify_settings_2
        resellers_settings_2.NOTIFY_FAULTS_TYPES = notify_faults_options_1

        resellers_settings_3 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings_3.NOTIFY_OPTIONS_SELF = notify_settings_2
        resellers_settings_3.NOTIFY_FAULTS_TYPES = notify_faults_options_2

        resellers = [
            ('res1', resellers_settings_1),
            ('res2', resellers_settings_1),
            ('res3', resellers_settings_2),
            ('res4', resellers_settings_3),
            ('res5', resellers_settings_3)]

        resellers_after_grouping = [
            (
                (notify_settings_1, notify_faults_options_1),
                [
                    ('res1', resellers_settings_1),
                    ('res2', resellers_settings_1),
                ]
            ),
            (
                (notify_settings_2, notify_faults_options_1),
                [
                    ('res3', resellers_settings_2),
                ]
            ),
            (
                (notify_settings_2, notify_faults_options_2),
                [
                    ('res4', resellers_settings_3),
                    ('res5', resellers_settings_3)
                ]
            ),
        ]

        with mock.patch.object(
                self.statsnotifier, '_get_resellers_options_filtered', return_value=resellers):

            result = []
            # open all generators so I can compare results
            for settings, grouped in self.statsnotifier._get_resellers_options_grouped():
                result.append((settings, list(grouped)))
            self.assertCountEqual(resellers_after_grouping, result)

    def test_get_customers_options_grouped(self):
        notify_settings_1 = NotifySettings(
            NOTIFY_MIN_FAULTS=1,
            NOTIFY_INTERVAL=120)

        notify_settings_2 = NotifySettings(
            NOTIFY_MIN_FAULTS=2,
            NOTIFY_INTERVAL=120)

        notify_faults_options_1 = NotifyFaultsOptions(
            NOTIFY_CPU=True,
            NOTIFY_IO=True,
            NOTIFY_IOPS=True,
            NOTIFY_MEMORY=True,
            NOTIFY_EP=True,
            NOTIFY_NPROC=True)

        notify_faults_options_2 = NotifyFaultsOptions(
            NOTIFY_CPU=True,
            NOTIFY_IO=True,
            NOTIFY_IOPS=True,
            NOTIFY_MEMORY=False,
            NOTIFY_EP=True,
            NOTIFY_NPROC=True)

        resellers_settings_1 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings_1.NOTIFY_OPTIONS_CUSTOMER = notify_settings_1
        resellers_settings_1.NOTIFY_FAULTS_TYPES = notify_faults_options_1

        resellers_settings_2 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings_2.NOTIFY_OPTIONS_CUSTOMER = notify_settings_2
        resellers_settings_2.NOTIFY_FAULTS_TYPES = notify_faults_options_1

        resellers_settings_3 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings_3.NOTIFY_OPTIONS_CUSTOMER = notify_settings_2
        resellers_settings_3.NOTIFY_FAULTS_TYPES = notify_faults_options_2

        resellers = [
            ('res1', resellers_settings_1),
            ('res2', resellers_settings_1),
            ('res3', resellers_settings_2),
            ('res4', resellers_settings_3),
            ('res5', resellers_settings_3)
        ]

        resellers_after_grouping = [
            (
                (notify_settings_1, notify_faults_options_1),
                [
                    ('res1', resellers_settings_1),
                    ('res2', resellers_settings_1),
                ]
            ),
            (
                (notify_settings_2, notify_faults_options_1),
                [
                    ('res3', resellers_settings_2),
                ]
            ),
            (
                (notify_settings_2, notify_faults_options_2),
                [
                    ('res4', resellers_settings_3),
                    ('res5', resellers_settings_3)
                ]
            ),
        ]

        with mock.patch.object(
                self.statsnotifier, '_get_customers_options_filtered', return_value=resellers):

            result = []
            # open all generators so I can compare results
            for settings, grouped in self.statsnotifier._get_customers_options_grouped():
                result.append((settings, list(grouped)))
            self.assertCountEqual(resellers_after_grouping, result)

    def test_iter_resellers_faults_data(self):
        notify_settings = NotifySettings(
            NOTIFY_MIN_FAULTS=1,
            NOTIFY_INTERVAL=120)

        notify_faults_options = NotifyFaultsOptions(
            NOTIFY_CPU=True,
            NOTIFY_IO=True,
            NOTIFY_IOPS=True,
            NOTIFY_MEMORY=True,
            NOTIFY_EP=True,
            NOTIFY_NPROC=True)

        resellers_settings = mock.Mock(spec_set=ResellerSettings)
        resellers_settings.NOTIFY_OPTIONS_CUSTOMER = notify_settings
        resellers_settings.NOTIFY_FAULTS_TYPES = notify_faults_options

        resellers_after_grouping = [
            (
                (notify_settings, notify_faults_options),
                [
                    (1001, resellers_settings),
                    (1002, resellers_settings),
                ]
            ),
            (
                (notify_settings, notify_faults_options),
                [
                    (1003, resellers_settings),
                ]
            ),
        ]

        with \
                mock.patch.object(self.statsnotifier, '_get_resellers_options_grouped',
                                  return_value=resellers_after_grouping), \
                mock.patch.object(self.statsnotifier, 'get_users_fault') as mock_get_users_fault, \
                mock.patch.object(self.statsnotifier, '_get_resellers_data'):
            resellers_faults = list(map(list, self.statsnotifier._iter_resellers_faults_data()))
            uids = [reseller_ids for _, reseller_ids in resellers_faults]

            self.assertCountEqual(uids, [[1001, 1002], [1003]])

            mock_get_users_fault.assert_has_calls([
                mock.call(period=120,
                          notify_min_faults=1,
                          notify_faults_types=notify_faults_options,
                          uids=[1000001001, 1000001002]),
                mock.call(period=120,
                          notify_min_faults=1,
                          notify_faults_types=notify_faults_options,
                          uids=[1000001003])])

    def test_iter_customers_faults_data(self):
        notify_settings = NotifySettings(
            NOTIFY_MIN_FAULTS=1,
            NOTIFY_INTERVAL=200)

        notify_settings1 = NotifySettings(
            NOTIFY_MIN_FAULTS=3,
            NOTIFY_INTERVAL=300)

        notify_faults_options = NotifyFaultsOptions(
            NOTIFY_CPU=False,
            NOTIFY_IO=True,
            NOTIFY_IOPS=True,
            NOTIFY_MEMORY=True,
            NOTIFY_EP=True,
            NOTIFY_NPROC=True)

        notify_faults_options1 = NotifyFaultsOptions(
            NOTIFY_CPU=True,
            NOTIFY_IO=True,
            NOTIFY_IOPS=True,
            NOTIFY_MEMORY=True,
            NOTIFY_EP=True,
            NOTIFY_NPROC=True)

        resellers_settings = mock.Mock(spec_set=ResellerSettings)
        resellers_settings.NOTIFY_OPTIONS_CUSTOMER = notify_settings
        resellers_settings.NOTIFY_FAULTS_TYPES = notify_faults_options

        resellers_settings1 = mock.Mock(spec_set=ResellerSettings)
        resellers_settings1.NOTIFY_OPTIONS_CUSTOMER = notify_settings1
        resellers_settings1.NOTIFY_FAULTS_TYPES = notify_faults_options1

        resellers_after_grouping = [
            (
                (notify_settings, notify_faults_options),
                [
                    (1001, resellers_settings),
                    (1002, resellers_settings),
                    (0,    resellers_settings)
                ]
            ),
            (
                (notify_settings1, notify_faults_options1),
                [
                    (1003, resellers_settings1),
                ]
            ),
        ]

        def mock_lve_id_list(lvp_id=0):
            if lvp_id == 0:
                return [10005]
            return [lvp_id, lvp_id * 1000]

        with \
                mock.patch.object(self.statsnotifier._proc_lve, 'lve_id_list', side_effect=mock_lve_id_list), \
                mock.patch.object(self.statsnotifier, '_get_customers_options_grouped', return_value=resellers_after_grouping), \
                mock.patch.object(self.statsnotifier._notifications, 'mark_users_notified') as mock_mark_users_notified, \
                mock.patch.object(self.statsnotifier._notifications, 'users_need_notification', return_value=True), \
                mock.patch.object(self.statsnotifier, 'get_users_fault') as mock_get_users_fault, \
                mock.patch.object(self.statsnotifier, 'get_users_data'):
            list(map(list, self.statsnotifier._iter_customers_faults_data()))

            mark_users_notified_calls = [
                mock.call([1001, 1002, 0]),
                mock.call([1003])]

            get_users_faults_calls = [
                mock.call(period=200,
                          notify_min_faults=1,
                          notify_faults_types=notify_faults_options,
                          uids=[1001, 1001000, 1002, 1002000, 10005]),
                mock.call(period=300,
                          notify_min_faults=3,
                          notify_faults_types=notify_faults_options1,
                          uids=[1003, 1003000])]

            mock_mark_users_notified.assert_has_calls(mark_users_notified_calls)
            mock_get_users_fault.assert_has_calls(get_users_faults_calls)

    def test_get_resellers_data(self):
        faults_data = [
            {'ID': 1000001001},
            {'ID': 1000001002},
            {'ID': 1000001003}]

        result = [
            {'ID': 1000001001, 'LOGIN': '1001', 'TOMAIL': '1001@local.com', 'LOCALE': 'en_US'},
            {'ID': 1000001002, 'LOGIN': '1002', 'TOMAIL': '1002@local.com', 'LOCALE': 'en_US'},
            {'ID': 1000001003, 'LOGIN': '1003', 'TOMAIL': '1003@local.com', 'LOCALE': 'en_US'}]

        def mock_cpapi(username, keyls=('mail', 'locale'), search_sys_users=True):
            self.assertFalse(search_sys_users)
            return [['{}@local.com'.format(username), 'en']]

        def mock_get_name(reseller_id):
            return str(reseller_id)

        with mock.patch.object(self.statsnotifier._name_map, 'get_name', side_effect=mock_get_name):
            with mock.patch('lvestats.plugins.generic.statsnotifier.cpapi.cpinfo', side_effect=mock_cpapi):
                info = self.statsnotifier._get_resellers_data(faults_data)

                self.assertCountEqual(info, result)

    def test_execute(self):
        with mock.patch.object(self.statsnotifier._name_map, 'link_xml_node') as mock_link_xml_node:
            with mock.patch.object(self.statsnotifier, '_check_users') as mock_check_users:
                with mock.patch.object(self.statsnotifier, '_check_resellers') as mock_check_resellers:
                    with mock.patch.object(self.statsnotifier, '_check_admin') as mock_check_admin:
                        self.statsnotifier.execute({})

                        mock_link_xml_node.assert_called_with(use_cache=False)
                        mock_check_users.assert_called()
                        mock_check_resellers.assert_called()
                        mock_check_admin.assert_called()

    @data([True,  True,  True],
          [True,  False, False],
          [False, False, False],
          [False, True,  False])
    @unpack
    def test_check_admin_mark_as_notified(self, notify_admin, admin_need_notification, mark_admin_notified):
        with \
                self.mock_admin_settings(notify_admin=notify_admin), \
                mock.patch.object(self.statsnotifier._notifications, 'admin_need_notification',
                                  return_value=admin_need_notification), \
                mock.patch.object(self.statsnotifier._proc_lve, 'lve_id_list', return_value=[]), \
                mock.patch.object(self.statsnotifier._notifications,
                                  'mark_admin_notified') as mock_mark_admin_notified:
            self.statsnotifier._check_admin()
            self.assertEqual(mock_mark_admin_notified.called, mark_admin_notified)

    @data(
        [
            dict(notify_reseller_on_customers_faults=NO_SETTINGS),
            ["CPU Limit"], # Moved EP Limit, now we do not show columns without faults
            ["EP Limit", "VMem Limit", "PMem Limit", "NPROC Limit", "IO Limit"],
            ["CPU Limit, %", "CPU Faults"],
            "lCPU", "CPUf"],
        [
            dict(notify_cpu=True,
                 notify_mem=False,
                 notify_nproc=False,
                 notify_iops=False,
                 notify_io=False,
                 notify_ep=True),
            ["CPU Limit"],
            ["EP Limit", "VMem Limit", "PMem Limit", "NPROC Limit", "IO Limit"],
            ["CPU Limit, %", "CPU Faults"],
            "lCPU", "CPUf"],
        [
            dict(notify_cpu=False,
                 notify_mem=True,
                 notify_nproc=True,
                 notify_iops=True,
                 notify_io=True,
                 notify_ep=False),
            ["VMem Limit"],
            ["EP Limit", "CPU Limit", "PMem Limit", "NPROC Limit", "IO Limit"],
            ["VMem Limit, KB", "VMem Faults"],
            "lVMem", "VMemF"])
    @unpack
    def test_generate_table_data(self, reseller_settings, assert_in, assert_not_in, table_headers,
                                 limit_key, fault_key):
        def mock_reseller_data(reseller_login):
            return {'TOMAIL': "{}@local.com".format(reseller_login), 'LOCALE': 'en', 'RESELLER_USERNAME': reseller_login}

        stats_html_template = """<table frame="box" rules="all">
    <thead>
        <tr>
            <th style="padding-left: 1em; padding-right: 1em; text-align: center">Username</th>
            <th style="padding-left: 1em; padding-right: 1em; text-align: center">Domain</th>
            <th style="padding-left: 1em; padding-right: 1em; text-align: center">Name</th>
            <th style="padding-left: 1em; padding-right: 1em; text-align: center">{}</th>
            <th style="padding-left: 1em; padding-right: 1em; text-align: center">{}</th>
        </tr>
    </thead>
    <tbody>
        <tr>
            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">john777</td>
            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">john.com</td>
            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">---</td>
            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">{}</td>
            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">{}</td>
        </tr>
        <tr>
            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">robot</td>
            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">whoismrrobot.com</td>
            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">Elliot Alderson</td>
            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">{}</td>
            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">{}</td>
        </tr>
    </tbody>
</table>"""
        expected_html = stats_html_template.format(*table_headers,
                                                   testcases_data.table_test_case3.input[1][limit_key],
                                                   testcases_data.table_test_case3.input[1][fault_key],
                                                   testcases_data.table_test_case3.input[2][limit_key],
                                                   testcases_data.table_test_case3.input[2][fault_key])
        self.statsnotifier.resellers = {'reseller1'}
        with \
                self.mock_admin_settings(notify_admin=True, notify_cpu=True, notify_mem=False, notify_nproc=False,
                                         notify_iops=False, notify_io=False, notify_ep=True), \
                mock.patch('lvestats.plugins.generic.statsnotifier.get_notification',
                           return_value=testcases_data.get_resellers_settings(**reseller_settings)),\
                mock.patch('lvestats.plugins.generic.statsnotifier.StatsNotifier.get_panel_reseller_data',
                           side_effect=mock_reseller_data):
            result = self.statsnotifier.prepare_resellers_data(testcases_data.table_test_case3.input[:])[0]
            plain_stats = result["STATS"]
            html_stats = result["STATS_HTML"]
        for test_value in assert_in:
            self.assertIn(test_value, plain_stats)
        for test_value in assert_not_in:
            self.assertNotIn(test_value, plain_stats)
        self.assertEqual(expected_html, html_stats)

    @data(
        [
            dict(notify_cpu=True, notify_mem=True, notify_nproc=True, notify_iops=True, notify_io=True, notify_ep=True),
            ["CPU Limit", "EP Limit", "VMem Limit", "PMem Limit", "NPROC Limit", "IO Limit", "IOPS Limit"],
            []])
    @unpack
    def test_generate_table_data_skip_limits_without_faults(self, reseller_settings, assert_in, assert_not_in):
        def mock_reseller_data(reseller_login):
            return {'TOMAIL': "{}@local.com".format(reseller_login), 'LOCALE': 'en'}
        self.statsnotifier.resellers = {'reseller1'}
        self.statsnotifier.lve_version = 7
        with \
                mock.patch('lvestats.plugins.generic.statsnotifier.get_notification',
                           return_value=testcases_data.get_resellers_settings(**reseller_settings)),\
                mock.patch('lvestats.plugins.generic.statsnotifier.StatsNotifier.get_panel_reseller_data', side_effect=mock_reseller_data):
            result = self.statsnotifier.prepare_resellers_data(testcases_data.table_test_case_all_limits_faults.input[:])[0]["STATS"]
        for test_value in assert_in:
            self.assertIn(test_value, result)
        for test_value in assert_not_in:
            self.assertNotIn(test_value, result)

    @data(
        ["admin@local.com", "reseller@local.com", "reseller", "admin@local.com"],
        ["admin@local.com", "someresl@local.com", "someresl", "someresl@local.com"]
    )
    @unpack
    def test_get_notify_from_mail(self, admin_notify_from_email, reseller_email, reseller, result):
        self.statsnotifier.resellers = {"someresl"}

        def mock_cpinfo(cpuser, keyls, search_sys_users):
            assert keyls == ('mail',)
            return [(reseller_email,)]

        with \
                self.mock_admin_settings(notify_from_email=admin_notify_from_email), \
                mock.patch('lvestats.plugins.generic.statsnotifier.cpapi.cpinfo', side_effect=mock_cpinfo):
            email = self.statsnotifier._get_notify_from_mail(reseller)
            self.assertEqual(email, result)

    @data(
        [0, 0, StatsNotifier.DEFAULT_PERIOD, StatsNotifier.DEFAULT_PERIOD],
        [-1, 100, StatsNotifier.DEFAULT_PERIOD, StatsNotifier.DEFAULT_PERIOD],
        [1, None, StatsNotifier.DEFAULT_PERIOD, StatsNotifier.DEFAULT_PERIOD],
        [1, False, StatsNotifier.DEFAULT_PERIOD, StatsNotifier.DEFAULT_PERIOD],
        [1, 30, 60*60, 30*60*60])
    @unpack
    def test_get_admin_settings(self, test_data_admin, test_data_user, result_admin, result_user):
        with mock.patch('lvestats.plugins.generic.statsnotifier.get_admin_notification',
                        return_value=testcases_data.get_admin_settings(
                            notify_admin_period=test_data_admin, notify_user_period=test_data_user)):
            settings = self.statsnotifier.get_admin_settings()
        self.assertEqual(settings.NOTIFY_OPTIONS_SELF.NOTIFY_INTERVAL, result_admin)
        self.assertEqual(settings.NOTIFY_OPTIONS_CUSTOMER.NOTIFY_INTERVAL, result_user)

    @data(
        [0, 0, StatsNotifier.DEFAULT_PERIOD, StatsNotifier.DEFAULT_PERIOD],
        [-1, 100, StatsNotifier.DEFAULT_PERIOD, StatsNotifier.DEFAULT_PERIOD],
        [1, None, StatsNotifier.DEFAULT_PERIOD, StatsNotifier.DEFAULT_PERIOD],
        [1, False, StatsNotifier.DEFAULT_PERIOD, StatsNotifier.DEFAULT_PERIOD],
        [1, 30, 60*60, 30*60*60])
    @unpack
    def test_get_reseller_settings(self, test_data_reseller, test_data_customer, result_reseller, result_customer):
        with mock.patch('lvestats.plugins.generic.statsnotifier.get_notification',
                        return_value=testcases_data.get_resellers_settings(
                            notify_reseller_period=test_data_reseller, notify_customer_period=test_data_customer)):
            settings = self.statsnotifier.get_reseller_settings("res1")
        self.assertEqual(settings.NOTIFY_OPTIONS_SELF.NOTIFY_INTERVAL, result_reseller)
        self.assertEqual(settings.NOTIFY_OPTIONS_CUSTOMER.NOTIFY_INTERVAL, result_customer)


if __name__ == '__main__':
    unittest.main()
