# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import unittest

import mock
from lvestats.lib.commons.litespeed import LiteSpeedException, LiteSpeedInvalidCredentials

from lvestats.plugins.generic.snapshot_saver import SnapshotHelper, LitespeedHelper


def mock_get_names(uid):
    try:
        return {500: ['user500'], 400: ['user400'], 300: ['user300']}[uid]
    except KeyError:
        return []


def mock_return_query(users):
    if 'user500' in users:
        return {"user500": ['fake query goes here']}
    return {}


def mock_proctitle(user):
    if user == "user500":
        return ['URL500:1', 'URL500:2']
    if user == "user400":
        return ['URL400']
    return []


def mock_litespeed_user_data(user):
    if user == "user500":
        return ['URL500:1_litespeed', 'URL500:2_litespeed']
    return []


mock_get_lveps_value = {500: {'TID': {600: {'CPU': '0%', 'MEM': '1', 'CMD': 'bash', 'PID': '600'},
                                      601: {'CPU': '0%', 'MEM': '3', 'CMD': 'python', 'PID': '601'}}},
                        400: {'TID': {700: {'CPU': '0%', 'MEM': '1', 'CMD': 'bash', 'PID': '700'}}}
                        }


class TestSnapshotHolder(unittest.TestCase):
    def setUp(self):
        self.helper = SnapshotHelper()
        self.litespeed = LitespeedHelper()

    @mock.patch("lvestats.plugins.generic.snapshot_saver.SnapshotHelper.get_names", side_effect=mock_get_names)
    @mock.patch("lvestats.plugins.generic.snapshot_saver.SnapshotHelper.retrieve_queries",
                side_effect=mock_return_query)
    @mock.patch('lvestats.lib.commons.proctitle.Proctitle.get_user_data', side_effect=mock_proctitle)
    @mock.patch('lvestats.plugins.generic.snapshot_saver.get_lveps', return_value=mock_get_lveps_value)
    def test_invalidate(self, *args):
        self.helper.invalidate([400, 500, 600])
        self.assertEqual(self.helper.username_dbquery_map, {'user500': ['fake query goes here']})
        self.assertEqual(self.helper.ps, mock_get_lveps_value)
        self.assertIsNotNone(self.helper.clpwd)
        self.assertIsNotNone(self.helper.username_dbquery_map)
        self.assertIsNotNone(self.helper.ps)
        # lets try to run init again, and see if that will re-initialize fields
        old_clpwd = self.helper.clpwd
        self.helper.invalidate([400, 500, 600])
        self.assertIsNot(old_clpwd, self.helper.clpwd)

    @mock.patch("lvestats.plugins.generic.snapshot_saver.SnapshotHelper.get_names", side_effect=mock_get_names)
    @mock.patch("lvestats.plugins.generic.snapshot_saver.SnapshotHelper.retrieve_queries",
                side_effect=mock_return_query)
    @mock.patch('lvestats.lib.commons.proctitle.Proctitle.get_user_data', side_effect=mock_proctitle)
    @mock.patch('lvestats.plugins.generic.snapshot_saver.get_lveps', return_value=mock_get_lveps_value)
    @mock.patch("lvestats.plugins.generic.snapshot_saver.LitespeedHelper.get_use_litespeed", return_value=False)
    def test_get_snapshot_data(self, *args):
        self.helper.invalidate([400, 500, 600])
        data = self.helper.get_snapshot_data(500, self.litespeed)
        self.assertEqual(data, ({600: {'CPU': '0%', 'MEM': '1', 'CMD': 'bash', 'PID': '600'},
                                 601: {'CPU': '0%', 'MEM': '3', 'CMD': 'python', 'PID': '601'}},
                                ['fake query goes here'], ['URL500:1', 'URL500:2']))
        data = self.helper.get_snapshot_data(400, self.litespeed)
        self.assertEqual(data, ({700: {'MEM': '1', 'CMD': 'bash', 'PID': '700', 'CPU': '0%'}}, [], ['URL400']))
        self.assertEqual(self.helper.get_snapshot_data(600, self.litespeed), ({}, [], []))
        self.assertEqual(self.helper.get_snapshot_data(700, self.litespeed), ({}, [], []))

    @mock.patch("lvestats.plugins.generic.snapshot_saver.SnapshotHelper.get_names", side_effect=mock_get_names)
    @mock.patch("lvestats.plugins.generic.snapshot_saver.LitespeedHelper.get_use_litespeed", return_value=True)
    @mock.patch('lvestats.lib.commons.proctitle.Proctitle.get_user_data', side_effect=mock_proctitle)
    @mock.patch('lvestats.lib.commons.litespeed.LiteSpeed.get_user_data', side_effect=mock_litespeed_user_data)
    @mock.patch('lvestats.plugins.generic.snapshot_saver.get_lveps', return_value=mock_get_lveps_value)
    def test_get_snapshot_data_litespeed(self, *args):
        """
        Litespeed works ok.
        """
        litespeed_info = LitespeedHelper()
        self.helper.invalidate([400, 500, 600])

        data = self.helper.get_snapshot_data(500, litespeed_info)
        self.assertEqual(data, ({600: {'CPU': '0%', 'MEM': '1', 'CMD': 'bash', 'PID': '600'},
                                 601: {'CPU': '0%', 'MEM': '3', 'CMD': 'python', 'PID': '601'}},
                                [], ['URL500:1_litespeed', 'URL500:2_litespeed']))
        data = self.helper.get_snapshot_data(400, litespeed_info)
        self.assertEqual(data, ({700: {'MEM': '1', 'CMD': 'bash', 'PID': '700', 'CPU': '0%'}}, [], []))
        self.assertEqual(self.helper.get_snapshot_data(600, self.litespeed), ({}, [], []))
        self.assertEqual(self.helper.get_snapshot_data(700, self.litespeed), ({}, [], []))

    @mock.patch("lvestats.plugins.generic.snapshot_saver.SnapshotHelper.get_names", side_effect=mock_get_names)
    @mock.patch("lvestats.plugins.generic.snapshot_saver.LitespeedHelper.get_use_litespeed", return_value=True)
    @mock.patch('lvestats.lib.commons.proctitle.Proctitle.get_user_data', side_effect=mock_proctitle)
    @mock.patch('lvestats.lib.commons.litespeed.LiteSpeed.get_user_data', side_effect=LiteSpeedException)
    @mock.patch('lvestats.plugins.generic.snapshot_saver.get_lveps', return_value=mock_get_lveps_value)
    def test_get_snapshot_data_litespeed_proctitle_fallback(self, *args):
        """
        litespeed raises an exception, get data from Proctitle
        """
        litespeed_info = LitespeedHelper()
        self.helper.invalidate([400, 500, 600])

        data = self.helper.get_snapshot_data(500, litespeed_info)
        self.assertEqual(data, ({600: {'CPU': '0%', 'MEM': '1', 'CMD': 'bash', 'PID': '600'},
                                 601: {'CPU': '0%', 'MEM': '3', 'CMD': 'python', 'PID': '601'}},
                                [], ['URL500:1', 'URL500:2']))
        data = self.helper.get_snapshot_data(400, litespeed_info)
        self.assertEqual(data, ({700: {'MEM': '1', 'CMD': 'bash', 'PID': '700', 'CPU': '0%'}}, [], ['URL400']))
        self.assertEqual(self.helper.get_snapshot_data(600, self.litespeed), ({}, [], []))
        self.assertEqual(self.helper.get_snapshot_data(700, self.litespeed), ({}, [], []))

    @mock.patch("lvestats.plugins.generic.snapshot_saver.SnapshotHelper.get_names", side_effect=mock_get_names)
    @mock.patch("lvestats.plugins.generic.snapshot_saver.LitespeedHelper.get_use_litespeed", return_value=True)
    @mock.patch('lvestats.lib.commons.proctitle.Proctitle.get_user_data', side_effect=mock_proctitle)
    @mock.patch('lvestats.lib.commons.litespeed.LiteSpeed.get_user_data', side_effect=LiteSpeedInvalidCredentials)
    @mock.patch('lvestats.plugins.generic.snapshot_saver.get_lveps', return_value=mock_get_lveps_value)
    def test_get_snapshot_data_litespeed_auth_fail_proctitle_fallback(self, *args):
        """
        litespeed auth fails, get data from Proctitle
        """
        with mock.patch("lvestats.plugins.generic.snapshot_saver.LitespeedHelper.dump_passwd") as mock_dump_passwd:
            litespeed_info = LitespeedHelper()
            self.helper.invalidate([400, 500, 600])

            data = self.helper.get_snapshot_data(500, litespeed_info)
            self.assertEqual(mock_dump_passwd.called, True)
            self.assertEqual(data, ({600: {'CPU': '0%', 'MEM': '1', 'CMD': 'bash', 'PID': '600'},
                                     601: {'CPU': '0%', 'MEM': '3', 'CMD': 'python', 'PID': '601'}},
                                    [], ['URL500:1', 'URL500:2']))
            data = self.helper.get_snapshot_data(400, litespeed_info)
            self.assertEqual(data, ({700: {'MEM': '1', 'CMD': 'bash', 'PID': '700', 'CPU': '0%'}}, [], ['URL400']))
            self.assertEqual(self.helper.get_snapshot_data(600, self.litespeed), ({}, [], []))
            self.assertEqual(self.helper.get_snapshot_data(700, self.litespeed), ({}, [], []))
