# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import datetime
import json
import os
import shutil
import time
import unittest
from io import StringIO

import ddt
import mock
from sqlalchemy import insert
import builtins

import test.test_lib_snapshot
from test.base.base_db_test import BaseDbTest
from lvestats.lib.commons import dateutil
from lvestats.lib.commons import proctitle
from lvestats.lib.snapshot import Snapshot
from lvestats.orm.incident import incident
from lvestats.plugins.generic import snapshot_saver
from lvestats.plugins.generic.analyzers import LVEUsage as LVEUsageOrig
from lvestats.plugins.generic.snapshot_saver import SnapshotSaver, SnapshotHelper, \
    LitespeedHelper, LiteSpeed

TEST_PATH = test.test_lib_snapshot.TEST_PATH


class LVEUsage(LVEUsageOrig):
    def __init__(self, **kwargs):
        super(LVEUsage, self).__init__(8)
        for (k, v) in kwargs.items():
            setattr(self, k, v)


EMPTY_LVE_DATA = {'lve_usage': {}}


class MockProctitle(proctitle.Proctitle):
    def __init__(self):
        shm_path = os.path.join(os.path.dirname(__file__), "data", "shm")
        super(MockProctitle, self).__init__(shm_path)


def get_all_user_domains_mock(username):
    user_dict = {
        "foouser": ["flaskbb.test247.cloudlinux.com"],
        "baruser": ["wordpress.test247.cloudlinux.com"]}
    return user_dict[username]


old_retrieve_queries = SnapshotHelper.retrieve_queries
old_get_names = SnapshotHelper.get_names
old_get_lveps = snapshot_saver.get_lveps

mock_retrieve_queries = {'foouser': [['mysql', '2015-01-19 11:44:00', 'SELECT 1;']]}


@ddt.ddt
class TestSnapshotSaver(BaseDbTest):
    @staticmethod
    def restore_methods():
        SnapshotHelper.get_names = old_get_names
        snapshot_saver.get_lveps = old_get_lveps

    @staticmethod
    def mock_methods():
        def mock_get_lveps():
            return {400: {'CPU': '27%',
                          'EP': '0',
                          'IO': '0',
                          'IOPS': 'N/A',
                          'MEM': '1',
                          'PNO': '3',
                          'TID': {1428: {'CPU': '25%',
                                         'IO': 'N/A',
                                         'IOPS': 'N/A',
                                         'MEM': '1',
                                         'PID': '4at'}},
                          'TNO': '3'},
                    500: {'CPU': '27%',
                          'EP': '0',
                          'IO': '0',
                          'IOPS': 'N/A',
                          'MEM': '1',
                          'PNO': '3',
                          'TID': {1429: {'CPU': '0%',
                                         'IO': 'N/A',
                                         'IOPS': 'N/A',
                                         'MEM': '1',
                                         'PID': 'su'}},
                          'TNO': '1'}}

        snapshot_saver.get_lveps = mock_get_lveps

        # noinspection PyUnusedLocal
        def mock_get_names(self, uid):
            try:
                return {500: ['foouser'], 400: ['baruser']}[uid]
            except KeyError:
                return []

        SnapshotHelper.get_names = mock_get_names

    def tearDown(self):
        self.patcher1.stop()
        if os.geteuid() != 0:
            self.patcher2.stop()

        self.restore_methods()
        self.plugin.session.close()
        BaseDbTest.tearDown(self)

        del os.environ['TZ']
        time.tzset()

    def setUp(self):
        os.environ['TZ'] = 'UTC'  # set the zero time zone
        time.tzset()

        BaseDbTest.setUp(self)
        shutil.rmtree(TEST_PATH, ignore_errors=True)
        self.mock_methods()
        self.lve_data = {
            'lve_usage': {
                500: LVEUsage(cpu_limit=2500, cpu_usage=2600, cpu_fault=1),
                400: LVEUsage(cpu_limit=2500)},
            'faults': {
                500: {"cpu_fault": 1}}}

        self.plugin = SnapshotSaver()
        self.plugin.now = 100
        self.plugin.server_id = 'localhost'
        self.plugin.set_db_engine(self.engine)
        self.plugin.init_session()

        self.patcher1 = mock.patch('lvestats.lib.snapshot.SNAPSHOT_PATH', TEST_PATH)
        self.patcher1.start()
        if os.geteuid() != 0:
            self.patcher2 = mock.patch('lvestats.lib.snapshot.drop_privileges', test.test_lib_snapshot.mock_drop_privileges)
            self.patcher2.start()

    @mock.patch.object(SnapshotHelper, "retrieve_queries", return_value=mock_retrieve_queries)
    def test_smoketest_snapshot(self, mock_retrieve_queries_result):
        """
        Simply runs all stuff with no checks
        """
        self.now = snapshot_saver.DEFAULT_PERIOD_BETWEEN_INCIDENTS
        with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
            self.plugin.execute(lve_data=self.lve_data)

    @mock.patch.object(SnapshotHelper, "retrieve_queries", return_value=mock_retrieve_queries)
    @mock.patch("lvestats.snapshots.reader.LVEReadSnaphot.open")
    def test_snapshot_reader(self, mock_open, mock_retrieve_queries_result):
        """
        Simply runs snapshot and reads it
        :return:
        """
        with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
            self.plugin.now = 100500
            self.plugin.execute(lve_data=self.lve_data)
            self.plugin.now += snapshot_saver.DEFAULT_PERIOD_BETWEEN_INCIDENTS + 1
            self.plugin.execute(lve_data=EMPTY_LVE_DATA)  # let the incident end

        from lvestats.snapshots.reader import LVEReadSnaphot
        start = dateutil.gm_to_local(datetime.datetime.utcfromtimestamp(100499))
        end = dateutil.gm_to_local(datetime.datetime.utcfromtimestamp(100501))
        pretty = LVEReadSnaphot(self.engine, start, end, 500, 'localhost', None, False)
        jsoned = LVEReadSnaphot(self.engine, start, end, 500, 'localhost', None, True)

        s = StringIO()
        mock_open.return_value = s
        pretty.run()
        self.assertTrue(len(s.getvalue()) > 0)

        s = StringIO()
        mock_open.return_value = s
        pretty.stats("1d")

        val = "Stats from %(from)s to %(to)s\n---\n\tfrom: %(from)s\n\tto: %(to)s\n\tincidents: " \
              "1\n\tsnapshots: 1\n\tduration: 0 sec.\n" % {'from': dateutil.ts_to_iso(100501 - 60 * 60 * 24),
                                                           'to': dateutil.ts_to_iso(100501)}
        self.assertEqual(s.getvalue(), val)

        s = StringIO()
        mock_open.return_value = s
        jsoned.run()
        self.assertTrue(len(s.getvalue()) > 0)

        d = json.loads(s.getvalue())
        self.assertEqual(1, len(d["data"]["snapshots"]))

    @mock.patch.object(SnapshotHelper, "retrieve_queries", return_value=mock_retrieve_queries)
    @mock.patch("time.time")
    @mock.patch("lvestats.snapshots.reader.LVEReadSnaphot.open")
    @mock.patch("lvestats.lib.commons.proctitle.Proctitle")
    @mock.patch('lvestats.plugins.generic.snapshot_saver.LitespeedHelper.get_use_litespeed', return_value=False)
    def test_snapshot_reader_http(self, mock_litespeed, mock_proctitle, mock_open, mock_time, mock_retrieve_queries_result):
        """
        Functional
        1) Save one snapshot with http data
        2) Read snapshot in JSON
        3) Check if http data is correct
        :param mock_proctitle:
        :param mock_open:
        :return:
        """
        mock_time.return_value = 1453981994.367
        mock_proctitle.return_value = MockProctitle()
        start_timestamp = 1453390913
        self.plugin.now = start_timestamp
        with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'), \
             mock.patch('lvestats.lib.commons.proctitle.get_all_user_domains', side_effect=get_all_user_domains_mock):
            self.plugin.execute(lve_data=self.lve_data)
            self.plugin.now += snapshot_saver.DEFAULT_PERIOD_BETWEEN_INCIDENTS + 1
            self.plugin.execute(lve_data=EMPTY_LVE_DATA)  # let the incident end

        from lvestats.snapshots.reader import LVEReadSnaphot
        start = dateutil.gm_to_local(datetime.datetime.utcfromtimestamp(start_timestamp - 1))
        end = dateutil.gm_to_local(datetime.datetime.utcfromtimestamp(start_timestamp + 1))
        jsoned = LVEReadSnaphot(self.engine, start, end, 500, 'localhost', None, True)
        s = StringIO()
        mock_open.return_value = s
        jsoned.run()
        self.assertEqual(len(json.loads(s.getvalue())["data"]["snapshots"][0]["snap_http"]), 3)
        self.assertEqual(
            sorted(json.loads(s.getvalue())["data"]["snapshots"][0]["snap_http"]),
            sorted([
                ['760686', 'flaskbb.test247.cloudlinux.com', "GET", "/category/1-test-category-1", "HTTP/1.0", "0.4"],
                ['760687', 'flaskbb.test247.cloudlinux.com', "GET", "/category/1-test-category-1", "HTTP/1.0", "3.4"],
                ['764193', 'flaskbb.test247.cloudlinux.com', "GET", "/category/1-test-category-1", "HTTP/1.0", "6.4"]]))

    def test_no_panel(self):
        self.plugin.now = snapshot_saver.DEFAULT_PERIOD_BETWEEN_INCIDENTS + 1
        try:
            with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
                self.plugin.execute(lve_data=self.lve_data)
        except Exception as e:
            self.fail(e)

    @mock.patch.object(SnapshotHelper, "retrieve_queries", return_value=mock_retrieve_queries)
    def test_max_snapshots_per_incident(self, mock_retrieve_queries_result):
        self.plugin.now = snapshot_saver.DEFAULT_PERIOD_BETWEEN_INCIDENTS + 2
        with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
            for _ in range(snapshot_saver.DEFAULT_MAX_SNAPSHOTS_PER_INCIDENT + 1000):
                self.plugin.execute(lve_data=self.lve_data)
                self.plugin.now += 5
                self.lve_data["faults"] = {500: {"cpu_fault": 2}}
        result = self.plugin.session.query(incident
                                           ).filter(incident.server_id == self.plugin.server_id,
                                                    incident.snapshot_count >
                                                    snapshot_saver.DEFAULT_MAX_SNAPSHOTS_PER_INCIDENT).all()
        self.assertEqual(len(result), 0)
        result = self.plugin.session.query(incident
                                           ).filter(incident.server_id == self.plugin.server_id,
                                                    ).all()
        self.assertEqual(result[0].snapshot_count, snapshot_saver.DEFAULT_MAX_SNAPSHOTS_PER_INCIDENT)

    def test_no_snapshot_if_no_incident(self):
        self.lve_data['lve_usage'][500].cpu_fault = 0
        self.lve_data["faults"] = {}
        with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
            for _ in range(snapshot_saver.DEFAULT_MAX_SNAPSHOTS_PER_INCIDENT + 10):
                self.plugin.execute(lve_data=self.lve_data)

        result = self.plugin.session.query(incident
                                           ).filter(incident.server_id == self.plugin.server_id,
                                                    ).all()
        self.assertEqual(len(result), 0)

    @mock.patch.object(SnapshotHelper, "retrieve_queries", return_value=mock_retrieve_queries)
    def test_distinguish_incidents(self, mock_retrieve_queries_result):
        incident_period = (snapshot_saver.DEFAULT_MAX_SNAPSHOTS_PER_INCIDENT + 10) * int(self.plugin._period / 5) + 1
        with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
            for _ in range(incident_period):
                self.plugin.execute(lve_data=self.lve_data)
                self.plugin.now += 5
                self.lve_data["faults"] = {500: {"cpu_fault": 20}}

        self.plugin.now += (snapshot_saver.DEFAULT_PERIOD_BETWEEN_INCIDENTS // self.plugin._period + 1) * self.plugin._period

        with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
            for _ in range(incident_period):
                self.plugin.execute(lve_data=self.lve_data)
                self.plugin.now += 5
                self.lve_data["faults"] = {500: {"io_fault": 1}}

            self.plugin.now += (snapshot_saver.DEFAULT_PERIOD_BETWEEN_INCIDENTS // self.plugin._period + 1) * self.plugin._period
            self.plugin.execute(lve_data=self.lve_data)
            self.lve_data["faults"] = {500: {"mep_fault": 1}}
            result = self.plugin.session.query(incident
                                               ).order_by(incident.incident_start_time
                                                          ).filter(incident.server_id == self.plugin.server_id,
                                                                   ).all()

        self.assertEqual(len(result), 3)
        self.assertEqual(result[0].snapshot_count, snapshot_saver.DEFAULT_MAX_SNAPSHOTS_PER_INCIDENT)
        self.assertEqual(result[1].incident_end_time,
                         result[1].incident_start_time + 5 * (incident_period - 1))
        self.assertIsNone(result[2].incident_end_time)

    @mock.patch.object(SnapshotHelper, "retrieve_queries")
    def test_get_snapshot_on_every_iteration(self, mock_retrieve_queries_result):
        with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
            self.plugin._period = 30
            self.plugin.now = 0
            mock_retrieve_queries_result.return_value = return_value = {'foouser': [['mysql', '2015-01-19 11:44:00', 'SELECT 1;']]}
            self.plugin.execute(lve_data=self.lve_data)
            self.assertNotIn(400, self.plugin._snapshots_data)
            self.assertEqual(self.plugin._snapshots_data[500][1], return_value["foouser"])
            result = self.plugin.session.query(incident
                                               ).order_by(incident.incident_start_time
                                                          ).filter(incident.server_id == self.plugin.server_id,
                                                                   ).all()
            self.assertNotEqual(self.lve_data["faults"], {})
            self.assertEqual(len(result), 0)

            self.plugin.now += 5
            mock_retrieve_queries_result.return_value = return_value = {'foouser': [['mysql', '2015-01-19 11:44:05', 'SELECT 1;']]}
            self.plugin.execute(lve_data=self.lve_data)
            self.assertNotIn(400, self.plugin._snapshots_data)
            self.assertNotEqual(self.plugin._snapshots_data[500][1], return_value["foouser"])
            result = self.plugin.session.query(incident
                                               ).order_by(incident.incident_start_time
                                                          ).filter(incident.server_id == self.plugin.server_id,
                                                                   ).all()
            self.assertNotEqual(self.lve_data["faults"], {})
            self.assertEqual(len(result), 0)

            self.plugin.now += 5
            self.lve_data["faults"] = {500: {"cpu_fault": 20, "io_fault": 1}, 400: {"mep_fault": 1}}
            mock_retrieve_queries_result.return_value = return_value = {
                'foouser': [['mysql', '2015-01-19 11:44:10', 'SELECT 1;']],
                'baruser': [['mysql', '2015-01-19 11:44:10', 'SELECT 1;']],
            }
            self.plugin.execute(lve_data=self.lve_data)
            self.assertEqual(self.plugin._snapshots_data[400][1], return_value["foouser"])
            self.assertNotEqual(self.plugin._snapshots_data[500][1], return_value["foouser"])
            result = self.plugin.session.query(incident
                                               ).order_by(incident.incident_start_time
                                                          ).filter(incident.server_id == self.plugin.server_id,
                                                                   ).all()
            self.assertNotEqual(self.lve_data["faults"], {})
            self.assertEqual(len(result), 0)

            self.plugin.now += 20
            self.plugin.execute(lve_data=self.lve_data)
            self.assertNotIn(400, self.plugin._snapshots_data)
            self.assertNotIn(500, self.plugin._snapshots_data)
            result = self.plugin.session.query(incident
                                               ).order_by(incident.incident_start_time
                                                          ).filter(incident.server_id == self.plugin.server_id,
                                                                   ).all()
            self.assertEqual(self.lve_data["faults"], {})
            self.assertEqual(len(result), 2)

    @mock.patch.object(SnapshotHelper, "retrieve_queries", return_value=mock_retrieve_queries)
    def test_not_splitting_incident(self, mock_retrieve_queries_result):
        self.plugin.now = 33333
        with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
            for _ in range((snapshot_saver.DEFAULT_MAX_SNAPSHOTS_PER_INCIDENT + 10) * int(self.plugin._period / 5)):
                self.plugin.execute(lve_data=self.lve_data)
                self.plugin.now += 5
                self.lve_data["faults"] = {500: {"mep_fault": 1}}
        result = self.plugin.session.query(incident
                                           ).order_by(incident.incident_start_time
                                                      ).filter(incident.server_id == self.plugin.server_id,
                                                               ).all()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].snapshot_count, snapshot_saver.DEFAULT_MAX_SNAPSHOTS_PER_INCIDENT)

    def test_finalize_incidents(self):
        server_id = self.plugin.server_id
        self.plugin.now = 1200
        self.plugin.period_between_incidents = 50
        incidents = [
            dict(uid=500, incident_start_time=1000, dump_time=1100, server_id=server_id),
            dict(uid=400, incident_start_time=1000, dump_time=1000, server_id=server_id),
            dict(uid=300, incident_start_time=1000, dump_time=1230, server_id=server_id),
            dict(uid=500, incident_start_time=100, incident_end_time=100, dump_time=100, server_id=server_id)]
        self.plugin.session.execute(insert(incident), incidents)
        self.plugin.session.commit()
        self.plugin.finalize_incidents()
        result = self.plugin.session.query(incident.uid, incident.incident_end_time
                                           ).filter(incident.server_id == self.plugin.server_id
                                                    ).order_by(incident.uid, incident.incident_start_time).all()
        self.assertEqual(result, [(300, None), (400, 1000), (500, 100), (500, 1100)])

    @mock.patch.object(SnapshotHelper, "retrieve_queries", return_value=mock_retrieve_queries)
    def test_change_dump_time_and_snapshot_count_1(self, mock_retrieve_queries_result):
        # test first incident
        with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
            self.plugin.execute(self.lve_data)
            inc_numb = 1
            result = self.plugin.session.query(incident.uid, incident.dump_time, incident.snapshot_count).all()
            self.assertEqual([(500, self.plugin.now, inc_numb)], result)

    @mock.patch.object(SnapshotHelper, "retrieve_queries", return_value=mock_retrieve_queries)
    def test_change_dump_time_and_snapshot_count_2(self, mock_retrieve_queries_result):
        with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
            self.plugin.execute(self.lve_data)
            self.plugin.now += self.plugin._period
            self.assertEqual(self.lve_data["faults"], {})
            self.lve_data["faults"] = {500: {"cpu_fault": 2}}
            self.plugin.execute(self.lve_data)
            inc_numb = 2
            result = self.plugin.session.query(incident.uid, incident.dump_time, incident.snapshot_count).all()
            self.assertEqual([(500, self.plugin.now, inc_numb)], result)

    @mock.patch.object(SnapshotHelper, "retrieve_queries", return_value=mock_retrieve_queries)
    def test_change_dump_time_and_snapshot_count_3(self, mock_retrieve_queries_result):
        with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
            for inc_numb in range(1, self.plugin.max_snapshots_per_incident + 1):
                self.plugin.now += self.plugin._period
                self.plugin.execute(self.lve_data)
                self.lve_data["faults"] = {500: {"io_fault": inc_numb}}

        result = self.plugin.session.query(incident.uid, incident.dump_time,
                                           incident.snapshot_count).all()
        self.assertEqual([(500, self.plugin.now, self.plugin.max_snapshots_per_incident)], result)

    @mock.patch.object(SnapshotHelper, "retrieve_queries", return_value=mock_retrieve_queries)
    def test_change_dump_time_and_snapshot_count_4(self, mock_retrieve_queries_result):
        start_time = self.plugin.now
        with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
            for inc_numb in range(1, self.plugin.max_snapshots_per_incident + 2):
                self.plugin.now += self.plugin._period
                self.plugin.execute(self.lve_data)
                self.lve_data["faults"] = {500: {"memphy_fault": inc_numb}}

        result = self.plugin.session.query(incident.uid, incident.dump_time,
                                           incident.snapshot_count).all()
        dump_time = start_time + self.plugin._period * self.plugin.max_snapshots_per_incident
        self.assertEqual(
            [(500, dump_time, self.plugin.max_snapshots_per_incident)], result)

    def test_save_old_incidents(self):
        self.plugin.now = 100
        self.plugin.period_between_incidents = 50
        self.plugin.get_incident(500)
        self.plugin.get_incident(400)
        self.plugin.get_incident(300)
        self.plugin.save_old_incidents()
        self.assertEqual(len(self.plugin.incidents_last_ts), 3)
        self.plugin.now = 200
        self.plugin.get_incident(500)
        self.plugin.get_incident(400)
        self.plugin.save_old_incidents()
        self.assertEqual(len(self.plugin.incidents_last_ts), 2)
        self.assertIn(400, self.plugin.incidents_last_ts)
        self.assertIn(500, self.plugin.incidents_last_ts)
        self.plugin.now = 251
        self.plugin.get_incident(500)
        self.plugin.save_old_incidents()
        self.assertEqual(len(self.plugin.incidents_last_ts), 1)
        self.assertIn(500, self.plugin.incidents_last_ts)
        self.plugin.now = 350
        self.plugin.save_old_incidents()
        self.assertEqual(len(self.plugin.incidents_last_ts), 0)
        result = self.plugin.session.query(incident).filter(incident.server_id == self.plugin.server_id
                                                            ).order_by(incident.uid).all()
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0].uid, 300)
        self.assertEqual(result[0].incident_start_time, 100)
        self.assertEqual(result[0].incident_end_time, 100)
        self.assertEqual(result[1].uid, 400)
        self.assertEqual(result[1].incident_start_time, 100)
        self.assertEqual(result[1].incident_end_time, 200)
        self.assertEqual(result[2].uid, 500)
        self.assertEqual(result[2].incident_start_time, 100)
        self.assertEqual(result[2].incident_end_time, 251)

    def test_get_incident(self):
        self.plugin.now = 100
        incident1 = self.plugin.get_incident(500)
        self.assertEqual(incident1.uid, 500)
        self.assertEqual(incident1.incident_start_time, 100)

        self.plugin.now = 150
        incident2 = self.plugin.get_incident(500)
        self.assertIs(incident1, incident2)

        self.plugin.now = 200
        incident3 = self.plugin.get_incident(400)
        self.assertEqual(incident3.uid, 400)
        self.assertEqual(incident3.incident_start_time, 200)
        self.assertIsNot(incident1, incident3)

    def test_init_session(self):
        # create plugin instance
        plugin = SnapshotSaver()
        plugin.now = time.time()
        plugin.set_db_engine(self.engine)

        self.assertTrue(plugin.first_run)
        plugin.init_session()
        self.assertFalse(plugin.first_run)

    @mock.patch.object(SnapshotHelper, "retrieve_queries", return_value=mock_retrieve_queries)
    def test_execute(self, mock_retrieve_queries_result):
        self.plugin.now = 100
        lve_usage = self.lve_data['lve_usage']
        lve_faults = {"cpu_fault": 1}
        with mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
            self.plugin.execute({'lve_usage': lve_usage, 'faults': {500: lve_faults}})
            self.assertEqual(len(self.plugin.incidents_last_ts), 1)
            self.plugin.now += 100
            self.plugin.execute({'lve_usage': lve_usage, 'faults': {500: lve_faults}})
            self.assertEqual(len(self.plugin.incidents_last_ts), 1)
            _incident = self.plugin.get_incident(500)
            self.assertEqual(_incident.snapshot_count, 2)
            self.assertEqual(_incident.dump_time, 200)

    @mock.patch.object(SnapshotHelper, "retrieve_queries", return_value=mock_retrieve_queries)
    def test_execute_snapshots_disabled(self, mock_retrieve_queries_result):
        self.plugin.now = 100
        self.plugin.set_config({"disable_snapshots": "true"})
        lve_usage = self.lve_data['lve_usage']
        lve_faults = {"cpu_fault": 1}
        self.plugin.execute({'lve_usage': lve_usage, 'faults': {500: lve_faults}})
        self.assertEqual(len(self.plugin.incidents_last_ts), 0)
        self.plugin.now += 100
        self.plugin.execute({'lve_usage': lve_usage, 'faults': {500: lve_faults}})
        self.assertEqual(len(self.plugin.incidents_last_ts), 0)
        _incident = self.plugin._incomplete_incidents_query().filter(
            incident.uid == 500).first()
        self.assertFalse(_incident)

    @mock.patch.object(SnapshotHelper, "retrieve_queries", return_value=mock_retrieve_queries)
    def test_save_snapshot(self, mock_retrieve_queries_result):
        self.plugin.now = 100
        _incident = dict(uid=500, incident_start_time=self.plugin.now, incident_end_time=None, snapshot_count=0)
        self.plugin._helper.invalidate(list(self.lve_data['lve_usage'].keys()))
        faults = {'cpu_fault': 5, 'mem_fault': 7}
        self.plugin.cache_snapshots({500: faults})
        self.plugin.save_snapshot(_incident, faults)
        self.assertEqual(_incident["snapshot_count"], 1)
        snapshots = Snapshot(_incident).get_incident_snapshots()
        self.assertEqual(snapshots,
                         [{u'server_id': u'localhost', u'uid': 500, u'incident_start_time': 100,
                           u'snap_faults': {u'cpu_fault': 5, u'mem_fault': 7},
                           u'snap_proc': {u'1429': {u'MEM': u'1', u'IOPS': u'N/A', u'PID': u'su', u'CPU': u'0%',
                                                    u'IO': u'N/A'}},
                           u'snap_sql': [[u'mysql', u'2015-01-19 11:44:00', u'SELECT 1;']], u'snap_http': [],
                           u'dump_time': 100}])
        self.plugin.now += 100
        faults = {'cpu_fault': 5}
        self.plugin.cache_snapshots({500: faults})
        self.plugin.save_snapshot(_incident, faults)
        self.assertEqual(_incident["snapshot_count"], 2)
        snapshots = Snapshot(_incident).get_incident_snapshots()
        self.assertEqual(snapshots,
                         [{u'server_id': u'localhost', u'uid': 500, u'incident_start_time': 100,
                           u'snap_faults': {u'cpu_fault': 5, u'mem_fault': 7},
                           u'snap_proc': {u'1429': {u'MEM': u'1', u'IOPS': u'N/A', u'PID': u'su', u'CPU': u'0%',
                                                    u'IO': u'N/A'}},
                           u'snap_sql': [[u'mysql', u'2015-01-19 11:44:00', u'SELECT 1;']], u'snap_http': [],
                           u'dump_time': 100},
                          {u'server_id': u'localhost', u'uid': 500, u'incident_start_time': 100,
                           u'snap_faults': {u'cpu_fault': 5},
                           u'snap_proc': {u'1429': {u'MEM': u'1', u'IOPS': u'N/A', u'PID': u'su', u'CPU': u'0%',
                                                    u'IO': u'N/A'}},
                           u'snap_sql': [[u'mysql', u'2015-01-19 11:44:00', u'SELECT 1;']], u'snap_http': [],
                           u'dump_time': 200}])

    @mock.patch('lvestats.plugins.generic.snapshot_saver.SnapshotHelper.get_snapshot_data')
    def test_bad_unicode(self, mock_get_snapshot_data):
        lve_usages = self.lve_data['lve_usage']
        self.plugin._helper.invalidate(list(lve_usages.keys()))
        mock_get_snapshot_data.return_value = (
            {
                4400: {'CPU': '26%', 'IO': 'N/A', 'MEM': '1', 'CMD': 'md5sum', 'IOPS': 'N/A'},
                4381: {'CPU': '0%', 'IO': 'N/A', 'MEM': '1', 'CMD': 'su', 'IOPS': 'N/A'},
                4382: {'CPU': '0%', 'IO': 'N/A', 'MEM': '1', 'CMD': 'bash', 'IOPS': 'N/A'}},
            [
                [b"Sleep", 1, b""],
                [b"Query", 5, b'test test \xa0 test']],
            [
                [b'123461', 'example%s.みんな'.encode() % b'\xa4', b'GET', '/русский.html'.encode(), b'HTTP/1.0', b'88.0'],
                [b'123475', "http://💩.la".encode(), b'GET', b'/\xa0\xa0/', b'HTTP/1.0', b'9.0']])
        inc = dict(
            uid=500,
            dump_time=30,
            incident_end_time=30,
            incident_start_time=27,
            server_id="qwert",
            snapshot_count=2)
        faults = {'cpu_fault': 1}
        self.plugin.cache_snapshots({500: faults})
        self.plugin.save_snapshot(inc, faults)
        saved_snapshot = Snapshot(inc).get_snapshots(self.plugin.now, self.plugin.now)
        self.assertEqual(len(saved_snapshot), 1)
        self.assertEqual(saved_snapshot[0]["snap_sql"], [[u'Sleep', 1, u''], [u'Query', 5, u'test test \ufffd test']])

        # pylint: disable=line-too-long
        self.assertEqual(saved_snapshot[0]["snap_http"],
                         [[u'123461', u'example\ufffd.\u307f\u3093\u306a', u'GET', u'/\u0440\u0443\u0441\u0441\u043a\u0438\u0439.html', u'HTTP/1.0', u'88.0'],
                          [u'123475', u'http://\U0001f4a9.la', u'GET', u'/\ufffd\ufffd/', u'HTTP/1.0', u'9.0']])

    @mock.patch('lvestats.plugins.generic.snapshot_saver.SnapshotHelper.get_snapshot_data')
    def test_good_unicode(self, mock_get_snapshot_data):
        lve_usages = self.lve_data['lve_usage']
        self.plugin._helper.invalidate(list(lve_usages.keys()))
        mock_get_snapshot_data.return_value = (
            {
                4400: {'CPU': '26%', 'IO': 'N/A', 'MEM': '1', 'CMD': 'md5sum', 'IOPS': 'N/A'}
            },
            [
                [u"Sleep", 1, b'SELECT SLEEP(1), "\xd0\xb0\xd0\xba\xd1\x81\xd0\xb5\xd1\x81\xd0\xbe\xd0\xb0\xd1\x80", '
                              b'"Jos\xc3\xa9 Beitr\xc3\xa4ge" FROM MyGuests']
            ], [])

        inc = dict(
            uid=500,
            dump_time=30,
            incident_end_time=30,
            incident_start_time=27,
            server_id="qwert",
            snapshot_count=2)
        faults = {'cpu_fault': 1}
        self.plugin.cache_snapshots({500: faults})
        self.plugin.save_snapshot(inc, faults)
        saved_snapshot = Snapshot(inc).get_snapshots(self.plugin.now, self.plugin.now)
        self.assertEqual(len(saved_snapshot), 1)
        self.assertEqual(saved_snapshot[0]["snap_sql"],
                         [[u'Sleep', 1, u'SELECT SLEEP(1), "\u0430\u043a\u0441\u0435\u0441\u043e\u0430\u0440", '
                                        u'"Jos\xe9 Beitr\xe4ge" FROM MyGuests']])

    @ddt.data(
        [False, True, False],
        [True, True, True],
        [False, False, False],
        [True, False, True],
        [None, False, False],
        [None, True, True])
    @ddt.unpack
    def test_force_webserver(self, force_litespeed, is_litespeed_runnnig, use_litespeed):
        litespeed_helper = LitespeedHelper()
        litespeed_helper.is_running = is_litespeed_runnnig
        litespeed_helper.force_litespeed = force_litespeed
        self.assertEqual(litespeed_helper.get_use_litespeed(), use_litespeed)
        litespeed_helper.broken_config = True
        self.assertEqual(litespeed_helper.get_use_litespeed(), False)

    @ddt.data(
        ['on', True],
        ['ON', True],
        ['1', True],
        ['off', False],
        ['oFF', False],
        ['0', False],
        ['auto', None])
    @ddt.unpack
    def test_webserver_config_option(self, option, force_litespeed_etalon):
        force_webserver = self.plugin._get_webserver_option({'litespeed': option})
        self.assertEqual(force_webserver, force_litespeed_etalon)

    def test_insert_new_incidents(self):
        self.plugin.now = 1000
        self.plugin.incidents_cache = {
            1: {
                'uid': 1,
                "incident_start_time": 100,
                "server_id": self.plugin.server_id,
                "snapshot_count": 5,
                "incident_end_time": None},
            2: {
                'uid': 2,
                "incident_start_time": 160,
                "server_id": self.plugin.server_id,
                "snapshot_count": 6,
                "incident_end_time": None},
            3: {
                'uid': 3,
                "incident_start_time": 130,
                "server_id": self.plugin.server_id,
                "snapshot_count": 1,
                "incident_end_time": None}}
        self.plugin.engine.execute(incident.__table__.insert(), list(self.plugin.incidents_cache.values()))
        result = self.plugin.session.query(
            incident.uid,
            incident.incident_start_time,
            incident.incident_end_time,
            incident.dump_time,
            incident.snapshot_count).order_by(incident.uid).all()
        self.assertListEqual(
            [
                (1, 100, None, None, 5),
                (2, 160, None, None, 6),
                (3, 130, None, None, 1)],
            result)

        self.plugin._insert_new_incidents({
            4: {"cpu_fault": 20},
            5: {"cpu_fault": 2},
            7: {"io_fault": 1}})
        result = self.plugin.session.query(
            incident.uid,
            incident.incident_start_time,
            incident.incident_end_time,
            incident.dump_time,
            incident.snapshot_count).order_by(incident.uid).all()
        self.assertListEqual(
            [
                (1, 100, None, None, 5),
                (2, 160, None, None, 6),
                (3, 130, None, None, 1),
                (4, 1000, None, None, 0),
                (5, 1000, None, None, 0),
                (7, 1000, None, None, 0)],
            result)
        self.assertEqual(len(self.plugin.incidents_cache), 6)
        self.assertEqual(self.plugin.incidents_cache[4]["uid"], 4)
        self.assertEqual(self.plugin.incidents_cache[4]["snapshot_count"], 0)
        self.assertEqual(self.plugin.incidents_cache[5]["incident_start_time"], 1000)
        self.assertEqual(self.plugin.incidents_cache[7]["incident_end_time"], None)
        self.assertEqual(self.plugin.incidents_cache[7]["server_id"], self.plugin.server_id)

    def test_increment_snapshot_count(self):
        self.plugin.now = 1250
        self.plugin.engine.execute(
            incident.__table__.insert(),
            [
                {
                    'uid': 1,
                    "incident_start_time": 100,
                    "server_id": self.plugin.server_id,
                    "snapshot_count": 5,
                    "dump_time": 250,
                    "incident_end_time": None},
                {
                    'uid': 2,
                    "incident_start_time": 100,
                    "server_id": "another_server",
                    "snapshot_count": 4,
                    "dump_time": 250,
                    "incident_end_time": None},
                {
                    'uid': 3,
                    "incident_start_time": 100,
                    "server_id": self.plugin.server_id,
                    "snapshot_count": 9,
                    "dump_time": 250,
                    "incident_end_time": 360},
                {
                    'uid': 4,
                    "incident_start_time": 170,
                    "server_id": self.plugin.server_id,
                    "snapshot_count": 0,
                    "dump_time": 250,
                    "incident_end_time": 200},
                {
                    'uid': 5,
                    "incident_start_time": 100,
                    "server_id": self.plugin.server_id,
                    "snapshot_count": 10,
                    "dump_time": 250,
                    "incident_end_time": None},
                {
                    'uid': 6,
                    "incident_start_time": 100,
                    "server_id": self.plugin.server_id,
                    "snapshot_count": 0,
                    "dump_time": 250,
                    "incident_end_time": None},
                {
                    'uid': 7,
                    "incident_start_time": 100,
                    "server_id": "another_server",
                    "snapshot_count": 0,
                    "dump_time": None,
                    "incident_end_time": None}])
        self.plugin._increment_snapshot_count(
            {
                1: {"cpu_fault": 20},
                2: {"cpu_fault": 2},
                3: {"io_fault": 1},
                4: {"cpu_fault": 20},
                5: {"cpu_fault": 2},
                6: {"io_fault": 1}})
        result = self.plugin.session.query(
            incident.uid,
            incident.incident_start_time,
            incident.incident_end_time,
            incident.dump_time,
            incident.snapshot_count).order_by(incident.uid).all()
        self.assertListEqual(
            [
                (1, 100, None, 1250, 6),
                (2, 100, None, 250, 4),
                (3, 100, 360, 250, 9),
                (4, 170, 200, 250, 0),
                (5, 100, None, 250, 10),
                (6, 100, None, 1250, 1),
                (7, 100, None, None, 0)],
            result)

    @ddt.data(
        (-1, 1),
        (0, 1),
        (1, 1),
        (5, 5),
        (9, 9),
        (10, 9))
    @ddt.unpack
    def test_config(self, value, result):
        self.plugin.set_config({'compresslevel': str(value)})
        self.assertEqual(self.plugin.compresslevel, result)

    @ddt.data(
        ("", False),
        ("admin:12345", False),
        ("admin:12345\n<br />\n<b>Warning</b>:  PHP Startup: PDO: Unable to initialize module", True))
    @ddt.unpack
    def test_incorrect_litespeed_config(self, test_value, result):
        open_original = builtins.open

        def mock_file(filename, value):
            def open_function(*args, **kwargs):
                if args[0] == filename:
                    return StringIO(value)
                else:
                    open_file = open_original(*args, **kwargs)
                    return open_file

            return open_function

        with mock.patch.object(LiteSpeed, "is_litespeed_running", return_value=True), \
                mock.patch("os.path.exists", return_value=True), \
                mock.patch("builtins.open", side_effect=mock_file(LiteSpeed.HTPASSWD_PATH, test_value)), \
                mock.patch('lvestats.plugins.generic.snapshot_saver.reboot_lock'):
            self.assertFalse(self.plugin.litespeed_info.broken_config)
            self.plugin.execute({})
            self.assertEqual(self.plugin.litespeed_info.broken_config, result)

    @ddt.data(
        ([], ['usermaindomain.com', 'www.usermaindomain.com']),
        (['www.domainalias.com'], ['www.domainalias.com', 'domainalias.com']),
        (['domainalias.com'], ['domainalias.com', 'www.domainalias.com']),
        ([], ['www.subdomain.user.com'])
    )
    @ddt.unpack
    def test_requests_to_be_saved(self, aliases, query_domains_list):
        domains = ['usermaindomain.com', 'subdomain.user.com']
        requests_data = [[1, query_domain, "", "", "", ""] for query_domain in query_domains_list]
        with mock.patch('lvestats.lib.commons.proctitle.get_all_user_domains', return_value=domains + aliases), \
                mock.patch.object(proctitle.Proctitle, "_get_all_data",
                                  return_value=requests_data):
            res = proctitle.Proctitle().get_user_data("user")
            res_domains = sorted(item[1] for item in res)
            self.assertCountEqual(res_domains, sorted(query_domains_list))


if __name__ == '__main__':
    unittest.main()
