# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import json
import os
import shutil
import unittest
from io import StringIO
import datetime

import mock
import ddt

import test.helpers as helpers
import test.test_lib_snapshot
from test.base.base_db_test import BaseDbTest
from lvestats.lib.snapshot import Snapshot
from lvestats.orm import incident
from lvestats.snapshots.reader import (LVEReadSnaphot, snapshot_reader_main,
                                       _calculate_period)
from lvestats.lib.parsers.lve_read_snapshot_argparse import lve_read_snapshot_parser


TEST_PATH = test.test_lib_snapshot.TEST_PATH


test_case_argparse = [
    ('--from 2016-01-01 00:00 --to 2016-01-01 23:59',
     (datetime.datetime(2016, 1, 1, 0, 0),
      datetime.datetime(2016, 1, 1, 23, 59))),

    ('--period=1h',
     (datetime.datetime(2000, 1, 1, 23, 0),
      datetime.datetime(2000, 1, 2, 0, 0))),

    ('--timestamp=1478773041',
     (1478773041.0,
      1478773041.999999)),

    ('--from 2016-01-001 00:00 --to 2016-01-01 23:59',
     (None, None)),  # for testing incorrect --from

    ('',
     (datetime.datetime(2000, 1, 1, 23, 50),
      datetime.datetime(2000, 1, 2, 0, 0))),  # for testing default period
]


@ddt.ddt
class LveReadSnapshotArgParse(unittest.TestCase):

    @ddt.data(*test_case_argparse)
    @ddt.unpack
    def test_argparse_period(self, args_str, start_end):
        argv_ = args_str.split()
        # temporary change timezone
        with helpers.change_timezone('Europe/London'):
            now = datetime.datetime(2000, 1, 2, 0, 0)  # after millennium 2 days
            parser = lve_read_snapshot_parser(now)
            opts = parser.parse_args(argv_)
            with helpers.nostdout():  # mute print
                self.assertEqual(start_end, _calculate_period(opts))


@ddt.ddt
class LveReadSnapshotStatsTestCase(BaseDbTest):
    def setUp(self):
        shutil.rmtree(TEST_PATH, ignore_errors=True)

        self.patcher1 = mock.patch('lvestats.lib.snapshot.SNAPSHOT_PATH', TEST_PATH)
        self.patcher1.start()

        if os.geteuid() != 0:
            self.patcher2 = mock.patch('lvestats.lib.snapshot.drop_privileges', test.test_lib_snapshot.mock_drop_privileges)
            self.patcher2.start()

        self.patcher3 = mock.patch('lvestats.lib.dbengine.clear_rows')
        self.patcher3.start()

        self.patcher4 = mock.patch('lvestats.lib.dbengine.logging.getLogger')
        self.patcher4.start()

        BaseDbTest.setUp(self)
        self.session = self.mk_db_session()

        snapshots = [
            dict(uid=500, server_id='localhost', incident_start_time=6, dump_time=6),
            dict(uid=500, server_id='localhost', incident_start_time=6, dump_time=8),
            dict(uid=500, server_id='localhost', incident_start_time=6, dump_time=11),
            dict(uid=500, server_id='localhost', incident_start_time=13, dump_time=13),
            dict(uid=500, server_id='localhost', incident_start_time=13, dump_time=15),
            dict(uid=500, server_id='localhost', incident_start_time=13, dump_time=17),
            dict(uid=500, server_id='localhost', incident_start_time=25, dump_time=25),
            dict(uid=512, server_id='localhost', incident_start_time=1, dump_time=1),
            dict(uid=512, server_id='localhost', incident_start_time=1, dump_time=3),
            dict(uid=512, server_id='localhost', incident_start_time=12, dump_time=12),
            dict(uid=512, server_id='localhost', incident_start_time=12, dump_time=15),
            dict(uid=512, server_id='localhost', incident_start_time=12, dump_time=25),
        ]
        incidents = [
            dict(uid=500, server_id='localhost', incident_start_time=6, dump_time=11, incident_end_time=11,
                 snapshot_count=3),
            dict(uid=500, server_id='localhost', incident_start_time=13, dump_time=17, incident_end_time=17,
                 snapshot_count=3),
            dict(uid=500, server_id='localhost', incident_start_time=25, dump_time=25, incident_end_time=25,
                 snapshot_count=1),
            dict(uid=512, server_id='localhost', incident_start_time=1, dump_time=3, incident_end_time=5,
                 snapshot_count=2),
            # Unfinished incident
            dict(uid=512, server_id='localhost', incident_start_time=12, dump_time=25, incident_end_time=None,
                 snapshot_count=3)
        ]
        sql = incident.__table__.insert().values(incidents)
        self.engine.execute(sql)

        for sn in snapshots:
            Snapshot(sn).save(sn)

    def tearDown(self):
        self.session.close()
        BaseDbTest.tearDown(self)

        self.patcher1.stop()
        if os.geteuid() != 0:
            self.patcher2.stop()

        self.patcher3.stop()
        self.patcher4.stop()

    @mock.patch("lvestats.snapshots.reader.LVEReadSnaphot.open")
    def test_no_incidents_timeunit(self, mock_open):
        s = StringIO()
        mock_open.return_value = s
        lve_read_snapshot = LVEReadSnaphot(self.engine, 10000, 30000, 500, 'localhost', None, False)
        lve_read_snapshot.stats("10m")
        self.assertEqual(len(s.getvalue().split('\n')), 2)

    @mock.patch("lvestats.snapshots.reader.LVEReadSnaphot.open")
    def test_list(self, mock_open):
        s = StringIO()
        mock_open.return_value = s
        # temporary change time zone
        with helpers.change_timezone('Europe/London'):
            lve_read_snapshot = LVEReadSnaphot(self.engine, 0, 30, 500, 'localhost', os.devnull, False)
            lve_read_snapshot.list()
        self.assertIn("\n"
                      "1970-01-01T01:00:06\n"
                      "1970-01-01T01:00:08\n"
                      "1970-01-01T01:00:11\n"
                      "1970-01-01T01:00:13\n"
                      "1970-01-01T01:00:15\n"
                      "1970-01-01T01:00:17\n"
                      "1970-01-01T01:00:25\n", s.getvalue())

    @mock.patch("lvestats.snapshots.reader.LVEReadSnaphot.open")
    def test_list_json(self, mock_open):
        s = StringIO()
        mock_open.return_value = s
        lve_read_snapshot = LVEReadSnaphot(self.engine, 0, 30, 500, 'localhost', os.devnull, True)
        lve_read_snapshot.list()
        self.assertEqual(
            json.loads('{"data": [6, 8, 11, 13, 15, 17, 25], "success": 1}'), json.loads(s.getvalue()))

    def test_main(self):
        args = {
            "config": {'db_type': 'sqlite1234', 'sqlite_db_path': ''},
            "argv_": ["-i", "500"]
        }
        with helpers.capture_all(snapshot_reader_main, **args) as output:
            self.assertEqual(1, output["result"])
            self.assertIn('sqlite1234', output["stdout"])

    def test_unit_auto(self):
        lve_read_snapshot = LVEReadSnaphot(None, 0, 30, 500, 'localhost', None, False)
        result = lve_read_snapshot.stats_by_incident(self.session)

        self.assertEqual(result[0], {"snapshots": 3, "incidents": 1, "from": 6, "to": 11, "duration": 5})
        self.assertEqual(result[1], {"snapshots": 3, "incidents": 1, "from": 13, "to": 17, "duration": 4})
        self.assertEqual(result[2], {"snapshots": 1, "incidents": 1, "from": 25, "to": 25, "duration": 0})

    def test_unit_fixed_time_unit_1(self):
        time_unit = 1
        lve_read_snaphot = LVEReadSnaphot(None, 0, 30, 500, 'localhost', None, False)
        result = lve_read_snaphot.stats_by_time_unit(self.session, time_unit)
        self.assertEqual(result, [
            {'incidents': 1, 'to': 6, 'duration': 0, 'from': 5, 'snapshots': 1},
            {'incidents': 1, 'to': 7, 'duration': 1, 'from': 6, 'snapshots': 1},
            {'incidents': 1, 'to': 8, 'duration': 1, 'from': 7, 'snapshots': 1},
            {'incidents': 1, 'to': 9, 'duration': 1, 'from': 8, 'snapshots': 1},
            {'incidents': 1, 'to': 10, 'duration': 1, 'from': 9, 'snapshots': 1},
            {'incidents': 1, 'to': 11, 'duration': 1, 'from': 10, 'snapshots': 1},
            {'incidents': 1, 'to': 12, 'duration': 0, 'from': 11, 'snapshots': 1},
            {'incidents': 1, 'to': 13, 'duration': 0, 'from': 12, 'snapshots': 1},
            {'incidents': 1, 'to': 14, 'duration': 1, 'from': 13, 'snapshots': 1},
            {'incidents': 1, 'to': 15, 'duration': 1, 'from': 14, 'snapshots': 1},
            {'incidents': 1, 'to': 16, 'duration': 1, 'from': 15, 'snapshots': 1},
            {'incidents': 1, 'to': 17, 'duration': 1, 'from': 16, 'snapshots': 1},
            {'incidents': 1, 'to': 18, 'duration': 0, 'from': 17, 'snapshots': 1},
            {'incidents': 1, 'to': 25, 'duration': 0, 'from': 24, 'snapshots': 1},
            {'incidents': 1, 'to': 26, 'duration': 0, 'from': 25, 'snapshots': 1}])

    def test_unit_fixed(self):
        time_unit = 10
        lve_read_snaphot = LVEReadSnaphot(None, 0, 30, 500, 'localhost', None, False)
        result = lve_read_snaphot.stats_by_time_unit(self.session, time_unit)

        self.assertEqual(result[0], {"snapshots": 2, "incidents": 1, "from": 0, "to": 10, "duration": 4})
        self.assertEqual(result[1], {"snapshots": 4, "incidents": 2, "from": 10, "to": 20, "duration": 5})
        self.assertEqual(result[2], {"snapshots": 1, "incidents": 1, "from": 20, "to": 30, "duration": 0})

    def test_unfinished_incident_unit_fixed(self):
        time_unit = 10

        lve_read_snapshot = LVEReadSnaphot(None, 0, 30, 512, 'localhost', None, False)
        result = lve_read_snapshot.stats_by_time_unit(self.session, time_unit)
        self.assertEqual(result[0], {"snapshots": 2, "incidents": 1, "from": 0, "to": 10, "duration": 4})
        self.assertEqual(result[1], {"snapshots": 2, "incidents": 1, "from": 10, "to": 20, "duration": 8})
        # orig duration -> 10, we consider if incomplete, until the end of time unit
        self.assertEqual(result[2], {"snapshots": 1, "incidents": 1, "from": 20, "to": 30, "duration": 5})

    def test_unfinished_incident_unit_auto(self):
        lve_read_snaphot = LVEReadSnaphot(None, 0, 30, 512, 'localhost', None, False)
        result = lve_read_snaphot.stats_by_incident(self.session)

        self.assertEqual(result[0], {"snapshots": 2, "incidents": 1, "from": 1, "to": 5, "duration": 5 - 1})
        self.assertEqual(result[1], {"snapshots": 3, "incidents": 1, "from": 12, "to": 25, "duration": 25 - 12})

    @ddt.data([
        {u'12256': {u'MEM': u'68', u'CMD': u'/home/user/index.php', u'PID': u'11768', u'CPU': u'N/A'},
         u'12072': {u'MEM': u'357', u'CMD': u'/usr/sbin/httpd -k start', u'PID': u'12049', u'CPU': u'N/A'},
         u'12049': {u'MEM': u'358', u'CMD': u'/usr/sbin/httpd -k start', u'PID': u'12049', u'CPU': u'N/A'},
         u'11768': {u'MEM': u'68', u'CMD': u'/home/user/index.php', u'PID': u'11768', u'CPU': u'5%'}},
        [{u'MEM': 136, u'CMD': u'/home/user/index.php', u'PID': u'11768', u'CPU': 5},
         {u'MEM': 715, u'CMD': u'/usr/sbin/httpd -k start', u'PID': u'12049', u'CPU': u'N/A'}]])
    @ddt.unpack
    def test_process_data_aggregate(self, snap_proc, result):
        self.assertEqual(LVEReadSnaphot._process_data_aggregate(snap_proc), result)


if __name__ == '__main__':
    unittest.main()
