# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import os.path
import time
import unittest
from datetime import datetime, timedelta

import ddt
import sqlalchemy
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker

from test.base.base_db_test import BaseDbTest
from lvestats.lib.lveinfolib import HistoryShow
from lvestats.lib.lveinfolib import OutputFormatter
from lvestats.lib.lveinfolib_gov import HistoryShowDBGov
from lvestats.orm.history import history_x60
from lvestats.plugins.generic.dbsaver_x60 import DBSaverX60
from lvestats.plugins.other.v1_db_migrator import V1TimeInterval, V1DBMigrator, V1History

DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')
TS_FILE = os.path.join(DATA_DIR, 'ts_file_test')
V1_DB_CONNECT_STRING = 'sqlite:///' + os.path.join(DATA_DIR, 'v1.db')
V1_BAD_DB_CONNECT_STRING = 'sqlite:///' + os.path.join(DATA_DIR, 'v1_badformat.db')


class V1TimeIntervalTest(unittest.TestCase):

    def setUp(self):
        self.V1Session = sessionmaker(bind=sqlalchemy.engine.create_engine(V1_DB_CONNECT_STRING))
        self.last_ts = datetime(2016, 0o4, 25, 13, 12, 40, 486445)

    @staticmethod
    def rm_ts():
        try:
            os.remove(TS_FILE)
        except IOError:
            pass

    def test_ts_not_initialized(self):
        self.rm_ts()
        v1t = V1TimeInterval(self.V1Session(), ts_file=TS_FILE)
        self.assertEqual(v1t.last_ts, self.last_ts)

    def test_ts_initalized(self):
        new_date = self.last_ts - timedelta(hours=1)
        v1t = V1TimeInterval(self.V1Session(), ts_file=TS_FILE)
        v1t.save_ts_to_file(new_date)
        self.assertEqual(new_date, v1t.last_ts)

    def test_getting_period(self):
        now = datetime.now()
        v1t = V1TimeInterval(self.V1Session(), ts_file=TS_FILE)
        v1t.save_ts_to_file(now)
        self.assertEqual(now, v1t.last_ts)
        from_ts, to_ts = v1t.get_period()
        self.assertEqual(from_ts, now - timedelta(hours=1))
        self.assertEqual(to_ts, now - timedelta(microseconds=1))
        self.assertFalse(v1t.is_too_old())
        v1t.save_timestamp(from_ts)
        self.assertEqual(from_ts, v1t.last_ts)
        for _ in range(30 * 24):
            from_ts, _ = v1t.get_period()
            v1t.save_timestamp(from_ts)
        self.assertTrue(v1t.is_too_old())  # 24 iterations, 30 days... data after that is old

    def test_bad_file(self):
        with open(TS_FILE, 'w') as f:
            f.write("HELLO")
        v1t = V1TimeInterval(self.V1Session(), ts_file=TS_FILE)
        self.assertEqual(self.last_ts, v1t.last_ts)

    def test_empty_database(self):
        v1t = V1TimeInterval(self.V1Session(),
                             ts_file='/tmp/no_such_file_gf4tyg',
                             server_id='no_such_id')
        self.assertTrue(v1t.is_too_old())
        # check very old time (hardcoded more than 100 years)
        self.assertGreater(datetime.now() - v1t.last_ts, timedelta(days=365*100))


def _create_v1(_id, created, weight, server_id,
               cpu, cpu_limit, cpu_max, ncpu,
               mep, mep_limit, mep_max, mep_fault,
               nproc, lnproc, nproc_max, nproc_fault,
               mem, mem_limit, mem_max, mem_fault,
               memphy, lmemphy, memphy_max, memphy_fault,
               io, io_limit, io_max,
               iops, liops, iops_max):
    r = V1History()
    r.id, r.created, r.weight, r.server_id = _id, created, weight, server_id
    r.cpu, r.cpu_limit, r.cpu_max, r.ncpu = cpu, cpu_limit, cpu_max, ncpu
    r.mep, r.mep_limit, r.mep_max, r.mep_fault = mep, mep_limit, mep_max, mep_fault
    r.nproc, r.lnproc, r.nproc_max, r.nproc_fault = nproc, lnproc, nproc_max, nproc_fault
    r.mem, r.mem_limit, r.mem_max, r.mem_fault = mem, mem_limit, mem_max, mem_fault
    r.memphy, r.lmemphy, r.memphy_max, r.memphy_fault = memphy, lmemphy, memphy_max, memphy_fault
    r.io, r.io_limit, r.io_max = io, io_limit, io_max
    r.iops, r.liops, r.iops_max = iops, liops, iops_max
    return r


v1row = _create_v1(10, datetime(2016, 5, 5, 1, 1, 1), 3, 'myhost',
                   10, 10, 10, 1,
                   5, 40, 15, 3,
                   30000, 35000, 33000, 100,
                   40000, 40500, 43000, 100,
                   50000, 53500, 53300, 100,
                   100, 100, 100,
                   100, 200, 100
                   )

v2row = [
    {'lnproc': 35000, 'iops': 100, 'io': 102400, 'cpu_limit': 4000, 'iops_fault': 0, 'id': 10, 'io_limit': 102400,
     'server_id': 'v2id', 'memphy_fault': 100, 'memphy': 50000, 'mep_limit': 40, 'mep_fault': 3, 'lmemphy': 53500,
     'io_fault': 1, 'nproc_fault': 100, 'mem': 40000, 'mem_fault': 100, 'mep': 5, 'nproc': 30000,
     'created': 1462410061, 'liops': 200, 'cpu_fault': 1, 'mem_limit': 40500, 'cpu': 4000},
    {'lnproc': 35000, 'iops': 100, 'io': 102400, 'cpu_limit': 4000, 'iops_fault': 0, 'id': 10, 'io_limit': 102400,
     'server_id': 'v2id', 'memphy_fault': 100, 'memphy': 50000, 'mep_limit': 40, 'mep_fault': 3, 'lmemphy': 53500,
     'io_fault': 1, 'nproc_fault': 100, 'mem': 40000, 'mem_fault': 100, 'mep': 5, 'nproc': 30000,
     'created': 1462410001, 'liops': 200, 'cpu_fault': 1, 'mem_limit': 40500, 'cpu': 4000},
    {'lnproc': 35000, 'iops': 100, 'io': 102400, 'cpu_limit': 4000, 'iops_fault': 0, 'id': 10, 'io_limit': 102400,
     'server_id': 'v2id', 'memphy_fault': 100, 'memphy': 50000, 'mep_limit': 40, 'mep_fault': 3, 'lmemphy': 53500,
     'io_fault': 1, 'nproc_fault': 100, 'mem': 40000, 'mem_fault': 100, 'mep': 5, 'nproc': 30000,
     'created': 1462409941, 'liops': 200, 'cpu_fault': 1, 'mem_limit': 40500, 'cpu': 4000}]


@ddt.ddt
class V1DBMigratorTest(BaseDbTest):
    migrator = None
    old_plugin_location = V1DBMigrator.PLUGIN_LOCATION

    def setUp(self):
        BaseDbTest.setUp(self)
        V1DBMigrator.PLUGIN_LOCATION = '/tmp/plugin.py'
        self.migrator = V1DBMigrator()
        self.migrator.v1_connect_string = V1_DB_CONNECT_STRING
        self.migrator.v2_server_id = 'v2id'
        self.migrator.v1_server_id = 'localhost'
        self.migrator.debug = False
        self.migrator._procs = 4
        self.migrator.control_time = False
        self.migrator.set_db_engine(self.engine)
        self.migrator.skip_on_error = False
        self.migrator._username_to_uid_cache = {'deepedhe': 1020, 'demoskop': 1031}
        self.migrator._conn = self.engine.connect()
        try:
            os.remove('/tmp/v1_migration_last.ts')
        except OSError:
            pass
        self.migrator.init_v1_db(ts='/tmp/v1_migration_last.ts')

    def tearDown(self):
        self.migrator._conn.close()
        BaseDbTest.tearDown(self)
        V1DBMigrator.PLUGIN_LOCATION = self.old_plugin_location

    def test_table_history_exist(self):
        self.migrator.v1_connect_string = str(self.engine.url)
        self.migrator.init_v1_db(ts='/tmp/v1_migration_last.ts')
        self.migrator.execute(dict())

    def test_database_not_exist(self):
        self.migrator.v1_connect_string = V1_DB_CONNECT_STRING + 'no_such_file'
        self.migrator.init_v1_db(ts='/tmp/v1_migration_last.ts')
        self.migrator.execute(dict())

    def test_convert_data_no_uid_for_db_gov(self):
        self.migrator.convert_username_to_uid = lambda _: None
        from_ts, to_ts = self.migrator.time_interval.get_period()
        self.migrator.convert_data(from_ts, to_ts)
        result = HistoryShow(dbengine=self.engine, period_from=from_ts, period_to=to_ts,
                             server_id=self.migrator.v2_server_id,
                             show_columns=['ID', 'aCPU']).proceed()
        self.assertListEqual([[1020, 0.0], [1031, 0.0], [34043, 0.0], [77775, 0.13115]], list(result))

        rows = HistoryShowDBGov(
            dbengine=self.engine,
            period_from=from_ts,
            period_to=to_ts,
            server_id=self.migrator.v2_server_id,
            show_columns=['ID', 'CPU'], cfg=self.cfg
        ).history_dbgov_show()
        self.assertEqual([], rows)

    def test_convert_io(self):
        v2row_ = {}
        self.migrator.convert_io(v1row, v2row_)
        self.assertEqual(v2row_['io'], 100 * 1024)
        self.assertEqual(v2row_['io_limit'], 100 * 1024)
        self.assertEqual(v2row_['io_fault'], 1)

    def test_convert_iops(self):
        v2row_ = {}
        self.migrator.convert_iops_faults(v1row, v2row_)
        self.assertEqual(v2row_['iops_fault'], 0)

    @ddt.data((4, 10, 25, 10, 2, (4000, 10000, 0)),
              (4, 25, 25, 10, 2, (10000, 10000, 0)),
              (4, 50, 25, 10, 2, (10000, 10000, 0)),
              (4, 25, 50, 25, 1, (10000, 10000, 1)),
              (16, 5, 8, 4, 2, (8000, 12800, 0)))
    @ddt.unpack
    def test_convert_cpu(self, procs, cpu, cpu_limit, cpu_max, ncpu, result):
        v2 = V1DBMigrator.convert_cpu_(procs, cpu, cpu_limit, cpu_max, ncpu)
        self.assertEqual(v2, result)

    def test_convert_row(self):
        result = self.migrator.convert_row(v1row, 4)
        self.assertEqual(v2row, result)

    def test_get_v1_gov_data(self):
        from_ts, to_ts = self.migrator.time_interval.get_period()
        data = list()
        for uid in self.migrator.time_interval.get_uids():
            username = self.migrator.convert_uid_to_username(uid)
            data.extend(self.migrator.get_v1_gov_data(from_ts, to_ts, username))
        self.assertEqual(len(data), 10)
        by_username = sorted(data, key=lambda f: f.username)
        self.assertEqual(by_username[0].username, 'deepedhe')
        self.assertEqual(by_username[9].username, 'demoskop')
        by_weight = sorted(data, key=lambda f: f.weight)
        self.assertEqual(by_weight[0].weight, 1)
        self.assertEqual(by_weight[4].weight, 5)
        self.assertEqual(by_weight[9].weight, 60)

    def test_get_v1_data(self):
        from_ts, to_ts = self.migrator.time_interval.get_period()
        data = list()
        for uid in self.migrator.time_interval.get_uids():
            data.extend(self.migrator.get_v1_data(from_ts, to_ts, uid))
        self.assertEqual(len(data), 25)
        by_id = sorted(data, key=lambda f: f.id)
        self.assertEqual((by_id[0].id, by_id[24].id), (0, 77775))
        by_created = sorted(data, key=lambda f: f.created)
        self.assertEqual(to_ts, by_created[24].created)
        self.assertEqual(datetime(2016, 0o4, 25, 13, 0), by_created[0].created)
        by_weight = sorted(data, key=lambda f: f.weight)
        self.assertEqual(60, by_weight[24].weight)
        self.assertEqual(1, by_weight[0].weight)

    def test_convert_data(self):
        from_ts, to_ts = self.migrator.time_interval.get_period()
        self.migrator.convert_data(from_ts, to_ts)
        result = HistoryShow(dbengine=self.engine, period_from=from_ts, period_to=to_ts,
                             server_id=self.migrator.v2_server_id,
                             show_columns=['ID', 'aCPU']).proceed()
        self.assertEqual([[1020, 0.0], [1031, 0.0], [34043, 0.0], [77775, 0.13115]], list(result))

        rows = HistoryShowDBGov(
            dbengine=self.engine,
            period_from=from_ts,
            period_to=to_ts,
            server_id=self.migrator.v2_server_id,
            show_columns=['ID', 'CPU'], cfg=self.cfg
        ).history_dbgov_show()
        output_formatted = OutputFormatter(fields=['ID', 'CPU'], orders=[(['CPU'], lambda x: round(x, 3))])  # pylint: disable=round-builtin
        output_formatted.set_rows(rows)
        result = output_formatted.get_corrected_list()
        self.assertEqual([[1020, 0.664], [1031, 0.122]], result)

    def test_error_convert_data_no_skip(self):
        self.migrator.skip_on_error = False
        from_ts, to_ts = self.migrator.time_interval.get_period()
        self.migrator.convert_data(from_ts, to_ts)
        self.migrator.time_interval.save_uid()
        with self.assertRaises(IntegrityError):
            self.migrator.convert_data(from_ts, to_ts)

    def test_error_convert_data_skip(self):
        from_ts, to_ts = self.migrator.time_interval.get_period()
        self.migrator.skip_on_error = True
        self.migrator.convert_data(from_ts, to_ts)
        self.migrator.time_interval.save_uid()
        self.migrator.convert_data(from_ts, to_ts)

    def test_cleanup(self):
        self.assertFalse(self.migrator.is_done)
        try:
            open(V1DBMigrator.PLUGIN_LOCATION, 'a').close()
        except IOError:
            pass
        self.migrator.cleanup()
        self.assertTrue(self.migrator.is_done)
        self.assertFalse(os.path.isfile(V1DBMigrator.PLUGIN_LOCATION))

    def test_cleanup_x60_table(self):
        dbsaver_x60 = DBSaverX60()
        dbsaver_x60.set_db_engine(self.engine)
        dbsaver_x60.server_id = self.migrator.v2_server_id
        self.save_history_x60_line("localhost", 500, 1)
        self.save_history_x60_line("server_id_1", 510, 2)
        self.save_history_x60_line(self.migrator.v2_server_id, 500, 3)
        self.save_history_x60_line(self.migrator.v2_server_id, 510, 4)
        self.save_history_x60_line(self.migrator.v2_server_id, 503, 5)
        self.save_history_x60_line(self.migrator.v2_server_id, 510, 6)

        history_x60_lines = list(self.engine.execute(history_x60.__table__.select()))
        self.assertEqual(len(history_x60_lines), 6)
        _, _, dt_x60_max, dt_x60_min = dbsaver_x60.get_max_min_ts()
        self.assertIsNotNone(dt_x60_max)
        self.assertIsNotNone(dt_x60_min)

        self.migrator.cleanup()
        history_x60_lines = list(self.engine.execute(history_x60.__table__.select()))
        self.assertEqual(len(history_x60_lines), 2)
        _, _, dt_x60_max, dt_x60_min = dbsaver_x60.get_max_min_ts()
        self.assertIsNone(dt_x60_max)
        self.assertIsNone(dt_x60_min)

    @staticmethod
    def _read_config(arg=None):
        if (arg == '/etc/sysconfig/lvestats2') or (arg is None):
            return {'server_id': 'v2host', 'dbtype': 'sqlite'}
        else:
            return {'v1_server_id': 'v1host', 'v1_connect_string': V1_DB_CONNECT_STRING, 'debug': 'T'}

    def test_set_config(self):
        from clcommon import clconfpars
        from lvestats.lib import config
        old_load = clconfpars.load
        old_isfile = os.path.isfile
        try:
            clconfpars.load = self._read_config
            os.path.isfile = lambda x: True
            config_ = config.read_plugin_config(config.read_config(), 'V1DBMigrator.cfg')
            self.migrator.set_config(config_)
            self.assertEqual(self.migrator.v1_server_id, 'v1host')
            self.assertEqual(self.migrator.v2_server_id, 'v2host')
            self.assertEqual(self.migrator.v1_connect_string, V1_DB_CONNECT_STRING)
            self.assertTrue(self.migrator.debug)
        finally:
            clconfpars.load = old_load
            os.path.isfile = old_isfile

    def test_invalid_database(self):
        self.migrator.v1_connect_string = V1_BAD_DB_CONNECT_STRING
        self.migrator.init_v1_db(ts='/tmp/v1_migration_last.ts')

        self.assertTrue(self.migrator._database_does_not_exist)

        self.migrator.now = time.time()
        self.migrator.execute(dict())


if __name__ == '__main__':
    unittest.main()
